// ************************************************************************** //
// IntersectingLineDoer.cpp : implementation of the CIntersectingLineDoer class
// Copyright(c)2011 Naiky Company. All rights reserved.
//
// Abstract:
//   ļʵCIntersectingLineDoer
// ************************************************************************** //
#include "StdAfx.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

using namespace ncmath;
using namespace nce;
using namespace intersectingline;

const CHAR c_szINTERSECT_AXIS_DIR[] = "INTERSECT_AXIS_DIR";
const CHAR c_szINTERSECT_AXIS_X[] = "INTERSECT_AXIS_X";
const CHAR c_szINTERSECT_AXIS_Y[] = "INTERSECT_AXIS_Y";
const CHAR c_szINTERSECT_STEELTUBEDIA[] = "INTERSECT_STEELTUBEDIA";
const CHAR c_szINTERSECT_CUTPRECISION[] = "INTERSECT_CUTPRECISION";
double _GetEllipsePerimeter(double nLongRadius_,double nShortRadius_);

CIntersectingLineDoer::CIntersectingLineDoer()
{
}

CIntersectingLineDoer::~CIntersectingLineDoer()
{
}

CIntersectingLinePath* CIntersectingLineDoer::TranslateCad2IntersectingPath(IN const CCadObject* pCadObjectSour_,
																			IN const IntersectingLineInfo& IntersectingLineInfo_)
{
	if (!pCadObjectSour_)
	{
		ASSERT(FALSE);
		return NULL;
	}

	switch (pCadObjectSour_->GetType())
	{
	case cadline:
		{
			return TranslateLine2IntersectingPath((CCadLine*)pCadObjectSour_, IntersectingLineInfo_);
		}
		break;

	case cadarc:
		{
			return TranslateArc2IntersectingPath((CCadArc*)pCadObjectSour_, IntersectingLineInfo_);
		}
		break;

	case cadellipsearc:
		{
			return TranslateEllipseArc2IntersectingPath((CCadEllipseArc*)pCadObjectSour_, IntersectingLineInfo_);
		}
		break;

	case cadcircle:
		{
			return TranslateCircle2IntersectingPath((CCadCircle*)pCadObjectSour_, IntersectingLineInfo_);
		}
		break;

	case cadellipse:
		{
			return TranslateEllipse2IntersectingPath((CCadEllipse*)pCadObjectSour_, IntersectingLineInfo_);
		}
		break;

	case cadrectangle:
	case cadpolyline:
	case cadpolygon:
	case cadstar:
		{
			return TranslateMultiSegments2IntersectingPath((CCadMultiSegments*)pCadObjectSour_, IntersectingLineInfo_);
		}
		break;

	default:
		{
			ASSERT(FALSE);
			return NULL;
		}
		break;
	}
}

void CIntersectingLineDoer::AddLinetoIntersectPath(CIntersectingLinePath* pIntersectLine_,
												   DPOINT2 ptStart_, DPOINT2 ptEnd_, double nPrecision_)
{
	if (!pIntersectLine_)
	{
		ASSERT(FALSE);
		return;
	}

	// ߶ˮƽĻֱģָҪע⣬Ϊ߹ܶơ ---  2011 9 1
	if (DOUBLE_EQU(ptStart_.x, ptEnd_.x) || DOUBLE_EQU(ptStart_.y, ptEnd_.y))
	{
		pIntersectLine_->AddtoPath(ptStart_);
		pIntersectLine_->AddtoPath(ptEnd_);
	}
	else
	{
		double _nLength = (ptEnd_ - ptStart_).GetLength();
		const int _nCirculations = (int)(_nLength / nPrecision_); // ָĶ
		if (_nCirculations <= 0)
			return;

		double _nStepLength = _nLength / _nCirculations; // 
		DPOINT2 _vec = ptEnd_ - ptStart_;
		_vec.Unify();
		for (int _i = 0; _i < _nCirculations; ++_i)
		{
			DPOINT2 _pt = ptStart_ + _vec * (_nStepLength * _i);
			pIntersectLine_->AddtoPath(_pt);
		}
		pIntersectLine_->AddtoPath(ptEnd_);
	}
}

void CIntersectingLineDoer::AddArctoIntersectPath(CIntersectingLinePath* pIntersectLine_,
												  DPOINT2 ptCenter_, double nRadius_,
												  double nStartAngle_, double nEndAngle_, double nPrecision_)
{
	if (!pIntersectLine_ || DOUBLE_EQU_ZERO(nRadius_) || DOUBLE_EQU(nStartAngle_, nEndAngle_))
	{
		ASSERT(FALSE);
		return;
	}

	int _nSign = (nEndAngle_ - nStartAngle_) >= 0 ? 1 : -1;
	double _nLength = nRadius_ * _nSign * (nEndAngle_ - nStartAngle_);
	ASSERT(DOUBLE_GREAT_ZERO(_nLength));
	const int _nCirculations = (int)(_nLength / nPrecision_);
	if (_nCirculations <= 0)
		return;

	double _nStepAngle = (nEndAngle_ - nStartAngle_) / _nCirculations; // ÿһСԲӦĽ
	for (int _i = 0; _i < _nCirculations; ++_i)
	{
		double _nAngle = nStartAngle_ + _nStepAngle * (_i);
		DPOINT2 _pt = ptCenter_ + DPOINT2(cos(_nAngle), sin(_nAngle)) * nRadius_;
		pIntersectLine_->AddtoPath(_pt);
	}
	DPOINT2 _ptEnd = ptCenter_ + DPOINT2(cos(nEndAngle_), sin(nEndAngle_)) * nRadius_;
	pIntersectLine_->AddtoPath(_ptEnd);
}

CIntersectingLinePath* CIntersectingLineDoer::TranslateLine2IntersectingPath(IN const nce::CCadLine* pCadLine_,
																			 IN const IntersectingLineInfo& IntersectInfo_)
{
	if (!pCadLine_)
	{
		ASSERT(FALSE);
		return NULL;
	}

	DPOINT2 _ptStart = pCadLine_->GetStartPoint();
	DPOINT2 _ptEnd = pCadLine_->GetEndPoint();
	CIntersectingLinePath* _pIntersectingLinePath = new CIntersectingLinePath;
	AddLinetoIntersectPath(_pIntersectingLinePath, _ptStart, _ptEnd, IntersectInfo_.nCutPrecision);

	return _pIntersectingLinePath;
}

CIntersectingLinePath* CIntersectingLineDoer::TranslateArc2IntersectingPath(IN const CCadArc* pCadArc_,
																			IN const IntersectingLineInfo& IntersectInfo_)
{
	if (!pCadArc_)
	{
		ASSERT(FALSE);
		return NULL;
	}

	DPOINT2 _ptCenter = pCadArc_->GetCenterPoint();
	double _nRadius = pCadArc_->GetRadius();
	double _nStartAngle = pCadArc_->GetStartAngle();
	double _nSweepAngle = pCadArc_->GetCentralAngle();
	double _nEndAngle;
	if (pCadArc_->IsNormalDir())
		_nEndAngle = _nStartAngle + _nSweepAngle;
	else
		_nEndAngle = _nStartAngle - _nSweepAngle;

	CIntersectingLinePath*  _pIntersectingLinePath = new CIntersectingLinePath;
	AddArctoIntersectPath(_pIntersectingLinePath, _ptCenter, _nRadius, _nStartAngle, _nEndAngle, IntersectInfo_.nCutPrecision);
	return _pIntersectingLinePath;
}

CIntersectingLinePath* CIntersectingLineDoer::TranslateCircle2IntersectingPath(IN const CCadCircle* pCircle_,
																			   IN const IntersectingLineInfo& IntersectingLineInfo_)
{
	// IntersectingLinePath ¬ΰ 2011-12-21
	if (!pCircle_ || DOUBLE_EQU_ZERO(pCircle_->GetRadius()))
	{
		ASSERT(FALSE);
		return NULL;
	}
	DPOINT2 _ptCenter = pCircle_->GetCenterPoint();
	double _nRadius = pCircle_->GetRadius();
	CIntersectingLinePath* _pIntersectingLinePath = new CIntersectingLinePath;
	if (pCircle_->IsNormalDir())
	{
		AddArctoIntersectPath(_pIntersectingLinePath, _ptCenter, _nRadius, 0., c_nPIE, IntersectingLineInfo_.nCutPrecision);
		AddArctoIntersectPath(_pIntersectingLinePath, _ptCenter, _nRadius, c_nPIE, 2 * c_nPIE, IntersectingLineInfo_.nCutPrecision);
	}
	else
	{
		AddArctoIntersectPath(_pIntersectingLinePath, _ptCenter, _nRadius, 0., -c_nPIE, IntersectingLineInfo_.nCutPrecision);
		AddArctoIntersectPath(_pIntersectingLinePath, _ptCenter, _nRadius, -c_nPIE, -2 * c_nPIE, IntersectingLineInfo_.nCutPrecision);
	}

	return _pIntersectingLinePath;
}

CIntersectingLinePath* CIntersectingLineDoer::TranslateEllipseArc2IntersectingPath(IN const CCadEllipseArc* pCadEllipseArc_,
																				   IN const IntersectingLineInfo& IntersectInfo_)
{
	if (!pCadEllipseArc_ || DOUBLE_EQU_ZERO(pCadEllipseArc_->GetLongRadius())
		|| DOUBLE_EQU_ZERO(pCadEllipseArc_->GetShortRadius()))
	{
		ASSERT(FALSE);
		return NULL;
	}

	const int _nCirculations = (int)(pCadEllipseArc_->GetLength() / IntersectInfo_.nCutPrecision);
	if (_nCirculations <= 0)
		return NULL;

	double _nSweepAngle = pCadEllipseArc_->GetCentralAngle();
	double _nStepAngle = _nSweepAngle / _nCirculations;
	CIntersectingLinePath* _pIntersectingLinePath = new CIntersectingLinePath;
	for (int _i = 0; _i < _nCirculations; ++_i)
	{
		double _nAngle = _nStepAngle * _i;
		DPOINT2 _pt = pCadEllipseArc_->GetPoint_XY(_nAngle);
		_pIntersectingLinePath->AddtoPath(_pt);
	}

	DPOINT2 _ptEnd = pCadEllipseArc_->GetEndPoint();
	_pIntersectingLinePath->AddtoPath(_ptEnd);

	return _pIntersectingLinePath;
}
// ԲתΪܶСֱ߶δ ¬ΰ 2011-12-20
CIntersectingLinePath* CIntersectingLineDoer::TranslateEllipse2IntersectingPath(IN const CCadEllipse* pEllipse_,
																				IN const IntersectingLineInfo& IntersectInfo_)
{
	if (!pEllipse_ || (pEllipse_->GetLongRadius() == 0) || (pEllipse_->GetShortRadius() == 0))
	{
		ASSERT(FALSE);
		return NULL;
	}

	const int _nCirculations = (int)(pEllipse_->GetLength() / IntersectInfo_.nCutPrecision);
	if (_nCirculations <= 0)
		return NULL;
	// IntersectingLinePath ¬ΰ 2011-12-21
	CIntersectingLinePath* _pIntersectingLinePath = new CIntersectingLinePath;
	if (pEllipse_->IsNormalDir())
	{
		double _nStepRadian = 2. * c_nPIE / _nCirculations;
		for (int _i = 0; _i < _nCirculations; _i++)
		{
			double _nAngle = _nStepRadian * _i;
			DPOINT2 _pt = pEllipse_->GetPoint_XY(_nAngle);
			_pIntersectingLinePath->AddtoPath(_pt);
		}
	}
	else
	{
		double _nStepRadian = -2. * c_nPIE / _nCirculations;
		for (int _i = 0; _i < _nCirculations; _i++)
		{
			double _nAngle = _nStepRadian * _i;
			_nAngle = RegularizeRadian(_nAngle);
			DPOINT2 _pt = pEllipse_->GetPoint_XY(_nAngle);
			_pIntersectingLinePath->AddtoPath(_pt);
		}
	}

	DPOINT2 _ptEnd = pEllipse_->GetPoint_XY(0.);
	_pIntersectingLinePath->AddtoPath(_ptEnd);

	return _pIntersectingLinePath;
}

CIntersectingLinePath* CIntersectingLineDoer::TranslateMultiSegments2IntersectingPath(IN const CCadMultiSegments* pMultiSegments_,
																					  IN const IntersectingLineInfo& IntersectingLineInfo_)
{
	if (!pMultiSegments_)
	{
		ASSERT(FALSE);
		return NULL;
	}

	// IntersectingLinePath ¬ΰ 2011-12-21
	CIntersectingLinePath* _pIntersectingLinePath = new CIntersectingLinePath;

	int _nCount = pMultiSegments_->GetSegmentCount();
	// ڷʱὫͼεڲݷתԴ˴Ҫ
	// truefalse ¬ΰ 2011-12-22
	int _nSegIndex = 0;
	while (_nSegIndex < _nCount)
	{
		SEGMENT _segment = pMultiSegments_->GetSegment(_nSegIndex++);
		if (_segment.nType == SEGMENT::LINE)
		{
			DPOINT2 _ptStart = _segment.ptStart;
			DPOINT2 _ptEnd = _segment.ptEnd;
			AddLinetoIntersectPath(_pIntersectingLinePath, _ptStart, _ptEnd, IntersectingLineInfo_.nCutPrecision);
		}
		else
		{
			DPOINT2 _ptStart = _segment.ptStart;
			DPOINT2 _ptEnd = _segment.ptEnd;
			double _nBulge = _segment.nBulge;
			DPOINT2 _ptCenter;
			double _nRadius;
			double _nStartAngle;
			double _nSweepAngle;
			SEGMENT::SolveArc(_ptStart, _ptEnd, _nBulge, &_ptCenter, &_nRadius, &_nStartAngle, &_nSweepAngle);
			double _nEndAngle = _nStartAngle + _nSweepAngle;
			AddArctoIntersectPath(_pIntersectingLinePath, _ptCenter, _nRadius, _nStartAngle, _nEndAngle,
				IntersectingLineInfo_.nCutPrecision);
		}
	}

	return _pIntersectingLinePath;
}

CCadObject* CIntersectingLineDoer::TranslateIntersectingPath2Cad(IN const CIntersectingLinePath* pIntersectingLinePathSour_)
{
	if (!pIntersectingLinePathSour_ || (pIntersectingLinePathSour_->m_vecPointVector.size() == 0))
	{
		ASSERT(FALSE);
		return NULL;
	}

	CCadPolyline* _pPolyline = new CCadPolyline;
	DPOINT2 _ptStart = pIntersectingLinePathSour_->m_ptStart;
	ASSERT(IsValidDouble(_ptStart.x) && IsValidDouble(_ptStart.y));
	_pPolyline->SetStartPoint(_ptStart);
	int _nSize = (int)pIntersectingLinePathSour_->m_vecPointVector.size();
	ASSERT(_nSize != 0);
	DPOINT2 _ptEnd = pIntersectingLinePathSour_->m_vecPointVector[0];
	DPOINT2 _ptLast = _ptEnd;
	int _nSegNo = 1;
	while (_nSegNo < _nSize)
	{
		DPOINT2 _pt = pIntersectingLinePathSour_->m_vecPointVector[_nSegNo];
		if (D2GetPointToLineDistance(_pt, _ptStart, _ptEnd) > 0.001)
		{
			_pPolyline->AddLine(_ptLast);
			_ptStart = _ptLast;
			_ptEnd = _pt;
		}
		_ptLast = _pt;

		_nSegNo++;
	}

	_ptEnd = pIntersectingLinePathSour_->m_vecPointVector[_nSize - 1];
	_pPolyline->AddLine(_ptEnd);

	delete pIntersectingLinePathSour_;
	pIntersectingLinePathSour_ = NULL;

	return _pPolyline;
}
void CIntersectingLineDoer::SetIntersectingInfotoCad(CCadObject* pCadObject_,
													 IN const IntersectingLineInfo& IntersectingLineInfo_)
{
	if (!pCadObject_ || IntersectingLineInfo_.dwSize != sizeof(IntersectingLineInfo))
	{
		ASSERT(FALSE);
		return;
	}

	if (pCadObject_->GetType() == cadgroup)
	{
		CCadGroup* _pGroup = (CCadGroup*)pCadObject_;
		for (CadObjectList::iterator _it = _pGroup->GetObjectsList()->begin();
			_it != _pGroup->GetObjectsList()->end(); _it++)
		{
			SetIntersectingInfotoCad(*_it, IntersectingLineInfo_);
		}
	}
	else if (pCadObject_->GetType() == cadtext)
	{
		CCadText* _pCadText = (CCadText*)pCadObject_;
		for (CadObjectList::iterator _it = _pCadText->GetTextGroup()->GetObjectsList()->begin();
			_it != _pCadText->GetTextGroup()->GetObjectsList()->end(); _it++)
		{
			SetIntersectingInfotoCad(*_it, IntersectingLineInfo_);
		}
	}
	else
	{
		DRECT _rectBound = pCadObject_->GetBoundRect();
		axis_t _nAxis = IntersectingLineInfo_.nAxis;
		if (_nAxis == axis::x && DOUBLE_GE(_rectBound.width, IntersectingLineInfo_.nSteelTubeDiameter))
		{
			_nAxis = axis::none;
		}
		if (_nAxis == axis::y && DOUBLE_GE(_rectBound.height, IntersectingLineInfo_.nSteelTubeDiameter))
		{
			_nAxis = axis::none;
		}

		if (_nAxis == axis::x)
		{
			pCadObject_->SetKey(c_szINTERSECT_AXIS_DIR, c_szINTERSECT_AXIS_X);
			pCadObject_->SetKey_DBL(c_szINTERSECT_STEELTUBEDIA, IntersectingLineInfo_.nSteelTubeDiameter);
			pCadObject_->SetKey_DBL(c_szINTERSECT_CUTPRECISION, IntersectingLineInfo_.nCutPrecision);
		}
		else if (_nAxis == axis::y)
		{
			pCadObject_->SetKey(c_szINTERSECT_AXIS_DIR, c_szINTERSECT_AXIS_Y);
			pCadObject_->SetKey_DBL(c_szINTERSECT_STEELTUBEDIA, IntersectingLineInfo_.nSteelTubeDiameter);
			pCadObject_->SetKey_DBL(c_szINTERSECT_CUTPRECISION, IntersectingLineInfo_.nCutPrecision);
		}
		else
		{
			pCadObject_->SetKey(c_szINTERSECT_AXIS_DIR, NULL);
			pCadObject_->SetKey(c_szINTERSECT_STEELTUBEDIA, NULL);
			pCadObject_->SetKey(c_szINTERSECT_CUTPRECISION, NULL);
		}
	}
}

bool CIntersectingLineDoer::GetIntersectingInfofromCad(IN const CCadObject* pCadObject_,
													   OUT IntersectingLineInfo& IntersectingLineInfo_)
{
	// ǷͼûϢ ¬ΰ 2011-12-20
	if (!pCadObject_)
	{
		ASSERT(FALSE);
		return false;
	}

	CString _str = pCadObject_->GetKey(c_szINTERSECT_AXIS_DIR);
	if (_str.IsEmpty())
		IntersectingLineInfo_.nAxis = axis::none;
	else if (_str == c_szINTERSECT_AXIS_X)
		IntersectingLineInfo_.nAxis = axis::x;
	else if (_str == c_szINTERSECT_AXIS_Y)
		IntersectingLineInfo_.nAxis = axis::y;
	else
		ASSERT(FALSE);

	double _nSteelTubeDiameter = pCadObject_->GetKey_DBL(c_szINTERSECT_STEELTUBEDIA);
	if (IsInvalidDouble(_nSteelTubeDiameter))
		IntersectingLineInfo_.nSteelTubeDiameter = 0;
	else
		IntersectingLineInfo_.nSteelTubeDiameter = _nSteelTubeDiameter;
	double _nCutPrecision = pCadObject_->GetKey_DBL(c_szINTERSECT_CUTPRECISION);
	if (IsInvalidDouble(_nCutPrecision))
		IntersectingLineInfo_.nCutPrecision = -1;
	else
		IntersectingLineInfo_.nCutPrecision = _nCutPrecision;

	// û ¬ΰ 2011-12-28
	if (IntersectingLineInfo_.nAxis == axis::none || DOUBLE_EQU_ZERO(IntersectingLineInfo_.nSteelTubeDiameter) ||
		DOUBLE_EQU_ZERO(IntersectingLineInfo_.nCutPrecision))
		return false;

	return true;
}

CCadObject* CIntersectingLineDoer::DoIntersecting(IN const CCadObject* pCadObject_)
{
	if (!pCadObject_)
	{
		ASSERT(FALSE);
		return NULL;
	}

	IntersectingLineInfo _IntersectingLineInfo;
	// ݣҪ¼ ¬ΰ 2011-12-21
	if (!GetIntersectingInfofromCad(pCadObject_, _IntersectingLineInfo))
	{
		return NULL;
	}

	CIntersectingLinePath* _pIntersectingLinePathSour =
		TranslateCad2IntersectingPath(pCadObject_, _IntersectingLineInfo);
	if (!_pIntersectingLinePathSour)
	{
		ASSERT(FALSE);
		return NULL;
	}

	CIntersectingLineCore* _pIntersectingLineCore = NULL;
	if (_IntersectingLineInfo.nAxis == axis::x)
		_pIntersectingLineCore = new CIntersectingLineCore_X;
	else if (_IntersectingLineInfo.nAxis == axis::y)
		_pIntersectingLineCore = new CIntersectingLineCore_Y;
	else
	{
		if (_pIntersectingLinePathSour)
		{
			delete _pIntersectingLinePathSour;
			_pIntersectingLinePathSour = NULL;
		}
		ASSERT(FALSE);
		return NULL;
	}

	CCadObject* _pCadObject = NULL;
	//˴ҪеĶκźͶα ¬ΰ 2011-12-21
	CIntersectingLinePath* _pIntersectingLinePathDest = NULL;
	if (_pIntersectingLineCore &&
		_pIntersectingLineCore->Initialze(&_IntersectingLineInfo) &&
		_pIntersectingLineCore->DoIntersectingLine(_pIntersectingLinePathSour,
		&_pIntersectingLinePathDest))
	{
		if (_pIntersectingLinePathDest)
		{
			_pCadObject = TranslateIntersectingPath2Cad(_pIntersectingLinePathDest);
			_pCadObject->SetParamIndex(pCadObject_->GetParamIndex());
		}
	}

	if (_pIntersectingLinePathSour)
	{
		delete _pIntersectingLinePathSour;
		_pIntersectingLinePathSour;
	}

	if (_pIntersectingLineCore)
	{
		delete _pIntersectingLineCore;
		_pIntersectingLineCore = NULL;
	}

	return _pCadObject;
}

////////////////////////////////////////////////////////////////////////////////
// Functions for Version 2 ----  2012-08-31
// 
#include "IntersectingLine.h"
#ifdef _DEBUG
	#pragma comment(lib, "IntersectingLineSd.lib")
#else
	#pragma comment(lib, "IntersectingLineS.lib")
#endif

CCadObject* CIntersectingLineDoer::GetIntersectingLine(const IntersectingInfo* pInfo_)
{
	if (!pInfo_ || pInfo_->nSize < sizeof(IntersectingInfo))
	{
		return NULL;
	}

	if (pInfo_->nType == IntersectingInfo::CircleCut)
	{
		return _GetIntersectingLine_1((IntersectingInfo_1*)pInfo_);
	}
	else if (pInfo_->nType == IntersectingInfo::PrimerCircleCut)
	{
		return _GetIntersectingLine_2((IntersectingInfo_2*)pInfo_);
	}
	else if (pInfo_->nType == IntersectingInfo::AuxiliaryCircleCut)
	{
		return _GetIntersectingLine_3((IntersectingInfo_3*)pInfo_);
	}
	else if (pInfo_->nType == IntersectingInfo::PrimerRectCut)
	{
		return _GetIntersectingLine_4((IntersectingInfo_4*)pInfo_);
	}
	else if (pInfo_->nType == IntersectingInfo::PrimerSymmetricCut)
	{
		return _GetIntersectingLine_5((IntersectingInfo_2*)pInfo_);
	}
	else if (pInfo_->nType == IntersectingInfo::RectCut)
	{
		return _GetIntersectingLine_6((IntersectingInfo_6*)pInfo_);
	}
	else if (pInfo_->nType == IntersectingInfo::EllipseCut)
	{
		return _GetIntersectingLine_7((IntersectingInfo_7*)pInfo_);
	}
	else
	{
		return NULL;
	}
}

CCadObject* CIntersectingLineDoer::_GetIntersectingLine_1(const IntersectingInfo_1* pInfo_)
{
	if (!pInfo_ || pInfo_->nSize != sizeof(IntersectingInfo_1) || pInfo_->nType != 1
		|| (pInfo_->nAxis != 'X' && pInfo_->nAxis != 'Y')
		|| pInfo_->nRadius <= 0
		|| pInfo_->nAngle <= 0 || 180 <= pInfo_->nAngle
		|| pInfo_->nPrecision < 0)
	{
		ASSERT(FALSE);
		return NULL;
	}

	// һֱͶӰжԲֱܵ ----  2012-08-31
	intersecting::CPolyline _Polyline;
	if (pInfo_->nAxis == 'X')
	{
		_Polyline.m_nStartX = DOUBLE_EQU(pInfo_->nAngle, 90) ? 0 : -1 * pInfo_->nRadius / tan(pInfo_->nAngle * c_nPIE / 180);
		_Polyline.m_nStartY = -1 * pInfo_->nRadius;
		_Polyline.m_vectorNodes.push_back(intersecting::CPolyline::NODE(
			DOUBLE_EQU(pInfo_->nAngle, 90) ? 0 : pInfo_->nRadius / tan(pInfo_->nAngle * c_nPIE / 180),
			pInfo_->nRadius));
	}
	else
	{
		_Polyline.m_nStartX = -1 * pInfo_->nRadius;
		_Polyline.m_nStartY = DOUBLE_EQU(pInfo_->nAngle, 90) ? 0 : -1 * pInfo_->nRadius / tan(pInfo_->nAngle * c_nPIE / 180);
		_Polyline.m_vectorNodes.push_back(intersecting::CPolyline::NODE(
			pInfo_->nRadius,
			DOUBLE_EQU(pInfo_->nAngle, 90) ? 0 : pInfo_->nRadius / tan(pInfo_->nAngle * c_nPIE / 180)));
	}

	//  ----  2012-08-31
	intersecting::InfoV1 _info;
	memset(&_info, 0, sizeof(intersecting::InfoV1));
	_info.nSize = sizeof(intersecting::InfoV1);
	_info.nVersion = 'V1';
	_info.nAxis = 0x80000000 | pInfo_->nAxis;
	_info.nRadius = pInfo_->nRadius;
	_info.nProjectAngle = 90;
	_info.nPrecision = min(max(2e-9, pInfo_->nPrecision), pInfo_->nRadius / 10) / 2;
	intersecting::CPolyline _PolylineOut;
	if (!intersecting::DoProjectPolyline(&_Polyline, &_PolylineOut, &_info))
	{
		return NULL;
	}

	// װΪߣֱͶӰжϵԳƵ ----  2012-08-31
	ASSERT(_PolylineOut.m_vectorNodes.size() > 0);
	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut.m_nStartX, _PolylineOut.m_nStartY));
	for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = _PolylineOut.m_vectorNodes[_i].nEndX;
		_node.ptEnd.y = _PolylineOut.m_vectorNodes[_i].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}
	for (size_t _i = _PolylineOut.m_vectorNodes.size(); _i > 0; _i--)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = pInfo_->nAxis == 'X' ? _PolylineOut.m_vectorNodes[_i - 1].nEndX
			: c_nPIE * pInfo_->nRadius - _PolylineOut.m_vectorNodes[_i - 1].nEndX;
		_node.ptEnd.y = pInfo_->nAxis == 'Y' ? _PolylineOut.m_vectorNodes[_i - 1].nEndY
			: c_nPIE * pInfo_->nRadius - _PolylineOut.m_vectorNodes[_i - 1].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}
	if (_PolylineOut.m_vectorNodes.size() > 0)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = pInfo_->nAxis == 'X' ? _PolylineOut.m_nStartX
			: c_nPIE * pInfo_->nRadius - _PolylineOut.m_nStartX;
		_node.ptEnd.y = pInfo_->nAxis == 'Y' ? _PolylineOut.m_nStartY
			: c_nPIE * pInfo_->nRadius - _PolylineOut.m_nStartY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}

	// ߶ ----  2012-08-31
	_ReducePolyline(_pCadPolyline, _info.nPrecision);

	return _pCadPolyline;
}

CCadObject* CIntersectingLineDoer::_GetIntersectingLine_2(const IntersectingInfo_2* pInfo_)
{
	// ɶԳƿʱõԶԳƿ͵ơ ---- ƺ 2013-09-28
	if (!pInfo_ || pInfo_->nSize != sizeof(IntersectingInfo_2)
		|| (pInfo_->nType != 2 && pInfo_->nType != 5)
		|| (pInfo_->nAxis != 'X' && pInfo_->nAxis != 'Y')
		|| pInfo_->nTrunkRadius <= 0
		|| pInfo_->nBranchRadius <= 0 || pInfo_->nBranchRadius > pInfo_->nTrunkRadius
		|| abs(pInfo_->nEccentricLength) > pInfo_->nTrunkRadius - pInfo_->nBranchRadius
		|| pInfo_->nIntersectingAngle <= 0 || 180 <= pInfo_->nIntersectingAngle
		|| pInfo_->nPrecision < 0)
	{
		ASSERT(FALSE);
		return NULL;
	}

	// һԲ ----  2012-08-31
	double _nCenterX = pInfo_->nAxis == 'X' ? 0 : pInfo_->nEccentricLength;
	double _nCenterY = pInfo_->nAxis == 'Y' ? 0 : pInfo_->nEccentricLength;
	double _nRaius = pInfo_->nBranchRadius;

	// ԲָΪС߶ ----  2012-09-15
	double _nPrecision = min(max(2e-9, pInfo_->nPrecision), pInfo_->nTrunkRadius / 10) / 2;
	double _nAngleStep = min(max(1e-9, _nPrecision), c_nPIE / 180.0) / _nRaius;
	int _nSteps = (int)(2.0 * c_nPIE / _nAngleStep + 0.5);
	_nAngleStep = 2.0 * c_nPIE / _nSteps;
	intersecting::CPolyline _Polyline;
	_Polyline.m_nStartX = _nCenterX + _nRaius * cos(0.0);
	_Polyline.m_nStartY = _nCenterY + _nRaius * sin(0.0);
	for (int _i = 1; _i <= _nSteps; _i++)
	{
		intersecting::CPolyline::NODE _node;
		_node.nEndX = _nCenterX + _nRaius * cos(_nAngleStep * _i);
		_node.nEndY = _nCenterY + _nRaius * sin(_nAngleStep * _i);
		_Polyline.m_vectorNodes.push_back(_node);
	}

	//  ----  2012-08-31
	intersecting::InfoV1 _info;
	memset(&_info, 0, sizeof(intersecting::InfoV1));
	_info.nSize = sizeof(intersecting::InfoV1);
	_info.nVersion = 'V1';
	_info.nAxis = 0x80000000 | pInfo_->nAxis;
	_info.nRadius = pInfo_->nTrunkRadius;
	_info.nProjectAngle = pInfo_->nIntersectingAngle;
	_info.nPrecision = _nPrecision * 10; // ʹڲľȿ
	intersecting::CPolyline _PolylineOut;
	if (!intersecting::DoProjectPolyline(&_Polyline, &_PolylineOut, &_info))
	{
		return NULL;
	}

	// װΪ ----  2012-08-31
	ASSERT(_PolylineOut.m_vectorNodes.size() > 0);
	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut.m_nStartX, _PolylineOut.m_nStartY));
	for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = _PolylineOut.m_vectorNodes[_i].nEndX;
		_node.ptEnd.y = _PolylineOut.m_vectorNodes[_i].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}

	// ߶ ----  2012-08-31
	_ReducePolyline(_pCadPolyline, _nPrecision);

	// Ӧһͼ ----  2012-09-03
	ASSERT(DOUBLE_EQU(_pCadPolyline->GetStartPoint().x, _pCadPolyline->GetEndPoint().x)
		&& DOUBLE_EQU(_pCadPolyline->GetStartPoint().y, _pCadPolyline->GetEndPoint().y));
	_pCadPolyline->SetClosed(true);

	return _pCadPolyline;
}

CCadObject* CIntersectingLineDoer::_GetIntersectingLine_3(const IntersectingInfo_3* pInfo_)
{
	if (!pInfo_ || pInfo_->nSize != sizeof(IntersectingInfo_3) || pInfo_->nType != 3
		|| (pInfo_->nAxis != 'X' && pInfo_->nAxis != 'Y')
		|| pInfo_->nTrunkRadius <= 0
		|| pInfo_->nBranchRadius <= 0 || pInfo_->nBranchRadius > pInfo_->nTrunkRadius
		|| abs(pInfo_->nEccentricLength) > pInfo_->nTrunkRadius - pInfo_->nBranchRadius
		|| pInfo_->nIntersectingAngle <= 0 || 180 <= pInfo_->nIntersectingAngle
		|| pInfo_->nPrecision < 0)
	{
		ASSERT(FALSE);
		return NULL;
	}

	// ϽȡһԲ ----  2012-09-03
	double _nStartAngle = asin(-1 * (pInfo_->nEccentricLength + pInfo_->nBranchRadius) / pInfo_->nTrunkRadius);
	double _nEndAngle = asin(-1 * (pInfo_->nEccentricLength - pInfo_->nBranchRadius) / pInfo_->nTrunkRadius);
	if (pInfo_->nAxis == 'Y')
	{
		_nStartAngle += c_nPIE / 2.0;
		_nEndAngle += c_nPIE / 2.0;
	}
	ASSERT(_nEndAngle > _nStartAngle);
	if (_nStartAngle < 0.0)
	{
		_nStartAngle += 2.0 * c_nPIE;
		_nEndAngle += 2.0 * c_nPIE;
	}

	CCadArc _Arc;
	_Arc.SetCenterPoint(DPOINT2(0.0, 0.0));
	_Arc.SetRadius(pInfo_->nTrunkRadius);
	_Arc.SetStartAngle(_nStartAngle);
	_Arc.SetCentralAngle(_nEndAngle - _nStartAngle);
	DPOINT2 _ptCenter = (_Arc.GetStartPoint() + _Arc.GetEndPoint()) / 2.0;
	TranslateCadObject(&_Arc, _ptCenter * -1);
	if (pInfo_->nAxis == 'Y')
	{
		_Arc.ReverseDir();
	}

	// ԲָΪС߶ ----  2012-09-15
	double _nPrecision = min(max(2e-9, pInfo_->nPrecision), pInfo_->nBranchRadius / 10) / 2;
	double _nAngleStep = min(max(1e-9, _nPrecision), c_nPIE / 180.0) / _Arc.GetRadius();
	int _nSteps = (int)(_Arc.GetCentralAngle() / _nAngleStep + 0.5);
	_nAngleStep = _Arc.GetCentralAngle() / _nSteps;
	intersecting::CPolyline _Polyline1; // ǰи
	_Polyline1.m_nStartX = _Arc.GetStartPoint().x;
	_Polyline1.m_nStartY = _Arc.GetStartPoint().y;
	__CUT(pInfo_->nAxis == 'X' ? _Polyline1.m_nStartY : _Polyline1.m_nStartX, -1 * pInfo_->nBranchRadius, pInfo_->nBranchRadius);
	for (int _i = 1; _i <= _nSteps; _i++)
	{
		intersecting::CPolyline::NODE _node;
		DPOINT2 _pt = _Arc.GetPoint(_i * _nAngleStep);
		_node.nEndX = _pt.x;
		_node.nEndY = _pt.y;
		__CUT(pInfo_->nAxis == 'X' ? _node.nEndY : _node.nEndX, -1 * pInfo_->nBranchRadius, pInfo_->nBranchRadius);
		_Polyline1.m_vectorNodes.push_back(_node);
	}
	_Arc.ReverseDir();
	intersecting::CPolyline _Polyline2; // и
	_Polyline2.m_nStartX = _Arc.GetStartPoint().x;
	_Polyline2.m_nStartY = _Arc.GetStartPoint().y;
	__CUT(pInfo_->nAxis == 'X' ? _Polyline2.m_nStartY : _Polyline2.m_nStartX, -1 * pInfo_->nBranchRadius, pInfo_->nBranchRadius);
	for (int _i = 1; _i <= _nSteps; _i++)
	{
		intersecting::CPolyline::NODE _node;
		DPOINT2 _pt = _Arc.GetPoint(_i * _nAngleStep);
		_node.nEndX = _pt.x;
		_node.nEndY = _pt.y;
		__CUT(pInfo_->nAxis == 'X' ? _node.nEndY : _node.nEndX, -1 * pInfo_->nBranchRadius, pInfo_->nBranchRadius);
		_Polyline2.m_vectorNodes.push_back(_node);
	}

	//  ----  2012-09-03
	intersecting::InfoV1 _info;
	memset(&_info, 0, sizeof(intersecting::InfoV1));
	_info.nSize = sizeof(intersecting::InfoV1);
	_info.nVersion = 'V1';
	_info.nAxis = 0x80000000 | pInfo_->nAxis;
	_info.nRadius = pInfo_->nBranchRadius;
	_info.nProjectAngle = pInfo_->nIntersectingAngle;
	_info.nPrecision = _nPrecision * 10.0; // ʹڲľȿ
	intersecting::CPolyline _PolylineOut1;
	if (!intersecting::DoProjectPolyline(&_Polyline1, &_PolylineOut1, &_info))
	{
		return NULL;
	}
	intersecting::CPolyline _PolylineOut2;
	_info.nProjectAngle = 180 - _info.nProjectAngle;
	if (!intersecting::DoProjectPolyline(&_Polyline2, &_PolylineOut2, &_info))
	{
		return NULL;
	}

	// װΪ ----  2012-09-03
	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	ASSERT(_PolylineOut1.m_vectorNodes.size() > 0);
	_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut1.m_nStartX, _PolylineOut1.m_nStartY));
	for (size_t _i = 0; _i < _PolylineOut1.m_vectorNodes.size(); _i++)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = _PolylineOut1.m_vectorNodes[_i].nEndX;
		_node.ptEnd.y = _PolylineOut1.m_vectorNodes[_i].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}
	ASSERT(_PolylineOut2.m_vectorNodes.size() > 0);
	ASSERT(DOUBLE_EQU(_pCadPolyline->GetEndPoint().x, _PolylineOut2.m_nStartX)
		&& DOUBLE_EQU(_pCadPolyline->GetEndPoint().y, _PolylineOut2.m_nStartY));
	for (size_t _i = 1; _i < _PolylineOut2.m_vectorNodes.size(); _i++)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = pInfo_->nAxis == 'X' ? _PolylineOut2.m_vectorNodes[_i].nEndX
			: c_nPIE * pInfo_->nBranchRadius - _PolylineOut2.m_vectorNodes[_i].nEndX;
		_node.ptEnd.y = pInfo_->nAxis == 'Y' ? _PolylineOut2.m_vectorNodes[_i].nEndY
			: c_nPIE * pInfo_->nBranchRadius - _PolylineOut2.m_vectorNodes[_i].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}

	// ߶ ----  2012-09-03
	_ReducePolyline(_pCadPolyline, _nPrecision);

	return _pCadPolyline;
}

CCadObject* CIntersectingLineDoer::_GetIntersectingLine_4(const IntersectingInfo_4* pInfo_)
{
	// ûгǿֱĿףƫľ෶ΧΪܰ뾶ȥ϶̱ߵһ ---- ƺ 2013-07-25
	double _nRange = pInfo_->nTrunkRadius - min(pInfo_->nLength, pInfo_->nWidth) / 2.0;
	ASSERT(_nRange >= 0.0);
	if (!pInfo_ || pInfo_->nSize != sizeof(IntersectingInfo_4) || pInfo_->nType != 4
		|| (pInfo_->nAxis != 'X' && pInfo_->nAxis != 'Y')
		|| pInfo_->nTrunkRadius <= 0
		|| pInfo_->nLength <= 0
		|| pInfo_->nWidth <= 0
		|| pInfo_->nTiltedAngle > c_nPIE
		|| pInfo_->nTiltedAngle < 0.0
		|| abs(pInfo_->nEccentricLength) > _nRange
		|| pInfo_->nIntersectingAngle <= 0
		|| pInfo_->nIntersectingAngle >= 180
		|| pInfo_->nPrecision < 0)
	{
		ASSERT(FALSE);
		return NULL;
	}

	// ȷĵ㣬ԽߵĽ ---- ƺ 2013-07-03
	double _nRectCenterX = pInfo_->nAxis == 'X' ? 0 : pInfo_->nEccentricLength;
	double _nRectCenterY = pInfo_->nAxis == 'Y' ? 0 : pInfo_->nEccentricLength;

	// ĵ½ǵ꣬볤бǹ ---- ƺ 2013-07-03
	double _nDiagonalLineLen = sqrt(pInfo_->nLength * pInfo_->nLength + pInfo_->nWidth * pInfo_->nWidth);
	double _nSideAngle = pInfo_->nAxis == 'X' ? asin(pInfo_->nWidth / _nDiagonalLineLen) : asin(pInfo_->nLength / _nDiagonalLineLen);
	double _nLeftPointAngle = c_nPIE + pInfo_->nTiltedAngle + _nSideAngle;				// ½ǵXļн
	DPOINT2 _ptFirstPoint;
	_ptFirstPoint.x = _nRectCenterX + _nDiagonalLineLen * cos(_nLeftPointAngle) / 2.0;	// һ
	_ptFirstPoint.y = _nRectCenterY + _nDiagonalLineLen * sin(_nLeftPointAngle) / 2.0;
	CCadRectangle _rect(_ptFirstPoint, pInfo_->nLength, pInfo_->nWidth, pInfo_->nTiltedAngle);// 
	if (pInfo_->nAxis == 'Y')
	{
		_rect.SetWidth(pInfo_->nWidth);
		_rect.SetHeight(pInfo_->nLength);
	}

	// ηָС߶ ---- ƺ 2013-07-03
	intersecting::CPolyline _Polyline;
	for (int _i = 0; _i < 4; _i++)
	{
		DPOINT2 _pt = _rect.GetVertex((_i + 1) % 4);
		_pt.x = pInfo_->nAxis == 'Y' ? min(max(-1.0 * pInfo_->nTrunkRadius, _pt.x), pInfo_->nTrunkRadius) : _pt.x;
		_pt.y = pInfo_->nAxis == 'X' ? min(max(-1.0 * pInfo_->nTrunkRadius, _pt.y), pInfo_->nTrunkRadius) : _pt.y;
		_Polyline.m_vectorNodes.push_back(intersecting::CPolyline::NODE(_pt.x, _pt.y));
	}
	_Polyline.m_nStartX = _Polyline.m_vectorNodes.back().nEndX;
	_Polyline.m_nStartY = _Polyline.m_vectorNodes.back().nEndY;

	//  ---- ƺ 2013-07-03
	intersecting::InfoV1 _info;
	memset(&_info, 0, sizeof(intersecting::InfoV1));
	_info.nSize = sizeof(intersecting::InfoV1);
	_info.nVersion = 'V1';
	_info.nAxis = 0x80000000 | pInfo_->nAxis;
	_info.nRadius = pInfo_->nTrunkRadius;
	_info.nProjectAngle = pInfo_->nIntersectingAngle;
	_info.nPrecision = max(1e-9, pInfo_->nPrecision / 2.0);
	intersecting::CPolyline _PolylineOut;
	if (!intersecting::DoProjectPolyline(&_Polyline, &_PolylineOut, &_info))
	{
		return NULL;
	}

	// װɶ ---- ƺ 2013-07-03
	ASSERT(_PolylineOut.m_vectorNodes.size() > 0);
	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut.m_nStartX, _PolylineOut.m_nStartY));
	for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = _PolylineOut.m_vectorNodes[_i].nEndX;
		_node.ptEnd.y = _PolylineOut.m_vectorNodes[_i].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}
	ASSERT(_pCadPolyline->GetNodeCount() > 0);

	// ߶ ---- ƺ 2013-07-03
	_ReducePolyline(_pCadPolyline, max(1e-9, pInfo_->nPrecision / 2.0));

	// ɵͼӦΪһͼ ---- ƺ 2013-07-03
	ASSERT(DOUBLE_EQU(_pCadPolyline->GetStartPoint().x, _pCadPolyline->GetEndPoint().x)
		&& DOUBLE_EQU(_pCadPolyline->GetStartPoint().y, _pCadPolyline->GetEndPoint().y));
	_pCadPolyline->SetClosed(true);

	return _pCadPolyline;
}

CCadObject* CIntersectingLineDoer::_GetIntersectingLine_5(const IntersectingInfo_2* pInfo_)
{
	// step1:ɵһס
	CCadObject* _pFirstHole = _GetIntersectingLine_2(pInfo_);
	ASSERT(_pFirstHole);

	// step2:ɵڶס
	// ת180㣬ȡһбǵĲǣƫľȡɵڶס 
	IntersectingInfo_2* _pInfo = const_cast<IntersectingInfo_2*>(pInfo_);
	_pInfo->nEccentricLength = 0.0 - _pInfo->nEccentricLength;
	_pInfo->nIntersectingAngle = 180.0 - _pInfo->nIntersectingAngle;
	CCadObject* _pSecondHole = _GetIntersectingLine_2(_pInfo);
	ASSERT(_pSecondHole);

	DPOINT2 _vecOffset(0.0, 0.0);
	if (_pInfo->nAxis == 'Y')
	{
		_vecOffset.x = c_nPIE * _pInfo->nTrunkRadius;
	}
	else
	{
		_vecOffset.y = c_nPIE * _pInfo->nTrunkRadius;
	}
	nce::TranslateCadObject(_pSecondHole, _vecOffset);

	// step3:ɵCadͼϳһͼΡ
	CCadGroup* _pCadGroup = new CCadGroup;
	_pCadGroup->PushBack(_pFirstHole);
	_pCadGroup->PushBack(_pSecondHole);
	return _pCadGroup;
}

CCadObject* CIntersectingLineDoer::_GetIntersectingLine_6(const IntersectingInfo_6* pInfo_)
{
	if (!pInfo_ || pInfo_->nSize != sizeof(IntersectingInfo_6) || pInfo_->nType != 6
		|| (/*pInfo_->nAxis != 'X' && */pInfo_->nAxis != 'Y')
		|| pInfo_->nWidth < pInfo_->nHeight
		|| pInfo_->nHeight <= 0.0
		|| pInfo_->nRadius <= 0.0
		|| pInfo_->nThickness <= 0.0
		|| pInfo_->nAngleAlpha <= 0.0 || 180.0 <= pInfo_->nAngleAlpha
		|| pInfo_->nAngleBeta != 90.0
		|| pInfo_->nPrecision < 0.0
		|| pInfo_->nWeldLength < 0.0)
	{
		ASSERT(FALSE);
		return NULL;
	}

	// 90ȵֱҲûк첹
	if (!pInfo_->bWeld || DOUBLE_EQU(pInfo_->nAngleAlpha, 90.0))
	{
		// ûк첹
		// һԭ㵽ܱص ----  2016-08-15
		ASSERT(pInfo_->nAxis == 'Y');
		intersecting::CPolyline _Polyline;
		_Polyline.m_nStartX = 0.0;
		_Polyline.m_nStartY = 0.0;
		_Polyline.m_vectorNodes.push_back(intersecting::CPolyline::NODE(pInfo_->nWidth / 2.0, 0.0));

		// ͶӰ ----  2016-08-15
		intersecting::InfoV2 _info;
		memset(&_info, 0, sizeof(intersecting::InfoV2));
		_info.nSize = sizeof(intersecting::InfoV2);
		_info.nAxis = 0x80000000 | pInfo_->nAxis;
		_info.nVersion = 'V2';
		_info.nWidth = pInfo_->nWidth;
		_info.nHeight = pInfo_->nHeight;
		_info.nRadius = pInfo_->nRadius;
		_info.nThickness = pInfo_->nThickness;
		_info.nProjectAngleAlpha = pInfo_->nAngleAlpha;
		_info.nProjectAngleTheta = 0.0; // չж
		_info.nPrecision = min(max(2e-9, pInfo_->nPrecision), pInfo_->nRadius / 10.0) / 2.0;
		intersecting::CPolyline _PolylineOut;
		if (!intersecting::DoProjectPolyline(&_Polyline, &_PolylineOut, &_info))
		{
			return NULL;
		}

		// װΪߣֻ1/2ͶӰжԳƹϵ ----  2016-08-15
		ASSERT(_PolylineOut.m_vectorNodes.size() > 0);
		CCadPolyline* _pCadPolyline = new CCadPolyline;
		ASSERT(_pCadPolyline);
		_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut.m_nStartX, _PolylineOut.m_nStartY));
		for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndX;
			_node.ptEnd.y = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		for (size_t _i = _PolylineOut.m_vectorNodes.size(); _i > 0; _i--)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = -1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndX + (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = -1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndX + (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = -1.0 * _PolylineOut.m_vectorNodes[_i].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		for (size_t _i = _PolylineOut.m_vectorNodes.size(); _i > 0; _i--)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = -1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndX + 2.0 * (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = +1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		if (_PolylineOut.m_vectorNodes.size() > 0)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = -1.0 * _PolylineOut.m_nStartX + 2.0 * (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = +1.0 * _PolylineOut.m_nStartY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}

		// ߶ ----  2012-08-31
		_ReducePolyline(_pCadPolyline, max(1e-9, _info.nPrecision / 2.0));

		return _pCadPolyline;
	}
	else if (pInfo_->bCurve)
	{
		// ߹⣬ǰ߼߹
		ASSERT(!DOUBLE_EQU(pInfo_->nAngleAlpha, 90.0));

		double _nWidth = pInfo_->nWidth;
		double _nHeight = pInfo_->nHeight;
		double _nArris = max(pInfo_->nRadius, pInfo_->nThickness);
		double _nTan = abs(tan(pInfo_->nAngleAlpha * c_nPIE / 180));
		double _nWeldWidth = _nArris / _nTan;
		double _nY = (pInfo_->nHeight - _nArris * 2.0) / _nTan;

		CCadGroup* _pCadGroup = new CCadGroup;
		CCadPolyline* _pCadPolyline = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldDowm1 = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldDowm2 = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldUp1 = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldUp2 = new CCadPolyline;
		ASSERT(_pCadGroup && _pCadPolyline && _pCadPolylineWeldDowm1
			&& _pCadPolylineWeldDowm2 && _pCadPolylineWeldUp1 && _pCadPolylineWeldUp2);

		CCadPolyline::ARCNODE _node;

		// ԭض
		ASSERT(pInfo_->nAxis == 'Y');
		intersecting::CPolyline _Polyline;
		_Polyline.m_nStartX = 0.0;
		_Polyline.m_nStartY = 0.0;
		_Polyline.m_vectorNodes.push_back(intersecting::CPolyline::NODE(pInfo_->nWidth / 2.0, 0.0));

		// ͶӰ ----  2016-08-15
		intersecting::InfoV2 _info;
		memset(&_info, 0, sizeof(intersecting::InfoV2));
		_info.nSize = sizeof(intersecting::InfoV2);
		_info.nAxis = 0x80000000 | pInfo_->nAxis;
		_info.nVersion = 'V2';
		_info.nWidth = pInfo_->nWidth;
		_info.nHeight = pInfo_->nHeight;
		_info.nRadius = pInfo_->nRadius;
		_info.nThickness = pInfo_->nThickness;
		_info.nProjectAngleAlpha = pInfo_->nAngleAlpha;
		_info.nProjectAngleTheta = 0.0; // չж
		_info.nPrecision = min(max(2e-9, pInfo_->nPrecision), pInfo_->nRadius / 10.0) / 2.0;
		intersecting::CPolyline _PolylineOut;
		if (!intersecting::DoProjectPolyline(&_Polyline, &_PolylineOut, &_info))
		{
			return NULL;
		}

		// װΪߣֻ1/2ͶӰжԳƹϵ ----  2016-08-15
		ASSERT(_PolylineOut.m_vectorNodes.size() > 0);

		_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut.m_nStartX, _PolylineOut.m_nStartY));
		for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndX;
			_node.ptEnd.y = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		for (size_t _i = _PolylineOut.m_vectorNodes.size(); _i > 0; _i--)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = -1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndX + (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = -1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndX + (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = -1.0 * _PolylineOut.m_vectorNodes[_i].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		for (size_t _i = _PolylineOut.m_vectorNodes.size(); _i > 0; _i--)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = -1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndX + 2.0 * (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = +1.0 * _PolylineOut.m_vectorNodes[_i - 1].nEndY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}
		if (_PolylineOut.m_vectorNodes.size() > 0)
		{
			CCadPolyline::ARCNODE _node;
			_node.ptEnd.x = -1.0 * _PolylineOut.m_nStartX + 2.0 * (pInfo_->nWidth + pInfo_->nHeight);
			_node.ptEnd.y = +1.0 * _PolylineOut.m_nStartY;
			_node.nBulge = 0.0;
			_pCadPolyline->GetNodes()->push_back(_node);
		}

		// ߶ ----  2012-08-31
		_ReducePolyline(_pCadPolyline, max(1e-9, _info.nPrecision / 2.0));

		// ڴ˴ߵλòҪҪʰѲҲƶ˴ƽ ---- η 2017-01-11
		double _nBottom = _pCadPolyline->GetBoundRect().GetBottom();
		double _nTop = _pCadPolyline->GetBoundRect().GetTop();

		// ·1
		_pCadPolylineWeldDowm1->SetStartPoint(DPOINT2(_nWidth + _nHeight, _nTop - _nWeldWidth));
		_node.ptEnd.x = _nWidth / 2.0 + _nHeight - _nArris;
		_node.ptEnd.y = _nTop - _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldDowm1->GetNodes()->push_back(_node);
		// ·2
		_pCadPolylineWeldDowm2->SetStartPoint(DPOINT2(_nWidth + _nHeight, _nTop - _nWeldWidth));
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris;
		_node.ptEnd.y = _nTop - _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldDowm2->GetNodes()->push_back(_node);

		// Ϸ1
		_pCadPolylineWeldUp1->SetStartPoint(DPOINT2(_nWidth * 2.0 + _nHeight * 2.0, _nBottom + _nWeldWidth));
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris;
		_node.ptEnd.y = _nBottom + _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldUp1->GetNodes()->push_back(_node);

		// Ϸ2
		_pCadPolylineWeldUp2->SetStartPoint(DPOINT2(0.0, _nBottom + _nWeldWidth));
		_node.ptEnd.x = _nWidth / 2.0 + _nArris;
		_node.ptEnd.y = _nBottom + _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldUp2->GetNodes()->push_back(_node);

		_pCadGroup->PushBack(_pCadPolylineWeldDowm1);
		_pCadGroup->PushBack(_pCadPolylineWeldDowm2);
		_pCadGroup->PushBack(_pCadPolyline);
		_pCadGroup->PushBack(_pCadPolylineWeldUp1);
		_pCadGroup->PushBack(_pCadPolylineWeldUp2);

		return _pCadGroup;
	}
	else if (DOUBLE_EQU_ZERO(pInfo_->nWeldLength))
	{
		// Ҫ첹,Ϊ0
		double _nOverlength = pInfo_->nWeldOverCut;		// Ϊ˱ƽйܱڵĹ
		double _nWidth = pInfo_->nWidth;
		double _nHeight = pInfo_->nHeight;
		double _nArris = max(pInfo_->nRadius, pInfo_->nThickness);
		double _nTan = abs(tan(pInfo_->nAngleAlpha * c_nPIE / 180));
		double _nWeldWidth = _nArris / _nTan;
		double _nY = (pInfo_->nHeight - _nArris * 2.0) / _nTan;

		CCadGroup* _pCadGroup = new CCadGroup;
		CCadPolyline* _pCadPolyline = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldDowm1 = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldDowm2 = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldUp1 = new CCadPolyline;
		CCadPolyline* _pCadPolylineWeldUp2 = new CCadPolyline;
		ASSERT(_pCadGroup && _pCadPolyline && _pCadPolylineWeldDowm1
			&& _pCadPolylineWeldDowm2 && _pCadPolylineWeldUp1 && _pCadPolylineWeldUp2);

		CCadPolyline::ARCNODE _node;

		// ·1
		_pCadPolylineWeldDowm1->SetStartPoint(DPOINT2(_nWidth + _nHeight, _nY + _nWeldWidth));
		_node.ptEnd.x = _nWidth / 2.0 + _nHeight - _nArris - _nOverlength;
		_node.ptEnd.y = _nY + _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldDowm1->GetNodes()->push_back(_node);
		// ·2
		_pCadPolylineWeldDowm2->SetStartPoint(DPOINT2(_nWidth + _nHeight, _nY + _nWeldWidth));
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris + _nOverlength;
		_node.ptEnd.y = _nY + _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldDowm2->GetNodes()->push_back(_node);

		// Ϸ1
		_pCadPolylineWeldUp1->SetStartPoint(DPOINT2(_nWidth * 2.0 + _nHeight * 2.0, _nWeldWidth));
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldUp1->GetNodes()->push_back(_node);

		// Ϸ2
		_pCadPolylineWeldUp2->SetStartPoint(DPOINT2(0.0, _nWeldWidth));
		_node.ptEnd.x = _nWidth / 2.0 + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolylineWeldUp2->GetNodes()->push_back(_node);

		_pCadPolyline->SetStartPoint(DPOINT2(0.0, 0.0));
		_node.ptEnd.x = _nWidth / 2.0 + _nArris + _nOverlength;
		_node.ptEnd.y = 0.0;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth / 2.0 + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth / 2.0 + _nHeight - _nArris - _nOverlength;
		_node.ptEnd.y = _nY + _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth / 2.0 + _nHeight - _nArris - _nOverlength;
		_node.ptEnd.y = _nY + _nWeldWidth * 2.0;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris + _nOverlength;
		_node.ptEnd.y = _nY + _nWeldWidth * 2.0;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris + _nOverlength;
		_node.ptEnd.y = _nY + _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris - _nOverlength;
		_node.ptEnd.y = 0.0;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 2.0 + _nHeight * 2.0;
		_node.ptEnd.y = 0.0;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);

		_pCadGroup->PushBack(_pCadPolylineWeldDowm1);
		_pCadGroup->PushBack(_pCadPolylineWeldDowm2);
		_pCadGroup->PushBack(_pCadPolyline);
		_pCadGroup->PushBack(_pCadPolylineWeldUp1);
		_pCadGroup->PushBack(_pCadPolylineWeldUp2);

		return _pCadGroup;
	}
	else
	{
		// Ҫ첹,಻Ϊ0ʱ
		ASSERT(!DOUBLE_EQU(pInfo_->nAngleAlpha, 90.0));

		double _nOverlength = pInfo_->nWeldOverCut;		// Ϊ˱ƽйܱڵĹ
		double _nWidth = pInfo_->nWidth;
		double _nHeight = pInfo_->nHeight;
		double _nArris = max(pInfo_->nRadius, pInfo_->nThickness);
		double _nTan = abs(tan(pInfo_->nAngleAlpha * c_nPIE / 180));
		double _nWeldWidth = _nArris / _nTan;
		double _nWeldLength = pInfo_->nWeldLength;
		double _nY = (pInfo_->nHeight - _nArris * 2.0) / _nTan;

		CCadGroup* _pCadGroup = new CCadGroup;
		CCadPolyline* _pCadPolyline1 = new CCadPolyline;
		CCadPolyline* _pCadPolyline2 = new CCadPolyline;
		ASSERT(_pCadPolyline1 && _pCadPolyline2);

		CCadPolyline::ARCNODE _node;

		_pCadPolyline1->SetStartPoint(DPOINT2(0.0, 0.0));
		_node.ptEnd.x = _nWidth * 0.5 + _nArris + _nOverlength;
		_node.ptEnd.y = 0;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 0.5 + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 0.5 + _nHeight - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth + _nY;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth + _nY;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris - _nOverlength;
		_node.ptEnd.y = 0;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 2.0 + _nHeight * 2.0;
		_node.ptEnd.y = 0;
		_node.nBulge = 0.0;
		_pCadPolyline1->GetNodes()->push_back(_node);

		_pCadPolyline2->SetStartPoint(DPOINT2(0.0, _nWeldWidth + _nWeldLength));
		_node.ptEnd.x = _nWidth * 0.5 + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 0.5 + _nHeight - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth + _nY + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 0.5 + _nHeight - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth * 2.0+ _nY + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth * 2.0+ _nY + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight + _nArris + _nOverlength;
		_node.ptEnd.y = _nWeldWidth + _nY + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 1.5 + _nHeight * 2.0 - _nArris - _nOverlength;
		_node.ptEnd.y = _nWeldWidth + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);
		_node.ptEnd.x = _nWidth * 2.0 + _nHeight * 2.0;
		_node.ptEnd.y = _nWeldWidth + _nWeldLength;
		_node.nBulge = 0.0;
		_pCadPolyline2->GetNodes()->push_back(_node);

		_pCadGroup->PushBack(_pCadPolyline1);
		_pCadGroup->PushBack(_pCadPolyline2);

		return _pCadGroup;
	}
}


CCadObject* CIntersectingLineDoer::_GetIntersectingLine_7(const IntersectingInfo_7* pInfo_)
{
	if (!pInfo_ || pInfo_->nSize != sizeof(IntersectingInfo_7) || pInfo_->nType != IntersectingInfo::EllipseCut
		|| (pInfo_->nAxis != 'Y')
		|| pInfo_->nShortRadius > pInfo_->nLongRadius
		|| pInfo_->nAngleAlpha <= 0.0 || 180.0 <= pInfo_->nAngleAlpha
		|| pInfo_->nPrecision < 0.0)
	{
		ASSERT(FALSE);
		return NULL;
	}

	ASSERT(pInfo_->nAxis == 'Y');
	double _nCutHeight = pInfo_->nShortRadius / tan(pInfo_->nAngleAlpha / 180 * c_nPIE);
	intersecting::CPolyline _Polyline;
	_Polyline.m_nStartX = -pInfo_->nShortRadius;
	_Polyline.m_nStartY = -_nCutHeight;
	_Polyline.m_vectorNodes.push_back(intersecting::CPolyline::NODE(pInfo_->nShortRadius, _nCutHeight));

	// ͶӰ ----  2017-01-20
	intersecting::InfoV3 _info;
	memset(&_info, 0, sizeof(intersecting::InfoV3));
	_info.nSize = sizeof(intersecting::InfoV3);
	_info.nAxis = 0x80000000 | pInfo_->nAxis;
	_info.nVersion = 'V3';
	_info.nLongRadius = pInfo_->nLongRadius;
	_info.nShortRadius = pInfo_->nShortRadius;
	_info.nProjectAngleAlpha = 90;
	_info.nProjectAngleTheta = 90;
	_info.nPrecision = min(max(2e-9, pInfo_->nPrecision), pInfo_->nShortRadius / 200) / 2.0;
	intersecting::CPolyline _PolylineOut;
	if (!intersecting::DoProjectPolyline(&_Polyline, &_PolylineOut, &_info))
	{
		return NULL;
	}

	//õһ
	ASSERT(_PolylineOut.m_vectorNodes.size() > 0);
	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	_pCadPolyline->SetStartPoint(DPOINT2(_PolylineOut.m_nStartX, _PolylineOut.m_nStartY));
	for (size_t _i = 0; _i < _PolylineOut.m_vectorNodes.size(); _i++)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndX;
		_node.ptEnd.y = +1.0 * _PolylineOut.m_vectorNodes[_i].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}

	//õԳƵһ
	double _nEllipsePerimeter = _GetEllipsePerimeter(pInfo_->nLongRadius, pInfo_->nShortRadius);
	double _nEllipsePerimeter_2 = _nEllipsePerimeter / 2;
	for (size_t _i = _PolylineOut.m_vectorNodes.size(); _i > 0; _i--)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = pInfo_->nAxis == 'X' ? _PolylineOut.m_vectorNodes[_i - 1].nEndX
			: _nEllipsePerimeter - _PolylineOut.m_vectorNodes[_i - 1].nEndX;
		_node.ptEnd.y = pInfo_->nAxis == 'Y' ? _PolylineOut.m_vectorNodes[_i - 1].nEndY
			: _nEllipsePerimeter - _PolylineOut.m_vectorNodes[_i - 1].nEndY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}
	if (_PolylineOut.m_vectorNodes.size() > 0)
	{
		CCadPolyline::ARCNODE _node;
		_node.ptEnd.x = pInfo_->nAxis == 'X' ? _PolylineOut.m_nStartX
			: _nEllipsePerimeter - _PolylineOut.m_nStartX;
		_node.ptEnd.y = pInfo_->nAxis == 'Y' ? _PolylineOut.m_nStartY
			: _nEllipsePerimeter - _PolylineOut.m_nStartY;
		_node.nBulge = 0.0;
		_pCadPolyline->GetNodes()->push_back(_node);
	}
	// ߶ ----  2017-01-20
	_ReducePolyline(_pCadPolyline, max(1e-9, _info.nPrecision / 2.0));

	// ȥΪ0߶Ρ
	double _nx = _pCadPolyline->GetStartPoint().x, _ny = _pCadPolyline->GetStartPoint().y;
	for (size_t _i = 0; _i < _pCadPolyline->GetNodes()->size(); _i++)
	{
		CCadPolyline::ARCNODE& _node = (*_pCadPolyline->GetNodes())[_i];
		if (DOUBLE_EQU(_node.ptEnd.x, _nx) && DOUBLE_EQU(_node.ptEnd.y, _ny))
		{
			_pCadPolyline->GetNodes()->erase(_pCadPolyline->GetNodes()->begin() + _i);
			_i--;
			continue;
		}
		_nx = _node.ptEnd.x;
		_ny = _node.ptEnd.y;
	}

	return _pCadPolyline;
	
}

void CIntersectingLineDoer::_ReducePolyline(CCadPolyline* pCadPolyline_, double nPrecision_)
{
	ASSERT(pCadPolyline_ && pCadPolyline_->GetNodeCount() > 0);
	ASSERT(pCadPolyline_->GetNodeCount() > 0);
	ASSERT(nPrecision_ >= 1e-9);

	// Nodeŵת ----  2012-05-05
	CCadPolyline::ArcNodeVector _vectorOutNodes;

	DPOINT2 _ptStart = pCadPolyline_->GetStartPoint();
	int _nSrcNodeCount = pCadPolyline_->GetNodeCount();
	int _nIndex = 0;
	while (_nIndex < _nSrcNodeCount)
	{
		// Բ
		if (DOUBLE_NOT_ZERO(pCadPolyline_->GetAt(_nIndex).nBulge))
		{
			_vectorOutNodes.push_back(pCadPolyline_->GetAt(_nIndex));
			_ptStart = _vectorOutNodes.back().ptEnd;
			_nIndex++;
			continue;
		}

		// ռ: ߶
		ASSERT(DOUBLE_EQU_ZERO(pCadPolyline_->GetAt(_nIndex).nBulge));
		reducer::CPolyline _Polyline;
		_Polyline.m_nStartX = _ptStart.x;
		_Polyline.m_nStartY = _ptStart.y;
		while (_nIndex < _nSrcNodeCount
			&& DOUBLE_EQU_ZERO(pCadPolyline_->GetAt(_nIndex).nBulge))
		{
			const DPOINT2& _ptEnd = pCadPolyline_->GetAt(_nIndex).ptEnd;
			_Polyline.m_vectorNodes.push_back(reducer::CPolyline::NODE(_ptEnd.x, _ptEnd.y));
			_nIndex++;
		}

		// CPolylineReducerо
		static const reducer::PolylineReducerParam _param = {
			sizeof(reducer::PolylineReducerParam),
			nPrecision_,// Ⱦ
			1.0e-1,		// ǶȾȣԼ6
			1.0e+9,		// ɾ򳤶ȣ޳
			0.5			// ɾǶȣԼ30
		};
		reducer::CPolyline _OutPolyline;
		reducer::CPolyline* _pOutPolyline = &_Polyline;
		if (reducer::ReducePolyline(&_Polyline, &_OutPolyline, &_param))
		{
			_pOutPolyline = &_OutPolyline;
		}

		// д
		ASSERT(_pOutPolyline);
		for (size_t _i = 0; _i < _pOutPolyline->m_vectorNodes.size(); _i++)
		{
			_vectorOutNodes.push_back(CCadPolyline::ARCNODE(
				DPOINT2(_pOutPolyline->m_vectorNodes[_i].nEndX, _pOutPolyline->m_vectorNodes[_i].nEndY),
				0.0));
		}

		// ¶ʱ¸ ----  2012-06-27
		ASSERT(_vectorOutNodes.size() >= _pOutPolyline->m_vectorNodes.size());
		if (_pOutPolyline->m_vectorNodes.size() > 0)
		{
			ASSERT(_vectorOutNodes.back().ptEnd.x == _pOutPolyline->m_vectorNodes.back().nEndX
				&& _vectorOutNodes.back().ptEnd.y == _pOutPolyline->m_vectorNodes.back().nEndY);
			_ptStart.x = _pOutPolyline->m_vectorNodes.back().nEndX;
			_ptStart.y = _pOutPolyline->m_vectorNodes.back().nEndY;
		}
	}

	//  ----  2012-05-05
	CCadPolyline::ArcNodeVector* _pvectorArcNodes = pCadPolyline_->GetNodes();
	ASSERT(_pvectorArcNodes);
	*_pvectorArcNodes = _vectorOutNodes;
}

double _GetEllipsePerimeter(double nLongRadius_,double nShortRadius_)
{
	double _nEccentricity = sqrt(pow(nLongRadius_, 2) - pow(nShortRadius_, 2)) / nLongRadius_;
	ASSERT((nLongRadius_ >= nShortRadius_) && DOUBLE_GREAT_ZERO(nShortRadius_));
	ASSERT((_nEccentricity >= 0.) && (_nEccentricity <= 1.0));
	if ((_nEccentricity < 0.) || (_nEccentricity > 1.0))
	{
		ASSERT(false);
		return 0.;
	}

	// ڶȫԲ֣Matlabellipke.mеõ
	double m = _nEccentricity * _nEccentricity;
	double a0 = 1.0;
	double b0 = sqrt(1.0 - m);
	double s0 = m;
	int i1 = 0;
	double a1 = 1.0;
	double mm = 1.0;
	while (mm > c_nTolerance)
	{
		a1 = (a0 + b0) * 0.5;
		double b1 = sqrt(a0 * b0);
		double c1 = (a0 - b0) * 0.5;
		double w1 = (2 << i1++) * (c1 * c1);
		mm = w1;
		s0 += w1;
		a0 = a1;
		b0 = b1;
	}

	double k = M_PI / (2.0 * a1);
	double e = k * (1.0 - s0 * 0.5);

	// Բܳ
	double _nPerimeter = 4.0 * nLongRadius_ * e;
	return _nPerimeter;
}
