#include "StdAfx.h"
#include "Intersection.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

using namespace ncmath;
#define DIVIDE_BY_ZERO_GUARD 1.E-20

// üεķֱߵĽ㣬˺ǴV9cvtģֲģѯ Wanglou 2010-07-30
bool ncmath::Intersect(IN const LineD2D& line1_, IN const LineD2D& line2_, OUT DPOINT2* pptIntersect_)
{
	double x1,y1,x2,y2;	// (x1,y1,x2,y2)һֱ
	double x3,y3,x4,y4;	// (x3,y3,x4,y4)ڶֱ
	double x5, y5;		// 
	x1 = line1_.pt1.x;	y1 = line1_.pt1.y;	x2 = line1_.pt2.x; y2 = line1_.pt2.y;
	x3 = line2_.pt1.x;	y3 = line2_.pt1.y;	x4 = line2_.pt2.x; y4 = line2_.pt2.y;

	double dvx12, dvy12;	// ֱ12ķʸ
	double dvx34, dvy34;	// ֱ34ķʸ
	double len12, len34;	// 1234ĳ
	double uvx12, uvy12;	// ֱ12ĵλʸ
	double uvx34, uvy34;	// ֱ34ĵλʸ
	double dvx13, dvy13;	// 13ʸ
	double prj;			// 13ʸֱ12ϵͶӰ
	double prjdvx, prjdvy;	// 13ʸֱ12ϵͶӰʸ
	double vvx, vvy;	// 312Ĵʸ
	double vvl;			// ߳
	double factor;			// 

	dvx12 = x2 - x1;		dvy12 = y2 - y1; // 12ʸ
	dvx34 = x4 - x3;		dvy34 = y4 - y3; // 34ʸ

	// ʸ()Ϊ0,ƽ޽
	if (DOUBLE_EQU_ZERO(dvx12 * dvy34 - dvx34 * dvy12))
		return false;

	len12 = sqrt(dvx12 * dvx12 + dvy12 * dvy12); // 12
	len34 = sqrt(dvx34 * dvx34 + dvy34 * dvy34); // 34

	if (DOUBLE_EQU_ZERO(len12) || DOUBLE_EQU_ZERO(len34))
		return false;

	// 3ֱ12,Ϊ (x3,y3)
	if (DOUBLE_EQU_ZERO((x3 - x1) * (y2 - y3) - (x2 - x3) * (y3 - y1)))
	{
		x5 = x3;
		y5 = y3;
	}
	else
	{
		uvx12 = dvx12 / len12;	uvy12 = dvy12 / len12; // 12λʸ
		uvx34 = dvx34 / len34;	uvy34 = dvy34 / len34; // 34λʸ
		dvx13 = x3 - x1;	dvy13 = y3 - y1; // 13ʸ
		prj = dvx13 * uvx12 + dvy13 * uvy12; // 1312ϵͶӰ
		prjdvx = prj * uvx12;	prjdvy = prj * uvy12; // 1312ϵͶӰʸ
		vvx = prjdvx - dvx13;	vvy = prjdvy - dvy13; // 312Ĵʸ
		vvl = sqrt(vvx * vvx + vvy * vvy); // 312Ĵ߳
		double nTemp = D2GetProjectionDistance(dvx34, dvy34, vvx, vvy);
		if (DOUBLE_EQU_ZERO(nTemp))
			return 0;
		factor = vvl / (nTemp + DIVIDE_BY_ZERO_GUARD); // ȱ

		x5 = x3 + dvx34 * factor;
		y5 = y3 + dvy34 * factor;
	}

	ASSERT(pptIntersect_ != NULL);
	pptIntersect_->x = x5;
	pptIntersect_->y = y5;
	return true;
}

namespace
{
	VectorD2D NormalizeVector(const VectorD2D& vecValue_)
	{
		double _nLength = vecValue_.GetLength();
		ASSERT(DOUBLE_GREAT_ZERO(_nLength));
		return VectorD2D(vecValue_.x / _nLength, vecValue_.y / _nLength);
	}
}

int ncmath::Intersect_BeelineWithBeeline(IN const DPOINT2& pt1_,
										 IN double nAngle1_,
										 IN const DPOINT2& pt2_,
										 IN double nAngle2_,
										 OUT DPOINT2* pptIntersect_ /* = NULL */)
{
	ASSERT(IsValidDouble(pt1_.x));
	ASSERT(IsValidDouble(pt1_.y));
	ASSERT(IsValidDouble(pt2_.x));
	ASSERT(IsValidDouble(pt2_.y));

	// һȡx = pt1.x + 1 ---  2010 09 07
	__point2<double> _pt11(pt1_.x, pt1_.y);
	__point2<double> _pt12(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);

	__point2<double> _pt21(pt2_.x, pt2_.y);
	__point2<double> _pt22(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);

	if (DOUBLE_EQU(nAngle1_, c_nPIE / 2) || DOUBLE_EQU(nAngle1_, 3 * c_nPIE / 2))
	{
		_pt12 .Set(pt1_.x, pt1_.y + 1);
	}
	else
	{
		_pt12.Set(pt1_.x + 1, tan(nAngle1_) + pt1_.y);
	}

	if (DOUBLE_EQU(nAngle2_, c_nPIE / 2) || DOUBLE_EQU(nAngle2_, 3 * c_nPIE / 2))
	{
		_pt22.Set(pt2_.x, pt2_.y + 1);
	}
	else
	{
		_pt22.Set(pt2_.x + 1, tan(nAngle2_) + pt2_.y);
	}

	// ԭ㵽ֱߵĶ˵ ---wanghong 2010 09 07
	VectorD2D _vec1x(_pt11), _vec1y(_pt12);
	VectorD2D _vec2x(_pt21), _vec2y(_pt22);
	// ֱߵ ---wanghong 2010 09 07
	VectorD2D _vecA(_vec1y - _vec1x), _vecB(_vec2y - _vec2x);
	// ڶֱߵĴֱ ---wanghong 2010 09 07
	VectorD2D _vecBPlumb(-_vecB.y, _vecB.x);

	if (DOUBLE_EQU_ZERO(NormalizeVector(_vecA) * NormalizeVector(_vecBPlumb)))
	{
		// ƽлغ ---wanghong 2010 09 07
		return 0;
	}

	double _n = (_vec2x * _vecBPlumb - _vec1x * _vecBPlumb) / (_vecA * _vecBPlumb);
	if (pptIntersect_ != NULL)
	{
		pptIntersect_->x = _vec1x.x + _n * _vecA.x;
		pptIntersect_->y = _vec1x.y + _n * _vecA.y;
	}

	return 1;
}

int ncmath::Intersect_BeelineWithBeeline(IN const DPOINT2& pt1_,
										 IN double nAngle1_,
										 IN const DPOINT2& pt2_,
										 IN double nAngle2_,
										 OUT std::vector<DPOINT2>& vectorIntersect_,
										 IN int nIntersectsNeed_ /* = 1 */)
{
	ASSERT(IsValidDouble(pt1_.x));
	ASSERT(IsValidDouble(pt1_.y));
	ASSERT(IsValidDouble(pt2_.x));
	ASSERT(IsValidDouble(pt2_.y));

	if (nIntersectsNeed_ == 0)
	{
		double _nCompare1 = c_nINVALID_DOUBLE;
		if (DOUBLE_EQU(nAngle1_, c_nPIE / 2) || DOUBLE_EQU(nAngle1_, 3 * c_nPIE / 2))
		{
			// бΪ90ʱȡxֵΪȽ ---  2010 09 19
			_nCompare1 = pt1_.x;
		}
		else
		{
			// xΪʱֱӦyֵ ---  2010 09 06
			_nCompare1 = pt1_.y - tan(nAngle1_) * pt1_.x;
		}

		double _nCompare2 = c_nINVALID_DOUBLE;
		if (DOUBLE_EQU(nAngle2_, c_nPIE / 2) || DOUBLE_EQU(nAngle2_, 3 * c_nPIE / 2))
		{
			_nCompare2 = pt2_.x;
		}
		else
		{
			_nCompare2 = pt2_.y - tan(nAngle2_) * pt2_.x;
		}

		// غϣ ---  2010 09 06
		if (DOUBLE_EQU(nAngle1_, nAngle2_) && DOUBLE_EQU(_nCompare1, _nCompare2))
		{
			return -1;
		}
		// ƽУ0 ---  2010 09 06
		else if (DOUBLE_EQU(nAngle1_, nAngle2_) && !DOUBLE_EQU(_nCompare1, _nCompare2))
		{
			return 0;
		}
		// ཻ1 ---  2010 09 06
		else
		{
			return 1;
		}
	}

	DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	if (!Intersect_BeelineWithBeeline(pt1_, nAngle1_, pt2_, nAngle2_, &_ptIntersect))
		return 0;

	vectorIntersect_.push_back(_ptIntersect);
	return 1;
}

int ncmath::Intersect_BeelineWithLine(IN const DPOINT2& ptBeeline_,
									  IN double nBeelineAngle_,
									  IN const DPOINT2& ptLineStart_,
									  IN const DPOINT2& ptLineEnd_,
									  OUT DPOINT2* pptIntersect_ /* = NULL */)
{
	ASSERT(IsValidDouble(ptBeeline_.x));
	ASSERT(IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptLineStart_.x));
	ASSERT(IsValidDouble(ptLineStart_.y));
	ASSERT(IsValidDouble(ptLineEnd_.x));
	ASSERT(IsValidDouble(ptLineEnd_.y));

	DPOINT2 _vecLine(ptLineEnd_ - ptLineStart_);
	ASSERT(DOUBLE_GREAT_ZERO(_vecLine.GetLength()));

	double _nAngleLine = D2GetNormalAngle(_vecLine);
	DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);

	// Ӧֱֱ߽ ---  2010 09 06
	int _nIntersect = Intersect_BeelineWithBeeline(ptBeeline_,
		nBeelineAngle_, ptLineStart_, _nAngleLine, &_ptIntersect);

	// ޽ֱӷ0 ---  2010 09 06
	if (_nIntersect == 0)
		return 0;

	// 㲻߶ϣ0 ---  2010 09 06
	if (!IsPointInLine(_ptIntersect, ptLineStart_, ptLineEnd_))
		return 0;

	if (pptIntersect_ != NULL)
		*pptIntersect_ = _ptIntersect;

	return 1;
}

int ncmath::Intersect_BeelineWithLine(IN const DPOINT2& ptBeeline_,
									  IN double nBeelineAngle_,
									  IN const DPOINT2& ptLineStart_,
									  IN const DPOINT2& ptLineEnd_,
									  OUT std::vector<DPOINT2>& vectorIntersect_,
									  IN int nIntersectsNeed_ /* = 1 */)
{
	ASSERT(IsValidDouble(ptBeeline_.x));
	ASSERT(IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptLineStart_.x));
	ASSERT(IsValidDouble(ptLineStart_.y));
	ASSERT(IsValidDouble(ptLineEnd_.x));
	ASSERT(IsValidDouble(ptLineEnd_.y));

	DPOINT2 _vecLine(ptLineEnd_ - ptLineStart_);
	ASSERT(DOUBLE_GREAT_ZERO(_vecLine.GetLength()));
	double _nAngleLine = D2GetNormalAngle(_vecLine);

	if (nIntersectsNeed_ == 0)
	{
		double _nNum = Intersect_BeelineWithBeeline(ptBeeline_,
			nBeelineAngle_, ptLineStart_, _nAngleLine, vectorIntersect_, 0);

		// ֱ߼߶ֱн㣬жϽǷ߶ ---  2010 09 06
		if (_nNum == 1)
		{
			DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
			return Intersect_BeelineWithLine(ptBeeline_,nBeelineAngle_,
								ptLineStart_, ptLineEnd_, &_ptIntersect);
		}

		// ޽ ---  2010 09 06
		return 0;
	}

	DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	if (!Intersect_BeelineWithLine(ptBeeline_,nBeelineAngle_,
					ptLineStart_, ptLineEnd_, &_ptIntersect))
	{
		return 0;
	}

	vectorIntersect_.push_back(_ptIntersect);
	return 1;
}

// úҪv9Уֻ޸ ---  2010 09 07
int ncmath::Intersect_BeelineWithCircle(IN const DPOINT2& ptBeeline_,
										IN double nBeelineAngle_,
										IN const DPOINT2& ptCenter_,
										IN double nRadius_,
										OUT DPOINT2* pptIntersect1_ /* = NULL */,
										OUT DPOINT2* pptIntersect2_ /* = NULL */)
{
	ASSERT(IsValidDouble(ptBeeline_.x));
	ASSERT(IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x));
	ASSERT(IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));

	double x1, y1, x2, y2;		// ֱ12
	double x3, y3;				// Բĵֱ12ͶӰ;
	double _nCenterToLineDis;	// Բĵֱ12ľ
	double _nPrjToIntersectDis;	// ͶӰ㵽ľ
	double _nLineLength;		// ָ߶εĳ
	double dvx1c, dvy1c;		// ӵ1Բĵֱߵķʸ
	double prjc;				// ӵ1Բĵֱ12ϵͶӰ

	// Ϊֱ˵㸳ֵ
	x1 = ptBeeline_.x;
	y1 = ptBeeline_.y;
	x2 = c_nINVALID_DOUBLE;
	y2 = c_nINVALID_DOUBLE;

	if (DOUBLE_EQU(nBeelineAngle_, c_nPIE / 2) ||
		DOUBLE_EQU(nBeelineAngle_, 3 * c_nPIE / 2))
	{
		x2 = ptBeeline_.x;
		y2 = ptBeeline_.y + 1;
	}
	else
	{
		x2 = ptBeeline_.x + 1;
		y2 = ptBeeline_.y + tan(nBeelineAngle_);
	}

	DPOINT2 _ptBeelineEnd(DPOINT2(x2, y2));
	_nLineLength = (_ptBeelineEnd - ptBeeline_).GetLength();
	DPOINT2 _vecLine(_ptBeelineEnd - ptBeeline_);

	if (DOUBLE_EQU_ZERO(nRadius_))
		return -1;

	// Բĵֱߵľڰ뾶ʱ,ֱԲûн
	_nCenterToLineDis = D2GetPointToLineDistance(ptCenter_, ptBeeline_, _ptBeelineEnd);
	if (DOUBLE_GREAT(_nCenterToLineDis, nRadius_))
		return 0;

	dvx1c = ptCenter_.x - x1;
	dvy1c = ptCenter_.y - y1;
	prjc = D2GetProjectionDistance(dvx1c, dvy1c, _vecLine.x, _vecLine.y);

	_vecLine.Unify();
	x3 = x1 + prjc * _vecLine.x;
	y3 = y1 + prjc * _vecLine.y;

	double nTemp = nRadius_ * nRadius_ - _nCenterToLineDis * _nCenterToLineDis;
	if (nTemp < 0)
		nTemp = 0;
	_nPrjToIntersectDis = sqrt(nTemp);

	DPOINT2 _ptIntersect1(x3 - _nPrjToIntersectDis * _vecLine.x, y3 - _nPrjToIntersectDis * _vecLine.y);
	DPOINT2 _ptIntersect2(x3 + _nPrjToIntersectDis * _vecLine.x, y3 + _nPrjToIntersectDis * _vecLine.y);

	//  ---  2010 09 07
	if (pptIntersect1_ != NULL && DOUBLE_EQU_ZERO((_ptIntersect1 - _ptIntersect2).GetLength()))
	{
		*pptIntersect1_ = _ptIntersect1;
		return 1;
	}

	if (pptIntersect1_ != NULL)
		*pptIntersect1_ = _ptIntersect1;

	if (pptIntersect2_ != NULL)
		*pptIntersect2_ = _ptIntersect2;

	return 2;
}

int ncmath::Intersect_BeelineWithCircle(IN const DPOINT2& ptBeeline_,
										IN double nBeelineAngle_,
										IN const DPOINT2& ptCenter_,
										IN double nRadius_,
										OUT std::vector<DPOINT2>& vectorIntersect_,
										IN int nIntersectsNeed_ /* = 2 */)
{
	ASSERT(IsValidDouble(ptBeeline_.x));
	ASSERT(IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x));
	ASSERT(IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));

	DPOINT2 _ptBeelineEnd(DPOINT2(ptBeeline_.x, ptBeeline_.y + 1));
	if (!DOUBLE_EQU(nBeelineAngle_, c_nPIE / 2) && !DOUBLE_EQU(nBeelineAngle_, 3 * c_nPIE / 2))
		_ptBeelineEnd = DPOINT2(ptBeeline_.x + 1, ptBeeline_.y + tan(nBeelineAngle_));

	if (nIntersectsNeed_ == 0)
	{
		double _nDis = D2GetPointToLineDistance(ptCenter_, ptBeeline_, _ptBeelineEnd);

		if (DOUBLE_EQU(_nDis, nRadius_))
		{
			return 1;
		}
		else if (DOUBLE_GREAT(_nDis, nRadius_))
		{
			return 0;
		}
		else
		{
			return 2;
		}
	}

	DPOINT2 _ptIntersect1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersect2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	double _nNum = Intersect_BeelineWithCircle(ptBeeline_, nBeelineAngle_,
					 ptCenter_, nRadius_, &_ptIntersect1, &_ptIntersect2);

	if (_nNum == 1)
	{
		vectorIntersect_.push_back(_ptIntersect1);
		return 1;
	}
	else if (_nNum == 2)
	{
		vectorIntersect_.push_back(_ptIntersect1);
		vectorIntersect_.push_back(_ptIntersect2);
		return 2;
	}
	else
	{
		return 0;
	}
}

int ncmath::Intersect_BeelineWithArc(IN const DPOINT2& ptBeeline_,
									 IN double nBeelineAngle_,
									 IN const DPOINT2& ptCenter_,
									 IN double nRadius_,
									 IN double nStartAngle_,
									 IN double nSweptAngle_,
									 OUT DPOINT2* pptIntersect1_ /* = NULL */,
									 OUT DPOINT2* pptIntersect2_ /* = NULL */)
{
	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * c_nPIE));

	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ֱͨԲӦԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 08
	int _nImageIntersect = Intersect_BeelineWithCircle(ptBeeline_, nBeelineAngle_, ptCenter_, nRadius_, &_ptIntersec1, &_ptIntersec2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist = IsPointInArc(_ptIntersec1, ptCenter_, nRadius_, nStartAngle_,
								fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{	// ʱظ --- wanghong 2010 09 08
		_bPt2Exist = IsPointInArc(_ptIntersec2, ptCenter_, nRadius_, nStartAngle_,
								fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	}
	//  --- wanghong 2010 09 08
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}
	// һ --- wanghong 2010 09 08
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_LineWithCircle(IN const DPOINT2& ptStart_,
									 IN const DPOINT2& ptEnd_,
									 IN const DPOINT2& ptCenter_,
									 IN double nRadius_,
									 OUT DPOINT2* pptIntersect1_/* = NULL*/,
									 OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptStart_.x));
	ASSERT(IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x));
	ASSERT(IsValidDouble(ptCenter_.x));
	ASSERT(IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));

	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ͨ߶ڵֱԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 08
	int _nImageIntersect = Intersect_BeelineWithCircle(ptStart_, ncmath::D2GetNormalAngle(ptEnd_ - ptStart_),
												ptCenter_, nRadius_, &_ptIntersec1, &_ptIntersec2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist = IsPointInLine(_ptIntersec1, ptStart_, ptEnd_);
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{	// ʱظ --- wanghong 2010 09 08
		_bPt2Exist = IsPointInLine(_ptIntersec2, ptStart_, ptEnd_);
	}
	//  --- wanghong 2010 09 08
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}
	// һ --- wanghong 2010 09 08
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_LineWithCircle(IN const DPOINT2& ptStart_,
									 IN const DPOINT2& ptEnd_,
									 IN const DPOINT2& ptCenter_,
									 IN double nRadius_,
									 OUT std::vector<DPOINT2>& vectorIntersect_,
									 IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptStart_.x));
	ASSERT(IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x));
	ASSERT(IsValidDouble(ptCenter_.x));
	ASSERT(IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_LineWithCircle(ptStart_, ptEnd_, ptCenter_, nRadius_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 08
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

int ncmath::Intersect_BeelineWithArc(IN const DPOINT2& ptBeeline_,
									 IN double nBeelineAngle_,
									 IN const DPOINT2& ptCenter_,
									 IN double nRadius_,
									 IN double nStartAngle_,
									 IN double nSweptAngle_,
									 OUT std::vector<DPOINT2>& vectorIntersect_,
									 IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * c_nPIE));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_BeelineWithArc(ptBeeline_, nBeelineAngle_, ptCenter_, nRadius_, 
											   nStartAngle_, nSweptAngle_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 08
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;
	
	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

int ncmath::Intersect_LineWithArc(IN const DPOINT2& ptStart_,
								  IN const DPOINT2& ptEnd_,
								  IN const DPOINT2& ptCenter_,
								  IN double nRadius_,
								  IN double nStartAngle_,
								  IN double nSweptAngle_,
								  OUT DPOINT2* pptIntersect1_/* = NULL*/,
								  OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ͨ߶ζӦֱԲӦԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 08
	int _nImageIntersect = Intersect_BeelineWithCircle(ptStart_, ncmath::D2GetNormalAngle(ptEnd_ - ptStart_),
												ptCenter_, nRadius_, &_ptIntersec1, &_ptIntersec2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist = IsPointInLine(_ptIntersec1, ptStart_, ptEnd_) &&
		IsPointInArc(_ptIntersec1, ptCenter_, nRadius_, nStartAngle_, fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{	// ʱظ --- wanghong 2010 09 08
		_bPt2Exist = IsPointInLine(_ptIntersec2, ptStart_, ptEnd_) &&
		IsPointInArc(_ptIntersec2, ptCenter_, nRadius_, nStartAngle_, fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	}
	//  --- wanghong 2010 09 08
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}
	// һ --- wanghong 2010 09 08
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_LineWithArc(IN const DPOINT2& ptStart_,
								  IN const DPOINT2& ptEnd_,
								  IN const DPOINT2& ptCenter_,
								  IN double nRadius_, 
								  IN double nStartAngle_,
								  IN double nSweptAngle_,
								  OUT std::vector<DPOINT2>& vectorIntersect_,
								  IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_LineWithArc(ptStart_, ptEnd_, ptCenter_, nRadius_, 
											nStartAngle_, nSweptAngle_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 08
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

int ncmath::Intersect_LineWithLine(IN const DPOINT2& ptStart1_,
								   IN const DPOINT2& ptEnd1_,
								   IN const DPOINT2& ptStart2_,
								   IN const DPOINT2& ptEnd2_,
								   OUT DPOINT2* pptIntersect_ /* = NULL */)
{
	DPOINT2 _vecLine1(ptEnd1_ - ptStart1_);
	DPOINT2 _vecLine2(ptEnd2_ - ptStart2_);

	ASSERT(DOUBLE_GREAT_ZERO(_vecLine1.GetLength()));
	ASSERT(DOUBLE_GREAT_ZERO(_vecLine2.GetLength()));

	double _nAngle1 = D2GetNormalAngle(_vecLine1);
	double _nAngle2 = D2GetNormalAngle(_vecLine2);

	DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	int _nNum = Intersect_BeelineWithBeeline(ptStart1_, _nAngle1,
							 ptStart2_, _nAngle2, &_ptIntersect);

	if (_nNum == 0 || !IsPointInLine(_ptIntersect, ptStart1_, ptEnd1_) ||
					  !IsPointInLine(_ptIntersect, ptStart2_, ptEnd2_))
	{
		return 0;
	}

	if (pptIntersect_)
		*pptIntersect_ = _ptIntersect;

	return 1;
}

int ncmath::Intersect_LineWithLine(IN const DPOINT2& ptStart1_,
								   IN const DPOINT2& ptEnd1_,
								   IN const DPOINT2& ptStart2_,
								   IN const DPOINT2& ptEnd2_,
								   OUT std::vector<DPOINT2>& vectorIntersect_,
								   IN int nIntersectsNeed_ /* = 1 */)
{
	DPOINT2 _vecLine1(ptEnd1_ - ptStart1_);
	DPOINT2 _vecLine2(ptEnd2_ - ptStart2_);

	ASSERT(DOUBLE_GREAT_ZERO(_vecLine1.GetLength()));
	ASSERT(DOUBLE_GREAT_ZERO(_vecLine2.GetLength()));

	if (nIntersectsNeed_ == 0)
	{
		// ų ---  2010 09 08
		LineD2D _Line1;
		_Line1.pt1 = ptStart1_;
		_Line1.pt2 = ptEnd1_;
		LineD2D _Line2;
		_Line2.pt1 = ptStart2_;
		_Line2.pt2 = ptEnd2_;

		RealRectD _Rect1 = _Line1.GetBoundRect();
		RealRectD _Rect2 = _Line2.GetBoundRect();

		if (!_Rect1.Intersect(_Rect2))
			return 0;

		//  
		// жP1P2Q1Q2ǣ
		// ( P1 - Q1 )  ( Q2 - Q1 ) * ( Q2 - Q1 )  ( P2 - Q1 ) >= 0
		// жQ1Q2P1P2ǣ
		// ( Q1 - P1 )  ( P2 - P1 ) * ( P2 - P1 )  ( Q2 - P1 ) >= 0
		//                                        ---  2010 09 08
		VectorD2D _vecQ1P1 = ptStart1_ - ptStart2_;
		VectorD2D _vecQ1Q2 = ptEnd2_ - ptStart2_;
		VectorD2D _vecQ1P2 = ptEnd1_ - ptStart2_;

		VectorD2D _vecP1Q1 = ptStart2_ - ptStart1_;
		VectorD2D _vecP1P2 = ptEnd1_ - ptStart1_;
		VectorD2D _vecP1Q2 = ptEnd2_ - ptStart1_;

		if (D2CrossMultiply(_vecQ1P1, _vecQ1Q2) * D2CrossMultiply(_vecQ1Q2, _vecQ1P2) < 0 ||
			D2CrossMultiply(_vecP1Q1, _vecP1P2) * D2CrossMultiply(_vecP1P2, _vecP1Q2) < 0)
			return 0;

		return 1;
	}

	DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	if (Intersect_LineWithLine(ptStart1_, ptEnd1_, ptStart2_, ptEnd2_, &_ptIntersect) == 0)
		return 0;

	vectorIntersect_.push_back(_ptIntersect);
	return 1;
}

// v9޸ ---  2010 09 08
int ncmath::Intersect_CircleWithCircle(IN const DPOINT2& ptCenter1_,
									   IN double nRadius1_,
									   IN const DPOINT2& ptCenter2_,
									   IN double nRadius2_,
									   OUT DPOINT2* pptIntersect1_ /* = NULL */,
									   OUT DPOINT2* pptIntersect2_ /* = NULL */)
{
	ASSERT(IsValidDouble(ptCenter1_.x));
	ASSERT(IsValidDouble(ptCenter1_.y));
	ASSERT(IsValidDouble(ptCenter2_.x));
	ASSERT(IsValidDouble(ptCenter2_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius1_));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius2_));

	double x3, y3;				// 12c1c2
	double _nCentersDis;		// ԲԲ֮ľ
	double d1;					// 1Բߵľ
	double l1;					// 
	double _nAngle1c1c2;		// 1,Բ1,Բ2ɵ,1c1c2

	// Բľ ---  2010 09 08
	_nCentersDis = D2GetDistance(ptCenter1_, ptCenter2_);

	// 룬ຬʱΪ ---  2010 09 08
	if (DOUBLE_GREAT(_nCentersDis, (nRadius1_ + nRadius2_)) ||
		DOUBLE_LESS(_nCentersDis, fabs(nRadius1_ - nRadius2_)))
	{
		return 0;
	}

	if (DOUBLE_EQU_ZERO(_nCentersDis) && DOUBLE_EQU(nRadius1_, nRadius2_))
		return -1;

	double nTemp = (nRadius1_ * nRadius1_ + _nCentersDis * _nCentersDis - nRadius2_ * nRadius2_) / (2 * nRadius1_ * _nCentersDis);
	__CUT(nTemp, -1, 1);
	_nAngle1c1c2 = acos(nTemp);
	d1 = nRadius1_ * cos(_nAngle1c1c2);
	l1 = nRadius1_ * sin(_nAngle1c1c2);

	// Բߵķʸ ---  2010 09 08
	DPOINT2 _vecCenters(ptCenter2_ - ptCenter1_);
	_vecCenters.Unify();

	// 1ԲϵͶӰ
	x3 = ptCenter1_.x + d1 * _vecCenters.x;
	y3 = ptCenter1_.y + d1 * _vecCenters.y;

	// λʸ31
	double uvx31 = -_vecCenters.y;
	double uvy31 = _vecCenters.x;

	// 
	DPOINT2 _ptIntersect1(x3 + l1 * uvx31, y3 + l1 * uvy31);
	DPOINT2 _ptIntersect2(x3 - l1 * uvx31, y3 - l1 * uvy31);

	//  ---  2010 09 08
	if (pptIntersect1_ && DOUBLE_EQU_ZERO((_ptIntersect1 - _ptIntersect2).GetLength()))
	{
		*pptIntersect1_ = _ptIntersect1;
		return 1;
	}

	if (pptIntersect1_)
		*pptIntersect1_ = _ptIntersect1;

	if (pptIntersect2_)
		*pptIntersect2_ = _ptIntersect2;

	return 2;
}

int ncmath::Intersect_CircleWithCircle(IN const DPOINT2& ptCenter1_,
									   IN double nRadius1_,
									   IN const DPOINT2& ptCenter2_,
									   IN double nRadius2_,
									   OUT std::vector<DPOINT2>& vectorIntersect_,
									   IN int nIntersectsNeed_ /* = 2 */)
{
	ASSERT(IsValidDouble(ptCenter1_.x));
	ASSERT(IsValidDouble(ptCenter1_.y));
	ASSERT(IsValidDouble(ptCenter2_.x));
	ASSERT(IsValidDouble(ptCenter2_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius1_));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius2_));

	if (nIntersectsNeed_ == 0)
	{
		// Բľ ---  2010 09 08
		double _nCentersDis = D2GetDistance(ptCenter1_, ptCenter2_);

		// 룬ຬʱΪ ---  2010 09 08
		if (DOUBLE_GREAT(_nCentersDis, (nRadius1_ + nRadius2_)) ||
			DOUBLE_LESS(_nCentersDis, fabs(nRadius1_ - nRadius2_)))
		{
			return 0;
		}
		// غ ---  2010 09 08
		else if (DOUBLE_EQU_ZERO(_nCentersDis) && DOUBLE_EQU(nRadius1_, nRadius2_))
		{
			return -1;
		}
		//  ---  2010 09 08
		else if (DOUBLE_EQU(_nCentersDis, nRadius1_) ||
				 DOUBLE_EQU(_nCentersDis, nRadius2_) ||
				 DOUBLE_EQU(_nCentersDis, nRadius1_ + nRadius2_))
		{
			return 1;
		}
		else
		{
			return 2;
		}
	}

	DPOINT2 _ptIntersect1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersect2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	int _nNum = Intersect_CircleWithCircle(ptCenter1_, nRadius1_,ptCenter2_,
								 nRadius2_, &_ptIntersect1, &_ptIntersect2);

	switch(_nNum)
	{
	case 2:
		{
			vectorIntersect_.push_back(_ptIntersect1);
			vectorIntersect_.push_back(_ptIntersect2);
		}
		break;

	case 1:
		{
			vectorIntersect_.push_back(_ptIntersect1);
		}
		break;

	default:
		break;
	}

	return _nNum;
}

int ncmath::Intersect_CircleWithArc(IN const DPOINT2& ptCircleCenter_,
									IN double nCircleRadius_,
									IN const DPOINT2& ptArcCenter_,
									IN double nArcRadius_,
									IN double nStartAngle_,
									IN double nSweptAngle_,
									OUT DPOINT2* pptIntersect1_ /* = NULL */,
									OUT DPOINT2* pptIntersect2_ /* = NULL */)
{
	ASSERT(IsValidDouble(ptCircleCenter_.x));
	ASSERT(IsValidDouble(ptCircleCenter_.y));
	ASSERT(IsValidDouble(ptArcCenter_.x));
	ASSERT(IsValidDouble(ptArcCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nCircleRadius_));
	ASSERT(DOUBLE_GREAT_ZERO(nArcRadius_));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) &&
		   DOUBLE_LESS(fabs(nSweptAngle_), 2 * c_nPIE));

	// ԲԲԲĽ ---  2010 09 08
	DPOINT2 _ptIntersect1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersect2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	int _nNum = Intersect_CircleWithCircle(ptCircleCenter_, nCircleRadius_,
				ptArcCenter_, nArcRadius_, &_ptIntersect1, &_ptIntersect2);

	switch(_nNum)
	{
	case 1:
		if (IsPointInArc(_ptIntersect1, ptArcCenter_, nArcRadius_,
			nStartAngle_, fabs(nSweptAngle_), nSweptAngle_ > 0))
		{
			*pptIntersect1_ = _ptIntersect1;
		}
		else
		{
			--_nNum;
		}

		break;

	case 2:
		{
			// жϽǷԲ ---  2010 09 08
			bool _bPt1InArc = IsPointInArc(_ptIntersect1, ptArcCenter_, nArcRadius_,
				nStartAngle_, fabs(nSweptAngle_), nSweptAngle_ > 0);

			bool _bPt2InArc = (IsPointInArc(_ptIntersect2, ptArcCenter_, nArcRadius_,
				nStartAngle_, fabs(nSweptAngle_), nSweptAngle_ > 0));

			if (_bPt1InArc && _bPt2InArc)
			{
				if (pptIntersect1_)
					*pptIntersect1_ = _ptIntersect1;
				if (pptIntersect2_)
					*pptIntersect2_ = _ptIntersect2;
			}
			else if (pptIntersect1_ && _bPt1InArc && !_bPt2InArc)
			{
				*pptIntersect1_ = _ptIntersect1;
				--_nNum;
			}
			else if (pptIntersect1_ && !_bPt1InArc && _bPt2InArc)
			{
				*pptIntersect1_ = _ptIntersect2;
				--_nNum;
			}
			else
			{
				_nNum = 0;
			}
		}
		break;

	case 0:
	default:
		_nNum = 0;
	}

	return _nNum;
}

namespace
{
	// жϸǶǷ(nLeft_, nRight_)ڣעnLeft_ܴnRight_ --- wanghong 2010 09 09
	inline bool InRange(double nLeft_, double nAngle_, double nRight_)
	{
		if (nLeft_ < nRight_)
			return nLeft_ < nAngle_ && nAngle_ < nRight_;

		return (nLeft_ < nAngle_ && nAngle_ < 2 * ncmath::c_nPIE) ||
			(0 < nAngle_ && nAngle_ < nRight_);
	}
	// ԸʼǺĽǣõʱеʼǺֹ --- wanghong 2010 09 09
	void GetLegalAngle(double& nStartAngle_, double& nEndAngle_, double nSweepAngle_)
	{
		if (DOUBLE_GREAT_ZERO(nSweepAngle_))
		{
			nEndAngle_ = nStartAngle_ + nSweepAngle_;
			if (DOUBLE_GE(nEndAngle_, 2 * ncmath::c_nPIE))
				nEndAngle_ -= 2 * ncmath::c_nPIE;
		}
		else
		{
			nEndAngle_ = nStartAngle_;
			nStartAngle_ = nStartAngle_ + nSweepAngle_;
			if (DOUBLE_LESS_ZERO(nStartAngle_))
				nStartAngle_ += 2 * ncmath::c_nPIE;
		}
	}
}

int ncmath::Intersect_CircleWithArc(IN const DPOINT2& ptCircleCenter_,
									IN double nCircleRadius_,
									IN const DPOINT2& ptArcCenter_,
									IN double nArcRadius_,
									IN double nStartAngle_,
									IN double nSweptAngle_,
									OUT std::vector<DPOINT2>& vectorIntersect_,
									IN int nIntersectsNeed_ /* = 2 */)
{
	ASSERT(IsValidDouble(ptCircleCenter_.x));
	ASSERT(IsValidDouble(ptCircleCenter_.y));
	ASSERT(IsValidDouble(ptArcCenter_.x));
	ASSERT(IsValidDouble(ptArcCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nCircleRadius_));
	ASSERT(DOUBLE_GREAT_ZERO(nArcRadius_));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) &&
		DOUBLE_LESS(fabs(nSweptAngle_), 2 * c_nPIE));

	DPOINT2 _ptIntersect1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersect2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	int _nNum = Intersect_CircleWithCircle(ptCircleCenter_, nCircleRadius_,
				ptArcCenter_, nArcRadius_, &_ptIntersect1, &_ptIntersect2);

	switch(_nNum)
	{
	case 2:
		{
			if (IsPointInArc(_ptIntersect1, ptArcCenter_, nArcRadius_,
				nStartAngle_, fabs(nSweptAngle_), nSweptAngle_ > 0))
			{
				vectorIntersect_.push_back(_ptIntersect1);
			}
			else
			{
				--_nNum;
			}

			if (IsPointInArc(_ptIntersect2, ptArcCenter_, nArcRadius_,
				nStartAngle_, fabs(nSweptAngle_), nSweptAngle_ > 0))
			{
				vectorIntersect_.push_back(_ptIntersect2);
			}
			else
			{
				--_nNum;
			}
		}
		break;

	case 1:
		{
			if (IsPointInArc(_ptIntersect1, ptArcCenter_, nArcRadius_,
				nStartAngle_, fabs(nSweptAngle_), nSweptAngle_ > 0))
			{
				vectorIntersect_.push_back(_ptIntersect1);
			}
			else
			{
				--_nNum;
			}
		}
		break;

	case -1:
		{
			vectorIntersect_.push_back(GetPointInCircle(ptArcCenter_, nArcRadius_, nStartAngle_));
			double _nEndAngle = RegularizeRadian(nStartAngle_ + nSweptAngle_);
			vectorIntersect_.push_back(GetPointInCircle(ptArcCenter_, nArcRadius_, _nEndAngle));
		}

	default:
		break;
	}

	return _nNum;
}

int ncmath::Intersect_CircleWithEllipse(IN const DPOINT2& ptCircleCenter_,
										IN double nRadius_,
										IN const DPOINT2& ptEllipseCenter_,
										IN double nURadius_,
										IN double nVRadius_,
										IN double nUAngle_,
										OUT std::vector<DPOINT2>& vectorIntersect_,
										IN int nIntersectsNeed_ /* = 4 */)
{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	// 潻
	DPOINT2 _ptIntersects[4];
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C
	int _nRet = Intersect_CircleWithEllipse(
				ptCircleCenter_, nRadius_, ptEllipseCenter_, nURadius_, nVRadius_, nUAngle_,
				_nNeed >= 1 ? &_ptIntersects[0] : NULL,
				_nNeed >= 2 ? &_ptIntersects[1] : NULL,
				_nNeed >= 3 ? &_ptIntersects[2] : NULL,
				_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}

int ncmath::Intersect_CircleWithEllipse(IN const DPOINT2& ptCircleCenter_,
										IN double nRadius_,
										IN const DPOINT2& ptEllipseCenter_,
										IN double nURadius_,
										IN double nVRadius_,
										IN double nUAngle_,
										OUT DPOINT2* pptIntersect1_/* = NULL */,
										OUT DPOINT2* pptIntersect2_/* = NULL */,
										OUT DPOINT2* pptIntersect3_/* = NULL */,
										OUT DPOINT2* pptIntersect4_/* = NULL */)
{
	ASSERT(IsValidDouble(ptCircleCenter_.x));
	ASSERT(IsValidDouble(ptCircleCenter_.y));
	ASSERT(IsValidDouble(nRadius_) && DOUBLE_GE_ZERO(nRadius_));
	ASSERT(IsValidDouble(ptEllipseCenter_.x));
	ASSERT(IsValidDouble(ptEllipseCenter_.y));
	ASSERT(IsValidDouble(nURadius_) && DOUBLE_GE_ZERO(nURadius_));
	ASSERT(IsValidDouble(nVRadius_) && DOUBLE_GE_ZERO(nVRadius_));
	ASSERT(IsValidDouble(nUAngle_));

	if (DOUBLE_EQU_ZERO(nRadius_) || (DOUBLE_EQU_ZERO(nURadius_) && DOUBLE_EQU_ZERO(nVRadius_)))
		return 0;

	int _nRet = 0;
	// Բ˻ԲµԲԲ󽻵ĺ Wanglou 2010-09-15
	if (DOUBLE_EQU(nURadius_, nVRadius_))
	{
		// ԲԲڲغϵ Wanglou 2010-09-16
		_nRet = Intersect_CircleWithCircle(ptCircleCenter_, nRadius_, ptEllipseCenter_,
				nURadius_, pptIntersect1_, pptIntersect2_);
	}
	// Բ˻ֱ ֱԲ󽻵ĺ Wanglou 2010-09-15
	else if (DOUBLE_EQU_ZERO(nVRadius_))
	{
		// ԲֱҲֻ Wanglou 2010-09-16
		DPOINT2 _ptStart, _ptEnd;
		_ptStart.x = ptEllipseCenter_.x + nURadius_ * cos(nUAngle_);
		_ptStart.y = ptEllipseCenter_.y + nURadius_ * sin(nUAngle_);
		_ptEnd.x = ptEllipseCenter_.x - nURadius_ * cos(nUAngle_);
		_ptEnd.y = ptEllipseCenter_.y - nURadius_ * sin(nUAngle_);
		_nRet = Intersect_LineWithCircle(_ptStart, _ptEnd, ptCircleCenter_, nRadius_,
				pptIntersect1_, pptIntersect2_);
	}
	else
	{
		// ͨµİԲԲ㽻 Wanglou 2010-09-16
		_nRet = Intersect_EllipseWithEllipse(ptEllipseCenter_, nURadius_, nVRadius_, nUAngle_, ptCircleCenter_,
				nRadius_, nRadius_, 0, pptIntersect1_, pptIntersect2_, pptIntersect3_, pptIntersect4_);
	}

	return _nRet;
}

int ncmath::Intersect_CircleWithEllipseArc(IN const DPOINT2& ptCircleCenter_,
										   IN double nRadius_,
										   IN const DPOINT2& ptEllipseCenter_,
										   IN double nURadius_,
										   IN double nVRadius_,
										   IN double nUAngle_,
										   IN double nStartAngle_,
										   IN double nSweptAngle_,
										   OUT std::vector<DPOINT2>& vectorIntersect_,
										   IN int nIntersectsNeed_ /* = 4 */)
{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	// 潻
	DPOINT2 _ptIntersects[4];
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C
	int _nRet = Intersect_CircleWithEllipseArc(
		ptCircleCenter_, nRadius_, ptEllipseCenter_, nURadius_, nVRadius_, nUAngle_, nStartAngle_, nSweptAngle_,
		_nNeed >= 1 ? &_ptIntersects[0] : NULL,
		_nNeed >= 2 ? &_ptIntersects[1] : NULL,
		_nNeed >= 3 ? &_ptIntersects[2] : NULL,
		_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}

int ncmath::Intersect_CircleWithEllipseArc(IN const DPOINT2& ptCircleCenter_,
										   IN double nRadius_,
										   IN const DPOINT2& ptEllipseCenter_,
										   IN double nURadius_,
										   IN double nVRadius_,
										   IN double nUAngle_,
										   IN double nStartAngle_,
										   IN double nSweptAngle_,
										   OUT DPOINT2* pptIntersect1_/* = NULL */,
										   OUT DPOINT2* pptIntersect2_/* = NULL */,
										   OUT DPOINT2* pptIntersect3_/* = NULL */,
										   OUT DPOINT2* pptIntersect4_/* = NULL */)
{
	ASSERT(IsValidDouble(nStartAngle_));
	ASSERT(IsValidDouble(nSweptAngle_));

	// ԲĽΪ0Ϊûн Wanglou 2010-09-16
	if (DOUBLE_EQU_ZERO(nSweptAngle_))
		return 0;

	// ԲԲĵ㲻һԲֻҪҪһԲԲĽ㣬ҪԲԲн
	// Wanglou 2010-09-16
	int _nNum = 0;
	DPOINT2 _ptIntersect[4];

	_nNum = Intersect_CircleWithEllipse(ptCircleCenter_, nRadius_, ptEllipseCenter_, nURadius_, nVRadius_,
			nUAngle_, &_ptIntersect[0], &_ptIntersect[1], &_ptIntersect[2], &_ptIntersect[3]);

	int _nCount = 0;
	for (int _nIndex = 0; _nIndex < _nNum; _nIndex++)
	{
		// ԲϵĽ㱣 Wanglou 2010-09-16
		if (IsPointInEllipseArc(_ptIntersect[_nIndex], ptEllipseCenter_, nURadius_, nVRadius_,
			nUAngle_, nStartAngle_, nSweptAngle_, true))
		{
			_ptIntersect[_nCount] = _ptIntersect[_nIndex];
			_nCount++;
		}
	}

	// ĵΪЧ
	for (int _i = _nCount; _i < 4; _i++)
	{
		_ptIntersect[_i] = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	}

	if (pptIntersect1_)
		*pptIntersect1_ = _ptIntersect[0];
	if (pptIntersect2_)
		*pptIntersect2_ = _ptIntersect[1];
	if (pptIntersect3_)
		*pptIntersect3_ = _ptIntersect[2];
	if (pptIntersect4_)
		*pptIntersect4_ = _ptIntersect[3];

	// غϣ-1
	int _nRet = _nCount;
	if (_nNum == -1)
		_nRet = -1;

	return _nRet;
}

int ncmath::Intersect_ArcWithArc(IN const DPOINT2& ptCenter1_,
								 IN double nRadius1_,
								 IN double nStartAngle1_,
								 IN double nSweptAngle1_,
								 IN const DPOINT2& ptCenter2_,
								 IN double nRadius2_,
								 IN double nStartAngle2_,
								 IN double nSweptAngle2_,
								 OUT DPOINT2* pptIntersect1_/* = NULL*/,
								 OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptCenter1_.x) && IsValidDouble(ptCenter1_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius1_));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle1_)) && DOUBLE_LE(fabs(nSweptAngle1_), 2 * ncmath::c_nPIE));
	ASSERT(IsValidDouble(ptCenter2_.x) && IsValidDouble(ptCenter2_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius2_));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle2_)) && DOUBLE_LE(fabs(nSweptAngle2_), 2 * ncmath::c_nPIE));

	// ԲӦԲغ£㽻 --- wanghong 2010 09 09
	if (ptCenter1_ == ptCenter2_ && DOUBLE_EQU(nRadius1_, nRadius2_))
	{
		double _nLegalStartAngle1 = nStartAngle1_, _nLegalEndAngle1 = c_nINVALID_DOUBLE;
		double _nLegalStartAngle2 = nStartAngle2_, _nLegalEndAngle2 = c_nINVALID_DOUBLE;
		GetLegalAngle(_nLegalStartAngle1, _nLegalEndAngle1, nSweptAngle1_);
		GetLegalAngle(_nLegalStartAngle2, _nLegalEndAngle2, nSweptAngle2_);

		// غϣȫغϣ --- wanghong 2010 09 09
		// ڳԲغϣͬʱһ㡢غĳεĶ˵ --- wanghong 2010 09 10
		bool _bStartInRange = false;
		if ((DOUBLE_EQU(_nLegalStartAngle1, _nLegalStartAngle2) && DOUBLE_EQU(_nLegalEndAngle1, _nLegalEndAngle2)) ||
			(_bStartInRange = InRange(_nLegalStartAngle1, _nLegalStartAngle2, _nLegalEndAngle1)) ||
			InRange(_nLegalStartAngle1, _nLegalEndAngle2, _nLegalEndAngle1))
		{
			if (pptIntersect1_ != NULL)
				*pptIntersect1_ = GetPointInCircle(ptCenter1_, nRadius1_, _bStartInRange ? _nLegalStartAngle2 : _nLegalStartAngle1);
			if (pptIntersect2_ != NULL)
				*pptIntersect2_ = GetPointInCircle(ptCenter1_, nRadius1_, _bStartInRange ? _nLegalEndAngle1 : _nLegalEndAngle2);
			return -1;
		}

		// Բǡûβ --- wanghong 2010 09 09
		if (DOUBLE_EQU(_nLegalStartAngle1, _nLegalEndAngle2) && DOUBLE_EQU(_nLegalEndAngle1, _nLegalStartAngle2))
		{
			if (pptIntersect1_ != NULL)
				*pptIntersect1_ = GetPointInCircle(ptCenter1_, nRadius1_, _nLegalStartAngle1);
			if (pptIntersect2_ != NULL)
				*pptIntersect2_ = GetPointInCircle(ptCenter1_, nRadius1_, _nLegalEndAngle1);
			return 2;
		}
		// Բһ --- wanghong 2010 09 09
		if (DOUBLE_EQU(_nLegalStartAngle1, _nLegalEndAngle2) || DOUBLE_EQU(_nLegalEndAngle1, _nLegalStartAngle2))
		{
			if (pptIntersect1_ != NULL)
			{
				*pptIntersect1_ = GetPointInCircle(ptCenter1_, nRadius1_,
					DOUBLE_EQU(_nLegalStartAngle1, _nLegalEndAngle2) ? _nLegalStartAngle1 : _nLegalEndAngle1);
			}
			return 1;
		}
		// ޽ --- wanghong 2010 09 09
		return 0;
	}

	DPOINT2 _ptIntersect1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersect2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ͨԲ1ӦԲԲ2󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 09
	int _nImageIntersect = Intersect_CircleWithArc(ptCenter1_, nRadius1_,
						ptCenter2_, nRadius2_, nStartAngle2_, nSweptAngle2_, &_ptIntersect1, &_ptIntersect2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist =
		IsPointInArc(_ptIntersect1, ptCenter1_, nRadius1_, nStartAngle1_, fabs(nSweptAngle1_), DOUBLE_GREAT_ZERO(nSweptAngle1_));
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{	// ʱظ --- wanghong 2010 09 09
		_bPt2Exist =
			IsPointInArc(_ptIntersect2, ptCenter1_, nRadius1_, nStartAngle1_, fabs(nSweptAngle1_), DOUBLE_GREAT_ZERO(nSweptAngle1_));
	}
	//  --- wanghong 2010 09 09
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersect1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersect2;
		return 2;
	}
	// һ --- wanghong 2010 09 09
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersect1 : _ptIntersect2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_ArcWithArc(IN const DPOINT2& ptCenter1_,
								 IN double nRadius1_,
								 IN double nStartAngle1_,
								 IN double nSweptAngle1_,
								 IN const DPOINT2& ptCenter2_,
								 IN double nRadius2_,
								 IN double nStartAngle2_,
								 IN double nSweptAngle2_,
								 OUT std::vector<DPOINT2>& vectorIntersect_,
								 IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptCenter1_.x) && IsValidDouble(ptCenter1_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius1_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle1_) && DOUBLE_LESS(nStartAngle1_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle1_)) && DOUBLE_LE(fabs(nSweptAngle1_), 2 * ncmath::c_nPIE));
	ASSERT(IsValidDouble(ptCenter2_.x) && IsValidDouble(ptCenter2_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius2_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle2_) && DOUBLE_LESS(nStartAngle2_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle2_)) && DOUBLE_LE(fabs(nSweptAngle2_), 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_ArcWithArc(ptCenter1_, nRadius1_, nStartAngle1_, nSweptAngle1_,
										ptCenter2_, nRadius2_, nStartAngle2_, nSweptAngle2_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 09
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect == -1 ? 2 : _nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

int ncmath::Intersect_ArcWithEllipse(IN const DPOINT2& ptArcCenter_,
									 IN double nRadius_,
									 IN double nStartAngle_,
									 IN double nSweptAngle_,
									 IN const DPOINT2& ptEllipseCenter_,
									 IN double nURadius_,
									 IN double nVRadius_,
									 IN double nUAngle_,
									 OUT DPOINT2* pptIntersect1_/* = NULL */,
									 OUT DPOINT2* pptIntersect2_/* = NULL */,
									 OUT DPOINT2* pptIntersect3_/* = NULL */,
									 OUT DPOINT2* pptIntersect4_/* = NULL */)
{
	ASSERT(IsValidDouble(nStartAngle_));
	ASSERT(IsValidDouble(nSweptAngle_));

	if (DOUBLE_EQU_ZERO(nSweptAngle_))
		return 0;

	int _nNum = 0;
	DPOINT2 _ptIntersect[4];

	_nNum = Intersect_CircleWithEllipse(ptArcCenter_, nRadius_, ptEllipseCenter_, nURadius_, nVRadius_,
			nUAngle_, &_ptIntersect[0], &_ptIntersect[1], &_ptIntersect[2], &_ptIntersect[3]);

	int _nCount = 0;
	for (int _nIndex = 0; _nIndex < _nNum; _nIndex++)
	{
		// ԲϵĽ㱣 Wanglou 2010-09-16
		if (IsPointInArc(_ptIntersect[_nIndex], ptArcCenter_, nRadius_, nStartAngle_, nSweptAngle_, true))
		{
			_ptIntersect[_nCount] = _ptIntersect[_nIndex];
			_nCount++;
		}
	}

	for (int _i = _nCount; _i < 4; _i++)
	{
		_ptIntersect[_i] = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	}

	if (pptIntersect1_)
		*pptIntersect1_ = _ptIntersect[0];
	if (pptIntersect2_)
		*pptIntersect2_ = _ptIntersect[1];
	if (pptIntersect3_)
		*pptIntersect3_ = _ptIntersect[2];
	if (pptIntersect4_)
		*pptIntersect4_ = _ptIntersect[3];

	// غϣ-1
	int _nRet = _nCount;
	if (_nNum == -1)
		_nRet = -1;

	return _nRet;
}

int ncmath::Intersect_ArcWithEllipse(IN const DPOINT2& ptArcCenter_,
									 IN double nRadius_,
									 IN double nStartAngle_,
									 IN double nSweptAngle_,
									 IN const DPOINT2& ptEllipseCenter_,
									 IN double nURadius_,
									 IN double nVRadius_,
									 IN double nUAngle_,
									 OUT std::vector<DPOINT2>& vectorIntersect_,
									 IN int nIntersectsNeed_/* = 4 */)
{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	// 潻
	DPOINT2 _ptIntersects[4];
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C
	int _nRet = Intersect_ArcWithEllipse(
		ptArcCenter_, nRadius_, nStartAngle_, nSweptAngle_, ptEllipseCenter_, nURadius_, nVRadius_, nUAngle_,
		_nNeed >= 1 ? &_ptIntersects[0] : NULL,
		_nNeed >= 2 ? &_ptIntersects[1] : NULL,
		_nNeed >= 3 ? &_ptIntersects[2] : NULL,
		_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}

int ncmath::Intersect_ArcWithEllipseArc(IN const DPOINT2& ptArcCenter_,
										IN double nRadius_,
										IN double nArcStartAngle_,
										IN double nArcSweptAngle_,
										IN const DPOINT2& ptEllipseCenter_,
										IN double nURadius_,
										IN double nVRadius_,
										IN double nUAngle_,
										IN double nEllipseArcStartAngle_,
										IN double nEllipseArcSweptAngle_,
										OUT DPOINT2* pptIntersect1_/* = NULL */,
										OUT DPOINT2* pptIntersect2_/* = NULL */,
										OUT DPOINT2* pptIntersect3_/* = NULL */,
										OUT DPOINT2* pptIntersect4_/* = NULL */)
{
	ASSERT(IsValidDouble(nArcStartAngle_));
	ASSERT(IsValidDouble(nArcSweptAngle_));
	ASSERT(IsValidDouble(nEllipseArcStartAngle_));
	ASSERT(IsValidDouble(nEllipseArcSweptAngle_));

	// ԲԲĽǻԲĽΪ0Ϊûн Wanglou 2010-09-16
	if (DOUBLE_EQU_ZERO(nArcSweptAngle_) || DOUBLE_EQU_ZERO(nEllipseArcSweptAngle_))
		return 0;

	int _nNum = 0;
	DPOINT2 _ptIntersect[4];

	_nNum = Intersect_CircleWithEllipse(ptArcCenter_, nRadius_, ptEllipseCenter_, nURadius_, nVRadius_,
			nUAngle_, &_ptIntersect[0], &_ptIntersect[1], &_ptIntersect[2], &_ptIntersect[3]);

	int _nCount = 0;
	for (int _nIndex = 0; _nIndex < _nNum; _nIndex++)
	{
		// ԲϵĽ㱣 Wanglou 2010-09-16
		if (IsPointInArc(_ptIntersect[_nIndex], ptArcCenter_, nRadius_, nArcStartAngle_, nArcSweptAngle_, true) &&
			IsPointInEllipseArc(_ptIntersect[_nIndex], ptEllipseCenter_, nURadius_, nVRadius_, nUAngle_,
								nEllipseArcStartAngle_, nEllipseArcSweptAngle_, true))
		{
			_ptIntersect[_nCount] = _ptIntersect[_nIndex];
			_nCount++;
		}
	}

	for (int _i = _nCount; _i < 4; _i++)
	{
		_ptIntersect[_i] = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	}

	if (pptIntersect1_)
		*pptIntersect1_ = _ptIntersect[0];
	if (pptIntersect2_)
		*pptIntersect2_ = _ptIntersect[1];
	if (pptIntersect3_)
		*pptIntersect3_ = _ptIntersect[2];
	if (pptIntersect4_)
		*pptIntersect4_ = _ptIntersect[3];

	// غϣ-1
	int _nRet = _nCount;
	if (_nNum == -1)
		_nRet = -1;

	return _nRet;
}

int ncmath::Intersect_ArcWithEllipseArc(IN const DPOINT2& ptArcCenter_,
										IN double nRadius_,
										IN double nArcStartAngle_,
										IN double nArcSweptAngle_,
										IN const DPOINT2& ptEllipseCenter_,
										IN double nURadius_,
										IN double nVRadius_,
										IN double nUAngle_,
										IN double nEllipseArcStartAngle_,
										IN double nEllipseArcSweptAngle_,
										OUT std::vector<DPOINT2>& vectorIntersect_,
										IN int nIntersectsNeed_ /* = 4 */)

{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	// 潻
	DPOINT2 _ptIntersects[4];
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C
	int _nRet = Intersect_ArcWithEllipseArc(
				ptArcCenter_, nRadius_, nArcStartAngle_, nArcSweptAngle_, ptEllipseCenter_,
				nURadius_, nVRadius_, nUAngle_, nEllipseArcStartAngle_, nEllipseArcSweptAngle_,
				_nNeed >= 1 ? &_ptIntersects[0] : NULL,
				_nNeed >= 2 ? &_ptIntersects[1] : NULL,
				_nNeed >= 3 ? &_ptIntersects[2] : NULL,
				_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}

namespace
{
	// ֪ acost + bsint = c(costsint)ؽĸ
	// ̲abͬʱΪ0 --- wanghong 2010 09 21
	int SolveTrgEquation(double nA_, double nB_, double nC_, DPOINT2 ptRes_[2])
	{
		ASSERT(!DOUBLE_EQU_ZERO(nA_) || !DOUBLE_EQU_ZERO(nB_));
		double _nA2 = nA_ * nA_;
		double _nB2 = nB_ * nB_;
		double _nC2 = nC_ * nC_;

		double _nDst = _nA2 + _nB2 - _nC2;
		// ʱ޽ --- wanghong 2010 09 25
		if (DOUBLE_LESS_ZERO(_nDst))
			return 0;

		// ¼ĸ --- wanghong 2010 09 25
		int _nIndex = 0;

		if (DOUBLE_EQU_ZERO(nB_))
		{
			ptRes_[_nIndex].x = nA_ * nC_ / (_nA2 + _nB2);
			// ΪҪжϽΪ0.0ҪʹݲˣּСΣ
			// Ϊȷʹsqrtʹfabs ȡֵ --- wanghong 2010 09 25
			ASSERT(DOUBLE_GE_ZERO(1.0 - ptRes_[_nIndex].x * ptRes_[_nIndex].x));
			ptRes_[_nIndex].y = sqrt(fabs(1.0 - ptRes_[_nIndex].x * ptRes_[_nIndex].x));
			_nIndex += 1;

			if (!DOUBLE_EQU_ZERO(ptRes_[_nIndex-1].y))
			{
				ptRes_[_nIndex].x = ptRes_[_nIndex-1].x;
				ptRes_[_nIndex].y = -ptRes_[_nIndex-1].y;
				_nIndex += 1;
			}

			return _nIndex;
		}
		// ʱ1 --- wanghong 2010 09 25
		if (DOUBLE_EQU_ZERO(_nDst))
		{
			ptRes_[_nIndex].x = nA_ * nC_ / (_nA2 + _nB2);
			ptRes_[_nIndex].y = (nC_ - nA_ * ptRes_[_nIndex].x) / nB_;

			return ++_nIndex;
		}
		// ʱ2 --- wanghong 2010 09 25
		double _nTmp = fabs(nB_) * sqrt(_nDst);
		ptRes_[_nIndex].x = (nA_ * nC_ + _nTmp) / (_nA2 + _nB2);
		ptRes_[_nIndex].y = (nC_ - nA_ * ptRes_[_nIndex].x) / nB_;
		_nIndex += 1;

		ptRes_[_nIndex].x = (nA_ * nC_ - _nTmp) / (_nA2 + _nB2);
		ptRes_[_nIndex].y = (nC_ - nA_ * ptRes_[_nIndex].x) / nB_;
		_nIndex += 1;

		return _nIndex;
	}
}

namespace
{
	// ֱϵһ --- wanghong 2010 09 19
	DPOINT2 GetSecondPoint(const DPOINT2& ptBeeline_, double nBeelineAngle_)
	{
		if (DOUBLE_EQU(nBeelineAngle_, ncmath::c_nPIE/2) || DOUBLE_EQU(nBeelineAngle_, ncmath::c_nPIE*3/2))
		{
			return DPOINT2(ptBeeline_.x, ptBeeline_.y + 1);
		}

		return DPOINT2(ptBeeline_.x + 1, ptBeeline_.y + tan(nBeelineAngle_));
	}
}

// UVϵý㣬תXYϵ --- wanghong 2010 09 13
int ncmath::Intersect_BeelineWithEllipse(IN const DPOINT2& ptBeeline_,
										 IN double nBeelineAngle_,
										 IN const DPOINT2& ptCenter_,
										 IN double nURadius_,
										 IN double nVRadius_,
										 IN double nUAngle_,
										 OUT DPOINT2* pptIntersect1_/* = NULL*/,
										 OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y)
		&& nBeelineAngle_ >= 0.0 && nBeelineAngle_ < 2.0 * c_nPIE);
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y)
		&& DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_)
		&& nUAngle_ >= 0.0 && nUAngle_ < 2.0 * c_nPIE);

	// ںļͨһ߷߷̽һԪη
	// ģһԪηʱĵȸбΪѣˣΪⲻҪ
	// ⣬ڴȽ⴦ˡ            ----  2010-09-25
	DPOINT2 _ptTangent1;
	DPOINT2 _ptTangent2;
	GetTangentPointOfEllipse(ptCenter_, nURadius_, nVRadius_, RegularizeIncludedRadian(nUAngle_),
		RegularizeIncludedRadian(nBeelineAngle_), &_ptTangent1, &_ptTangent2);
	ASSERT(IsValidDouble(_ptTangent1.x) && IsValidDouble(_ptTangent1.y)
		&& IsValidDouble(_ptTangent2.x) && IsValidDouble(_ptTangent2.y)
		&& DOUBLE_NOT_ZERO((_ptTangent1 - _ptTangent2).GetLength()));
	double _nDist1 = D2GetPointToBeelineDistance(_ptTangent1, ptBeeline_, RegularizeIncludedRadian(nBeelineAngle_));
	if (DOUBLE_EQU_ZERO(_nDist1))
	{
		*pptIntersect1_ = _ptTangent1;
		*pptIntersect2_ = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
		return 1;
	}
	double _nDist2 = D2GetPointToBeelineDistance(_ptTangent1, ptBeeline_, RegularizeIncludedRadian(nBeelineAngle_));
	if (DOUBLE_EQU_ZERO(_nDist2))
	{
		*pptIntersect1_ = _ptTangent2;
		*pptIntersect2_ = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
		return 1;
	}

	// 任UVϵ(Բ任Ϊ׼Բ) ----  2010-09-21
	DPOINT2 _ptBeelineUV = CoorTransXYToUV(ptBeeline_, ptCenter_, nUAngle_);
	double _nBeelineAngleUV = RegularizeIncludedRadian(nBeelineAngle_ - nUAngle_);
	double _a = nURadius_;
	double _b = nVRadius_;
	double _a2 = _a * _a;
	double _b2 = _b * _b;

	// UVϵ󽻵 ----  2010-09-21
	DPOINT2 _ptIntersects[2] =
	{
		DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE),
		DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE),
	};

	ASSERT(_nBeelineAngleUV >= 0.0 && _nBeelineAngleUV < c_nPIE);
	if (DOUBLE_LE(_nBeelineAngleUV, 0.25 * c_nPIE) || DOUBLE_GE(_nBeelineAngleUV, 0.75 * c_nPIE))
	{
		// ֱXнǽСʱֱ߷̿дΪ y = kx + d, ֱԲ̿
		// ⽻㡣
		// ֱ߷ y = kx + d, Բ x^2/a^2 + y^2/b^2 = 1 ɵ÷
		// (b^2/a^2 + k^2)*x^2 + (2*k*d)*x + (d^2 - b^2) = 0 ----  2010-09-21
		double _k = tan(_nBeelineAngleUV);
		double _d = _ptBeelineUV.y - _k * _ptBeelineUV.x;
		double _nA = _b2 / _a2 + _k * _k;
		double _nB = 2.0 * _k * _d;
		double _nC = _d * _d - _b2;
		double _x1;
		double _x2;
		int _n = SolveQuadraticEquation(_nA, _nB, _nC, &_x1, &_x2);
		ASSERT(_n == 0 || _n == 1 || _n == 2);

		if (_n >= 1)
		{
			_ptIntersects[0].x = _x1;
			_ptIntersects[0].y = _k * _x1 + _d;
			ASSERT(IsPointInEllipse(_ptIntersects[0], DPOINT2(0.0, 0.0), _a, _b, 0.0));
		}

		if (_n >= 2)
		{
			_ptIntersects[1].x = _x2;
			_ptIntersects[1].y = _k * _x2 + _d;
			ASSERT(IsPointInEllipse(_ptIntersects[1], DPOINT2(0.0, 0.0), _a, _b, 0.0));
		}
	}
	else
	{
		// ֱXнǽϴ󣬴ʱֱ߷̿дΪ x = ky + d, ֱԲ̿
		// ⽻㡣
		// ֱ߷ x = ky + d, Բ x^2/a^2 + y^2/b^2 = 1 ɵ÷
		// (a^2/b^2 + k^2)*y^2 + (2*k*d)*y + (d^2 - a^2) = 0 ----  2010-09-21
		double _k = tan(0.5 * c_nPIE - _nBeelineAngleUV);
		double _d = _ptBeelineUV.x - _k * _ptBeelineUV.y;
		double _nA = _a2 / _b2 + _k * _k;
		double _nB = 2.0 * _k * _d;
		double _nC = _d * _d - _a2;
		double _y1;
		double _y2;
		int _n = SolveQuadraticEquation(_nA, _nB, _nC, &_y1, &_y2);
		ASSERT(_n == 0 || _n == 1 || _n == 2);

		if (_n >= 1)
		{
			_ptIntersects[0].x = _k * _y1 + _d;
			_ptIntersects[0].y = _y1;
			ASSERT(IsPointInEllipse(_ptIntersects[0], DPOINT2(0.0, 0.0), _a, _b, 0.0));
		}

		if (_n >= 2)
		{
			_ptIntersects[1].x = _k * _y2 + _d;
			_ptIntersects[1].y = _y2;
			ASSERT(IsPointInEllipse(_ptIntersects[1], DPOINT2(0.0, 0.0), _a, _b, 0.0));
		}
	}

	// صXY ----  2010-09-21
	int _nIntersect = 0;

	if (IsValidDouble(_ptIntersects[0].x) && IsValidDouble(_ptIntersects[0].y))
	{
		_nIntersect++;
		_ptIntersects[0] = CoorTransUVToXY(_ptIntersects[0], ptCenter_, nUAngle_);
	}

	if (IsValidDouble(_ptIntersects[1].x) && IsValidDouble(_ptIntersects[1].y))
	{
		_nIntersect++;
		_ptIntersects[1] = CoorTransUVToXY(_ptIntersects[1], ptCenter_, nUAngle_);
	}

	//  ----  2010-09-21
	*pptIntersect1_ = _ptIntersects[0];
	*pptIntersect2_ = _ptIntersects[1];

	return _nIntersect;

	// ´⣬֮취ƺٻָ֮ ----  2010-09-21
	ASSERT(FALSE);

	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GE_ZERO(nBeelineAngle_) && DOUBLE_LESS(nBeelineAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));

	// ֱȡABתUVϵUVϵ --- wanghong 2010 09 13
	DPOINT2 _ptBeelineA = CoorTransXYToUV(ptBeeline_, ptCenter_, nUAngle_);
	DPOINT2 _ptBeelineB = CoorTransXYToUV(GetSecondPoint(ptBeeline_, nBeelineAngle_), ptCenter_, nUAngle_);
	// ԭ㵽AB --- wanghong 2010 09 13
	VectorD2D _vecBeelineA(_ptBeelineA.x, _ptBeelineA.y);
	VectorD2D _vecBeelineB(_ptBeelineB.x, _ptBeelineB.y);
	// ֱABԼֱ --- wanghong 2010 09 13
	VectorD2D _vecAB = _vecBeelineB - _vecBeelineA;
	VectorD2D _vecABPlumb(-_vecAB.y, _vecAB.x);
	// [NcMath/Doc/˵-󽻺]еֱԲ󽻹ʽ㽻 --- wanghong 2010 09 19
	DPOINT2 _ptRes[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	// Ϊ˱ϴǽǷ̵ϵһ --- wanghong 2010 09 25
	double nDiv = _vecBeelineA * _vecABPlumb;
	int _nCount = 0;
	if (DOUBLE_EQU_ZERO(nDiv))
	{	// ʱʼΪ nURadius_ * nVRadius_ --- wanghong 2010 09 25
		_nCount = SolveTrgEquation(_vecABPlumb.x / nVRadius_, _vecABPlumb.y / nURadius_, .0, _ptRes);
	}
	else
	{	// ʱʼΪ _vecBeelineA * _vecABPlumb --- wanghong 2010 09 25
		_nCount = SolveTrgEquation(nURadius_ * _vecABPlumb.x / nDiv, nVRadius_ * _vecABPlumb.y / nDiv, 1.0, _ptRes);
	}

	if (_nCount == 0)
	{
		return 0;
	}
	// UVϵµĽתXYϵ --- wanghong 2010 09 13
	if (pptIntersect1_ != NULL)
		*pptIntersect1_ = CoorTransUVToXY(DPOINT2(nURadius_ * _ptRes[0].x, nVRadius_ * _ptRes[0].y), ptCenter_, nUAngle_);

	if (_nCount == 2)
	{
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = CoorTransUVToXY(DPOINT2(nURadius_ * _ptRes[1].x, nVRadius_ * _ptRes[1].y), ptCenter_, nUAngle_);
	}

	return _nCount;
}

int ncmath::Intersect_BeelineWithEllipse(IN const DPOINT2& ptBeeline_, 
										 IN double nBeelineAngle_,
										 IN const DPOINT2& ptCenter_, 
										 IN double nURadius_, 
										 IN double nVRadius_, 
										 IN double nUAngle_,
										 OUT std::vector<DPOINT2>& vectorIntersect_,
										 IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_BeelineWithEllipse(ptBeeline_, nBeelineAngle_,
												ptCenter_, nURadius_, nVRadius_, nUAngle_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 13
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

namespace
{
	bool IsPointInEllipseArc_impl(IN const DPOINT2& ptTarget_,
								  IN const DPOINT2& ptCenter_,
								  IN double nSlopeAngle_,
								  IN double nStartAngle_,
								  IN double nCentralAngle_,
								  IN bool bIsNormalDir_)
	{
		// ԲĿǶ --- wanghong 2010 9 16
		double _nTargetAngle = D2GetNormalAngle(ptTarget_ - ptCenter_);

		// ע⣬ҪʼǲnStartAngle_ --- wanghong 2010 09 16
		return IsAngleInSector(_nTargetAngle, nStartAngle_ + nSlopeAngle_, nCentralAngle_, bIsNormalDir_);
	}
}

int ncmath::Intersect_BeelineWithEllipseArc(IN const DPOINT2& ptBeeline_,
											IN double nBeelineAngle_,
											IN const DPOINT2& ptCenter_,
											IN double nURadius_,
											IN double nVRadius_,
											IN double nUAngle_,
											IN double nStartAngle_,
											IN double nSweptAngle_,
											OUT DPOINT2* pptIntersect1_/* = NULL*/,
											OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * ncmath::c_nPIE));

	// ֱͨԲڵԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 13
	DPOINT2 _ptIntersect1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersect2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);

	int _nImageIntersect = Intersect_BeelineWithEllipse(ptBeeline_, nBeelineAngle_,
		ptCenter_, nURadius_, nVRadius_, nUAngle_, &_ptIntersect1, &_ptIntersect2);

	if (_nImageIntersect < 1)
		return _nImageIntersect;
	
	bool _bPt1Exist =
		IsPointInEllipseArc_impl(_ptIntersect1, ptCenter_, nUAngle_,
			nStartAngle_, fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{
		_bPt2Exist =
			IsPointInEllipseArc_impl(_ptIntersect2, ptCenter_, nUAngle_,
				nStartAngle_, fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	}
	//  --- wanghong 2010 09 13
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersect1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersect2;
		return 2;
	}
	// һ --- wanghong 2010 09 13
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersect1 : _ptIntersect2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_BeelineWithEllipseArc(IN const DPOINT2& ptBeeline_,
											IN double nBeelineAngle_,
											IN const DPOINT2& ptCenter_,
											IN double nURadius_,
											IN double nVRadius_,
											IN double nUAngle_,
											IN double nStartAngle_,
											IN double nSweptAngle_,
											OUT std::vector<DPOINT2>& vectorIntersect_,
											IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptBeeline_.x) && IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nURadius_) && DOUBLE_GREAT_ZERO(nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_BeelineWithEllipseArc(ptBeeline_, nBeelineAngle_, ptCenter_, nURadius_, nVRadius_, nUAngle_,
													nStartAngle_, nSweptAngle_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 13
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

int ncmath::Intersect_LineWithEllipse(IN const DPOINT2& ptStart_,
									  IN const DPOINT2& ptEnd_,
									  IN const DPOINT2& ptCenter_,
									  IN double nURadius_,
									  IN double nVRadius_,
									  IN double nUAngle_,
									  OUT DPOINT2* pptIntersect1_/* = NULL*/,
									  OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ͨ߶ζӦֱԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 14
	int _nImageIntersect = Intersect_BeelineWithEllipse(ptStart_,ncmath::D2GetNormalAngle(ptEnd_ - ptStart_),
		ptCenter_, nURadius_, nVRadius_, nUAngle_, &_ptIntersec1, &_ptIntersec2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist = IsPointInLine(_ptIntersec1, ptStart_, ptEnd_);
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{
		_bPt2Exist = IsPointInLine(_ptIntersec2, ptStart_, ptEnd_);
	}
	//  --- wanghong 2010 09 14
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}
	// һ --- wanghong 2010 09 14
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_LineWithEllipse(IN const DPOINT2& ptStart_,
									  IN const DPOINT2& ptEnd_,
									  IN const DPOINT2& ptCenter_,
									  IN double nURadius_,
									  IN double nVRadius_,
									  IN double nUAngle_,
									  OUT std::vector<DPOINT2>& vectorIntersect_,
									  IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));

	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_LineWithEllipse(ptStart_, ptEnd_, ptCenter_, nURadius_, nVRadius_, nUAngle_,
												_ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 14
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

int ncmath::Intersect_LineWithEllipseArc(IN const DPOINT2& ptStart_,
										 IN const DPOINT2& ptEnd_,
										 IN const DPOINT2& ptCenter_,
										 IN double nURadius_,
										 IN double nVRadius_,
										 IN double nUAngle_,
										 IN double nStartAngle_,
										 IN double nSweptAngle_,
										 OUT DPOINT2* pptIntersect1_/* = NULL*/,
										 OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * ncmath::c_nPIE));
	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ͨ߶ζӦֱԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 14
	int _nImageIntersect = Intersect_BeelineWithEllipseArc(ptStart_,ncmath::D2GetNormalAngle(ptEnd_ - ptStart_),
		ptCenter_, nURadius_, nVRadius_, nUAngle_, nStartAngle_, nSweptAngle_, &_ptIntersec1, &_ptIntersec2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist = IsPointInLine(_ptIntersec1, ptStart_, ptEnd_);
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{
		_bPt2Exist = IsPointInLine(_ptIntersec2, ptStart_, ptEnd_);
	}
	//  --- wanghong 2010 09 14
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}
	// һ --- wanghong 2010 09 14
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}

	return 0;
}

int ncmath::Intersect_LineWithEllipseArc(IN const DPOINT2& ptStart_,
										 IN const DPOINT2& ptEnd_,
										 IN const DPOINT2& ptCenter_,
										 IN double nURadius_,
										 IN double nVRadius_,
										 IN double nUAngle_,
										 IN double nStartAngle_,
										 IN double nSweptAngle_,
										 OUT std::vector<DPOINT2>& vectorIntersect_,
										 IN int nIntersectsNeed_/* = 2*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nVRadius_) && DOUBLE_GE(nURadius_, nVRadius_));
	ASSERT(DOUBLE_GE_ZERO(nUAngle_) && DOUBLE_LESS(nUAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * ncmath::c_nPIE));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * ncmath::c_nPIE));
	
	DPOINT2 _ptIntersect[2] = {DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE), DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE)};
	int _nIntersect = Intersect_LineWithEllipseArc(ptStart_, ptEnd_,
		ptCenter_, nURadius_, nVRadius_, nUAngle_, nStartAngle_, nSweptAngle_, _ptIntersect, _ptIntersect + 1);
	// nIntersectsNeed_ ȡ-1ʱҪ󷵻еĽ㣬ֵͬΪ2 --- wanghong 2010 09 14
	if (nIntersectsNeed_ <= -1)
		nIntersectsNeed_ = 2;

	int _nMinIntersect = min(_nIntersect, nIntersectsNeed_);
	std::copy(_ptIntersect, _ptIntersect + _nMinIntersect, std::back_inserter(vectorIntersect_));

	return _nIntersect;
}

// ¿ռڵĺǹIntersect_EllipseWithEllipseõľֲ ----  2010-09-17
namespace Intersect_EllipseWithEllipse_Implement
{
	// ԲӦڶ̰ᣬ[0, 2PI) ----  2010-09-17
	struct Ellipse
	{
		DPOINT2 ptCenter;
		double nURadius;
		double nVRadius;
		double nUAngle;
	};

	// Բʼ[0, 2PI)ɨ(0, 2PI)ʾʱɨʼǱʾ
	// ԲǵĽǣʼǡɨǾֱָ۽ǣԲеĽ
	// ֵTheta                                         ----  2010-09-17
	struct EllipseArc : public Ellipse
	{
		double nStartAngle;
		double nSweptAngle;
	};

	// һԲֳĸԲ(yx)xе㡢yе㽫
	// ԲΪĸΡ                                ----  2010-09-17
	void DivideEllipseIntoFourMonotonousArc(
		IN const Ellipse& Ellipse_,
		OUT EllipseArc* pEllipseArc1_,
		OUT EllipseArc* pEllipseArc2_,
		OUT EllipseArc* pEllipseArc3_,
		OUT EllipseArc* pEllipseArc4_)
	{
		// ڲ ----  2010-09-17
		ASSERT(pEllipseArc1_ && pEllipseArc2_ && pEllipseArc3_ && pEllipseArc4_);

		// Բ ----  2010-09-17
		DPOINT2 _ptC = Ellipse_.ptCenter;
		double _a = Ellipse_.nURadius;
		double _b = Ellipse_.nVRadius;
		double _nTheta = Ellipse_.nUAngle;
		ASSERT(IsValidDouble(_ptC.x) && IsValidDouble(_ptC.y)
			&& _a >= _b && DOUBLE_GREAT_ZERO(_b)
			&& _nTheta >= 0 && _nTheta < 2.0 * c_nPIE);

		// ĸ߽㣬Ӧĸ ----  2010-09-17
		DPOINT2 _ptTangents[4];
		GetTangentPointOfEllipse(_ptC, _a, _b, RegularizeIncludedRadian(_nTheta), 0.0,
			&_ptTangents[0], &_ptTangents[2]);
		GetTangentPointOfEllipse(_ptC, _a, _b, RegularizeIncludedRadian(_nTheta), 0.5 * c_nPIE,
			&_ptTangents[1], &_ptTangents[3]);
		double _nAngles[4];
		for (int _i = 0; _i < 4; _i++)
		{
			DPOINT2 _vec = _ptTangents[_i] - _ptC;
			ASSERT(DOUBLE_GREAT_ZERO(_vec.GetLength()));
			_nAngles[_i] = D2GetNormalAngle(_vec);
		}

		// ȥԲǺֵ ----  2010-09-17
		for (int _i = 0; _i < 4; _i++)
		{
			double _nAngle = _nAngles[_i] - _nTheta;
			NormalizeRadian(_nAngle);
			_nAngles[_i] = _nAngle;
		}

		// 򣬰ǴС ----  2010-09-17
		for (int _i = 0; _i < 4; _i++)
		{
			for (int _j = _i + 1; _j < 4; _j++)
			{
				ASSERT(!DOUBLE_EQU(_nAngles[_j], _nAngles[_i]));
				if (_nAngles[_j] < _nAngles[_i])
				{
					double _nTemp = _nAngles[_i];
					_nAngles[_i] = _nAngles[_j];
					_nAngles[_j] = _nTemp;
				}
			}
		}

		// ΪԲβ ----  2010-09-17
		EllipseArc* _pEllipseArcs[4] =
		{
			pEllipseArc1_,
			pEllipseArc2_,
			pEllipseArc3_,
			pEllipseArc4_,
		};
		for (int _i = 0; _i < 4; _i++)
		{
			_pEllipseArcs[_i]->ptCenter = _ptC;
			_pEllipseArcs[_i]->nURadius = _a;
			_pEllipseArcs[_i]->nVRadius = _b;
			_pEllipseArcs[_i]->nUAngle = _nTheta;
			_pEllipseArcs[_i]->nStartAngle = _nAngles[_i];
			_pEllipseArcs[_i]->nSweptAngle = RegularizeRadian(_nAngles[(_i + 1) % 4] - _nAngles[_i]);
		}
	}

	// ֵεĽ㣬ڵñǰѾȷڴxֻ
	// һ㡣                                       ----  2010-09-18
	void Intersect_MonotonousEllipseArcWithMonotonousEllipseArc_OnlyOne(
		IN const EllipseArc& EllipseArc1_,
		IN const EllipseArc& EllipseArc2_,
		IN const double& nXLeft_,
		IN const double& nXRight_,
		OUT DPOINT2* pptIntersect_)
	{
		ASSERT(IsValidDouble(nXLeft_) && IsValidDouble(nXRight_) && DOUBLE_LESS(nXLeft_, nXRight_));
		ASSERT(pptIntersect_);

		// Բ ----  2010-09-18
		DPOINT2 _ptC1 = EllipseArc1_.ptCenter;
		double _a1 = EllipseArc1_.nURadius;
		double _b1 = EllipseArc1_.nVRadius;
		double _nTheta1 = EllipseArc1_.nUAngle;
		double _nStart1 = EllipseArc1_.nStartAngle;
		double _nSwept1 = EllipseArc1_.nSweptAngle;
		ASSERT(IsValidDouble(_ptC1.x) && IsValidDouble(_ptC1.y)
			&& _a1 >= _b1 && DOUBLE_GREAT_ZERO(_b1)
			&& _nTheta1 >= 0 && _nTheta1 < 2.0 * c_nPIE
			&& _nStart1 >= 0 && _nStart1 < 2.0 * c_nPIE
			&& DOUBLE_GREAT_ZERO(_nSwept1) && _nSwept1 < 2.0 * c_nPIE);
		DPOINT2 _ptC2 = EllipseArc2_.ptCenter;
		double _a2 = EllipseArc2_.nURadius;
		double _b2 = EllipseArc2_.nVRadius;
		double _nTheta2 = EllipseArc2_.nUAngle;
		double _nStart2 = EllipseArc2_.nStartAngle;
		double _nSwept2 = EllipseArc2_.nSweptAngle;
		ASSERT(IsValidDouble(_ptC2.x) && IsValidDouble(_ptC2.y)
			&& _a2 >= _b2 && DOUBLE_GREAT_ZERO(_b2)
			&& _nTheta2 >= 0 && _nTheta2 < 2.0 * c_nPIE
			&& _nStart2 >= 0 && _nStart2 < 2.0 * c_nPIE
			&& DOUBLE_GREAT_ZERO(_nSwept2) && _nSwept2 < 2.0 * c_nPIE);

		// Ҷ˵ ----  2010-09-18
		DPOINT2 _ptLeft1;
		DPOINT2 _ptLeft2;
		DPOINT2 _ptRight1;
		DPOINT2 _ptRight2;
		int _n;
		_n = Intersect_BeelineWithEllipseArc(DPOINT2(nXLeft_, 0.0), 0.5 * c_nPIE,
			_ptC1, _a1, _b1, _nTheta1, _nStart1, _nSwept1,
			&_ptLeft1);
		ASSERT(_n == 1);
		_n = Intersect_BeelineWithEllipseArc(DPOINT2(nXLeft_, 0.0), 0.5 * c_nPIE,
			_ptC2, _a2, _b2, _nTheta2, _nStart2, _nSwept2,
			&_ptLeft2);
		ASSERT(_n == 1);
		_n = Intersect_BeelineWithEllipseArc(DPOINT2(nXRight_, 0.0), 0.5 * c_nPIE,
			_ptC1, _a1, _b1, _nTheta1, _nStart1, _nSwept1,
			&_ptRight1);
		ASSERT(_n == 1);
		_n = Intersect_BeelineWithEllipseArc(DPOINT2(nXRight_, 0.0), 0.5 * c_nPIE,
			_ptC2, _a2, _b2, _nTheta2, _nStart2, _nSwept2,
			&_ptRight2);
		ASSERT(_n == 1);

		// ҶyֵΪ ----  2010-09-18
		double _nDyLeft = _ptLeft1.y - _ptLeft2.y;
		double _nDyRight = _ptRight1.y - _ptRight2.y;
		ASSERT(DOUBLE_NOT_ZERO(_nDyLeft) && DOUBLE_NOT_ZERO(_nDyRight)
			&& DOUBLE_LESS_ZERO(_nDyLeft * _nDyRight));

		//  ----  2010-09-18
		DPOINT2 _ptMid1;
		DPOINT2 _ptMid2;
		while (DOUBLE_GREAT(_ptRight1.x, _ptLeft1.x))
		{
			double _nMid = (_ptRight1.x + _ptLeft1.x) / 2.0;
			int _n = Intersect_BeelineWithEllipseArc(DPOINT2(_nMid, 0.0), 0.5 * c_nPIE,
				_ptC1, _a1, _b1, _nTheta1, _nStart1, _nSwept1, &_ptMid1);
			ASSERT(_n == 1);
			_n = Intersect_BeelineWithEllipseArc(DPOINT2(_nMid, 0.0), 0.5 * c_nPIE,
				_ptC2, _a2, _b2, _nTheta2, _nStart2, _nSwept2, &_ptMid2);
			ASSERT(_n == 1);
			if (DOUBLE_EQU(_ptMid1.y, _ptMid2.y))
			{
				break;
			}
			else if ((_ptLeft1.y - _ptLeft2.y) * (_ptMid1.y - _ptMid2.y) > 0.0)
			{
				_ptLeft1 = _ptMid1;
				_ptLeft2 = _ptMid2;
			}
			else
			{
				_ptRight1 = _ptMid1;
				_ptRight2 = _ptMid2;
			}
		}

		ASSERT(DOUBLE_EQU(_ptMid1.x, _ptMid2.x));
		*pptIntersect_ = (_ptMid1 + _ptMid2) / 2.0;
	}

	// ֪ԲĽ㡣ⲻغϵغд
	//   취§ҵģ ͼϵͳʵԲ󽻿㷨о
	// ʵ֡ ѧ ͵ ͼѧѧ 2000 3 μģ
	// ˲׸                                       ----  2010-09-18
	int Intersect_MonotonousEllipseArcWithMonotonousEllipseArc(
		IN const EllipseArc& EllipseArc1_,
		IN const EllipseArc& EllipseArc2_,
		OUT DPOINT2* pptIntersect1_ = NULL,
		OUT DPOINT2* pptIntersect2_ = NULL)
	{
		// ڲ ----  2010-09-18
		ASSERT(pptIntersect1_ && pptIntersect2_
			|| pptIntersect1_ && !pptIntersect2_
			|| !pptIntersect1_ && !pptIntersect2_);

		// Ϊ·˳ȲЧֵ ----  2010-09-18
		if (pptIntersect1_)
		{
			pptIntersect1_->x = c_nINVALID_DOUBLE;
			pptIntersect1_->y = c_nINVALID_DOUBLE;
		}

		if (pptIntersect2_)
		{
			pptIntersect2_->x = c_nINVALID_DOUBLE;
			pptIntersect2_->y = c_nINVALID_DOUBLE;
		}

		// Բ ----  2010-09-18
		DPOINT2 _ptC1 = EllipseArc1_.ptCenter;
		double _a1 = EllipseArc1_.nURadius;
		double _b1 = EllipseArc1_.nVRadius;
		double _nTheta1 = EllipseArc1_.nUAngle;
		double _nStart1 = EllipseArc1_.nStartAngle;
		double _nSwept1 = EllipseArc1_.nSweptAngle;
		ASSERT(IsValidDouble(_ptC1.x) && IsValidDouble(_ptC1.y)
			&& _a1 >= _b1 && DOUBLE_GREAT_ZERO(_b1)
			&& _nTheta1 >= 0 && _nTheta1 < 2.0 * c_nPIE
			&& _nStart1 >= 0 && _nStart1 < 2.0 * c_nPIE
			&& DOUBLE_GREAT_ZERO(_nSwept1) && _nSwept1 < 2.0 * c_nPIE);
		DPOINT2 _ptC2 = EllipseArc2_.ptCenter;
		double _a2 = EllipseArc2_.nURadius;
		double _b2 = EllipseArc2_.nVRadius;
		double _nTheta2 = EllipseArc2_.nUAngle;
		double _nStart2 = EllipseArc2_.nStartAngle;
		double _nSwept2 = EllipseArc2_.nSweptAngle;
		ASSERT(IsValidDouble(_ptC2.x) && IsValidDouble(_ptC2.y)
			&& _a2 >= _b2 && DOUBLE_GREAT_ZERO(_b2)
			&& _nTheta2 >= 0 && _nTheta2 < 2.0 * c_nPIE
			&& _nStart2 >= 0 && _nStart2 < 2.0 * c_nPIE
			&& DOUBLE_GREAT_ZERO(_nSwept2) && _nSwept2 < 2.0 * c_nPIE);

		// Ҷ˵ ----  2010-09-18
		DPOINT2 _ptLeft1 = GetPointInEllipse(_ptC1, _a1, _b1, _nTheta1, _nStart1);
		DPOINT2 _ptRight1 = GetPointInEllipse(_ptC1, _a1, _b1, _nTheta1, RegularizeRadian(_nStart1 + _nSwept1));
		if (_ptLeft1.x > _ptRight1.x)
		{
			DPOINT2 _ptTmp = _ptLeft1;
			_ptLeft1 = _ptRight1;
			_ptRight1 = _ptTmp;
		}
		DPOINT2 _ptLeft2 = GetPointInEllipse(_ptC2, _a2, _b2, _nTheta2, _nStart2);
		DPOINT2 _ptRight2 = GetPointInEllipse(_ptC2, _a2, _b2, _nTheta2, RegularizeRadian(_nStart2 + _nSwept2));
		if (_ptLeft2.x > _ptRight2.x)
		{
			DPOINT2 _ptTmp = _ptLeft2;
			_ptLeft2 = _ptRight2;
			_ptRight2 = _ptTmp;
		}

		// صxȻ޽ ----  2010-09-18
		if (DOUBLE_GREAT(_ptLeft1.x, _ptRight2.x) || DOUBLE_GREAT(_ptLeft2.x, _ptRight1.x))
		{
			return 0;
		}

		// ȷxΧ ----  2010-09-18
		DPOINT2 _ptTemp;
		int _nTemp;
		// 
		_nTemp = Intersect_BeelineWithEllipseArc(_ptLeft1, 0.5 * c_nPIE,
			_ptC2, _a2, _b2, _nTheta2, _nStart2, _nSwept2,
			&_ptTemp);
		ASSERT(_nTemp == 0 || _nTemp == 1);
		if (_nTemp == 1)
		{
			_ptLeft2 = _ptTemp;
		}
		else
		{
			_nTemp = Intersect_BeelineWithEllipseArc(_ptLeft2, 0.5 * c_nPIE,
				_ptC1, _a1, _b1, _nTheta1, _nStart1, _nSwept1,
				&_ptTemp);
			ASSERT(_nTemp == 1);
			_ptLeft1 = _ptTemp;
		}
		// Ҷ
		_nTemp = Intersect_BeelineWithEllipseArc(_ptRight1, 0.5 * c_nPIE,
			_ptC2, _a2, _b2, _nTheta2, _nStart2, _nSwept2,
			&_ptTemp);
		ASSERT(_nTemp == 0 || _nTemp == 1);
		if (_nTemp == 1)
		{
			_ptRight2 = _ptTemp;
		}
		else
		{
			_nTemp = Intersect_BeelineWithEllipseArc(_ptRight2, 0.5 * c_nPIE,
				_ptC1, _a1, _b1, _nTheta1, _nStart1, _nSwept1,
				&_ptTemp);
			ASSERT(_nTemp == 1);
			_ptRight1 = _ptTemp;
		}

		// صyȻ޽ ----  2010-09-18
		if (DOUBLE_GREAT(_ptLeft1.y, _ptLeft2.y) && DOUBLE_GREAT(_ptLeft1.y, _ptRight2.y)
			&& DOUBLE_GREAT(_ptRight1.y, _ptLeft2.y) && DOUBLE_GREAT(_ptRight1.y, _ptRight2.y)
			|| DOUBLE_GREAT(_ptLeft2.y, _ptLeft1.y) && DOUBLE_GREAT(_ptLeft2.y, _ptRight1.y)
			&& DOUBLE_GREAT(_ptRight2.y, _ptLeft1.y) && DOUBLE_GREAT(_ptRight2.y, _ptRight1.y))
		{
			return 0;
		}

		// ׼ںͷֵ ----  2010-09-18
		int _nIntersect = 0;
		DPOINT2* _pIntersects[2] =
		{
			pptIntersect1_,
			pptIntersect2_,
		};

		// ˵yֵ ----  2010-09-18
		double _nDyLeft = _ptLeft1.y - _ptLeft2.y;
		double _nDyRight = _ptRight1.y - _ptRight2.y;
		if (DOUBLE_EQU_ZERO(_nDyLeft) || DOUBLE_EQU_ZERO(_nDyRight))
		{
			// yֵΪ0ϵó ----  2010-09-18
			if (DOUBLE_EQU_ZERO(_nDyLeft))
			{
				*_pIntersects[_nIntersect] = _ptLeft1;
				_nIntersect++;
			}

			if (DOUBLE_EQU_ZERO(_nDyRight))
			{
				*_pIntersects[_nIntersect] = _ptRight1;
				_nIntersect++;
			}
		}
		else if (_nDyLeft * _nDyRight < 0.0)
		{
			// yֵţҽһ ----  2010-09-18
			if (_pIntersects[_nIntersect])
			{
				Intersect_MonotonousEllipseArcWithMonotonousEllipseArc_OnlyOne(EllipseArc1_, EllipseArc2_,
					_ptLeft1.x, _ptRight1.x, _pIntersects[_nIntersect]);
			}

			_nIntersect++;
		}
		else
		{
			// yֵͬţǲ ----  2010-09-18
#define MyNormalize(x_, b_) { if (DOUBLE_EQU(x_, 0.5 * c_nPIE)) x_ *= (b_ > 0.0) ? 1.0 : -1.0; else if (x_ > 0.5 * c_nPIE) x_ -= c_nPIE; }
			double _nAngleLeft1 = GetTangentAngleOfEllipse(_ptC1, _a1, _b1, RegularizeIncludedRadian(_nTheta1), _ptLeft1);
			MyNormalize(_nAngleLeft1, _ptRight1.y - _ptLeft1.y);
			double _nAngleLeft2 = GetTangentAngleOfEllipse(_ptC2, _a2, _b2, RegularizeIncludedRadian(_nTheta2), _ptLeft2);
			MyNormalize(_nAngleLeft2, _ptRight2.y - _ptLeft2.y);
			double _nAngleRight1 = GetTangentAngleOfEllipse(_ptC1, _a1, _b1, RegularizeIncludedRadian(_nTheta1), _ptRight1);
			MyNormalize(_nAngleRight1, _ptRight1.y - _ptLeft1.y);
			double _nAngleRight2 = GetTangentAngleOfEllipse(_ptC2, _a2, _b2, RegularizeIncludedRadian(_nTheta2), _ptRight2);
			MyNormalize(_nAngleRight2, _ptRight2.y - _ptLeft2.y);

			if (DOUBLE_EQU_ZERO(_nAngleLeft1 - _nAngleLeft2)
				|| DOUBLE_EQU_ZERO(_nAngleRight1 - _nAngleRight2)
				|| (_nAngleLeft1 - _nAngleLeft2) * (_nAngleRight1 - _nAngleRight2) > 0.0)
			{
				// ǲͬţ޽ ----  2010-09-18
			}
			else
			{
				// ǲţͬxֵݴ˴yֵжǷн ----  2010-09-18
				DPOINT2 _ptL1 = _ptLeft1;
				DPOINT2 _ptL2 = _ptLeft2;
				DPOINT2 _ptR1 = _ptRight1;
				DPOINT2 _ptR2 = _ptRight2;
				DPOINT2 _ptMid1;
				DPOINT2 _ptMid2;
				// ַͬxֵ ----  2010-09-18
				while (DOUBLE_GREAT(_ptR1.x, _ptL1.x))
				{
					double _nMidX = (_ptL1.x + _ptR1.x) / 2.0;
					int _n = Intersect_BeelineWithEllipseArc(DPOINT2(_nMidX, 0.0), 0.5 * c_nPIE,
						_ptC1, _a1, _b1, _nTheta1, _nStart1, _nSwept1, &_ptMid1);
					ASSERT(_n == 1);
					_n = Intersect_BeelineWithEllipseArc(DPOINT2(_nMidX, 0.0), 0.5 * c_nPIE,
						_ptC2, _a2, _b2, _nTheta2, _nStart2, _nSwept2, &_ptMid2);
					ASSERT(_n == 1);

					double _nAngleMid1 = GetTangentAngleOfEllipse(_ptC1, _a1, _b1, RegularizeIncludedRadian(_nTheta1), _ptMid1);
					MyNormalize(_nAngleMid1, _ptRight1.y - _ptLeft1.y);
					double _nAngleMid2 = GetTangentAngleOfEllipse(_ptC2, _a2, _b2, RegularizeIncludedRadian(_nTheta2), _ptMid2);
					MyNormalize(_nAngleMid2, _ptRight2.y - _ptLeft2.y);
					if (DOUBLE_EQU(_nAngleMid1, _nAngleMid2))
					{
						break;
					}
					else if ((_nAngleLeft1 - _nAngleLeft2) * (_nAngleMid1 - _nAngleMid2) > 0.0)
					{
						_ptL1 = _ptMid1;
						_ptL2 = _ptMid2;
					}
					else
					{
						_ptR1 = _ptMid1;
						_ptR2 = _ptMid2;
					}
				}
				ASSERT(DOUBLE_EQU(_ptMid1.x, _ptMid2.x));

				// ȴyֵжϽ ----  2010-09-18
				double _nDyMid = _ptMid1.y - _ptMid2.y;
				if (DOUBLE_EQU_ZERO(_nDyMid))
				{
					// yֵΪ㣬Ϊ
					if (_pIntersects[_nIntersect])
					{
						*_pIntersects[_nIntersect] = _ptMid1;
					}

					_nIntersect++;
				}
				else if (_nDyMid * _nDyLeft > 0.0)
				{
					// yֵͬţ޽
				}
				else
				{
					// yֵţ㣬߸һ
					if (_pIntersects[_nIntersect])
					{
						Intersect_MonotonousEllipseArcWithMonotonousEllipseArc_OnlyOne(EllipseArc1_, EllipseArc2_,
							_ptLeft1.x, _ptMid1.x, _pIntersects[_nIntersect]);
					}
					_nIntersect++;

					if (_pIntersects[_nIntersect])
					{
						Intersect_MonotonousEllipseArcWithMonotonousEllipseArc_OnlyOne(EllipseArc1_, EllipseArc2_,
							_ptMid1.x, _ptRight1.x, _pIntersects[_nIntersect]);
					}
					_nIntersect++;
				}
			}
#undef MyNormalize
		}

		ASSERT(_nIntersect >= 0 && _nIntersect <= 2);
		return _nIntersect;
	}
}

// ************************************************************************** //
// ԲԲ㡣
// 
// ********************  Version 2  *******************
// 취
//   (1)޳غڷغ²账
//   (2)Բֳ4ε(xy)תΪ4X4⣻
//   (3)xüüΪͬһx䡣
//   (4)ݻ˵㴦yֵϵбǷҽһ⣺yֵ˻Ϊ㣬
//      ֵн˻Ϊҽһ⣻˻Ϊ(5)һжϡ
//   (5)ݻ˵㴦yֵ(༴ǲ)ϵбǷ޽⣺ ǲ
//      Ǹ޽㣻(6)һжϡ
//   (6)ݻyֵ(ֵ)ͬyֵϵб(0,1,2) Զַ
//      yֵ(ֵ)ͬ㣬 yֵΪ㣬˴Ϊ㣻yֵ
//      ˵yֵͬţ޽㣻㣬ֱڴ˴ࡣ
//   (7)ϾΪб𣬶н䣬öַ֮
// עð취ŵ - ȷʧӦеĽ⣻ûйڸӵѧ㣬ͨ
// ֽΪһ⣻ӶȺܺãʹöֵӶΪlogN
// ޽ԿΪӾκӾ޽޽
//   㷨§ҵġͼϵͳʵԲ󽻿㷨оʵ֡ 
// ѧ ͵ ͼѧѧ 2000 3  μġ
//                                                        ----  2010-09-18
// 
// ********************  Version 1  *******************
// 취
//   (1)任任Բ1UVϵϣԲ1任Ϊ׼Բ
//   (2)Բ2ֳNݣNԲͨǷԲ1ͬбûǷ
//      Բ1нн򱣴
//   (3)нԲöֱƽ㡣
// עð취 = ַֻ + 㡣ֶΰ취ǶԲеĽǶֵTheta
// ֶ֣εѡ1ȣ180֮һȡ
//   ÷ȱǣ׶ʧн㡢Ͻ㣻 ŵǣ㷨򵥣
// Կɿ(Ƕֵ)
//   ð취ܿܶʧЧ㣬ҵǣĽ㶼ȷĽ㡣
//                                                        ----  2010-09-14
// ************************************************************************** //
int ncmath::Intersect_EllipseWithEllipse(
	IN const DPOINT2& ptCenter1_, IN double nURadius1_, IN double nVRadius1_, IN double nUAngle1_,
	IN const DPOINT2& ptCenter2_, IN double nURadius2_, IN double nVRadius2_, IN double nUAngle2_,
	OUT DPOINT2* pptIntersect1_/* = NULL*/,
	OUT DPOINT2* pptIntersect2_/* = NULL*/,
	OUT DPOINT2* pptIntersect3_/* = NULL*/,
	OUT DPOINT2* pptIntersect4_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptCenter1_.x) && IsValidDouble(ptCenter1_.y)
		&& nURadius1_ >= nVRadius1_ && DOUBLE_GREAT_ZERO(nVRadius1_)
		&& nUAngle1_ >= 0 && nUAngle1_ < 2.0 * c_nPIE);
	ASSERT(IsValidDouble(ptCenter2_.x) && IsValidDouble(ptCenter2_.y)
		&& nURadius2_ >= nVRadius2_ && DOUBLE_GREAT_ZERO(nVRadius2_)
		&& nUAngle2_ >= 0 && nUAngle2_ < 2.0 * c_nPIE);
	ASSERT(pptIntersect1_ && pptIntersect2_ && pptIntersect3_ && pptIntersect4_
		|| pptIntersect1_ && pptIntersect2_ && pptIntersect3_ && !pptIntersect4_
		|| pptIntersect1_ && pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_
		|| pptIntersect1_ && !pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_
		|| !pptIntersect1_ && !pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_);

	// ΪеĶ·˳˴ȶԳڲЧֵ ----  2010-09-15
	DPOINT2* _pptOuts[4];
	_pptOuts[0] = pptIntersect1_;
	_pptOuts[1] = pptIntersect2_;
	_pptOuts[2] = pptIntersect3_;
	_pptOuts[3] = pptIntersect4_;
	for (int _i = 0; _i < 4; _i++)
	{
		if (_pptOuts[_i])
		{
			*_pptOuts[_i] = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
		}
	}

	// Բʽغ ----  2010-09-14
	if (DOUBLE_EQU(ptCenter1_.x, ptCenter2_.x) && DOUBLE_EQU(ptCenter1_.y, ptCenter2_.y)
		&& DOUBLE_EQU(nURadius1_, nURadius2_) && DOUBLE_EQU(nVRadius1_, nVRadius2_)
		&& (DOUBLE_EQU(nUAngle1_, nUAngle2_) || DOUBLE_EQU(nUAngle1_ + c_nPIE, nUAngle2_) || DOUBLE_EQU(nUAngle1_, nUAngle2_ + c_nPIE)))
	{
		return -1;
	}

	// ˻ΪԲʽغ ----  2010-09-15
	if (DOUBLE_EQU(ptCenter1_.x, ptCenter2_.x) && DOUBLE_EQU(ptCenter1_.y, ptCenter2_.y)
		&& DOUBLE_EQU(nURadius1_, nVRadius1_) && DOUBLE_EQU(nURadius2_, nVRadius2_)
		&& DOUBLE_EQU(nURadius1_, nURadius2_))
	{
		return -1;
	}

	// Բຬ޽ ----  2010-09-16
	double _nCenterDist = (ptCenter2_ - ptCenter1_).GetLength();
	if (DOUBLE_LESS(nURadius1_ + nURadius2_, _nCenterDist)
		|| DOUBLE_LESS(_nCenterDist + nURadius1_, nVRadius2_)
		|| DOUBLE_LESS(_nCenterDist + nURadius2_, nVRadius1_))
	{
		return 0;
	}

	using namespace Intersect_EllipseWithEllipse_Implement;
	typedef Intersect_EllipseWithEllipse_Implement::Ellipse Ellipse;
	typedef Intersect_EllipseWithEllipse_Implement::EllipseArc EllipseArc;

	// ֽΪĶε ----  2010-09-18
	Ellipse _Ellipse1;
	_Ellipse1.ptCenter = ptCenter1_;
	_Ellipse1.nURadius = nURadius1_;
	_Ellipse1.nVRadius = nVRadius1_;
	_Ellipse1.nUAngle = nUAngle1_;
	Ellipse _Ellipse2;
	_Ellipse2.ptCenter = ptCenter2_;
	_Ellipse2.nURadius = nURadius2_;
	_Ellipse2.nVRadius = nVRadius2_;
	_Ellipse2.nUAngle = nUAngle2_;
	EllipseArc _Ellipse1Arcs[4];
	EllipseArc _Ellipse2Arcs[4];
	DivideEllipseIntoFourMonotonousArc(_Ellipse1,
		&_Ellipse1Arcs[0], &_Ellipse1Arcs[1],&_Ellipse1Arcs[2],&_Ellipse1Arcs[3]);
	DivideEllipseIntoFourMonotonousArc(_Ellipse2,
		&_Ellipse2Arcs[0], &_Ellipse2Arcs[1],&_Ellipse2Arcs[2],&_Ellipse2Arcs[3]);

	//  ----  2010-09-18
	std::vector<DPOINT2> _vectorIntersects;
	for (int _i = 0; _i < 4; _i++)
	{
		for (int _j = 0; _j < 4; _j++)
		{
			DPOINT2 _ptIntersect1;
			DPOINT2 _ptIntersect2;
			int _n = Intersect_MonotonousEllipseArcWithMonotonousEllipseArc(
				_Ellipse1Arcs[_i], _Ellipse2Arcs[_j], &_ptIntersect1, &_ptIntersect2);

			if (_n >= 1)
			{
				_vectorIntersects.push_back(_ptIntersect1);
			}

			if (_n >= 2)
			{
				_vectorIntersects.push_back(_ptIntersect2);
			}
		}
	}

	// ָʱڽ㴦ָ ----  2010-09-18
	for (std::vector<DPOINT2>::iterator _it = _vectorIntersects.begin();
		_it != _vectorIntersects.end(); _it++)
	{
		for (std::vector<DPOINT2>::iterator _itTemp = _it + 1;
			_itTemp != _vectorIntersects.end(); )
		{
			DPOINT2& _pt1 = *_it;
			DPOINT2& _pt2 = *_itTemp;
			if (DOUBLE_EQU(_pt1.x, _pt2.x) && DOUBLE_EQU(_pt1.y, _pt2.y))
			{
				_itTemp = _vectorIntersects.erase(_itTemp);
			}
			else
			{
				_itTemp++;
			}
		}
	}
	ASSERT(_vectorIntersects.size() <= 4);

	//  ----  2010-09-18
	for (UINT _i = 0; _i < 4; _i++)
	{
		if (_pptOuts[_i])
		{
			if (_vectorIntersects.size() > _i)
			{
				*_pptOuts[_i] = _vectorIntersects[_i];
			}
			else
			{
				_pptOuts[_i]->x = c_nINVALID_DOUBLE;
				_pptOuts[_i]->y = c_nINVALID_DOUBLE;
			}
		}
	}

	// ؽ ----  2010-09-18
	return (int)(_vectorIntersects.size());
}

int ncmath::Intersect_EllipseWithEllipse(
	IN const DPOINT2& ptCenter1_, IN double nURadius1_, IN double nVRadius1_, IN double nUAngle1_,
	IN const DPOINT2& ptCenter2_, IN double nURadius2_, IN double nVRadius2_, IN double nUAngle2_,
	OUT std::vector<DPOINT2>& vectorIntersect_,
	IN int nIntersectsNeed_/* = 4*/)
{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	//  ----  2010-09-14
	DPOINT2 _ptIntersects[4];
	// Ҫٸ ----  2010-09-14
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C ----  2010-09-14
	int _nRet = Intersect_EllipseWithEllipse(
		ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_,
		ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_,
		_nNeed >= 1 ? &_ptIntersects[0] : NULL,
		_nNeed >= 2 ? &_ptIntersects[1] : NULL,
		_nNeed >= 3 ? &_ptIntersects[2] : NULL,
		_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ ----  2010-09-14
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}

// ͨԲԲ ----  2010-09-14
int ncmath::Intersect_EllipseWithEllipseArc(
	IN const DPOINT2& ptCenter1_, IN double nURadius1_, IN double nVRadius1_, IN double nUAngle1_,
	IN const DPOINT2& ptCenter2_, IN double nURadius2_, IN double nVRadius2_, IN double nUAngle2_,
	IN double nStartAngle_, IN double nSweptAngle_,
	OUT DPOINT2* pptIntersect1_/* = NULL*/,
	OUT DPOINT2* pptIntersect2_/* = NULL*/,
	OUT DPOINT2* pptIntersect3_/* = NULL*/,
	OUT DPOINT2* pptIntersect4_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptCenter1_.x) && IsValidDouble(ptCenter1_.y)
		&& nURadius1_ >= nVRadius1_ && DOUBLE_GREAT_ZERO(nVRadius1_)
		&& nUAngle1_ >= 0 && nUAngle1_ < 2.0 * c_nPIE);
	ASSERT(IsValidDouble(ptCenter2_.x) && IsValidDouble(ptCenter2_.y)
		&& nURadius2_ >= nVRadius2_ && DOUBLE_GREAT_ZERO(nVRadius2_)
		&& nUAngle2_ >= 0 && nUAngle2_ < 2.0 * c_nPIE
		&& nStartAngle_ >= 0 && nStartAngle_ < 2.0 * c_nPIE
		&& DOUBLE_GREAT_ZERO(nSweptAngle_) && nSweptAngle_ < 2.0 * c_nPIE);
	ASSERT(pptIntersect1_ && pptIntersect2_ && pptIntersect3_ && pptIntersect4_
		|| pptIntersect1_ && pptIntersect2_ && pptIntersect3_ && !pptIntersect4_
		|| pptIntersect1_ && pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_
		|| pptIntersect1_ && !pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_
		|| !pptIntersect1_ && !pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_);

	// Intersect_EllipseWithEllipseԲĽ ----  2010-09-14
	DPOINT2 _ptIntersects[4];
	int _nRet = Intersect_EllipseWithEllipse(
		ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_,
		ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_,
		&_ptIntersects[0],
		&_ptIntersects[1],
		&_ptIntersects[2],
		&_ptIntersects[3]);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// Բ˵ ----  2010-09-25
	DPOINT2 _ptStart = GetPointInEllipse(ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle_);
	DPOINT2 _ptEnd = GetPointInEllipse(ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, RegularizeRadian(nStartAngle_ + nSweptAngle_));

	// жĽǷԲϣ ----  2010-09-14
	DPOINT2* _pptOuts[4];
	_pptOuts[0] = pptIntersect1_;
	_pptOuts[1] = pptIntersect2_;
	_pptOuts[2] = pptIntersect3_;
	_pptOuts[3] = pptIntersect4_;
	int _nIntersect = 0;
	for (int _i = 0; _i < _nRet && _i < 4; _i++)
	{
		// _i ----  2010-09-14
		DPOINT2 _ptCurIntersect = _ptIntersects[_i];

		// ýԲĵ ----  2010-09-14
		DPOINT2 _vec = _ptCurIntersect - ptCenter2_;
		double _nAngle = atan2(_vec.y, _vec.x);

		// ǶԲĳļн ----  2010-09-14
		_nAngle = _nAngle - nUAngle2_;

		// ǶԲʼǵļн ----  2010-09-14
		_nAngle = _nAngle - nStartAngle_;

		// Բ ----  2010-09-14
		NormalizeRadian(_nAngle);
		if (_nAngle >= 0.0 && _nAngle <= nSweptAngle_
			|| DOUBLE_EQU_ZERO((_ptCurIntersect - _ptStart).GetLength())
			|| DOUBLE_EQU_ZERO((_ptCurIntersect - _ptEnd).GetLength()))
		{
			if (_pptOuts[_nIntersect])
			{
				*_pptOuts[_nIntersect] = _ptCurIntersect;
			}

			_nIntersect++;
		}
	}

	// ĳڲЧֵ ----  2010-09-15
	for (int _i = _nIntersect; _i < 4; _i++)
	{
		if (_pptOuts[_i])
		{
			*_pptOuts[_i] = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
		}
	}

	return _nIntersect;
}

int ncmath::Intersect_EllipseWithEllipseArc(
	IN const DPOINT2& ptCenter1_, IN double nURadius1_, IN double nVRadius1_, IN double nUAngle1_,
	IN const DPOINT2& ptCenter2_, IN double nURadius2_, IN double nVRadius2_, IN double nUAngle2_,
	IN double nStartAngle_, IN double nSweptAngle_,
	OUT std::vector<DPOINT2>& vectorIntersect_,
	IN int nIntersectsNeed_/* = 4*/)
{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	//  ----  2010-09-14
	DPOINT2 _ptIntersects[4];
	// Ҫٸ ----  2010-09-14
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C ----  2010-09-14
	int _nRet = Intersect_EllipseWithEllipseArc(
		ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_,
		ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle_, nSweptAngle_,
		_nNeed >= 1 ? &_ptIntersects[0] : NULL,
		_nNeed >= 2 ? &_ptIntersects[1] : NULL,
		_nNeed >= 3 ? &_ptIntersects[2] : NULL,
		_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ ----  2010-09-14
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}

int ncmath::Intersect_EllipseArcWithEllipseArc(
	IN const DPOINT2& ptCenter1_, IN double nURadius1_, IN double nVRadius1_, IN double nUAngle1_,
	IN double nStartAngle1_, IN double nSweptAngle1_,
	IN const DPOINT2& ptCenter2_, IN double nURadius2_, IN double nVRadius2_, IN double nUAngle2_,
	IN double nStartAngle2_, IN double nSweptAngle2_,
	OUT DPOINT2* pptIntersect1_/* = NULL*/,
	OUT DPOINT2* pptIntersect2_/* = NULL*/,
	OUT DPOINT2* pptIntersect3_/* = NULL*/,
	OUT DPOINT2* pptIntersect4_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptCenter1_.x) && IsValidDouble(ptCenter1_.y)
		&& nURadius1_ >= nVRadius1_ && DOUBLE_GREAT_ZERO(nVRadius1_)
		&& nUAngle1_ >= 0 && nUAngle1_ < 2.0 * c_nPIE
		&& nStartAngle1_ >= 0 && nStartAngle1_ < 2.0 * c_nPIE
		&& DOUBLE_GREAT_ZERO(nSweptAngle1_) && nSweptAngle1_ < 2.0 * c_nPIE);
	ASSERT(IsValidDouble(ptCenter2_.x) && IsValidDouble(ptCenter2_.y)
		&& nURadius2_ >= nVRadius2_ && DOUBLE_GREAT_ZERO(nVRadius2_)
		&& nUAngle2_ >= 0 && nUAngle2_ < 2.0 * c_nPIE
		&& nStartAngle2_ >= 0 && nStartAngle2_ < 2.0 * c_nPIE
		&& DOUBLE_GREAT_ZERO(nSweptAngle2_) && nSweptAngle2_ < 2.0 * c_nPIE);
	ASSERT(pptIntersect1_ && pptIntersect2_ && pptIntersect3_ && pptIntersect4_
		|| pptIntersect1_ && pptIntersect2_ && pptIntersect3_ && !pptIntersect4_
		|| pptIntersect1_ && pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_
		|| pptIntersect1_ && !pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_
		|| !pptIntersect1_ && !pptIntersect2_ && !pptIntersect3_ && !pptIntersect4_);

	// Intersect_EllipseWithEllipseԲĽ ----  2010-09-14
	DPOINT2 _ptIntersects[4];
	int _nRet = Intersect_EllipseWithEllipse(
		ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_,
		ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_,
		&_ptIntersects[0],
		&_ptIntersects[1],
		&_ptIntersects[2],
		&_ptIntersects[3]);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	DPOINT2* _pptOuts[4];
	_pptOuts[0] = pptIntersect1_;
	_pptOuts[1] = pptIntersect2_;
	_pptOuts[2] = pptIntersect3_;
	_pptOuts[3] = pptIntersect4_;
	int _nIntersect = 0;

	if (_nRet == -1)
	{
		// ԲغرۣΪΪԲԲغϲƳ
		// Բһغϡ                                  ----  2010-09-25
		DPOINT2 _ptStart1 = GetPointInEllipse(ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_, nStartAngle1_);
		DPOINT2 _ptEnd1 = GetPointInEllipse(ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_,
			RegularizeRadian(nStartAngle1_ + nSweptAngle1_));
		DPOINT2 _ptStart2 = GetPointInEllipse(ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_);
		DPOINT2 _ptEnd2 = GetPointInEllipse(ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_,
			RegularizeRadian(nStartAngle2_ + nSweptAngle2_));
		if (!IsPointInEllipseArc(_ptStart2, ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_, nStartAngle1_, nSweptAngle1_)
			&& !IsPointInEllipseArc(_ptEnd2, ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_, nStartAngle1_, nSweptAngle1_)
			&& !IsPointInEllipseArc(_ptStart1, ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_, nSweptAngle2_)
			&& !IsPointInEllipseArc(_ptEnd1, ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_, nSweptAngle2_))
		{
			// ޽
			_nIntersect = 0;
		}
		else if (DOUBLE_EQU_ZERO((_ptStart1 - _ptEnd2).GetLength())
			&& DOUBLE_EQU_ZERO((_ptEnd1 - _ptStart2).GetLength()))
		{
			// 
			_nIntersect = 2;

			if (_pptOuts[0])
				*_pptOuts[0] = _ptStart1;

			if (_pptOuts[1])
				*_pptOuts[1] = _ptEnd1;
		}
		else if (DOUBLE_EQU_ZERO((_ptStart1 - _ptEnd2).GetLength())
			&& !IsPointInEllipseArc(_ptEnd1, ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_, nSweptAngle2_))
		{
			// 1ͷ2β
			_nIntersect = 1;

			if (_pptOuts[0])
				*_pptOuts[0] = _ptStart1;
		}
		else if (DOUBLE_EQU_ZERO((_ptEnd1 - _ptStart2).GetLength())
			&& !IsPointInEllipseArc(_ptStart1, ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_, nSweptAngle2_))
		{
			// 1β2ͷ
			_nIntersect = 1;

			if (_pptOuts[0])
				*_pptOuts[0] = _ptEnd1;
		}
		else
		{
			// غ
			_nIntersect = -1;
		}
	}
	else
	{
		// жĽǷͬʱԲ1Բ2ϣ ----  2010-09-14
		ASSERT(_nRet >= 0 && _nRet <= 4);
		for (int _i = 0; _i < _nRet && _i < 4; _i++)
		{
			// _i ----  2010-09-14
			DPOINT2 _pt = _ptIntersects[_i];
			ASSERT(IsValidDouble(_pt.x) && IsValidDouble(_pt.y));
			ASSERT(IsPointInEllipse(_pt, ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_)
				&& IsPointInEllipse(_pt, ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_));

			// ͬʱԲ ----  2010-09-14
			if (IsPointInEllipseArc(_pt, ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_, nStartAngle1_, nSweptAngle1_)
				&& IsPointInEllipseArc(_pt, ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_, nSweptAngle2_))
			{
				if (_pptOuts[_nIntersect])
				{
					*_pptOuts[_nIntersect] = _pt;
				}

				_nIntersect++;
			}
		}
	}

	// ĳڲЧֵ ----  2010-09-15
	for (int _i = _nIntersect; _i < 4; _i++)
	{
		if ((_i >= 0 && _i < 4) && _pptOuts[_i])
		{
			*_pptOuts[_i] = DPOINT2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
		}
	}

	return _nIntersect;
}

int ncmath::Intersect_EllipseArcWithEllipseArc(
	IN const DPOINT2& ptCenter1_, IN double nURadius1_, IN double nVRadius1_, IN double nUAngle1_,
	IN double nStartAngle1_, IN double nSweptAngle1_,
	IN const DPOINT2& ptCenter2_, IN double nURadius2_, IN double nVRadius2_, IN double nUAngle2_,
	IN double nStartAngle2_, IN double nSweptAngle2_,
	OUT std::vector<DPOINT2>& vectorIntersect_,
	IN int nIntersectsNeed_/* = 4*/)
{
	ASSERT(vectorIntersect_.empty());
	ASSERT(nIntersectsNeed_ >= -1);

	//  ----  2010-09-14
	DPOINT2 _ptIntersects[4];
	// Ҫٸ ----  2010-09-14
	int _nNeed = (nIntersectsNeed_ < 0 || nIntersectsNeed_ >= 4) ? 4 : nIntersectsNeed_;

	// C ----  2010-09-14
	int _nRet = Intersect_EllipseArcWithEllipseArc(
		ptCenter1_, nURadius1_, nVRadius1_, nUAngle1_, nStartAngle1_, nSweptAngle1_,
		ptCenter2_, nURadius2_, nVRadius2_, nUAngle2_, nStartAngle2_, nSweptAngle2_,
		_nNeed >= 1 ? &_ptIntersects[0] : NULL,
		_nNeed >= 2 ? &_ptIntersects[1] : NULL,
		_nNeed >= 3 ? &_ptIntersects[2] : NULL,
		_nNeed >= 4 ? &_ptIntersects[3] : NULL);
	ASSERT(_nRet >= -1 && _nRet <= 4);

	// ռ ----  2010-09-14
	vectorIntersect_.clear();
	for (int _i = 0; _i < _nRet && _i < _nNeed; _i++)
	{
		ASSERT(IsValidDouble(_ptIntersects[_i].x) && IsValidDouble(_ptIntersects[_i].y));
		vectorIntersect_.push_back(_ptIntersects[_i]);
	}

	return _nRet;
}
