// ************************************************************************** //
// MicroJointTarget.cpp : implementation of the CMicroJointTarget class
// Copyright(c)2016 WeiHong Company. All rights reserved.
//
// Abstract:
//   ļʵCMicroJointTargetࡣ
// *********************************************** ---- ǿ 2016-03-09***** //
#include "stdafx.h"
#include "IntroPosCalculator.h"
#include "ExplodeMicrJointAction.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

using namespace nce;
using namespace std;
using namespace ncmath;

CS_T s_szManualSetFailed = _DEFCS(_CSV(16347),
	_CHINESE("|E|˴΢㣬ѡ!")
	_ENGLISH("|E|This place can't set microjoint,Please select again!"));
CS_T s_szNoSelectedObject = _DEFCS(_CSV(16347),
	_CHINESE("|E|ûѡжԶ΢ʧܣ")
	_ENGLISH("|E|No seleceted objects, Auto micro-joint failed!"));
CS_T s_szAutoSetFailed = _DEFCS(_CSV(16347),
	_CHINESE("|E|ûпԶ΢ͼΣԶ΢ʧܣ")
	_ENGLISH("|E|No object can micro joint,Auto micro-joint failed!"));

const CHAR c_szMJ[] = "MJ%d";
const CHAR c_szMJFormat[] = "(%d,%lf,%lf)";
const CHAR c_szScanCamGXXNodes[] = "ScanCamGxxNodes";

IMPLEMENT_DYNAMIC(CMicroJointTarget, CMouseMsgTarget)
BEGIN_MESSAGE_MAP(CMicroJointTarget, CMouseMsgTarget)
	ON_COMMAND(IDCMD_MICROJOINTTARGET_MICROJOINT, OnMicroJoint)
	ON_UPDATE_COMMAND_UI(IDCMD_MICROJOINTTARGET_MICROJOINT, OnUpdateMicroJoint)
	ON_COMMAND(IDCMD_MICROJOINTTARGET_CLEARMICROJOINT, OnClearMicroJoint)
	ON_UPDATE_COMMAND_UI(IDCMD_MICROJOINTTARGET_CLEARMICROJOINT, OnUpdateClearMicroJoint)
	ON_COMMAND(IDCMD_MICROJOINTTARGET_EXPLODEMICROJOINT, OnExplodeMicroJoint)
	ON_UPDATE_COMMAND_UI(IDCMD_MICROJOINTTARGET_EXPLODEMICROJOINT, OnUpdateExplodeMicroJoint)
	ON_TARGET_NOTIFY(TN_CADDIR_CHANGE, OnCadDirChanged)
END_MESSAGE_MAP()

static bool FindObjectByID(IN CCadObject* pCadObj_, IN int nID_, OUT CCadObject** ppResult_);
static double GetCadLength(IN const CCadObject* pCadObj_);

CMicroJointTarget* CMicroJointTarget::ms_pMicroJointTarget = NULL;
CMicroJointTarget* GetMicroJointTarget()
{
	return CMicroJointTarget::ms_pMicroJointTarget;
}

CMicroJointTarget::CMicroJointTarget()
{
	ASSERT(ms_pMicroJointTarget == NULL);
	ms_pMicroJointTarget = this;
	m_nMicroJointLength = 1;
	m_bNoPause = FALSE;
	LOAD_PARAM(this,CMicroJointTarget);
	Reset();
	m_hCursor = AfxGetApp()->LoadCursor(IDC_CUR_ADDBRIDGE);
}

CMicroJointTarget::~CMicroJointTarget()
{
	SAVE_PARAM(this, CMicroJointTarget);
}

CCadObject* CMicroJointTarget::CreateToolPath(const nce::CCadObject* pCadObject_)
{
	CStringA _strKey;
	_strKey.Format(c_szMJ, 0);
	if (pCadObject_->GetKey(_strKey).IsEmpty())
	{
		return NULL;
	}

	switch (pCadObject_->GetType())
	{
	case cadline:
		return m_bNoPause ? _CreateLineToolPath2((CCadLine*)pCadObject_) : _CreateLineToolPath((CCadLine*)pCadObject_);
		break;
	case cadcircle:
		return m_bNoPause ? _CreateCircleToolPath2((CCadCircle*)pCadObject_) : _CreateCircleToolPath((CCadCircle*)pCadObject_);
		break;
	case cadarc:
		return m_bNoPause ? _CreateArcToolPath2((CCadArc*)pCadObject_) : _CreateArcToolPath((CCadArc*)pCadObject_);
		break;
	case cadellipse:
		return m_bNoPause ? _CreateEllipseToolPath2((CCadEllipse*)pCadObject_) : _CreateEllipseToolPath((CCadEllipse*)pCadObject_);
		break;
	case cadellipsearc:
		return m_bNoPause ? _CreateEllipseArcToolPath2((CCadEllipseArc*)pCadObject_) : _CreateEllipseArcToolPath((CCadEllipseArc*)pCadObject_);
		break;
	case cadpath:
	case cadpolygon:
	case cadpolyline:
	case cadstar:
	case cadrectangle:
		return m_bNoPause ? _CreateMultiSegmentToolPath2((CCadMultiSegments*)pCadObject_) : _CreateMultiSegmentToolPath((CCadMultiSegments*)pCadObject_);
		break;
	}

	return NULL;
}

CCadObject* CMicroJointTarget::_CreateLineToolPath(const nce::CCadLine* pCadLine_)
{
	CCadGroup* _pCadGroup = new CCadGroup;

	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadLine_->GetKey(_strKeyMJ);
	DPOINT2 _ptStartPoint = pCadLine_->GetStartPoint();

	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		DPOINT2 _ptEndPoint = (pCadLine_->GetEndPoint() - pCadLine_->GetStartPoint()) * _nRatio + pCadLine_->GetStartPoint();

		if (_ptStartPoint != _ptEndPoint)
		{
			CCadLine* _pCadLine = new CCadLine(_ptStartPoint, _ptEndPoint);
			_pCadGroup->PushBack(_pCadLine);
		}

		_ptStartPoint = pCadLine_->GetStartPoint() + (pCadLine_->GetEndPoint() - pCadLine_->GetStartPoint()) * (_nRatio + _nLength / pCadLine_->GetLength());

		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadLine_->GetKey(_strKeyMJ);
	}

	if (_ptStartPoint != pCadLine_->GetEndPoint())
	{
		_pCadGroup->PushBack(new CCadLine(_ptStartPoint, pCadLine_->GetEndPoint()));
	}

	if (_pCadGroup->GetNumOfChild() == 0)
	{
		delete _pCadGroup;
		_pCadGroup = NULL;
	}
	return _pCadGroup;
}

CCadObject* CMicroJointTarget::_CreateCircleToolPath(const CCadCircle* pCadCircle_)
{
	CCadGroup* _pCadGroup = new CCadGroup;

	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadCircle_->GetKey(_strKeyMJ);
	double _nStartAngle = 0;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));

		double _nCenterAngle = 2 * c_nPIE * _nRatio - _nStartAngle;
		if (!DOUBLE_EQU_ZERO(_nCenterAngle))
		{
			//޸˳ʱ£΢Բӹ켣ȷ޸----2017-4-27 
			if(!pCadCircle_->IsNormalDir())
			{
				CCadArc* _pCadArc = new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),
					_nStartAngle+_nCenterAngle, _nCenterAngle, pCadCircle_->IsNormalDir());
				_pCadGroup->PushFront(_pCadArc);
			}
			else
			{
				CCadArc* _pCadArc = new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),
					_nStartAngle, _nCenterAngle, pCadCircle_->IsNormalDir());
				_pCadGroup->PushBack(_pCadArc);
			}
		}

		_nStartAngle = _nStartAngle + _nCenterAngle + (_nLength / pCadCircle_->GetLength()) * 2 * c_nPIE;

		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadCircle_->GetKey(_strKeyMJ);
	}

	if (DOUBLE_LESS(_nStartAngle, 2 * c_nPIE))
	{
		//޸˳ʱ£΢Բӹ켣ȷ޸----2017-4-27 
		if(!pCadCircle_->IsNormalDir())
		{
			_pCadGroup->PushFront(new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),0, 2 * c_nPIE - _nStartAngle, pCadCircle_->IsNormalDir()));
		}
		else
		{
			_pCadGroup->PushBack(new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),_nStartAngle, 2 * c_nPIE - _nStartAngle, pCadCircle_->IsNormalDir()));
		}
	}

	if (_pCadGroup->GetNumOfChild() == 0)
	{
		delete _pCadGroup;
		_pCadGroup = NULL;
	}
	return _pCadGroup;
}

CCadObject* CMicroJointTarget::_CreateArcToolPath(const nce::CCadArc* pCadArc_)
{
	CCadGroup* _pCadGroup = new CCadGroup;

	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadArc_->GetKey(_strKeyMJ);
	double _nStartAngle = pCadArc_->GetStartAngle();
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		double _nCenterAngle = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		if (pCadArc_->IsNormalDir())
		{
			_nCenterAngle = (pCadArc_->GetCentralAngle() * _nRatio + pCadArc_->GetStartAngle()) - _nStartAngle;
		}
		//˳ʱ뷽򻡶ȼ,֮ǰʱ뷽򷴹㣬ΪstartangleǴӴСı仯
		else
		{
			_nCenterAngle = pCadArc_->GetCentralAngle() * _nRatio - (pCadArc_->GetStartAngle() - _nStartAngle);
		}

		NormalizeRadian(_nCenterAngle);
		if (!DOUBLE_EQU_ZERO(_nCenterAngle))
		{
				CCadArc* _pCadArc = new CCadArc(pCadArc_->GetCenterPoint(), pCadArc_->GetRadius(),
					_nStartAngle, _nCenterAngle, pCadArc_->IsNormalDir());
					_pCadGroup->PushBack(_pCadArc);
		}
		if(pCadArc_->IsNormalDir())
		{
			_nStartAngle = _nStartAngle + _nCenterAngle + (_nLength / pCadArc_->GetLength()) * pCadArc_->GetCentralAngle();
		}
		//Ϊ˳ʱ뷽Ƕ𽥼СһǶӦһǶȼȥӦƫƽǶ
		else
		{
			_nStartAngle = _nStartAngle - _nCenterAngle - (_nLength / pCadArc_->GetLength()) * pCadArc_->GetCentralAngle();
		}

		NormalizeRadian(_nStartAngle);

		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadArc_->GetKey(_strKeyMJ);
	}

		if (!DOUBLE_EQU(_nStartAngle, pCadArc_->GetEndAngle()))
		{
			double _nCenterAngle = 0.;
			if (pCadArc_->IsNormalDir())
			{
				_nCenterAngle = pCadArc_->GetEndAngle() - _nStartAngle;
			}
			else
			{
				_nCenterAngle = _nStartAngle- pCadArc_->GetEndAngle();
			}
			
			if (_nCenterAngle < 0.)
			{
				_nCenterAngle += 2 * c_nPIE;
			}
			_pCadGroup->PushBack(new CCadArc(pCadArc_->GetCenterPoint(), pCadArc_->GetRadius(),_nStartAngle, _nCenterAngle, pCadArc_->IsNormalDir()));
		}

	if (_pCadGroup->GetNumOfChild() == 0)
	{
		delete _pCadGroup;
		_pCadGroup = NULL;
	}
	return _pCadGroup;
}

CCadObject* CMicroJointTarget::_CreateMultiSegmentToolPath(const nce::CCadMultiSegments* pCadMultiSegments_)
{
	CCadGroup* _pCadGroup = new CCadGroup;
	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadMultiSegments_->GetKey(_strKeyMJ);
	DPOINT2 _ptStartPoint = pCadMultiSegments_->GetStartPoint();
	int _nSegmentCount = pCadMultiSegments_->GetSegmentCount(); // ߵܶ ---- yangxiaobin 2017-05-16
	int _nTailSegCount = 0;
	double _nTaliRatio = 0.;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));

		if (_nSeg >= _nSegmentCount)
		{
			TRACE("EXCEPTION: MJ %s out of _nSegmentCount=%d at CMicroJointTarget:%d.\n", _strValueTemp, _nSegmentCount, __LINE__);
			// ASSERT(_nSeg < _nSegmentCount);
			break;
		}

		if(_nTailSegCount >= _nSegmentCount)
		{
			TRACE("EXCEPTION: MJ %s out of _nSegmentCount=%d at CMicroJointTarget:%d.\n", _strValueTemp, _nSegmentCount, __LINE__);
			// ASSERT(_nTailSegCount < _nSegmentCount);
			break;
		}

		CCadPolyline* _pCadPolyline = new CCadPolyline;
		_pCadPolyline->SetStartPoint(_ptStartPoint);
		while(_nTailSegCount < _nSeg)
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
			// ͷǶԲ߶ι̣Ϊ˷ǾֱΪⲿֲڣ㵽֮ǰ΢Ĳ
			if (_Segment.GetLength() * (1 - _nTaliRatio) > 1e-6)
			{
				if (_Segment.nType == SEGMENT::LINE)
				{
					_pCadPolyline->AddLine(_Segment.ptEnd);
				}
				else if (_Segment.nType == SEGMENT::ARC)
				{
					DPOINT2 _ptCenter;
					double _nRadius,_nStartAngle,_nSweptAngele;
					SEGMENT::SolveArc(_Segment.ptStart,
						_Segment.ptEnd,
						_Segment.nBulge,
						&_ptCenter,
						&_nRadius,
						&_nStartAngle,
						&_nSweptAngele);

					// ʵʼӹֹƳȲʣԲ뾶Ȼܳ͹1e-10
					// Ҫ͹ȣ㹻С͵߶δ
					if (abs(tan((1 - _nTaliRatio) * _nSweptAngele / 4)) > 1e-9)
					{
						_pCadPolyline->AddArc(_Segment.ptEnd, tan((1 - _nTaliRatio) * _nSweptAngele / 4));
					}
					else
					{
						_pCadPolyline->AddLine(_Segment.ptEnd);
					}
				}
				else
				{
					ASSERT(FALSE);
				}
			}
			// ̣߶ΣֱӼȥǻֱӱΪ߶μȥ,ȡСԲ
			else
			{
				_pCadPolyline->AddLine(_Segment.ptEnd);

			}
			_nTaliRatio = 0;
			_nTailSegCount++;
		}
		SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
		if (_Segment.nType == SEGMENT::LINE)
		{
			_pCadPolyline->AddLine(_Segment.GetPoint(_nRatio));
		}
		else if (_Segment.nType == SEGMENT::ARC)
		{
			DPOINT2 _ptCenter;
			double _nRadius,_nStartAngle,_nSweptAngele;
			SEGMENT::SolveArc(_Segment.ptStart,
				_Segment.ptEnd,
				_Segment.nBulge,
				&_ptCenter,
				&_nRadius,
				&_nStartAngle,
				&_nSweptAngele);

			_pCadPolyline->AddArc(_Segment.GetPoint(_nRatio), tan(_nSweptAngele * (_nRatio - _nTaliRatio) / 4));
		}
		else
		{
			ASSERT(FALSE);
		}

		if (_pCadPolyline->GetNodeCount() <= 0)
		{
			delete _pCadPolyline;
			_pCadPolyline = NULL;
		}
		else
		{
			_pCadGroup->PushBack(_pCadPolyline);
		}
		_nTaliRatio = _nRatio;

		double _nNeedLength = _Segment.GetLength() * _nRatio + _nLength;
		while(DOUBLE_GREAT(_nLength, 0) && _nTailSegCount < pCadMultiSegments_->GetSegmentCount())
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
			double _nSegLength =_Segment.GetLength();
			if (_nSegLength > _nNeedLength)
			{
				_ptStartPoint =_Segment.GetPoint(_nNeedLength / _nSegLength);
				_nTaliRatio = _nNeedLength / _nSegLength;
				break;
			}

			_nNeedLength -= _nSegLength;
			_nTailSegCount++;
		}

		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadMultiSegments_->GetKey(_strKeyMJ);
	}

	if (_ptStartPoint != pCadMultiSegments_->GetEndPoint()
		&& _nTailSegCount < pCadMultiSegments_->GetSegmentCount())
	{
		CCadPolyline* _pCadPolyine = new CCadPolyline;
		_pCadPolyine->SetStartPoint(pCadMultiSegments_->GetSegment(_nTailSegCount).GetPoint(_nTaliRatio));
		bool _bFirst = true;
		while(_nTailSegCount < pCadMultiSegments_->GetSegmentCount())
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);

			// һһ΢Ϻʣµı߽
			if (_bFirst)
			{
				// γȲܺԣʵʵӽȥ
				if(_Segment.GetLength() * (1 - _nTaliRatio) > 1e-6)
				{
					if (_Segment.nType == SEGMENT::LINE)
					{
						_pCadPolyine->AddLine(_Segment.ptEnd);
					}
					else if (_Segment.nType == SEGMENT::ARC)
					{
						DPOINT2 _ptCenter;
						double _nRadius,_nStartAngle,_nSweptAngele;
						SEGMENT::SolveArc(_Segment.ptStart,
							_Segment.ptEnd,
							_Segment.nBulge,
							&_ptCenter,
							&_nRadius,
							&_nStartAngle,
							&_nSweptAngele);
						// ʵʼӹֹƳȲʣԲ뾶Ȼܳ͹1e-10
						// Ҫ͹ȣ㹻С͵߶δ
						if (abs(tan(_nSweptAngele * (1-_nTaliRatio) / 4)) > 1e-9)
						{
							_pCadPolyine->AddArc(_Segment.ptEnd, tan(_nSweptAngele * (1-_nTaliRatio) / 4));
						}
						else
						{
							_pCadPolyine->AddLine(_Segment.ptEnd);
						}
					}
					else
					{
						ASSERT(FALSE);
					}
				}
				else
				{
					//ͬϣ̣߶ΣֱӼȥǻֱӱΪ߶μȥ
					_pCadPolyine->AddLine(_Segment.ptEnd);
				}
				_bFirst = false;
			}
			
			// ľֱӰμӽȥ
			else
			{
				if (_Segment.nType == SEGMENT::LINE)
				{
					_pCadPolyine->AddLine(_Segment.ptEnd);
				}
				else if (_Segment.nType == SEGMENT::ARC)
				{
					_pCadPolyine->AddArc(_Segment.ptEnd, _Segment.nBulge);
				}
				else
				{
					ASSERT(FALSE);
				}
			}

			_nTailSegCount++;
		}
		if (_pCadPolyine->GetNodeCount() <= 0)
		{
			delete _pCadPolyine;
			_pCadPolyine = NULL;
		}
		else
		{
			_pCadGroup->PushBack(_pCadPolyine);
		}
	}

	if (_pCadGroup->GetNumOfChild() == 0)
	{
		delete _pCadGroup;
		_pCadGroup = NULL;
	}
	return _pCadGroup;
}


CCadObject* CMicroJointTarget::_CreateEllipseToolPath(const CCadEllipse* pCadEllipse_)
{
	CCadPolyline* _pCadPolyline = TranslateObjectToPolyline(pCadEllipse_);
	_pCadPolyline->SetParamString(pCadEllipse_->GetParamString());
	CCadMultiSegments* _pCadMultiSegments = (CCadMultiSegments*)(_pCadPolyline);
	// ԲıǰԲһģҲǴʣתɶ
	// ߵʱҪתɶĶκźͱ ---- DingQiang 2016-04-27
	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadEllipse_->GetKey(_strKeyMJ);
	DPOINT2 _ptStartPoint = _pCadMultiSegments->GetStartPoint();
	int _nTailSegCount = 0;
	double _nTaliRatio = 0.;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		double _nTargetLength = _pCadMultiSegments->GetLength() * _nRatio;
		double _nLengthTemp = 0.;
		int _nSegTemp = 0;
		while(_nLengthTemp < _nTargetLength && _nSegTemp < _pCadMultiSegments->GetSegmentCount())
		{
			SEGMENT _segment = _pCadMultiSegments->GetSegment(_nSegTemp);
			if (_nLengthTemp + _segment.GetLength() >= _nTargetLength)
			{
				double _nRatio = (_nTargetLength - _nLengthTemp) / _segment.GetLength();
				_strValueTemp.Format(c_szMJFormat, _nSegTemp, _nRatio, _nLength);
				_pCadMultiSegments->SetKey(_strKeyMJ, _strValueTemp);
				break;
			}
			_nLengthTemp += _segment.GetLength();
			_nSegTemp++;
		}
		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = _pCadMultiSegments->GetKey(_strKeyMJ);
	}

	CCadObject* _pObjectRet = _CreateMultiSegmentToolPath(_pCadMultiSegments);
	delete _pCadPolyline;
	_pCadPolyline = NULL;
	return _pObjectRet;
}

CCadObject* CMicroJointTarget::_CreateEllipseArcToolPath(const nce::CCadEllipseArc* pCadEllipseArc_)
{
	CCadPolyline* _pCadPolyline = TranslateObjectToPolyline(pCadEllipseArc_);
	_pCadPolyline->SetParamString(pCadEllipseArc_->GetParamString());
	CCadMultiSegments* _pCadMultiSegments = (CCadMultiSegments*)(_pCadPolyline);
	// ԲıǰԲһģҲǴʣתɶ
	// ߵʱҪתɶĶκźͱ ---- DingQiang 2016-04-27
	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadEllipseArc_->GetKey(_strKeyMJ);
	DPOINT2 _ptStartPoint = _pCadMultiSegments->GetStartPoint();
	int _nTailSegCount = 0;
	double _nTaliRatio = 0.;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		double	_nTargetLength = _pCadMultiSegments->GetLength() * _nRatio;
		double _nLengthTemp = 0.;
		int _nSegTemp = 0;
		while(_nLengthTemp < _nTargetLength && _nSegTemp < _pCadMultiSegments->GetSegmentCount())
		{
			SEGMENT _segment = _pCadMultiSegments->GetSegment(_nSegTemp);
			if (_nLengthTemp + _segment.GetLength() >= _nTargetLength)
			{
				double _nRatio = (_nTargetLength - _nLengthTemp) / _segment.GetLength();
				_strValueTemp.Format(c_szMJFormat, _nSegTemp, _nRatio, _nLength);
				_pCadMultiSegments->SetKey(_strKeyMJ, _strValueTemp);
				break;
			}
			_nLengthTemp += _segment.GetLength();
			_nSegTemp++;
		}
		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = _pCadMultiSegments->GetKey(_strKeyMJ);
	}

	CCadObject* _pObjectRet = _CreateMultiSegmentToolPath(_pCadMultiSegments);
	delete _pCadPolyline;
	_pCadPolyline = NULL;
	return _pObjectRet;
}

// _CreateLineToolPathĺ_CreateLineToolPath2ǰѵ·Ϊ
// иʽ
//   һ߶Σ5΢11εPathе13579ǲ
// ģҪ0246810αΪҪġ
//                                                        ----  2016-10-24
CCadObject* CMicroJointTarget::_CreateLineToolPath2(const nce::CCadLine* pCadLine_)
{
	ASSERT(pCadLine_);
	if (DOUBLE_LE_ZERO(pCadLine_->GetLength()))
	{
		return NULL;
	}

	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	DPOINT2 _ptStart = pCadLine_->GetStartPoint();
	DPOINT2 _ptEnd = pCadLine_->GetEndPoint();
	_pCadPolyline->SetStartPoint(_ptStart);

	CStringA _strGXXNodes;
	DPOINT2 _ptSegStart = _ptStart;
	for (int _nMJIndex = 0; _nMJIndex < 10000; _nMJIndex++)
	{
		CStringA _strMJKey;
		_strMJKey.Format(c_szMJ, _nMJIndex);
		CStringA _strMJValue = pCadLine_->GetKey(_strMJKey);
		if (_strMJValue.IsEmpty())
			break;

		int _nSeg = 0;
		double _nRatio = 0.0;
		double _nLength = 0.0;
		VERIFY(3 == sscanf_s(_strMJValue, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		ASSERT(_nSeg == 0);
		DPOINT2 _ptGxxEnd = _ptStart + (_ptEnd - _ptStart) * _nRatio;
		if (!DOUBLE_EQU(_ptGxxEnd.x, _ptSegStart.x) || !DOUBLE_EQU(_ptGxxEnd.y, _ptSegStart.y))
		{
			_pCadPolyline->AddLine(_ptGxxEnd);
			_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
		}

		DPOINT2 _ptG00End = _ptStart + (_ptEnd - _ptStart) * (_nRatio + _nLength / pCadLine_->GetLength());
		if (!DOUBLE_EQU(_ptG00End.x, _ptGxxEnd.x) || !DOUBLE_EQU(_ptG00End.y, _ptGxxEnd.y))
		{
			_pCadPolyline->AddLine(_ptG00End);
		}

		_ptSegStart = _ptG00End;
	}

	if (!DOUBLE_EQU(_ptSegStart.x, _ptEnd.x) || !DOUBLE_EQU(_ptSegStart.y, _ptEnd.y))
	{
		_pCadPolyline->AddLine(_ptEnd);
		_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
	}

	_pCadPolyline->SetParamIndex(pCadLine_->GetParamIndex());
	_pCadPolyline->SetParamString(pCadLine_->GetParamString());
	_pCadPolyline->SetKey(c_szScanCamGXXNodes, _strGXXNodes);
	return _pCadPolyline;
}

CCadObject* CMicroJointTarget::_CreateCircleToolPath2(const CCadCircle* pCadCircle_)
{
	ASSERT(pCadCircle_);
	if (DOUBLE_LE_ZERO(pCadCircle_->GetLength()))
	{
		return NULL;
	}

	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	DPOINT2 _ptStart = pCadCircle_->GetPoint(0.0);
	DPOINT2 _ptEnd = _ptStart;
	_pCadPolyline->SetStartPoint(_ptStart);

	CStringA _strGXXNodes;
	double _nSegStartAngle = 0.0;
	for (int _nMJIndex = 0; _nMJIndex < 10000; _nMJIndex++)
	{
		CStringA _strMJKey;
		_strMJKey.Format(c_szMJ, _nMJIndex);
		CStringA _strMJValue = pCadCircle_->GetKey(_strMJKey);
		if (_strMJValue.IsEmpty())
			break;

		int _nSeg = 0;
		double _nRatio = 0.0;
		double _nLength = 0.0;
		VERIFY(3 == sscanf_s(_strMJValue, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		ASSERT(_nSeg == 0);
		double _nGxxEndAngle = 2 * c_nPIE * _nRatio * (pCadCircle_->IsNormalDir() ? 1.0 : -1.0);
		double _nGxxCenterAngle = _nGxxEndAngle - _nSegStartAngle;
		if (DOUBLE_GREAT_ZERO(_nGxxCenterAngle))
		{
			DPOINT2 _ptGxxEnd = pCadCircle_->GetPoint(RegularizeRadian(_nGxxEndAngle));
			double _nGxxBulge = tan(_nGxxCenterAngle / 4.0);
			_pCadPolyline->AddArc(_ptGxxEnd, _nGxxBulge);
			_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
		}

		double _nG00CenterAngle = (_nLength / pCadCircle_->GetLength()) * 2 * c_nPIE * (pCadCircle_->IsNormalDir() ? 1.0 : -1.0);
		double _nG00EngAngle = _nGxxEndAngle + _nG00CenterAngle;
		if (DOUBLE_GREAT_ZERO(_nG00CenterAngle))
		{
			DPOINT2 _ptG00End = pCadCircle_->GetPoint(RegularizeRadian(_nG00EngAngle));
			double _nG00Bulge = tan(_nG00CenterAngle / 4.0);
			_pCadPolyline->AddArc(_ptG00End, _nG00Bulge);
		}

		_nSegStartAngle = _nG00EngAngle;
	}

	if (DOUBLE_LESS(_nSegStartAngle, 2 * c_nPIE))
	{
		double _nBulge = tan((2 * c_nPIE - _nSegStartAngle) * (pCadCircle_->IsNormalDir() ? 1.0 : -1.0) / 4.0);
		_pCadPolyline->AddArc(_ptEnd, _nBulge);
		_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
	}

	_pCadPolyline->SetParamIndex(pCadCircle_->GetParamIndex());
	_pCadPolyline->SetParamString(pCadCircle_->GetParamString());
	_pCadPolyline->SetKey(c_szScanCamGXXNodes, _strGXXNodes);
	return _pCadPolyline;
}

CCadObject* CMicroJointTarget::_CreateArcToolPath2(const nce::CCadArc* pCadArc_)
{
	ASSERT(pCadArc_);
	if (DOUBLE_LE_ZERO(pCadArc_->GetLength()))
	{
		return NULL;
	}

	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	DPOINT2 _ptStart = pCadArc_->GetStartPoint();
	DPOINT2 _ptEnd = pCadArc_->GetEndPoint();
	_pCadPolyline->SetStartPoint(_ptStart);

	CStringA _strGXXNodes;
	double _nSegStartAngle = 0;
	for (int _nMJIndex = 0; _nMJIndex < 10000; _nMJIndex++)
	{
		CStringA _strMJKey;
		_strMJKey.Format(c_szMJ, _nMJIndex);
		CStringA _strMJValue = pCadArc_->GetKey(_strMJKey);
		if (_strMJValue.IsEmpty())
			break;

		int _nSeg = 0;
		double _nRatio = 0.0;
		double _nLength = 0.0;
		VERIFY(3 == sscanf_s(_strMJValue, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		ASSERT(_nSeg == 0);
		double _nGxxEndAngle = pCadArc_->GetCentralAngle() * _nRatio;
		double _nGxxCenterAngle = _nGxxEndAngle - _nSegStartAngle;
		if (DOUBLE_GREAT_ZERO(_nGxxCenterAngle))
		{
			DPOINT2 _ptGxxEnd = pCadArc_->GetPoint(_nGxxEndAngle);
			double _nGxxBulge = tan(0.25 * _nGxxCenterAngle * (pCadArc_->IsNormalDir() ? 1.0 : -1.0));
			_pCadPolyline->AddArc(_ptGxxEnd, _nGxxBulge);
			_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
		}

		double _nG00CenterAngle = (_nLength / pCadArc_->GetLength()) * pCadArc_->GetCentralAngle();
		double _nG00EngAngle = _nGxxEndAngle + _nG00CenterAngle;
		if (DOUBLE_GREAT_ZERO(_nG00CenterAngle))
		{
			DPOINT2 _ptG00End = pCadArc_->GetPoint(_nG00EngAngle);
			double _nG00Bulge = tan(0.25 * _nG00CenterAngle * (pCadArc_->IsNormalDir() ? 1.0 : -1.0));
			_pCadPolyline->AddArc(_ptG00End, _nG00Bulge);
		}

		_nSegStartAngle = _nG00EngAngle;
	}

	if (DOUBLE_LESS(_nSegStartAngle, pCadArc_->GetCentralAngle()))
	{
		double _nBulge = tan(0.25 * (pCadArc_->GetCentralAngle() - _nSegStartAngle) * (pCadArc_->IsNormalDir() ? 1.0 : -1.0));
		_pCadPolyline->AddArc(_ptEnd, _nBulge);
		_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
	}

	_pCadPolyline->SetParamIndex(pCadArc_->GetParamIndex());
	_pCadPolyline->SetParamString(pCadArc_->GetParamString());
	_pCadPolyline->SetKey(c_szScanCamGXXNodes, _strGXXNodes);
	return _pCadPolyline;
}

CCadObject* CMicroJointTarget::_CreateMultiSegmentToolPath2(const nce::CCadMultiSegments* pCadMultiSegments_)
{
	// ը΢·
	// TODO: Ӧðⲿ߼ȡDoer
	CadObjectList _listCads;
	if (!_ExplodeMultiSegmentMicroJoint(pCadMultiSegments_, &_listCads))
	{
		return NULL;
	}

	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	CStringA _strGXXNodes;
	for (CadObjectList::iterator _it = _listCads.begin(); _it != _listCads.end(); _it++)
	{
		CCadObject* _pCad = *_it;
		ASSERT(_pCad && _pCad->GetType() == cadpolyline);
		CCadPolyline* _pPol = (CCadPolyline*)_pCad;
		ASSERT(_pPol->GetNodeCount() > 0);
		if (_pCadPolyline->GetNodeCount() <= 0)
			_pCadPolyline->SetStartPoint(_pPol->GetStartPoint());
		else
			_pCadPolyline->AddLine(_pPol->GetStartPoint());

		for (int _i = 0; _i < _pPol->GetNodeCount(); _i++)
		{
			const CCadPolyline::ARCNODE& _node = _pPol->GetAt(_i);
			_pCadPolyline->AddArc(_node.ptEnd, _node.nBulge);
			_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
		}
	}

	_pCadPolyline->SetParamIndex(pCadMultiSegments_->GetParamIndex());
	_pCadPolyline->SetParamString(pCadMultiSegments_->GetParamString());
	_pCadPolyline->SetKey(c_szScanCamGXXNodes, _strGXXNodes);
	return _pCadPolyline;
}

CCadObject* CMicroJointTarget::_CreateEllipseToolPath2(const nce::CCadEllipse* pCadEllipse_)
{
	CadObjectList _listCads;
	if (!_ExplodeEllipseMicroJoint(pCadEllipse_, &_listCads))
	{
		return NULL;
	}

	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	CStringA _strGXXNodes;
	for (CadObjectList::iterator _it = _listCads.begin(); _it != _listCads.end(); _it++)
	{
		CCadObject* _pCad = *_it;
		ASSERT(_pCad && _pCad->GetType() == cadpolyline); // TODO: Բ΢ըпܻ
		CCadPolyline* _pPol = (CCadPolyline*)_pCad;
		ASSERT(_pPol->GetNodeCount() > 0);
		if (_pCadPolyline->GetNodeCount() <= 0)
			_pCadPolyline->SetStartPoint(_pPol->GetStartPoint());
		else
			_pCadPolyline->AddLine(_pPol->GetStartPoint());

		for (int _i = 0; _i < _pPol->GetNodeCount(); _i++)
		{
			const CCadPolyline::ARCNODE& _node = _pPol->GetAt(_i);
			_pCadPolyline->AddArc(_node.ptEnd, _node.nBulge);
			_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
		}
	}

	_pCadPolyline->SetParamIndex(pCadEllipse_->GetParamIndex());
	_pCadPolyline->SetParamString(pCadEllipse_->GetParamString());
	_pCadPolyline->SetKey(c_szScanCamGXXNodes, _strGXXNodes);
	return _pCadPolyline;
}

CCadObject* CMicroJointTarget::_CreateEllipseArcToolPath2(const nce::CCadEllipseArc* pCadEllipseArc_)
{
	CadObjectList _listCads;
	if (!_ExplodeEllipseArcMicroJoint(pCadEllipseArc_, &_listCads))
	{
		return NULL;
	}

	CCadPolyline* _pCadPolyline = new CCadPolyline;
	ASSERT(_pCadPolyline);
	CStringA _strGXXNodes;
	for (CadObjectList::iterator _it = _listCads.begin(); _it != _listCads.end(); _it++)
	{
		CCadObject* _pCad = *_it;
		ASSERT(_pCad && _pCad->GetType() == cadpolyline); // TODO: Բ΢ըпܻ
		CCadPolyline* _pPol = (CCadPolyline*)_pCad;
		ASSERT(_pPol->GetNodeCount() > 0);
		if (_pCadPolyline->GetNodeCount() <= 0)
			_pCadPolyline->SetStartPoint(_pPol->GetStartPoint());
		else
			_pCadPolyline->AddLine(_pPol->GetStartPoint());

		for (int _i = 0; _i < _pPol->GetNodeCount(); _i++)
		{
			const CCadPolyline::ARCNODE& _node = _pPol->GetAt(_i);
			_pCadPolyline->AddArc(_node.ptEnd, _node.nBulge);
			_strGXXNodes.AppendFormat("%d,", _pCadPolyline->GetNodeCount() - 1);
		}
	}

	_pCadPolyline->SetParamIndex(pCadEllipseArc_->GetParamIndex());
	_pCadPolyline->SetParamString(pCadEllipseArc_->GetParamString());
	_pCadPolyline->SetKey(c_szScanCamGXXNodes, _strGXXNodes);
	return _pCadPolyline;
}

BOOL CMicroJointTarget::OnLButtonDown(UINT nFlags_, DPOINT2 point_)
{
	if (!IsActive())
	{
		return FALSE;
	}

	CFileTarget* _pFileTarget = GetFileTarget();
	if (!_pFileTarget)
	{
		ASSERT(FALSE);
		return FALSE;
	}
	CCadGroup* _pCurrentGroup = _pFileTarget->GetCurrentCadGroup();
	if (_pCurrentGroup == NULL)
		return FALSE;

	SendTargetMessage(ID_TARGET_CATCH, TM_CATCH_REQUEST, CR_CATCH, (LPARAM)&point_);
	CCadObject* _pCadObject = NULL;
	Pos _pos;
	if (FindCadPositionByPoint(_pCurrentGroup, point_, &_pCadObject, _pos))
	{
		if (_pCadObject->GetType() == caddot
			|| _pCadObject->GetType() == cadtext)
		{
			NceShowMessage(_GETCS(s_szManualSetFailed));
			return TRUE;
		}

		CadObjectList _CadObjectListBefore;
		CadObjectList _CadObjectListAfter;
		CCadObject* _pCadObjectSrc = _pCadObject->Clone();
		if (!SetMicroJoint(_pCadObjectSrc, _pos.nID, _pos.nOffect, m_nMicroJointLength))
		{
			delete _pCadObjectSrc;
			_pCadObjectSrc = NULL;
			return TRUE;
		}

		_CadObjectListBefore.push_back(_pCadObject);
		_CadObjectListAfter.push_back(_pCadObjectSrc);
		if (CAction::NoAction != CMicroJointAction::Do(&_CadObjectListBefore, &_CadObjectListAfter))
		{
			UpdateAllVisions();
		}
	}
	else
	{
		NceShowMessage(_GETCS(s_szManualSetFailed));
	}
	return TRUE;
}

BOOL CMicroJointTarget::OnRButtonUp(UINT nFlags_, DPOINT2 point_)
{
	if (!IsActive())
	{
		return FALSE;
	}
	Exit();
	return TRUE;
}

BOOL CMicroJointTarget::OnMouseMove(UINT nFlags_, DPOINT2 point_)
{
	if (!IsActive())
	{
		return FALSE;
	}
	CFileTarget* _pFileTarget = GetFileTarget();
	if (!_pFileTarget)
	{
		ASSERT(FALSE);
		return FALSE;
	}
	CCadGroup* _pCurrentGroup = _pFileTarget->GetCurrentCadGroup();
	if (_pCurrentGroup == NULL)
		return FALSE;

	SendTargetMessage(ID_TARGET_CATCH, TM_CATCH_REQUEST, CR_CATCH, (LPARAM)&point_);

	return TRUE;
}

void CMicroJointTarget::Reset()
{
}

void CMicroJointTarget::OnMicroJoint()
{
	CMicroJointDlg_R2 _dlg;
	if (_dlg.DoModal() != IDOK)
	{
		return;
	}
	m_bNoPause = _dlg.GetIfNoPause();

	// TODOӦðϢ浽ļУ޸ЩʱȥдļĶǡ
	// UpdateAllVisionsǱҪģΪCToolPathTarget::Updateлȡʹ
	// m_bNoPauseֵ(΢ʵַʽ)               ----  2016-10-25
	UpdateAllVisions();

	if (_dlg.GetMode() == 0)
	{
		m_nMicroJointLength = _dlg.GetAutoMicroJointLength();
		m_nMicroJointCounts = _dlg.GetAutoMicroJointCounts();
		AutoMicroJointByCounts();
	}
	else if (_dlg.GetMode() == 1)
	{
		m_nMicroJointLength = _dlg.GetAutoMicroJointLength();
		m_nMicroJointSpace = _dlg.GetAutoMicroJointSpace();
		AutoMicroJointBySpace();
	}
	else if (_dlg.GetMode() == 2)
	{
		m_nMicroJointLength = _dlg.GetManualMicroJointLength();
		ManualMicroJoint();
	}
}

void CMicroJointTarget::ManualMicroJoint()
{
	if (IsActive())
		Exit();

	Activate();
}

void CMicroJointTarget::AutoMicroJointBySpace()
{
	CSelectTarget* _pSelectTarget = GetSelectTarget();
	if (!_pSelectTarget)
		return;
	CadObjectList* _plistSelectObjects = const_cast<CadObjectList*>(_pSelectTarget->GetSelectedCadObjects());
	if (_plistSelectObjects->size() == 0)
	{
		NceShowMessage(_GETCS(s_szNoSelectedObject));
		return;
	}

	CadObjectList _CadObjectListBefore;
	CadObjectList _CadObjectListAfter;

	for (CadObjectList::iterator _it = _plistSelectObjects->begin(); _it != _plistSelectObjects->end(); _it++)
	{
		CCadObject* _pNewCadObject = NULL;
		if (_AutoMicroJointBySpace(*_it, &_pNewCadObject))
		{
			_CadObjectListBefore.push_back(*_it);
			_CadObjectListAfter.push_back(_pNewCadObject);
		}
	}

	if (_CadObjectListAfter.size() <= 0)
	{
		NceShowMessage(_GETCS(s_szAutoSetFailed));
		return;
	}

	if (_CadObjectListAfter.size() >= 0
		&& CAction::NoAction != CMicroJointAction::Do(&_CadObjectListBefore, &_CadObjectListAfter))
	{
		UpdateAllVisions();
	}
}

void CMicroJointTarget::AutoMicroJointByCounts()
{
	CSelectTarget* _pSelectTarget = GetSelectTarget();
	if (!_pSelectTarget)
		return;
	CadObjectList* _plistSelectObjects = const_cast<CadObjectList*>(_pSelectTarget->GetSelectedCadObjects());
	if (_plistSelectObjects->size() == 0)
	{
		NceShowMessage(_GETCS(s_szNoSelectedObject));
		return;
	}

	CadObjectList _CadObjectListBefore;
	CadObjectList _CadObjectListAfter;

	for (CadObjectList::iterator _it = _plistSelectObjects->begin(); _it != _plistSelectObjects->end(); _it++)
	{
		CCadObject* _pNewCadObject = NULL;
		if (_AutoMicroJointByCounts(*_it, &_pNewCadObject))
		{
			_CadObjectListBefore.push_back(*_it);
			_CadObjectListAfter.push_back(_pNewCadObject);
		}
	}

	if (_CadObjectListAfter.size() <= 0)
	{
		NceShowMessage(_GETCS(s_szAutoSetFailed));
		return;
	}

	if (_CadObjectListAfter.size() >= 0
		&& CAction::NoAction != CMicroJointAction::Do(&_CadObjectListBefore, &_CadObjectListAfter))
	{
		UpdateAllVisions();
	}
}

bool CMicroJointTarget::_AutoMicroJointBySpace(IN const nce::CCadObject* pCadObjectSrc_, OUT nce::CCadObject** ppCadObjectDes_)
{
	switch(pCadObjectSrc_->GetType())
	{
	case cadgroup:
		return false;
	case cadline:
	case cadcircle:
	case cadarc:
	case cadellipse:
	case cadellipsearc:
		return _AutoMicroJointBySpaceOneSeg(pCadObjectSrc_, ppCadObjectDes_);
	case cadpath:
	case cadpolygon:
	case cadpolyline:
	case cadstar:
	case cadrectangle:
		return _AutoMicroJointBySpaceMulSeg((const CCadMultiSegments*)pCadObjectSrc_, ppCadObjectDes_);
	}

	return false;
}

bool CMicroJointTarget::_AutoMicroJointBySpaceOneSeg(IN const nce::CCadObject* pCadObjectSrc_, OUT nce::CCadObject** ppCadObjectDes_)
{
	ASSERT(pCadObjectSrc_ && ppCadObjectDes_ && *ppCadObjectDes_ == NULL);
	double _nCadLength = GetCadLength(pCadObjectSrc_);
	if (_nCadLength <= m_nMicroJointSpace + m_nMicroJointLength)
	{
		// 1΢ͷ ---- DingQiang 2016-05-18
		*ppCadObjectDes_ = NULL;
		return false;
	}

	CCadObject* _pCadObject = pCadObjectSrc_->Clone();
	ClearMicroJoint(_pCadObject);
	double _nRemindLength = _nCadLength;
	CStringA _strSetKey;
	CStringA _strSetValue;
	int _nCountIndex = 0;
	double _nRatio;
	double _nHeadLength = m_nMicroJointSpace;
	while (_nRemindLength > m_nMicroJointSpace + m_nMicroJointLength)
	{
		_nRatio = _nHeadLength / _nCadLength;
		_strSetKey.Format(c_szMJ, _nCountIndex);
		_strSetValue.Format(c_szMJFormat, 0, _nRatio, m_nMicroJointLength);
		_pCadObject->SetKey(_strSetKey, _strSetValue);
		_nCountIndex++;
		_nRemindLength -= m_nMicroJointSpace + m_nMicroJointLength;
		_nHeadLength += m_nMicroJointLength + m_nMicroJointSpace;
	}

	*ppCadObjectDes_ = _pCadObject;
	return true;
}

bool CMicroJointTarget::_AutoMicroJointBySpaceMulSeg(IN const nce::CCadMultiSegments* pCadCadMultiSegmentsSrc_, OUT nce::CCadObject** ppCadObjectDes_)
{
	ASSERT(pCadCadMultiSegmentsSrc_ && ppCadObjectDes_ && *ppCadObjectDes_ == NULL);
	double _nCadLength = pCadCadMultiSegmentsSrc_->GetLength();
	if (_nCadLength <= m_nMicroJointSpace + m_nMicroJointLength)
	{
		// 1΢ͷ ---- DingQiang 2016-05-18
		*ppCadObjectDes_ = NULL;
		return false;
	}

	CCadObject* _pCadObject = pCadCadMultiSegmentsSrc_->Clone();
	ClearMicroJoint(_pCadObject);
	double _nRemindLength = _nCadLength;
	CStringA _strSetTemp;
	int _nCountIndex = 0;
	double _nRatio;
	double _nHeadLength = m_nMicroJointSpace;
	while (_nRemindLength > m_nMicroJointSpace + m_nMicroJointLength)
	{
		double _nAddlength = 0;
		int _nSeg = 0;
		while(_nAddlength < _nHeadLength && _nSeg < pCadCadMultiSegmentsSrc_->GetSegmentCount())
		{
			SEGMENT _segment = pCadCadMultiSegmentsSrc_->GetSegment(_nSeg);
			if (_nAddlength + _segment.GetLength() >= _nHeadLength)
			{
				_nRatio = (_nHeadLength - _nAddlength) / _segment.GetLength();
				break;
			}
			_nAddlength += _segment.GetLength();
			_nSeg++;
		}

		if (_nSeg >= pCadCadMultiSegmentsSrc_->GetSegmentCount())
		{
			ASSERT(FALSE);
			break;
		}

		CStringA _strKeyMJ;
		_strKeyMJ.Format(c_szMJ, _nCountIndex);
		CStringA _strValue;
		_strValue.Format(c_szMJFormat, _nSeg, _nRatio, m_nMicroJointLength);
		_pCadObject->SetKey(_strKeyMJ, _strValue);
		_nCountIndex++;
		_nRemindLength -= m_nMicroJointSpace + m_nMicroJointLength;
		_nHeadLength += m_nMicroJointLength + m_nMicroJointSpace;
	}

	*ppCadObjectDes_ = _pCadObject;
	return true;
}

bool CMicroJointTarget::_AutoMicroJointByCounts(IN const nce::CCadObject* pCadObjectSrc_, OUT nce::CCadObject** ppCadObjectDes_)
{
	switch(pCadObjectSrc_->GetType())
	{
	case cadgroup:
		return false;
	case cadline:
	case cadcircle:
	case cadarc:
	case cadellipse:
	case cadellipsearc:
		return _AutoMicroJointByCountsOneSeg(pCadObjectSrc_, ppCadObjectDes_);
	case cadpath:
	case cadpolygon:
	case cadpolyline:
	case cadstar:
	case cadrectangle:
		return _AutoMicroJointByCountsMulSeg((const CCadMultiSegments*)pCadObjectSrc_, ppCadObjectDes_);
	}

	return false;
}

bool CMicroJointTarget::_AutoMicroJointByCountsOneSeg(IN const nce::CCadObject* pCadObjectSrc_, OUT nce::CCadObject** ppCadObjectDes_)
{
	ASSERT(pCadObjectSrc_ && ppCadObjectDes_ && *ppCadObjectDes_ == NULL);
	double _nCadLength = GetCadLength(pCadObjectSrc_);
	if (_nCadLength <= m_nMicroJointLength * m_nMicroJointCounts)
	{
		// ͷ ---- DingQiang 2016-05-18
		*ppCadObjectDes_ = NULL;
		return false;
	}

	CCadObject* _pCadObject = pCadObjectSrc_->Clone();
	ClearMicroJoint(_pCadObject);

	// ΢ɵĶ ---- DingQiang 2016-05-18
	int _nNewCounts = m_nMicroJointCounts;
	if (!_pCadObject->IsClose())
	{
		_nNewCounts++;
	}

	double _nEachSegLength = (_nCadLength - m_nMicroJointLength * m_nMicroJointCounts) / _nNewCounts;
	int _nCountIndex = 0;
	CStringA _strSetKey;
	CStringA _strSetValue;
	double _nRatio = -1;
	double _nHeadLength = _nEachSegLength;
	while(_nCountIndex < m_nMicroJointCounts)
	{
		_nRatio = _nHeadLength / _nCadLength;
		_nRatio = _nHeadLength / _nCadLength;
		_strSetKey.Format(c_szMJ, _nCountIndex);
		_strSetValue.Format(c_szMJFormat, 0, _nRatio, m_nMicroJointLength);
		_pCadObject->SetKey(_strSetKey, _strSetValue);
		_nHeadLength += _nEachSegLength + m_nMicroJointLength;
		_nCountIndex++;
	}

	*ppCadObjectDes_ = _pCadObject;
	return true;
}

bool CMicroJointTarget::_AutoMicroJointByCountsMulSeg(IN const nce::CCadMultiSegments* pCadObjectSrc_, OUT nce::CCadObject** ppCadObjectDes_)
{
	ASSERT(pCadObjectSrc_ && ppCadObjectDes_ && *ppCadObjectDes_ == NULL);
	double _nCadLength = GetCadLength(pCadObjectSrc_);
	if (_nCadLength <= m_nMicroJointLength * m_nMicroJointCounts)
	{
		// ͷ ---- DingQiang 2016-05-18
		*ppCadObjectDes_ = NULL;
		return false;
	}

	CCadObject* _pCadObject = pCadObjectSrc_->Clone();
	ClearMicroJoint(_pCadObject);

	// ΢ɵĶ ---- DingQiang 2016-05-18
	int _nNewCounts = m_nMicroJointCounts;
	if (!_pCadObject->IsClose())
	{
		_nNewCounts++;
	}

	double _nEachSegLength = (_nCadLength - m_nMicroJointLength * m_nMicroJointCounts) / _nNewCounts;
	int _nCountIndex = 0;
	CStringA _strSetTemp;
	double _nRatio = -1;
	double _nHeadLength = _nEachSegLength;
	while(_nCountIndex < m_nMicroJointCounts)
	{
		double _nAddlength = 0;
		int _nSeg = 0;
		while(_nAddlength < _nHeadLength && _nSeg < pCadObjectSrc_->GetSegmentCount())
		{
			SEGMENT _segment = pCadObjectSrc_->GetSegment(_nSeg);
			if (_nAddlength + _segment.GetLength() >= _nHeadLength)
			{
				_nRatio = (_nHeadLength - _nAddlength) / _segment.GetLength();
				break;
			}
			_nAddlength += _segment.GetLength();
			_nSeg++;
		}
		if (_nSeg >= pCadObjectSrc_->GetSegmentCount())
		{
			break;
		}
		CStringA _strKeyMJ;
		_strKeyMJ.Format(c_szMJ, _nCountIndex);
		CStringA _strValue;
		_strValue.Format(c_szMJFormat, _nSeg, _nRatio, m_nMicroJointLength);
		_pCadObject->SetKey(_strKeyMJ, _strValue);
		_nCountIndex++;
		_nHeadLength += _nEachSegLength + m_nMicroJointLength;
	}

	*ppCadObjectDes_ = _pCadObject;
	return true;
}

void CMicroJointTarget::OnUpdateMicroJoint(CCmdUI* pCmdUI_)
{
	CFileTarget* _pFileTarget = GetFileTarget();
	if (!_pFileTarget)
		return;
	CCadGroup* _pCadGroup = _pFileTarget->GetCurrentCadGroup();
	BOOL _bEnable = FALSE;
	if (_pCadGroup && (_pCadGroup->GetNumOfChild() != 0))
		_bEnable = TRUE;
	pCmdUI_->Enable(_bEnable);
}

void CMicroJointTarget::OnClearMicroJoint()
{
	CSelectTarget* _pSelectTarget = GetSelectTarget();
	if (!_pSelectTarget)
		return;
	CadObjectList* _plistSelectObjects = const_cast<CadObjectList*>(_pSelectTarget->GetSelectedCadObjects());

	CadObjectList _CadObjectListBefore;
	CadObjectList _CadObjectListAfter;

	for (CadObjectList::iterator _it = _plistSelectObjects->begin(); _it != _plistSelectObjects->end(); _it++)
	{
		if (CanClearMicroJoint(*_it))
		{
			CCadObject* _pNewObject = (*_it)->Clone();
			_CadObjectListBefore.push_back(*_it);
			if(ClearMicroJoint(_pNewObject))
			{
				_CadObjectListAfter.push_back(_pNewObject);
			}
		}
	}

	if (_CadObjectListAfter.size() >= 0
		&& CAction::NoAction != CMicroJointAction::Do(&_CadObjectListBefore, &_CadObjectListAfter))
	{
		UpdateAllVisions();
	}
}

void CMicroJointTarget::OnUpdateClearMicroJoint(CCmdUI* pCmdUI_)
{
	CSelectTarget* _pSelectTarget = GetSelectTarget();
	if (!_pSelectTarget)
		return;
	CadObjectList* _plistSelectObjects = const_cast<CadObjectList*>(_pSelectTarget->GetSelectedCadObjects());
	ASSERT(_plistSelectObjects);
	BOOL _bEnable = FALSE;
	if (_plistSelectObjects->size() > 0)
		_bEnable = TRUE;
	pCmdUI_->Enable(_bEnable);
}

void CMicroJointTarget::OnExplodeMicroJoint()
{
	CSelectTarget* _pSelectTarget = GetSelectTarget();
	if (!_pSelectTarget)
		return;
	CadObjectList* _plistSelectObjects = const_cast<CadObjectList*>(_pSelectTarget->GetSelectedCadObjects());

	ExplodeMicroJointItemList _listExplodeItems;
	for (CadObjectList::iterator _it = _plistSelectObjects->begin(); _it != _plistSelectObjects->end(); _it++)
	{

		if (CanExplodeMicroJoint(*_it))
		{
			CCadObject* _pCad = *_it;
			ASSERT(_pCad);
			ExplodeMicroJointItem _item = {_pCad};
			if (ExplodeMicroJoint(_item.pSrcCadObject, &_item.listDstCadObjects))
			{
				_listExplodeItems.push_back(_item);
			}
		}
	}

	if (_listExplodeItems.empty())
	{
		return;
	}

	CExplodeMicroJointAction::Do(&_listExplodeItems);
	UpdateAllVisions();
}

void CMicroJointTarget::OnUpdateExplodeMicroJoint(CCmdUI* pCmdUI_)
{
	CSelectTarget* _pSelectTarget = GetSelectTarget();
	if (!_pSelectTarget)
		return;
	CadObjectList* _plistSelectObjects = const_cast<CadObjectList*>(_pSelectTarget->GetSelectedCadObjects());
	ASSERT(_plistSelectObjects);
	BOOL _bEnable = FALSE;
	if (_plistSelectObjects->size() > 0)
		_bEnable = TRUE;
	pCmdUI_->Enable(_bEnable);
}

bool CMicroJointTarget::FindCadPositionByPoint(IN nce::CCadGroup* pCadGroup_, IN const DPOINT2 ptCurrent_,
											   IN nce::CCadObject** pCadObject_, Pos& PosInfo_)
{
	CIntroPosCalculator _IntroPosCalculator;
	_IntroPosCalculator.Init(pCadGroup_, &ptCurrent_, GetSelectTarget()->GetSelectToleranceLP());
	IntroPos _nIntroPos;
	if (!_IntroPosCalculator.CalculatePosition(&_nIntroPos, false))
		return false;

	ASSERT(*pCadObject_ == NULL);
	if (!FindObjectByID(pCadGroup_, _nIntroPos.nID, pCadObject_))
		return false;

	PosInfo_.nID = _nIntroPos.nSegmentNo;
	PosInfo_.nOffect = _nIntroPos.nRate;
	return true;
}

// ıӹ ----  2017-03-06
void CMicroJointTarget::OnCadDirChanged(WPARAM wParam_, LPARAM lParam_)
{
	CadObjectList* _pCadObjectList = (CadObjectList*)wParam_;
	for (CadObjectList::iterator _it = _pCadObjectList->begin(); _it != _pCadObjectList->end(); ++_it)
	{
		OnCadDirChanged(*_it);
	}
}

void CMicroJointTarget::OnCadDirChanged(nce::CCadObject* pCadObject_)
{
	// ıӹ
	ASSERT(pCadObject_);
	switch(pCadObject_->GetType())
	{
	case cadgroup:
		{
			CCadGroup* _CadGroup = (CCadGroup*)pCadObject_;
			for (int _i = 0; _i < _CadGroup->GetNumOfChild(); ++_i)
			{
				CCadObject* _pCadObject = _CadGroup->GetAt(_i);
				OnCadDirChanged(_pCadObject);
			}
		}
		break;

	case cadtext:
		{
			CCadText* _pCadText = (CCadText*)pCadObject_;
			OnCadDirChanged(_pCadText->GetTextGroup());
		}
		break;

	case cadline:
		{
			int _nIndex = 0;
			CStringA _strKey;
			_strKey.Format(c_szMJ, _nIndex);
			CStringA _strValueTemp = pCadObject_->GetKey(_strKey);
			if(!_strValueTemp.IsEmpty())
			{
				// ռ΢
				std::vector<MJItem> _vectMJItem;
				int _nIndex = 0;
				CStringA _strKeyMJ;
				_strKeyMJ.Format(c_szMJ, _nIndex);
				CStringA _strValueTemp = pCadObject_->GetKey(_strKeyMJ);
				while(!_strValueTemp.IsEmpty())
				{
					MJItem _item;
					double _nRatio = 0.;
					double _nLength = 0.;
					VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_item.nSeg, &_nRatio, &_nLength));
					_item.nRatio = _nRatio;
					_item.nLength = _nLength;

					_vectMJItem.push_back(_item);
					_nIndex++;
					_strKeyMJ.Format(c_szMJ, _nIndex);
					_strValueTemp = pCadObject_->GetKey(_strKeyMJ);
				}

				std::sort(_vectMJItem.begin(), _vectMJItem.end());
				std::vector<MJItem> _vectChangeMJItems;
				double _nCadLen = GetCadLength(pCadObject_);
				for (size_t _i = 0; _i < _vectMJItem.size(); _i++)
				{
					MJItem _MJItem = _vectMJItem[_i];
					MJItem _MJChangeItem;
					ASSERT(_MJItem.nSeg == 0);
					ASSERT(_MJItem.nRatio >= 0.0 && _MJItem.nRatio <= 1.0);
					double _nRatio = _MJItem.nLength / _nCadLen;
					_MJChangeItem.nSeg = 0;
					_MJChangeItem.nRatio = 1.0 - _MJItem.nRatio - _nRatio;
					if (_MJChangeItem.nRatio < 0)
						_MJChangeItem.nRatio = 0;
					_MJChangeItem.nLength = _MJItem.nLength;
					_vectChangeMJItems.push_back(_MJChangeItem);
				}
				std::sort(_vectChangeMJItems.begin(), _vectChangeMJItems.end());
				for (size_t _i = 0; _i < _vectChangeMJItems.size(); _i++)
				{
					MJItem _MJItem = _vectChangeMJItems[_i];
					CStringA _strSetKey;
					_strSetKey.Format(c_szMJ, _i);
					CStringA _strValue;
					_strValue.Format(c_szMJFormat, _MJItem.nSeg, _MJItem.nRatio, _MJItem.nLength);
					pCadObject_->SetKey(_strSetKey, _strValue);
				}
			}
		}
		break;

	default:
		{
			if ((pCadObject_->GetType() & 0xFF00) != cadmultisegments)
				return;
			int _nIndex = 0;
			CStringA _strKey;
			_strKey.Format(c_szMJ, _nIndex);
			CStringA _strValueTemp = pCadObject_->GetKey(_strKey);
			if(!_strValueTemp.IsEmpty())
			{
				// ռ΢
				std::vector<MJItem> _vectMJItem;
				int _nIndex = 0;
				CStringA _strKeyMJ;
				_strKeyMJ.Format(c_szMJ, _nIndex);
				CStringA _strValueTemp = pCadObject_->GetKey(_strKeyMJ);
				while(!_strValueTemp.IsEmpty())
				{
					MJItem _item;
					double _nRatio = 0.;
					double _nLength = 0.;
					VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_item.nSeg, &_nRatio, &_nLength));
					_item.nRatio = _nRatio;
					_item.nLength = _nLength;

					_vectMJItem.push_back(_item);
					_nIndex++;
					_strKeyMJ.Format(c_szMJ, _nIndex);
					_strValueTemp = pCadObject_->GetKey(_strKeyMJ);
				}

				std::sort(_vectMJItem.begin(), _vectMJItem.end());
				std::vector<MJItem> _vectChangeMJItems;
				CCadMultiSegments* _pCadMS = (CCadMultiSegments*)pCadObject_;
				UINT _nSegCount = _pCadMS->GetSegmentCount();
				ASSERT(_nSegCount > 0);
				for (size_t _i = 0; _i < _vectMJItem.size(); _i++)
				{
					MJItem _MJItem = _vectMJItem[_i];
					MJItem _MJChangeItem;
					ASSERT(_MJItem.nSeg >= 0 && _MJItem.nSeg < _nSegCount);
					ASSERT(_MJItem.nRatio >= 0.0 && _MJItem.nRatio <= 1.0);
					double _nItemLen = _pCadMS->GetSegment((_nSegCount - 1) - _MJItem.nSeg).GetLength();
					double _nRatio = _MJItem.nLength / _nItemLen;
					_MJChangeItem.nSeg = (_nSegCount - 1) - _MJItem.nSeg;
					_MJChangeItem.nRatio = 1.0 - _MJItem.nRatio - _nRatio;
					if (_MJChangeItem.nRatio < 0)
						_MJChangeItem.nRatio = 0;
					_MJChangeItem.nLength = _MJItem.nLength;
					_vectChangeMJItems.push_back(_MJChangeItem);
				}
				std::sort(_vectChangeMJItems.begin(), _vectChangeMJItems.end());

				// ϲ΢ ---- DingQiang 2016-03-11
				//CombineMicroJoint(pCadObject_, _vectChangeMJItems);

				for (size_t _i = 0; _i < _vectChangeMJItems.size(); _i++)
				{
					MJItem _MJItem = _vectChangeMJItems[_i];
					CStringA _strSetKey;
					_strSetKey.Format(c_szMJ, _i);
					CStringA _strValue;
					_strValue.Format(c_szMJFormat, _MJItem.nSeg, _MJItem.nRatio, _MJItem.nLength);
					pCadObject_->SetKey(_strSetKey, _strValue);
				}
			}
		}
		break;
	}
}

// ΢
bool CMicroJointTarget::SetMicroJoint(CCadObject* pCadObject_, UINT nSeg_, double nRatio_, double nLength_)
{
	MJItem _SrcItem;
	_SrcItem.nSeg = nSeg_;
	_SrcItem.nRatio = nRatio_;
	_SrcItem.nLength = nLength_;

	// ռ΢
	std::vector<MJItem> _vectMJItem;
	int _nIndex = 0;
	CStringA _strKeyMJ;
	_strKeyMJ.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadObject_->GetKey(_strKeyMJ);
	while(!_strValueTemp.IsEmpty())
	{
		MJItem _item;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_item.nSeg, &_nRatio, &_nLength));
		_item.nRatio = _nRatio;
		_item.nLength = _nLength;

		_vectMJItem.push_back(_item);
		_nIndex++;
		_strKeyMJ.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadObject_->GetKey(_strKeyMJ);
	}

	std::sort(_vectMJItem.begin(), _vectMJItem.end());
	bool _bInsert = false;
	for (std::vector<MJItem>::iterator _it = _vectMJItem.begin(); _it != _vectMJItem.end(); _it++)
	{
		MJItem _MJItem = *_it;
		// Ѿ
		if (_MJItem == _SrcItem)
		{
			if (DOUBLE_EQU(_MJItem.nLength, _SrcItem.nLength))
			{
				return false;
			}
			else
			{
				_it->nLength = _SrcItem.nLength;
				return true;
			}
		}

		if (_SrcItem < _MJItem)
		{
			_vectMJItem.insert(_it, _SrcItem);
			_bInsert = true;
			break;
		}
	}
	if (!_bInsert)
	{
		_vectMJItem.push_back(_SrcItem);
	}

	// ϲ΢ ---- DingQiang 2016-03-11
	CombineMicroJoint(pCadObject_, _vectMJItem);

	for (size_t _i = 0; _i < _vectMJItem.size(); _i++)
	{
		MJItem _MJItem = _vectMJItem[_i];
		CStringA _strSetKey;
		_strSetKey.Format(c_szMJ, _i);
		CStringA _strValue;
		_strValue.Format(c_szMJFormat, _MJItem.nSeg, _MJItem.nRatio, _MJItem.nLength);
		pCadObject_->SetKey(_strSetKey, _strValue);
	}

	return true;
}

bool CMicroJointTarget::CanClearMicroJoint(const CCadObject* pCadObject_)
{
	bool _bRet = false;
	ASSERT(pCadObject_);

	if (pCadObject_->GetType() == cadgroup)
	{
		CCadGroup* _pCadGroup = (CCadGroup*)pCadObject_;
		for (int _i = 0; _i < _pCadGroup->GetNumOfChild(); _i++)
		{
			_bRet |= CanClearMicroJoint(_pCadGroup->GetAt(_i));
		}
	}
	else
	{
		int _nIndex = 0;
		CStringA _strKey;
		_strKey.Format(c_szMJ, _nIndex);
		CStringA _strValueTemp = pCadObject_->GetKey(_strKey);
		double _nStartAngle = 0;
		while(!_strValueTemp.IsEmpty())
		{
			return true;
		}
	}

	return _bRet;
}

bool CMicroJointTarget::ClearMicroJoint(nce::CCadObject* pCadObject_)
{
	bool _bRet = false;
	ASSERT(pCadObject_);

	if (pCadObject_->GetType() == cadgroup)
	{
		CCadGroup* _pCadGroup = (CCadGroup*)pCadObject_;
		for (int _i = 0; _i < _pCadGroup->GetNumOfChild(); _i++)
		{
			_bRet |= ClearMicroJoint(_pCadGroup->GetAt(_i));
		}
	}
	else
	{
		int _nIndex = 0;
		CStringA _strKey;
		_strKey.Format(c_szMJ, _nIndex);
		CStringA _strValueTemp = pCadObject_->GetKey(_strKey);
		double _nStartAngle = 0;
		while(!_strValueTemp.IsEmpty())
		{
			pCadObject_->SetKey(_strKey, NULL);
			_bRet = true;
			_nIndex++;
			_strKey.Format(c_szMJ, _nIndex);
			_strValueTemp = pCadObject_->GetKey(_strKey);
		}
	}

	return _bRet;
}

bool CMicroJointTarget::CanExplodeMicroJoint(const nce::CCadObject* pCadObject_)
{
	bool _bRet = false;
	ASSERT(pCadObject_);

	if (pCadObject_->GetType() == cadgroup)
	{
		CCadGroup* _pCadGroup = (CCadGroup*)pCadObject_;
		for (int _i = 0; _i < _pCadGroup->GetNumOfChild(); _i++)
		{
			_bRet |= CanExplodeMicroJoint(_pCadGroup->GetAt(_i));
		}
	}
	else
	{
		int _nIndex = 0;
		CStringA _strKey;
		_strKey.Format(c_szMJ, _nIndex);
		CStringA _strValueTemp = pCadObject_->GetKey(_strKey);
		while(!_strValueTemp.IsEmpty())
		{
			return true;
		}
	}

	return _bRet;
}

bool CMicroJointTarget::ExplodeMicroJoint(const CCadObject* pCadObjectSrc_, CadObjectList* pCadObjectListDes_)
{
	if (pCadObjectSrc_->GetType() == cadgroup)
	{
		CCadGroup* _pCadGroup = (CCadGroup*)pCadObjectSrc_;
		*pCadObjectListDes_ = *_pCadGroup->GetObjectsList();
	}
	else
	{
		switch(pCadObjectSrc_->GetType())
		{
			case cadline:
				return _ExplodeLineMicroJoint((CCadLine*)pCadObjectSrc_, pCadObjectListDes_);
				break;
			case cadcircle:
				return _ExplodeCircleMicroJoint((CCadCircle*)pCadObjectSrc_, pCadObjectListDes_);
				break;
			case cadarc:
				return _ExplodeArcMicroJoint((CCadArc*)pCadObjectSrc_, pCadObjectListDes_);
				break;
			case cadellipse:
				return _ExplodeEllipseMicroJoint((CCadEllipse*)pCadObjectSrc_, pCadObjectListDes_);
				break;
			case cadellipsearc:
				return _ExplodeEllipseArcMicroJoint((CCadEllipseArc*)pCadObjectSrc_, pCadObjectListDes_);
				break;
			case cadpath:
			case cadpolygon:
			case cadpolyline:
			case cadstar:
			case cadrectangle:
				return _ExplodeMultiSegmentMicroJoint((const CCadMultiSegments*)pCadObjectSrc_, pCadObjectListDes_);
				break;
		}
	}

	return !pCadObjectListDes_->empty();
}

bool CMicroJointTarget::_ExplodeLineMicroJoint(const nce::CCadLine* pCadLine_, nce::CadObjectList* pCadObjectListDes_)
{
	size_t _nOldSize = pCadObjectListDes_->size();
	int _nIndex = 0;
	CStringA _strKey;
	_strKey.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadLine_->GetKey(_strKey);
	DPOINT2 _ptStartPoint = pCadLine_->GetStartPoint();
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		DPOINT2 _ptEndPoint = (pCadLine_->GetEndPoint() - pCadLine_->GetStartPoint()) * _nRatio + pCadLine_->GetStartPoint();

		if (_ptStartPoint != _ptEndPoint)
		{
			CCadLine* _pCadLine = new CCadLine(_ptStartPoint, _ptEndPoint);

			_pCadLine->Select(pCadLine_->IsSelected());
			_pCadLine->SetParamIndex(pCadLine_->GetParamIndex());
			_pCadLine->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadLine_));
			pCadObjectListDes_->push_back(_pCadLine);
		}

		_ptStartPoint = pCadLine_->GetStartPoint() + (pCadLine_->GetEndPoint() - pCadLine_->GetStartPoint()) * (_nRatio + _nLength / pCadLine_->GetLength());

		_nIndex++;
		_strKey.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadLine_->GetKey(_strKey);
	}

	if (_ptStartPoint != pCadLine_->GetEndPoint())
	{
		CCadLine* _pCadLine = new CCadLine(_ptStartPoint, pCadLine_->GetEndPoint());
		_pCadLine->Select(pCadLine_->IsSelected());
		_pCadLine->SetParamIndex(pCadLine_->GetParamIndex());
		_pCadLine->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadLine_));
		pCadObjectListDes_->push_back(_pCadLine);
	}

	return pCadObjectListDes_->size() > _nOldSize;
}

bool CMicroJointTarget::_ExplodeCircleMicroJoint(const nce::CCadCircle* pCadCircle_, nce::CadObjectList* pCadObjectListDes_)
{
	size_t _nOldSize = pCadObjectListDes_->size();

	int _nIndex = 0;
	CStringA _strKey;
	_strKey.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadCircle_->GetKey(_strKey);
	double _nStartAngle = 0;
	bool _bFirstSegment = true;
	CCadArc* _pFirstArc = NULL;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));

		double _nCenterAngle = 2 * c_nPIE * _nRatio - _nStartAngle;
		if (!DOUBLE_EQU_ZERO(_nCenterAngle))
		{
			if (_bFirstSegment)
			{
				_pFirstArc = new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),
					_nStartAngle, _nCenterAngle, pCadCircle_->IsNormalDir());
				_pFirstArc->Select(pCadCircle_->IsSelected());
				_pFirstArc->SetParamIndex(pCadCircle_->GetParamIndex());
				_pFirstArc->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadCircle_));
			}
			else
			{
				CCadArc* _pCadArc = new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),
					_nStartAngle, _nCenterAngle, pCadCircle_->IsNormalDir());
				_pCadArc->Select(pCadCircle_->IsSelected());
				_pCadArc->SetParamIndex(pCadCircle_->GetParamIndex());
				_pCadArc->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadCircle_));
				pCadObjectListDes_->push_back(_pCadArc);
			}
		}

		_bFirstSegment = false;

		_nStartAngle = _nStartAngle + _nCenterAngle + (_nLength / pCadCircle_->GetLength()) * 2 * c_nPIE;

		_nIndex++;
		_strKey.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadCircle_->GetKey(_strKey);
	}

	if (!DOUBLE_EQU(_nStartAngle, 2 * c_nPIE))
	{
		double _nCenterAngle = 2 * c_nPIE - _nStartAngle;
		if (_pFirstArc)
		{
			_nCenterAngle += _pFirstArc->GetCentralAngle();
			delete _pFirstArc;
			_pFirstArc = NULL;
		}
		CCadArc* _pCadArc = new CCadArc(pCadCircle_->GetCenterPoint(), pCadCircle_->GetRadius(),
			_nStartAngle, _nCenterAngle, pCadCircle_->IsNormalDir());
		_pCadArc->Select(pCadCircle_->IsSelected());
		_pCadArc->SetParamIndex(pCadCircle_->GetParamIndex());
		_pCadArc->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadCircle_));
		pCadObjectListDes_->push_back(_pCadArc);
	}
	else
	{
		if (_pFirstArc)
		{
			pCadObjectListDes_->push_back(_pFirstArc);
		}
	}

	return pCadObjectListDes_->size() > _nOldSize;
}

bool CMicroJointTarget::_ExplodeArcMicroJoint(const nce::CCadArc* pCadArc_, nce::CadObjectList* pCadObjectListDes_)
{
	size_t _nOldSize = pCadObjectListDes_->size();
	int _nIndex = 0;
	CStringA _strKey;
	_strKey.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadArc_->GetKey(_strKey);
	double _nStartAngle = pCadArc_->GetStartAngle();
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));

		double _nCenterAngle = (pCadArc_->GetCentralAngle() * _nRatio + pCadArc_->GetStartAngle()) - _nStartAngle;
		NormalizeRadian(_nCenterAngle);
		if (!DOUBLE_EQU_ZERO(_nCenterAngle))
		{
			CCadArc* _pCadArc = new CCadArc(pCadArc_->GetCenterPoint(), pCadArc_->GetRadius(),
				_nStartAngle, _nCenterAngle, pCadArc_->IsNormalDir());
			_pCadArc->Select(pCadArc_->IsSelected());
			_pCadArc->SetParamIndex(pCadArc_->GetParamIndex());
			_pCadArc->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadArc_));
			pCadObjectListDes_->push_back(_pCadArc);
		}

		_nStartAngle = _nStartAngle + _nCenterAngle + (_nLength / pCadArc_->GetLength()) * pCadArc_->GetCentralAngle();
		NormalizeRadian(_nStartAngle);
		_nIndex++;
		_strKey.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadArc_->GetKey(_strKey);
	}

	if (!DOUBLE_EQU(_nStartAngle, pCadArc_->GetEndAngle()))
	{
		double _nCenterAngle = pCadArc_->GetEndAngle() - _nStartAngle;
		if (_nCenterAngle < 0.)
		{
			_nCenterAngle += 2 * c_nPIE;
		}
		CCadArc* _pCadArc = new CCadArc(pCadArc_->GetCenterPoint(), pCadArc_->GetRadius(),
			_nStartAngle, _nCenterAngle, pCadArc_->IsNormalDir());
		_pCadArc->Select(pCadArc_->IsSelected());
		_pCadArc->SetParamIndex(pCadArc_->GetParamIndex());
		_pCadArc->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadArc_));
		pCadObjectListDes_->push_back(_pCadArc);
	}

	return pCadObjectListDes_->size() > _nOldSize;
}

bool CMicroJointTarget::_ExplodeEllipseMicroJoint(const nce::CCadEllipse* pCadEllipse_, nce::CadObjectList* pCadObjectListDes_)
{
	CCadPolyline* _pCadPolyline = TranslateObjectToPolyline(pCadEllipse_);
	_pCadPolyline->SetParamString(pCadEllipse_->GetParamString());
	_pCadPolyline->Select(pCadEllipse_->IsSelected());
	_pCadPolyline->SetParamIndex(pCadEllipse_->GetParamIndex());
	CCadMultiSegments* _pCadMultiSegments = (CCadMultiSegments*)(_pCadPolyline);
	// ԲıǰԲһģҲǴʣתɶ
	// ߵʱҪתɶĶκźͱ ---- DingQiang 2016-04-27
	int _nIndex = 0;
	CStringA _strKey;
	_strKey.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = _pCadMultiSegments->GetKey(_strKey);
	DPOINT2 _ptStartPoint = _pCadMultiSegments->GetStartPoint();
	int _nTailSegCount = 0;
	double _nTaliRatio = 0.;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		double _nTargetLength = _pCadMultiSegments->GetLength() * _nRatio;
		double _nLengthTemp = 0.;
		int _nSegTemp = 0;
		while(_nLengthTemp < _nTargetLength && _nSegTemp < _pCadMultiSegments->GetSegmentCount())
		{
			SEGMENT _segment = _pCadMultiSegments->GetSegment(_nSegTemp);
			if (_nLengthTemp + _segment.GetLength() >= _nTargetLength)
			{
				_nRatio = (_nTargetLength - _nLengthTemp) / _segment.GetLength();
				_strValueTemp.Format(c_szMJFormat, _nSegTemp, _nRatio, _nLength);
				_pCadMultiSegments->SetKey(_strKey, _strValueTemp);
				break;
			}
			_nLengthTemp += _segment.GetLength();
			_nSegTemp++;
		}
		_nIndex++;
		_strKey.Format(c_szMJ, _nIndex);
		_strValueTemp = _pCadMultiSegments->GetKey(_strKey);
	}

	bool _bRet = _ExplodeMultiSegmentMicroJoint(_pCadMultiSegments, pCadObjectListDes_);
	delete _pCadPolyline;
	_pCadPolyline = NULL;
	return _bRet;
}

bool CMicroJointTarget::_ExplodeEllipseArcMicroJoint(const nce::CCadEllipseArc* pCadEllipseArc_, nce::CadObjectList* pCadObjectListDes_)
{
	CCadPolyline* _pCadPolyline = TranslateObjectToPolyline(pCadEllipseArc_);
	_pCadPolyline->SetParamString(pCadEllipseArc_->GetParamString());
	_pCadPolyline->Select(pCadEllipseArc_->IsSelected());
	_pCadPolyline->SetParamIndex(pCadEllipseArc_->GetParamIndex());
	CCadMultiSegments* _pCadMultiSegments = (CCadMultiSegments*)(_pCadPolyline);
	// ԲıǰԲһģҲǴʣתɶ
	// ߵʱҪתɶĶκźͱ ---- DingQiang 2016-04-27
	int _nIndex = 0;
	CStringA _strKey;
	_strKey.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = _pCadMultiSegments->GetKey(_strKey);
	DPOINT2 _ptStartPoint = _pCadMultiSegments->GetStartPoint();
	int _nTailSegCount = 0;
	double _nTaliRatio = 0.;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));
		double _nTargetLength = _pCadMultiSegments->GetLength() * _nRatio;
		double _nLengthTemp = 0.;
		int _nSegTemp = 0;
		while(_nLengthTemp < _nTargetLength && _nSegTemp < _pCadMultiSegments->GetSegmentCount())
		{
			SEGMENT _segment = _pCadMultiSegments->GetSegment(_nSegTemp);
			if (_nLengthTemp + _segment.GetLength() >= _nTargetLength)
			{
				_nRatio = (_nTargetLength - _nLengthTemp) / _segment.GetLength();
				_strValueTemp.Format(c_szMJFormat, _nSegTemp, _nRatio, _nLength);
				_pCadMultiSegments->SetKey(_strKey, _strValueTemp);
				break;
			}
			_nLengthTemp += _segment.GetLength();
			_nSegTemp++;
		}
		_nIndex++;
		_strKey.Format(c_szMJ, _nIndex);
		_strValueTemp = _pCadMultiSegments->GetKey(_strKey);
	}

	bool _bRet = _ExplodeMultiSegmentMicroJoint(_pCadMultiSegments, pCadObjectListDes_);
	delete _pCadPolyline;
	_pCadPolyline = NULL;
	return _bRet;
}

bool CMicroJointTarget::_ExplodeMultiSegmentMicroJoint(const nce::CCadMultiSegments* pCadMultiSegments_, nce::CadObjectList* pCadObjectListDes_)
{
	size_t _nOldSize = pCadObjectListDes_->size();
	int _nIndex = 0;
	CStringA _strKey;
	_strKey.Format(c_szMJ, _nIndex);
	CStringA _strValueTemp = pCadMultiSegments_->GetKey(_strKey);
	DPOINT2 _ptStartPoint = pCadMultiSegments_->GetStartPoint();
	int _nSegmentCount = pCadMultiSegments_->GetSegmentCount();
	int _nTailSegCount = 0;
	double _nTaliRatio = 0.;
	bool _bFirstSegment = true;
	CCadPolyline* _pCadFirstPolyline = NULL;
	while(!_strValueTemp.IsEmpty())
	{
		int _nSeg = 0;
		double _nRatio = 0.;
		double _nLength = 0.;
		VERIFY(3 == sscanf_s(_strValueTemp, c_szMJFormat, &_nSeg, &_nRatio, &_nLength));

		if(_nTailSegCount >= _nSegmentCount)
		{
			TRACE("EXCEPTION: _nTailSegCount=%d out of _nSegmentCount=%d at CMicroJointTarget:%d.\n", _nTailSegCount, _nSegmentCount, __LINE__);
			// ASSERT(_nTailSegCount < _nSegmentCount);
			break;
		}

		CCadPolyline* _pCadPolyline = new CCadPolyline;
		_pCadPolyline->SetStartPoint(_ptStartPoint);
		while(_nTailSegCount < _nSeg)
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
			if (_Segment.nType == SEGMENT::LINE)
			{
				_pCadPolyline->AddLine(_Segment.ptEnd);
				_nTaliRatio = 0;
			}
			else if (_Segment.nType == SEGMENT::ARC)
			{
				DPOINT2 _ptCenter;
				double _nRadius,_nStartAngle,_nSweptAngele;
				SEGMENT::SolveArc(_Segment.ptStart,
					_Segment.ptEnd,
					_Segment.nBulge,
					&_ptCenter,
					&_nRadius,
					&_nStartAngle,
					&_nSweptAngele);
				_pCadPolyline->AddArc(_Segment.ptEnd, tan((1 - _nTaliRatio) * _nSweptAngele / 4));
				_nTaliRatio = 0;
			}
			else
			{
				ASSERT(FALSE);
			}

			_nTailSegCount++;
			
			// TFS#3741ĳЩͼθ쳣MJ΢κ>=߶Ǽ򻯻ϲͼʱδMJ£
			// ʱ_nTailSegCount++ᵼºִGetSegmentĴԽ
			// ڴжһ¶ߵĶ쳣MJڣǰ쳣MJʾ ---- yangxiaobin 2017-05-16
			if (_nTailSegCount >= _nSegmentCount)
			{
				TRACE("EXCEPTION: _nTailSegCount=%d out of _nSegmentCount=%d at CMicroJointTarget:%d!\n", _nTailSegCount, _nSegmentCount, __LINE__);
				// ASSERT(_nTailSegCount < _nSegmentCount);
				break;
			}
		}
		// TFS#3741ֹ쳣MJµGetSegmentԽ    ---- yangxiaobin 2017-05-16
		// ΢SegmentΪͷ(SS)---΢ʼ(MJS)--΢ĩ(MJE)----β(SE)
		double _nNeedLength = 0.0;
		if (_nTailSegCount < _nSegmentCount) 
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
			_nNeedLength = _Segment.GetLength() * _nRatio + _nLength; //MJEڵλ
			if (_Segment.nType == SEGMENT::LINE)
			{
				_pCadPolyline->AddLine(_Segment.GetPoint(_nRatio));
			}
			else if (_Segment.nType == SEGMENT::ARC)
			{
				DPOINT2 _ptCenter;
				double _nRadius,_nStartAngle,_nSweptAngele;
				SEGMENT::SolveArc(_Segment.ptStart,
					_Segment.ptEnd,
					_Segment.nBulge,
					&_ptCenter,
					&_nRadius,
					&_nStartAngle,
					&_nSweptAngele);

				_pCadPolyline->AddArc(_Segment.GetPoint(_nRatio), tan(_nSweptAngele * (_nRatio - _nTaliRatio) / 4));
			}
			else
			{
				ASSERT(FALSE);
			}
		}

		if (_pCadPolyline->GetNodeCount() <= 0)
		{
			delete _pCadPolyline;
			_pCadPolyline = NULL;
		}
		else
		{
			if (_bFirstSegment && pCadMultiSegments_->IsClose())
			{
				_pCadFirstPolyline = _pCadPolyline;
			}
			else
			{
				_pCadPolyline->Select(pCadMultiSegments_->IsSelected());
				_pCadPolyline->SetParamIndex(pCadMultiSegments_->GetParamIndex());
				_pCadPolyline->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadMultiSegments_));
				pCadObjectListDes_->push_back(_pCadPolyline);
			}
		}
		_nTaliRatio = _nRatio;

		_bFirstSegment = false;

		while(DOUBLE_GREAT(_nLength, 0) && _nTailSegCount < pCadMultiSegments_->GetSegmentCount())
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
			double _nSegLength =_Segment.GetLength();
			if (_nSegLength > _nNeedLength)
			{
				_ptStartPoint =_Segment.GetPoint(_nNeedLength / _nSegLength);
				_nTaliRatio = _nNeedLength / _nSegLength;
				break;
			}

			_nNeedLength -= _nSegLength;
			_nTailSegCount++;
		}

		_nIndex++;
		_strKey.Format(c_szMJ, _nIndex);
		_strValueTemp = pCadMultiSegments_->GetKey(_strKey);
	}

	if (_ptStartPoint != pCadMultiSegments_->GetEndPoint()
		&& _nTailSegCount < pCadMultiSegments_->GetSegmentCount())
	{
		CCadPolyline* _pCadPolyine = new CCadPolyline;
		_pCadPolyine->SetStartPoint(pCadMultiSegments_->GetSegment(_nTailSegCount).GetPoint(_nTaliRatio));
		bool _bFirst = true;
		while(_nTailSegCount < pCadMultiSegments_->GetSegmentCount())
		{
			SEGMENT _Segment = pCadMultiSegments_->GetSegment(_nTailSegCount);
			if (_Segment.nType == SEGMENT::LINE)
			{
				_pCadPolyine->AddLine(_Segment.ptEnd);
				_bFirst = false;
			}
			else if (_Segment.nType == SEGMENT::ARC)
			{
				if (_bFirst)
				{
					DPOINT2 _ptCenter;
					double _nRadius,_nStartAngle,_nSweptAngele;
					SEGMENT::SolveArc(_Segment.ptStart,
						_Segment.ptEnd,
						_Segment.nBulge,
						&_ptCenter,
						&_nRadius,
						&_nStartAngle,
						&_nSweptAngele);
					_pCadPolyine->AddArc(_Segment.ptEnd, tan(_nSweptAngele * (1-_nTaliRatio) / 4));
					_bFirst = false;
				}
				else
				{
					_pCadPolyine->AddArc(_Segment.ptEnd, _Segment.nBulge);
				}
			}
			else
			{
				ASSERT(FALSE);
			}
			_nTailSegCount++;
		}
		if (_pCadPolyine->GetNodeCount() <= 0)
		{
			delete _pCadPolyine;
			_pCadPolyine = NULL;
			if (_pCadFirstPolyline)
			{
				_pCadFirstPolyline->Select(pCadMultiSegments_->IsSelected());
				_pCadFirstPolyline->SetParamIndex(pCadMultiSegments_->GetParamIndex());
				_pCadFirstPolyline->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadMultiSegments_));
				pCadObjectListDes_->push_back(_pCadFirstPolyline);
			}
		}
		else
		{
			if (_pCadFirstPolyline)
			{
				ASSERT(_pCadFirstPolyline->GetStartPoint() == _pCadPolyine->GetEndPoint());
				for (int i = 0; i < _pCadFirstPolyline->GetSegmentCount(); i++)
				{
					CCadPolyline::ARCNODE _Segment = _pCadFirstPolyline->GetAt(i);
					if (_Segment.nBulge == 0.)
					{
						_pCadPolyine->AddLine(_Segment.ptEnd);
					}
					else
					{
						_pCadPolyine->AddArc(_Segment.ptEnd, _Segment.nBulge);
					}
				}
			}
			_pCadPolyine->Select(pCadMultiSegments_->IsSelected());
			_pCadPolyine->SetParamIndex(pCadMultiSegments_->GetParamIndex());
			_pCadPolyine->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadMultiSegments_));
			pCadObjectListDes_->push_back(_pCadPolyine);
		}
	}
	else if (_pCadFirstPolyline)
	{
		_pCadFirstPolyline->Select(pCadMultiSegments_->IsSelected());
		_pCadFirstPolyline->SetParamIndex(pCadMultiSegments_->GetParamIndex());
		_pCadFirstPolyline->SetParamString(_GetCadObjectParamExceptMicroJoint(pCadMultiSegments_));
		pCadObjectListDes_->push_back(_pCadFirstPolyline);
	}

	return pCadObjectListDes_->size() > _nOldSize;
}

CStringA CMicroJointTarget::_GetCadObjectParamExceptMicroJoint(const nce::CCadObject* pCadObject_)
{
	CCadObject* _pCadObject = pCadObject_->Clone();
	ClearMicroJoint(_pCadObject);
	CString _strParam = _pCadObject->GetParamString();
	delete _pCadObject;
	_pCadObject = NULL;
	return _strParam;
}

void CMicroJointTarget::CombineMicroJoint(nce::CCadObject* pCadObject_, std::vector<MJItem>& vectMJItem_)
{
	switch(pCadObject_->GetType())
	{
	case cadline:
		_CombineLineMicroJoint((CCadLine*)pCadObject_, vectMJItem_);
		break;
	case cadcircle:
		_CombineCircleMicroJoint((CCadCircle*)pCadObject_, vectMJItem_);
		break;
	case cadarc:
		_CombineArcMicroJoint((CCadArc*)pCadObject_, vectMJItem_);
		break;
	case cadellipse:
		_CombineEllipseMicroJoint((CCadEllipse*)pCadObject_, vectMJItem_);
		break;
	case cadellipsearc:
		_CombineEllipseArcMicroJoint((CCadEllipseArc*)pCadObject_, vectMJItem_);
		break;
	case cadpolygon:
	case cadpolyline:
	case cadstar:
	case cadrectangle:
		_CombineMultiSegmentMicroJoint((CCadMultiSegments*)pCadObject_, vectMJItem_);
		break;
	case caddot:
		ASSERT(FALSE);
		break;
	}
}

void CMicroJointTarget::_CombineLineMicroJoint(nce::CCadLine* pCadLine_, std::vector<MJItem>& vectMJItem_)
{
	ASSERT(vectMJItem_.size() > 0);

	double _nTotleLength = pCadLine_->GetLength();
	vector<MJItem>::iterator _itLast = vectMJItem_.begin();
	for (vector<MJItem>::iterator _itThis = vectMJItem_.begin(); _itThis != vectMJItem_.end(); _itThis++)
	{
		if (_itThis == vectMJItem_.begin())
		{
			if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
			{
				_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
				if (DOUBLE_EQU(_itThis->nLength, 0))
				{
					vectMJItem_.erase(_itThis);
					_itLast = _itThis = vectMJItem_.begin();
				}
				break;
			}

			continue;
		}

		// ΢Ϊ0ߵǰ΢ǰһ΢Χȥǰ΢ ---- DingQiang 2016-03-11
		if (_itThis->nRatio < _itLast->nRatio + _itLast->nLength / _nTotleLength)
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		// ʣ벻΢ȣ΢ȵΪʣĳ ---- DingQiang 2016-03-11
		if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
		{
			_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
		}

		if (DOUBLE_EQU(_itThis->nLength, 0))
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		_itLast = _itThis;
	}
}

void CMicroJointTarget::_CombineCircleMicroJoint(nce::CCadCircle* pCadCircle_, std::vector<MJItem>& vectMJItem_)
{
	ASSERT(vectMJItem_.size() > 0);

	double _nTotleLength = pCadCircle_->GetLength();
	vector<MJItem>::iterator _itLast = vectMJItem_.begin();
	for (vector<MJItem>::iterator _itThis = vectMJItem_.begin(); _itThis != vectMJItem_.end(); _itThis++)
	{
		if (_itThis == vectMJItem_.begin())
		{
			if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
			{
				_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
				if (DOUBLE_EQU(_itThis->nLength, 0))
				{
					vectMJItem_.erase(_itThis);
					_itLast = _itThis = vectMJItem_.begin();
				}
				break;
			}

			continue;
		}

		// ΢Ϊ0ߵǰ΢ǰһ΢Χȥǰ΢ ---- DingQiang 2016-03-11
		if (_itThis->nRatio < _itLast->nRatio + _itLast->nLength / _nTotleLength)
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		// ʣ벻΢ȣ΢ȵΪʣĳ ---- DingQiang 2016-03-11
		if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
		{
			_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
		}

		if (DOUBLE_EQU(_itThis->nLength, 0))
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		_itLast = _itThis;
	}
}

void CMicroJointTarget::_CombineArcMicroJoint(nce::CCadArc* pCadArc_, std::vector<MJItem>& vectMJItem_)
{
	ASSERT(vectMJItem_.size() > 0);

	double _nTotleLength = pCadArc_->GetLength();
	vector<MJItem>::iterator _itLast = vectMJItem_.begin();
	for (vector<MJItem>::iterator _itThis = vectMJItem_.begin(); _itThis != vectMJItem_.end(); _itThis++)
	{
		if (_itThis == vectMJItem_.begin())
		{
			if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
			{
				_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
				if (DOUBLE_EQU(_itThis->nLength, 0))
				{
					vectMJItem_.erase(_itThis);
					_itLast = _itThis = vectMJItem_.begin();
				}
				break;
			}

			continue;
		}

		// ǰ΢ǰһ΢Χȥǰ΢ ---- DingQiang 2016-03-11
		if (_itThis->nRatio < _itLast->nRatio + _itLast->nLength / _nTotleLength)
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		// ʣ벻΢ȣ΢ȵΪʣĳ ---- DingQiang 2016-03-11
		if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
		{
			_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
		}

		if (DOUBLE_EQU(_itThis->nLength, 0))
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		_itLast = _itThis;
	}
}

void CMicroJointTarget::_CombineMultiSegmentMicroJoint(nce::CCadMultiSegments* pMultiSegment_, std::vector<MJItem>& vectMJItem_)
{
	ASSERT(vectMJItem_.size() > 0);
	int _nSegmentCounts = pMultiSegment_->GetSegmentCount();
	double _nTotleLength = pMultiSegment_->GetLength();
	vector<MJItem>::iterator _itLast = vectMJItem_.begin();
	for (vector<MJItem>::iterator _itThis = vectMJItem_.begin(); _itThis != vectMJItem_.end(); _itThis++)
	{
		double _nHeadLength;
		double _nTailLength;
		_GetMultiSenmentDivdeLength(pMultiSegment_, _itThis->nSeg, _itThis->nRatio, _nHeadLength, _nTailLength);
		if (_itThis == vectMJItem_.begin())
		{
			if (_itThis->nLength > _nTailLength)
			{
				_itThis->nLength = _nTailLength;
				if (DOUBLE_EQU(_itThis->nLength, 0))
				{
					vectMJItem_.erase(_itThis);
					_itLast = _itThis = vectMJItem_.begin();
				}
				break;
			}

			continue;
		}


		// жһ΢ǲһ΢Ŀճ
		MJItem _nLastEnd;
		bool _bFind = false;
		double _nLastSegRemained = pMultiSegment_->GetSegment(_itLast->nSeg).GetLength() * (1 - _itLast->nRatio);
		if (_nLastSegRemained >= _itLast->nLength)
		{
			_nLastEnd.nSeg = _itLast->nSeg;
			_nLastEnd.nRatio = _itLast->nRatio + _itLast->nLength / pMultiSegment_->GetSegment(_itLast->nSeg).GetLength();
			_bFind = true;
		}
		else
		{
			// һβ㹻
			double _nTempLength = _itLast->nLength - _nLastSegRemained;
			int _nSegNo = _itLast->nSeg + 1;
			while(_nSegNo < pMultiSegment_->GetSegmentCount())
			{
				if (_nTempLength < pMultiSegment_->GetSegment(_nSegNo).GetLength())
				{
					_nLastEnd.nSeg = _nSegNo;
					_nLastEnd.nRatio = _nTempLength / pMultiSegment_->GetSegment(_nSegNo).GetLength();
					_bFind = true;
					break;
				}
				_nSegNo++;
				if (_nSegNo < pMultiSegment_->GetSegmentCount())
				{
					_nTempLength -= pMultiSegment_->GetSegment(_nSegNo).GetLength();
				}
			}
		}
		ASSERT(_bFind);

		// ǰ΢ǰһ΢Χȥǰ΢ ---- DingQiang 2016-03-11
		if (!_bFind || *_itThis < _nLastEnd )
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		// ʣ벻΢ȣ΢ȵΪʣĳ ---- DingQiang 2016-03-11
		if (_itThis->nLength > _nTailLength)
		{
			_itThis->nLength = _nTailLength;
		}

		if (DOUBLE_EQU(_itThis->nLength, 0))
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		_itLast = _itThis;
	}
}


void CMicroJointTarget::_GetMultiSenmentDivdeLength(IN nce::CCadMultiSegments* pMultiSegment_, IN UINT nSegNo_, IN double nRatio_,
													OUT double& nHeadLength_, OUT double& nTailLength_)
{
	UINT _nSegIndex = 0;
	nHeadLength_ = 0.;
	if (nSegNo_ >= (UINT)pMultiSegment_->GetSegmentCount())
	{
		ASSERT(FALSE);
		return;
	}

	while (_nSegIndex < nSegNo_)
	{
		SEGMENT _segment = pMultiSegment_->GetSegment(_nSegIndex++);
		nHeadLength_ += _segment.GetLength();
	}
	ASSERT(_nSegIndex == nSegNo_);
	SEGMENT _segment = pMultiSegment_->GetSegment(_nSegIndex);
	nHeadLength_ += _segment.GetLength() * nRatio_;
	nTailLength_ = pMultiSegment_->GetLength() - nHeadLength_;
}

void CMicroJointTarget::_CombineEllipseMicroJoint(nce::CCadEllipse* pCadEllipse_, std::vector<MJItem>& vectMJItem_)
{
	ASSERT(vectMJItem_.size() > 0);

	double _nTotleLength = pCadEllipse_->GetLength();
	vector<MJItem>::iterator _itLast = vectMJItem_.begin();
	for (vector<MJItem>::iterator _itThis = vectMJItem_.begin(); _itThis != vectMJItem_.end(); _itThis++)
	{
		if (_itThis == vectMJItem_.begin())
		{
			if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
			{
				_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
				if (DOUBLE_EQU(_itThis->nLength, 0))
				{
					vectMJItem_.erase(_itThis);
					_itLast = _itThis = vectMJItem_.begin();
				}
				break;
			}

			continue;
		}

		// ΢Ϊ0ߵǰ΢ǰһ΢Χȥǰ΢ ---- DingQiang 2016-03-11
		if (_itThis->nRatio < _itLast->nRatio + _itLast->nLength / _nTotleLength)
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		// ʣ벻΢ȣ΢ȵΪʣĳ ---- DingQiang 2016-03-11
		if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
		{
			_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
		}

		if (DOUBLE_EQU(_itThis->nLength, 0))
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		_itLast = _itThis;
	}
}

void CMicroJointTarget::_CombineEllipseArcMicroJoint(nce::CCadEllipseArc* pCadEllipseArc_, std::vector<MJItem>& vectMJItem_)
{
	ASSERT(vectMJItem_.size() > 0);

	double _nTotleLength = pCadEllipseArc_->GetLength();
	vector<MJItem>::iterator _itLast = vectMJItem_.begin();
	for (vector<MJItem>::iterator _itThis = vectMJItem_.begin(); _itThis != vectMJItem_.end(); _itThis++)
	{
		if (_itThis == vectMJItem_.begin())
		{
			if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
			{
				_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
				if (DOUBLE_EQU(_itThis->nLength, 0))
				{
					vectMJItem_.erase(_itThis);
					_itLast = _itThis = vectMJItem_.begin();
				}
				break;
			}

			continue;
		}

		// ǰ΢ǰһ΢Χȥǰ΢ ---- DingQiang 2016-03-11
		if (_itThis->nRatio < _itLast->nRatio + _itLast->nLength / _nTotleLength)
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		// ʣ벻΢ȣ΢ȵΪʣĳ ---- DingQiang 2016-03-11
		if (_itThis->nLength > _nTotleLength * (1 - _itThis->nRatio))
		{
			_itThis->nLength = _nTotleLength * (1 - _itThis->nRatio);
		}

		if (DOUBLE_EQU(_itThis->nLength, 0))
		{
			vectMJItem_.erase(_itThis);
			_itThis = _itLast;
			continue;
		}

		_itLast = _itThis;
	}
}

static bool FindObjectByID(IN CCadObject* pCadObj_, IN int nID_, OUT CCadObject** ppResult_)
{
	if (pCadObj_->GetID() == nID_)
	{
		*ppResult_ = pCadObj_;
		return true;
	}

	if (pCadObj_->GetType() == cadgroup)
	{
		CCadGroup* _pCadGroup = dynamic_cast<CCadGroup*>(pCadObj_);
		ASSERT(_pCadGroup);
		for (int _i = 0; _i < _pCadGroup->GetNumOfChild(); _i++)
		{
			CCadObject* _pCad = _pCadGroup->GetAt(_i);
			if (FindObjectByID(_pCad, nID_, ppResult_))
				return true;
		}
	}

	return false;
}

static double GetCadLength(IN const CCadObject* pCadObj_)
{
	switch(pCadObj_->GetType())
	{
	case cadgroup:
		return -1;
	case cadline:
		return ((const CCadLine*)pCadObj_)->GetLength();
	case cadcircle:
		return ((const CCadCircle*)pCadObj_)->GetLength();
	case cadarc:
		return ((const CCadArc*)pCadObj_)->GetLength();
	case cadellipse:
		return ((const CCadEllipse*)pCadObj_)->GetLength();
	case cadellipsearc:
		return ((const CCadEllipseArc*)pCadObj_)->GetLength();
	case cadpolygon:
		return ((const CCadPolygon*)pCadObj_)->GetLength();
	case cadpolyline:
		return ((const CCadPolyline*)pCadObj_)->GetLength();
	case cadstar:
		return ((const CCadStar*)pCadObj_)->GetLength();
	case cadrectangle:
		return ((const CCadRectangle*)pCadObj_)->GetLength();
	}

	return false;
}
