#include "stdafx.h"
//#include "Geo_Head.h"
//#include "Triangulation.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

namespace leadingline{

Point_2 g_ptCurPosition;

string PointTypeString[]=
{
	"PT_BEGIN",
	"PT_END",
	"PT_REGULARLEFT",
	"PT_REGULARRIGHT",
	"PT_MERGE",
	"PT_SPLIT",
	"PT_NULL"
};

turn_dir_t leadingline::GetTurnType(Point_2 ptPre,Point_2 ptCur,Point_2 ptNext)//VectorD2D vec1_, VectorD2D vec2_
{
	Point_2 vec1_(ptCur.x-ptPre.x ,ptCur.y - ptPre.y);
	Point_2 vec2_(ptNext.x-ptCur.x ,ptNext.y - ptCur.y);
	double sinVal = vec1_.x*vec2_.y - vec1_.y*vec2_.x;
	return GetTurnType(vec1_,vec2_);
}

turn_dir_t leadingline::GetTurnType(Point_2 vec1_,Point_2 vec2_)
{
	double sinVal = vec1_.x*vec2_.y - vec1_.y*vec2_.x;

	if (abs(sinVal)<1.e-9) //DOUBLE_EQU_ZERO(res) 
	{
		//same direction or opposite direction
		double cosVal = vec1_.x*vec2_.x + vec1_.y*vec2_.y;
		if (cosVal>0)
		{
			return turn_same;
		}
		else if (cosVal<0)
		{
			return turn_opposite;
		}
		else
		{
			_ASSERT(0); //It's impossible here
			return turn_null;
		}
	}
	else if (sinVal>0)
	{
		return turn_left;
	}
	else if (sinVal<0)
	{
		return turn_right;
	}
	_ASSERT(0);
	return turn_null;
}

void leadingline::TransformLineExpressiong(IN Point_2 lineStart_,IN Point_2 lineEnd_,OUT double& A,OUT double& B,OUT double& C)
{
	A = lineEnd_.y - lineStart_.y;
	B = lineStart_.x - lineEnd_.x;
	C = A*lineStart_.x + B*lineStart_.y;
}


bool leadingline::GetLineIntersect(IN Point_2 line1start_, IN Point_2 line1end_, IN Point_2 line2start_, IN Point_2 line2end_,OUT Point_2& ret_)
{
	double A1,A2,B1,B2,C1,C2;
	TransformLineExpressiong(line1start_,line1end_,A1,B1,C1);
	TransformLineExpressiong(line2start_,line2end_,A2,B2,C2);
	double det = A1*B2 - A2*B1;
	if(DOUBLE_EQU_ZERO(det)){			//Lines are parallel
		return false;
	}else{
		ret_.x = (B2*C1 - B1*C2)/det;
		ret_.y = (A1*C2 - A2*C1)/det;
	}
	return true;
}

double leadingline::Direction(Point_2 Pi,Point_2 Pj,Point_2 Pk)
{
	return (Pj.x - Pi.x)*(Pk.y - Pi.y) - (Pk.x - Pi.x)*(Pj.y - Pi.y);
}

bool leadingline::OnSegment(const Segment& seg, const Point_2& pt)
{
	if ((min(seg.ptStart.x,seg.ptEnd.x) <= pt.x && pt.x <= max(seg.ptStart.x,seg.ptEnd.x))
		&&(min(seg.ptStart.y,seg.ptEnd.y) <= pt.y && pt.y <= max(seg.ptStart.y,seg.ptEnd.y))
		)
	{
		return true;
	}
	else 
		return false;
}

bool leadingline::IsIntersect(const Segment& s1,const Segment& s2)
{
	double d1,d2,d3,d4;
	d1 = Direction(s2.ptStart,s2.ptEnd,s1.ptStart);
	d2 = Direction(s2.ptStart,s2.ptEnd,s1.ptEnd);
	d3 = Direction(s1.ptStart,s1.ptEnd,s2.ptStart);
	d4 = Direction(s1.ptStart,s1.ptEnd,s2.ptEnd);
	if (DOUBLE_EQU_ZERO(d1) && OnSegment(s2,s1.ptStart))
	{
		return true;
	}
	else if (DOUBLE_EQU_ZERO(d2) && OnSegment(s2,s1.ptEnd))
	{
		return true;
	}
	else if (DOUBLE_EQU_ZERO(d3) && OnSegment(s1,s2.ptStart))
	{
		return true;
	}
	else if (DOUBLE_EQU_ZERO(d4) && OnSegment(s1,s2.ptEnd))
	{
		return true;
	}
	else if (((d1>0 && d2<0) || (d1<0 && d2>0))
		&&((d3>0 && d4<0) || (d3<0 && d4>0)))
	{
		return true;
	}
	return false;
}

bool leadingline::RawSegmentInterSection(const Segment& s1,const Segment& s2,Point_2& ptOut)
{
	if (IsIntersect(s1,s2))
	{
		return GetLineIntersect(s1.ptStart,s1.ptEnd,s2.ptStart,s2.ptEnd,ptOut);
	}
	return false;
}

bool leadingline::operator<(const Point_2& pt1,const Point_2& pt2)
{
	if (abs(pt1.y - pt2.y)<1.0e-9)
	{
		if (abs(pt1.x - pt2.x)<1.0e-9)
		{
			return false;
		}
		else if (pt1.x < pt2.x)
		{
			return true;
		}
		else
			return false;
	}
	else if (pt1.y > pt2.y)
	{
		return true;
	}
	else 
		return false;
}

bool leadingline::operator==(const Point_2& pt1,const Point_2& pt2)
{
	return (DOUBLE_EQU_ZERO(pt1.x-pt2.x) && DOUBLE_EQU_ZERO(pt1.y-pt2.y));
}

bool leadingline::operator<(const DoubleEdge& eg1,const DoubleEdge& eg2)
{
	return (eg1.m_ptStart) < (eg2.m_ptStart) ;
}


bool leadingline::PointUpperThan(const Point_2& pt1,const Point_2& pt2)
{
	if (abs(pt1.y-pt2.y)<1.e-9)
	{
		if (pt1.x<pt2.x)
		{
			return true;
		}
		else
			return false;
	}
	else if (pt1.y>pt2.y)
	{
		return true;
	}
	else
		return false;
}

bool leadingline::PointLeftOf(const Point_2& pt1,const Point_2& pt2)
{
	if (abs(pt1.x-pt2.x)<1.e-9)
	{
		if (pt1.y<pt2.y)
		{
			return true;
		}
		else
			return false;
	}
	else if (pt1.x<pt2.x)
	{
		return true;
	}
	else
		return false;
}
bool leadingline::operator<(const DoubleEdgeRep& eg1,const DoubleEdgeRep& eg2)
{
	return (*eg1.m_pEdge)<(*eg2.m_pEdge);
}

int leadingline::CompareDoubleEdge(const DoubleEdge* node1,const DoubleEdge* node2)
{
	//-1 0 1 ,Сڣڣ
	if (node1->m_edgeId == node2->m_edgeId)
	{
		return 0;
	}
	else
	{
		DoubleEdge* pNext1 = node1->m_pNext;
		DoubleEdge* pNext2 = node2->m_pNext;
		Point_2 pt1Start = node1->m_ptStart;
		Point_2 pt1End   = pNext1->m_ptStart;

		Point_2 pt2Start = node2->m_ptStart;
		Point_2 pt2End   = pNext2->m_ptStart;

		double curY = g_ptCurPosition.y;
		double edge1x ,edge2x;

		if ( abs(pt1Start.y-pt1End.y)<1.e-9 )
		{
			edge1x = pt1Start.x;
		}
		else
		{
			edge1x =( (curY-pt1Start.y)*(pt1End.x-pt1Start.x)/(pt1End.y-pt1Start.y) ) + pt1Start.x ;
		}

		if ( abs(pt2Start.y-pt2End.y)<1.e-9 ) 
		{
			edge2x = pt2Start.x;
		}
		else
		{
			edge2x =( (curY-pt2Start.y)*(pt2End.x-pt2Start.x)/(pt2End.y-pt2Start.y) ) + pt2Start.x ;
		}

		if (edge1x<edge2x)
		{
			return -1;
		}
		else
			return 1;
	}

}

// compute angle according to the relative coordinates (X, Y).
// the range of return value is (0 ~ 2PI).
/*
*(X,Y)xĵĽǶ
*/
double _compute_angle(double X, double Y)
{
	double _Radius = sqrt(powersum2(X, Y));

	if (DOUBLE_LE_ZERO(_Radius))
		return 0.0;

	ASSERT(_Radius > 0.0);
	double _temp = X / _Radius;
	__CUT(_temp, -1., 1.);
	double _angle = acos(_temp);

	// First or second quadrant
	if (DOUBLE_GE_ZERO(Y)) 
		return _angle;

	// Third or fourth quadrant	
	return 2.0 * c_nPI - _angle; 
}

void  RetrieveArcInf(IN const Point_2& ptStart_,
					 IN const Point_2&ptEnd_,
					 IN const double& nBulge_,
					 OUT Point_2* pptCenter_,
					 OUT double* pnStartAngle_,
					 OUT double* pnEndAngle_,
					 OUT double* pnRadius_)
{
	Point_2 _ptFromTo(ptEnd_.x-ptStart_.x,ptEnd_.y-ptStart_.y);
	double _nLen = sqrt(_ptFromTo.x*_ptFromTo.x+_ptFromTo.y*_ptFromTo.y);
	ASSERT(DOUBLE_GE_ZERO(_nLen));
	if (DOUBLE_EQU_ZERO(_nLen))
	{
		*pptCenter_ = ptStart_;
		*pnStartAngle_ = 0.;
		*pnEndAngle_ = 0.;
		*pnRadius_ = 0.;
		return;
	}

	// Inner angle
	double _alpha = atan(nBulge_) * 4.;
	ASSERT(DOUBLE_NOT_ZERO(_alpha));

	// 1) Radius
	double _nRadius = _nLen * 0.5 / sin(_alpha * 0.5);
	*pnRadius_ = fabs(_nRadius);

	// 2) Circle centre
	// 2-1) proportional translate
	ASSERT(DOUBLE_NOT_ZERO(_nLen));
	double _nK = _nRadius / _nLen;

	Point_2 _ptByPropTrans;
	_ptByPropTrans.x = _ptFromTo.x * _nK;
	_ptByPropTrans.y = _ptFromTo.y * _nK;

	// 2-2) revolving transform
	double _theta = (c_nPI - _alpha) * 0.5;
	Point_2 _ptByRevTrans;
	_ptByRevTrans.x = cos(_theta) * _ptByPropTrans.x - sin(_theta) * _ptByPropTrans.y;
	_ptByRevTrans.y = sin(_theta) * _ptByPropTrans.x + cos(_theta) * _ptByPropTrans.y;

	// 2-3) Get circle center 
	pptCenter_->x = ptStart_.x + _ptByRevTrans.x;
	pptCenter_->y = ptStart_.y + _ptByRevTrans.y;

	// 3) Start angle
	*pnStartAngle_ = _compute_angle(ptStart_.x - pptCenter_->x, ptStart_.y - pptCenter_->y);

	// 4) End angle
	*pnEndAngle_ = *pnStartAngle_ + _alpha;

	return;
}


double  TransformArcToSegments(IN const Point_2& ptStart_,
								   IN const Point_2& ptEnd_,
								   IN const double& nBulge_,
								   IN const double& nGap_,
								   OUT vector<Point_2>& vectorOut_)
{
	Point_2 _center;
	double _nAngleStart, _nAngleEnd, _nRadius;

	RetrieveArcInf(ptStart_, ptEnd_, nBulge_, &_center, &_nAngleStart, &_nAngleEnd, &_nRadius);

	double _nDegreeStep = 4*asin(sqrt(nGap_/(2*_nRadius)));

	
	double _nAngleCur = _nAngleStart;
	Point_2 _curVertex;
	double _delta ;

	vectorOut_.push_back(ptStart_);
	int nTotalStep ;
	if (_nAngleStart>_nAngleEnd) //˳ʱ
	{
		_delta = _nAngleStart - _nAngleEnd;
		nTotalStep = (int)(_delta/fabs(_nDegreeStep));
		if (nTotalStep==0)
		{
			++nTotalStep;
		}
		_nDegreeStep = _delta/nTotalStep;
		if (nBulge_<0)
		{
			_nDegreeStep = -_nDegreeStep;
		}	
		//_nAngleEnd += 2.0 * c_nPI;
		//while (_nAngleCur + _nDegreeStep > _nAngleEnd)
		for (int i=1;i<nTotalStep;i++)
		{
			_nAngleCur += _nDegreeStep;
			_curVertex.x = _center.x + _nRadius*cos(_nAngleCur);
			_curVertex.y = _center.y + _nRadius*sin(_nAngleCur);
			vectorOut_.push_back(_curVertex);
		}
	}
	else //ʱ
	{
		_delta = _nAngleEnd - _nAngleStart;
		nTotalStep = (int)(_delta/fabs(_nDegreeStep));
		if (nTotalStep==0)
		{
			++nTotalStep;
		}
		_nDegreeStep = _delta/nTotalStep;
		if (nBulge_<0)
		{
			_nDegreeStep = -_nDegreeStep;
		}	
		//while (_nAngleCur + _nDegreeStep < _nAngleEnd)
		for (int i=1;i<nTotalStep;i++)
		{
			_nAngleCur += _nDegreeStep;
			_curVertex.x = _center.x + _nRadius*cos(_nAngleCur);
			_curVertex.y = _center.y + _nRadius*sin(_nAngleCur);
			vectorOut_.push_back(_curVertex);
		}
	}
	return _nRadius*_delta;
}

void TransformCircleToSegments(IN const Point_2& ptStart_,
									  IN const Point_2& ptCenter_,
									  IN const double& nRadius_,
									  IN const double& nGap_,
									  OUT vector<Point_2>& vectorOut_)
{
	double _nAngleStart;

	_nAngleStart = atan( (ptStart_.y-ptCenter_.y)/(ptStart_.x - ptCenter_.x) );

	double _nDegreeStep  = 4*asin(sqrt(nGap_/(2*nRadius_)));//= nGap_/nRadius_;

	double _nAngle = 0;
	Point_2 _curVertex;

	while (_nAngle + _nDegreeStep < 2*c_nPI)
	{
		_nAngle += _nDegreeStep;
		_curVertex.x = ptCenter_.x + nRadius_*cos(_nAngleStart+_nAngle);
		_curVertex.y = ptCenter_.y + nRadius_*sin(_nAngleStart+_nAngle);
		vectorOut_.push_back(_curVertex);
	}
	_curVertex.x = ptStart_.x ;
	_curVertex.y = ptStart_.y ;
	vectorOut_.push_back(_curVertex);
}

bool operator ==(const EdgeIndex& EIdx1,const EdgeIndex& EIdx2 )
{
	if (EIdx1.MajorIndex == EIdx2.MajorIndex && EIdx1.MinorIndex == EIdx2.MinorIndex)
	{
		return true;
	}
	return false;
}

 DoubleEdgeFactory* DoubleEdgeFactory::s_pEdgeFactory = NULL;

DoubleEdge* DoubleEdgeFactory::ConstructLineEdge(const Point_2& ptStart,DoubleEdge* pNext)
{
	DoubleEdge* pNewEdge = new DoubleEdge(ptStart,pNext);
	if (pNewEdge)
	{
		m_EdgeCache.push_back(pNewEdge);
	}
	pNewEdge->m_edgeId.MajorIndex = GetNextMajorIdCnt();
	pNewEdge->m_edgeId.MinorIndex = -1;	
	return pNewEdge;
}

//ԲɢС߶
int DoubleEdgeFactory::ConstructArcEdge(const Point_2& ptStart,const Point_2& ptEnd,double nBulge,OUT DoubleEdge*& pStartEdge,OUT DoubleEdge*& pEndEdge)
{
	vector<Point_2> vectorOut;
	double arcLen = TransformArcToSegments(ptStart,ptEnd,nBulge,0.01,vectorOut);
	int nSegCnt = (int)vectorOut.size();
	pStartEdge = ConstructArcSegmentEdge(ptStart,EAET_FIRST);
	DoubleEdge* pPreEdge = pStartEdge;
	DoubleEdge* pCurEdge;

	for (int i=1;i<nSegCnt-1;i++)
	{
		pCurEdge = ConstructArcSegmentEdge(vectorOut[i],EAET_MIDDLE);
		pPreEdge->m_pNext = pCurEdge;
		pCurEdge->m_pPre = pPreEdge;
		pPreEdge = pCurEdge;
	}

	pCurEdge = ConstructArcSegmentEdge(vectorOut[nSegCnt-1],EAET_LAST);
	pPreEdge->m_pNext = pCurEdge;
	pCurEdge->m_pPre = pPreEdge;
	pEndEdge = pCurEdge;
	ArcInfo arcInfo;
	arcInfo.nArcSegNum = nSegCnt;
	arcInfo.ptArcStart = ptStart;
	arcInfo.ptArcEnd = ptEnd;
	arcInfo.nLength = arcLen;
	arcInfo.nBulge = nBulge;
	m_pTriangulation->m_mpIdxToArcInfo[pCurEdge->m_edgeId.MajorIndex] = arcInfo;
	return nSegCnt;
}

DoubleEdge* DoubleEdgeFactory::ConstructArcSegmentEdge(const Point_2& ptStart,ArcEdgeType ArcEdgeT,DoubleEdge* pNext)
{
	DoubleEdge* pNewEdge = new DoubleEdge(ptStart,pNext);
	if (pNewEdge)
	{
		m_EdgeCache.push_back(pNewEdge);
	}
	switch(ArcEdgeT)
	{
	case EAET_FIRST:
		{
			ClearMinorIdCount();
			pNewEdge->m_edgeId.MajorIndex = GetNextMajorIdCnt();
			pNewEdge->m_edgeId.MinorIndex = GetNextMinorIdCnt();
		}
		break;
	case EAET_MIDDLE:
		{
			pNewEdge->m_edgeId.MajorIndex = GetCurrentMajorIdCnt();
			pNewEdge->m_edgeId.MinorIndex = GetNextMinorIdCnt();
		}
		break;
	case EAET_LAST:
		{
			pNewEdge->m_edgeId.MajorIndex = GetCurrentMajorIdCnt();
			pNewEdge->m_edgeId.MinorIndex = GetNextMinorIdCnt();
		}
		break;
	default:
		ASSERT(0);
	}
	return pNewEdge;
}

DoubleEdgeFactory* DoubleEdgeFactory::GetDoubleEdgeFactory()
{
	if (s_pEdgeFactory==NULL)
	{
		s_pEdgeFactory = new DoubleEdgeFactory;
		s_pEdgeFactory->ClearMajorIdCount();
		s_pEdgeFactory->ClearMinorIdCount();
	}
	return s_pEdgeFactory;		
}

void DoubleEdgeFactory::ReleaseEdgeFactory()
{
	if (s_pEdgeFactory)
	{
		delete s_pEdgeFactory;
		s_pEdgeFactory = NULL;
	}
}

bool DoubleEdgeFactory::ReleaseEdges()
{
	list<DoubleEdge*>::iterator Iter = m_EdgeCache.begin(),
		IterEnd = m_EdgeCache.end();
	while (Iter!=IterEnd)
	{
		DoubleEdge* pEdge = (*Iter);
		if (pEdge)
		{
			delete pEdge;
		}
		++Iter;
	}
	m_EdgeCache.clear();
	return true;
}

void DoubleEdgeFactory::ClearMajorIdCount()
{
	GetDoubleEdgeFactory()->m_MajorIdCount = -1;
}

int DoubleEdgeFactory::GetCurrentMajorIdCnt()
{
	return GetDoubleEdgeFactory()->m_MajorIdCount;
}

int DoubleEdgeFactory::GetNextMajorIdCnt()
{
	return GetDoubleEdgeFactory()->m_MajorIdCount+=1;
}

void DoubleEdgeFactory::ClearMinorIdCount()
{
	GetDoubleEdgeFactory()->m_minorIdCount = -1;
}

int DoubleEdgeFactory::GetCurrentMinorIdCnt()
{
	return GetDoubleEdgeFactory()->m_minorIdCount;
}

int DoubleEdgeFactory::GetNextMinorIdCnt()
{
	return GetDoubleEdgeFactory()->m_minorIdCount+=1;
}

DoubleEdgeFactory* GeoGetDoubleEdgeFactory()
{
	return DoubleEdgeFactory::GetDoubleEdgeFactory();
}

void GeoReleaseDoubleEdgeFactory()
{
	DoubleEdgeFactory::ReleaseEdgeFactory();
}

Surface::Surface()
{
	m_bIsYMonotone = false;

	m_ptEdge = NULL;
	m_nEdgeCount = 0;
}


Surface::Surface(DoubleEdge* ptEdge,bool IsYMonoTone ):m_ptEdge(ptEdge),m_bIsYMonotone(IsYMonoTone)
{
	m_nEdgeCount = 0;
	DoubleEdge* pCurEdge = m_ptEdge;

	do 
	{
		++m_nEdgeCount;
		pCurEdge = pCurEdge->m_pNext;
	} while (pCurEdge!=m_ptEdge);
}

void Surface::CalcPointType()
{
	DoubleEdge* pStartEdge = m_ptEdge;
	DoubleEdge* pPre = pStartEdge;
	DoubleEdge* pCur ;
	DoubleEdge* pNext ;

	int cnt = 0;
	do 
	{
		pCur = pPre->m_pNext;
		pNext = pCur->m_pNext;
		++cnt;
		pCur->m_eType = JudgePointType(pPre->m_ptStart,pCur->m_ptStart,pNext->m_ptStart);
		pPre = pCur;
	} while (pPre!=pStartEdge);
}

PointType Surface::JudgePointType(const Point_2& ptPre,const Point_2& ptCur,const Point_2& ptNext)
{
	if (PointUpperThan(ptCur,ptPre) && PointUpperThan(ptCur,ptNext))
	{
		turn_dir_t turn = GetTurnType(ptPre,ptCur,ptNext);
		if (turn==turn_left)
		{
			return PT_BEGIN;
		}
		else
			return PT_SPLIT;
			
	}
	if (PointUpperThan(ptPre,ptCur) && PointUpperThan(ptCur,ptNext))
	{
		return PT_REGULARLEFT;
	}
	if (PointUpperThan(ptCur,ptPre) && PointUpperThan(ptNext,ptCur))
	{
		return PT_REGULARRIGHT;
	}
	if (PointUpperThan(ptPre,ptCur) && PointUpperThan(ptNext,ptCur))
	{
		turn_dir_t turn = GetTurnType(ptPre,ptCur,ptNext);
		if (turn==turn_left)
		{
			return PT_END;
		}
		else
			return PT_MERGE;
	}
	//ASSERT(0);
	return PT_NULL;
}

bool Surface::MakeSurface(const Point_2& ptStart_, const vector<Point_2>& vecInput ,const vector<double>& vecBulge)  
{
	GeoGetDoubleEdgeFactory()->ClearMajorIdCount();
	GeoGetDoubleEdgeFactory()->ClearMinorIdCount();
	int num = (int)vecInput.size();

	m_nEdgeCount = 0;
	m_ptCurStart = ptStart_;

	for (int i=0;i<num;i++)
	{
		AddTail(vecInput[i],vecBulge[i]);
	}
	double area = Area();
	//if (area < 0)
	//{
	//	ReverseSurface();
	//}
	CalcPointType();
#if (_DEBUG_LEVEL>=1)
	OutPutEdgeInfo();
#endif
#if (_DEBUG_LEVEL==2)
	ofstream Ofile;
	Ofile.open("D:\\InputPolygon.txt");
	OutPutEdgeForDebug(Ofile);
	Ofile.close();
#endif 
	if (area<0)
	{
		return false;
	}
	else
	{
		return true;
	}
}

//m_ptEdge Ϊ㣬ptΪյ㣺ֱ߶ΣֱӼ룻Բȴɢټ
void Surface::AddTail(const Point_2& pt,const double& nBulge)
{
	if (DOUBLE_EQU_ZERO(nBulge))
	{	
		DoubleEdge* pNewEdge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(m_ptCurStart);
		m_ptCurStart = pt;
		if (m_nEdgeCount==0)
		{
			m_ptEdge = pNewEdge;
			m_ptEdge->m_pNext = m_ptEdge->m_pPre = m_ptEdge;
		}
		else
		{
			pNewEdge->m_pNext = m_ptEdge->m_pNext;
			m_ptEdge->m_pNext->m_pPre = pNewEdge;
			m_ptEdge->m_pNext = pNewEdge;
			pNewEdge->m_pPre = m_ptEdge;
			m_ptEdge = pNewEdge;
		}
		m_nEdgeCount++;
	}
	else
	{
		DoubleEdge* ptArcStartEdge = NULL;
		DoubleEdge* ptArcEndEdge   = NULL;
		int segCnt = GeoGetDoubleEdgeFactory()->ConstructArcEdge(m_ptCurStart,pt,nBulge,ptArcStartEdge,ptArcEndEdge);
		m_ptCurStart = pt;
		if (m_nEdgeCount==0)
		{
			ptArcStartEdge->m_pPre = ptArcEndEdge;
			ptArcEndEdge->m_pNext = ptArcStartEdge;
			m_ptEdge = ptArcEndEdge;	
		}
		else
		{
			ptArcEndEdge->m_pNext = m_ptEdge->m_pNext;
			m_ptEdge->m_pNext->m_pPre = ptArcEndEdge;
			m_ptEdge->m_pNext = ptArcStartEdge;
			ptArcStartEdge->m_pPre = m_ptEdge;
			m_ptEdge = ptArcEndEdge;	
		}

		m_nEdgeCount += segCnt;
	}

}

void Surface::OutPutEdgeInfo(ostream& os)
{
	DoubleEdge* pCurEdge = m_ptEdge;
	for (int i=0;i<m_nEdgeCount;i++)
	{
		os<<"ID is: "<<pCurEdge->m_edgeId.MajorIndex<<" Start is: ";
		pCurEdge->m_ptStart.OutPut(os);
		os<<"Type is "<<PointTypeString[pCurEdge->m_eType]<<endl;
		pCurEdge = pCurEdge->m_pNext;
	}
}

void Surface::OutPutEdge(ostream& os)
{
	DoubleEdge* pCurEdge = m_ptEdge;
#if (_DEBUG_LEVEL>=1)
	os<<"Polygon is : "<<endl;
#endif
	for (int i=0;i<m_nEdgeCount;i++)
	{
		//os<<"ID is: "<<pCurEdge->m_nId<<" Start is: ";
		pCurEdge->m_ptStart.OutPut(os);
		//os<<"Type is "<<PointTypeString[pCurEdge->m_eType]<<endl;
		pCurEdge = pCurEdge->m_pNext;
	}
	os<<endl;
}

void Surface::OutPutEdgeForDebug(ostream& os )
{
	DoubleEdge* pCurEdge = m_ptEdge;
	DoubleEdge* pNextEdge ;
	for (int i=0;i<m_nEdgeCount;i++)
	{
		pNextEdge = pCurEdge->m_pNext;
		os<<pCurEdge->m_ptStart.x<<" "<<pCurEdge->m_ptStart.y<<endl;
		pCurEdge = pCurEdge->m_pNext;
	}
}

void Surface::ReverseSurface()
{
	DoubleEdge* pCurEdge = m_ptEdge;
	stack<DoubleEdge*> st;

	for (int i=0;i<m_nEdgeCount;i++)
	{
		st.push(pCurEdge);
		pCurEdge = pCurEdge->m_pNext;
	}
	while (!st.empty())
	{
		pCurEdge = st.top();
		st.pop();
		DoubleEdge* pTmp = pCurEdge->m_pNext;
		pCurEdge->m_pNext = pCurEdge->m_pPre;
		pCurEdge->m_pPre = pTmp;
		//pCurEdge->m_nId = m_nEdgeCount-1-pCurEdge->m_nId ;

	}
}


bool Surface::DoTriangulate(vector<Surface*>& vecOutput)
{
	if (m_bIsYMonotone==false)
	{
		return false;
	}
	if (m_nEdgeCount==3)
	{
		Surface* pNewSurface = new Surface;
		pNewSurface->m_bIsYMonotone = true;
		pNewSurface->m_ptEdge = m_ptEdge;
		pNewSurface->m_nEdgeCount = m_nEdgeCount;
		vecOutput.push_back(pNewSurface);
		return true;
	}

	set<DoubleEdgeRep> setEdges;
	DoubleEdge* pCurEdge = m_ptEdge;
	for (int i=0;i<m_nEdgeCount;i++)
	{
		setEdges.insert(DoubleEdgeRep(pCurEdge));
		pCurEdge = pCurEdge->m_pNext;
	}

	set<DoubleEdgeRep>::iterator Iter ;

	//calculate the edge side type
	Iter = setEdges.begin();
	DoubleEdgeRep& refEdge = (*Iter);
	DoubleEdge* pEdge = refEdge.m_pEdge;
	Point_2 ptStart = pEdge->m_ptStart;
	DoubleEdge* pNextEdge ;
	Point_2 ptEnd ;
	do 
	{
		pNextEdge = pEdge->m_pNext;
		ptEnd = pNextEdge->m_ptStart;
		if (ptStart<ptEnd) //left side
		{
			pEdge->m_eEdgeSide = ELeftEdge;
			pEdge = pNextEdge;
			ptStart = ptEnd;
		}
		else
			break;
	} while (1);
	do 
	{
		pNextEdge = pEdge->m_pNext;
		ptEnd = pNextEdge->m_ptStart;
		if (ptEnd<ptStart) //right side
		{
			pEdge->m_eEdgeSide = ERightEdge;
			pEdge = pNextEdge;
			ptStart = ptEnd;
		}
		else
			break;
	} while (1);
	//~calculate the edge side type

	stack<DoubleEdgeRep> st;
	for (int i=0;i<2;i++)
	{
		Iter = setEdges.begin();
		DoubleEdgeRep& CurEdge = (*Iter);
		st.push(CurEdge);
		setEdges.erase(CurEdge);
	}

	while (setEdges.size()>0)
	{			
		Iter = setEdges.begin();
		DoubleEdgeRep& CurEdge = (*Iter);
		DoubleEdge* pCurEdge = CurEdge.m_pEdge;
		setEdges.erase(CurEdge);

		if (pCurEdge->m_pNext->m_pNext==pCurEdge->m_pPre)
		{
			Surface* pNewSurface = new Surface(pCurEdge,true);
			vecOutput.push_back(pNewSurface);
			continue;
		}
		
		DoubleEdgeRep lastInStack = st.top();

		//same side
		if (lastInStack.m_pEdge->m_eEdgeSide == pCurEdge->m_eEdgeSide)
		{
			st.pop(); //ջ
			DoubleEdgeRep EdgeCanConnect;

			if (pCurEdge->m_eEdgeSide == ELeftEdge)
			{		
				while(!st.empty()) 
				{
					EdgeCanConnect = st.top();
					Point_2 ptCur = pCurEdge->m_ptStart;
					Point_2 ptLastInStack = lastInStack.m_pEdge->m_ptStart;
					Point_2 ptCanConnect = EdgeCanConnect.m_pEdge->m_ptStart;
					Point_2 vec1(ptLastInStack.x - ptCur.x, ptLastInStack.y - ptCur.y);
					Point_2 vec2(ptCanConnect.x - ptCur.x, ptCanConnect.y - ptCur.y);
					turn_dir_t turnDir = GetTurnType(vec1,vec2);
					if ( turnDir == turn_right ) //пԷָȥ
					{
						//DoubleEdge* pCurToCanConnect = new DoubleEdge(ptCur);
						//DoubleEdge* pCanConnectToCur = new DoubleEdge(ptCanConnect);
						DoubleEdge* pCurToCanConnect = GeoGetDoubleEdgeFactory()->ConstructLineEdge(ptCur);
						DoubleEdge* pCanConnectToCur = GeoGetDoubleEdgeFactory()->ConstructLineEdge(ptCanConnect);

						DoubleEdge* pPreCur = pCurEdge->m_pPre;
						DoubleEdge* pPreCanConnect = EdgeCanConnect.m_pEdge->m_pPre;

						ConnectThreeEdge(lastInStack.m_pEdge,pCurToCanConnect,EdgeCanConnect.m_pEdge);//ǰָ
						ConnectThreeEdge(pPreCanConnect,pCanConnectToCur,pCurEdge);

						Surface* pNewSurface = new Surface(pCurToCanConnect,true);
						vecOutput.push_back(pNewSurface);
					
						lastInStack = DoubleEdgeRep(pCanConnectToCur);
						st.pop();
					}
					else
					{
						break;
					}						
				} 				
				st.push(lastInStack);
				st.push(DoubleEdgeRep(pCurEdge));
			}
			else if (pCurEdge->m_eEdgeSide == ERightEdge)
			{
				while(!st.empty()) 
				{
					EdgeCanConnect = st.top();
					Point_2 ptCur = pCurEdge->m_ptStart;
					Point_2 ptLastInStack = lastInStack.m_pEdge->m_ptStart;
					Point_2 ptCanConnect = EdgeCanConnect.m_pEdge->m_ptStart;
					Point_2 vec1(ptLastInStack.x - ptCur.x, ptLastInStack.y - ptCur.y);
					Point_2 vec2(ptCanConnect.x - ptCur.x, ptCanConnect.y - ptCur.y);
					turn_dir_t turnDir = GetTurnType(vec1,vec2);
					if ( turnDir == turn_left )
					{
						//DoubleEdge* pCurToCanConnect = new DoubleEdge(ptCur);
						//DoubleEdge* pCanConnectToCur = new DoubleEdge(ptCanConnect);
						DoubleEdge* pCurToCanConnect = GeoGetDoubleEdgeFactory()->ConstructLineEdge(ptCur);
						DoubleEdge* pCanConnectToCur = GeoGetDoubleEdgeFactory()->ConstructLineEdge(ptCanConnect);

						DoubleEdge* pPreCur = pCurEdge->m_pPre;
						DoubleEdge* pPreCanConnect = EdgeCanConnect.m_pEdge->m_pPre;

						ConnectThreeEdge(pPreCanConnect,pCanConnectToCur,pCurEdge);//ǰָ
						ConnectThreeEdge(pPreCur,pCurToCanConnect,EdgeCanConnect.m_pEdge);

						Surface* pNewSurface = new Surface(pCanConnectToCur,true);
						vecOutput.push_back(pNewSurface);

						lastInStack = EdgeCanConnect;//עҲǲһģ
						pCurEdge = pCurToCanConnect;
						pCurEdge->m_eEdgeSide = ERightEdge;

						st.pop();
					}
					else
					{
						break;
					}
				} 		
				st.push(lastInStack);
				st.push(DoubleEdgeRep(pCurEdge));
			}
		}
		else //curEdge is on the another side
		{	
			stack<DoubleEdgeRep> newSt;

			while (!st.empty())
			{
				newSt.push(st.top()); st.pop();
			}			

			DoubleEdgeRep topEdge = newSt.top();
			newSt.pop();

			while(!newSt.empty())
			{
				DoubleEdgeRep EdgeCanConnect = newSt.top();
				newSt.pop();

				DoubleEdge* pPreCur = pCurEdge->m_pPre;
				DoubleEdge* pPreCanConnect = EdgeCanConnect.m_pEdge->m_pPre;

				//DoubleEdge* pCurToCanConnect = new DoubleEdge(pCurEdge->m_ptStart);
				//DoubleEdge* pCanConnectToCur = new DoubleEdge(EdgeCanConnect.m_pEdge->m_ptStart);
				DoubleEdge* pCurToCanConnect = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pCurEdge->m_ptStart);
				DoubleEdge* pCanConnectToCur = GeoGetDoubleEdgeFactory()->ConstructLineEdge(EdgeCanConnect.m_pEdge->m_ptStart);

				ConnectThreeEdge(pPreCur,pCurToCanConnect,EdgeCanConnect.m_pEdge);
				ConnectThreeEdge(pPreCanConnect,pCanConnectToCur,pCurEdge);

				DoubleEdge* pNewSurfaceEdge ;
				
				if (pCurEdge->m_eEdgeSide == ELeftEdge)
				{													
					pNewSurfaceEdge = pCurToCanConnect;	
				}
				else
				{
					pNewSurfaceEdge = pCanConnectToCur;
					pCurEdge = pCurToCanConnect;
					pCurEdge->m_eEdgeSide = ERightEdge;
				}
				Surface* pNewSurface = new Surface(pNewSurfaceEdge,true);
				vecOutput.push_back(pNewSurface);	
			}		
			if (pCurEdge->m_eEdgeSide==ELeftEdge)
			{
				//pCurEdge->m_pPre->m_eEdgeSide = ELeftEdge;
				st.push(DoubleEdgeRep(pCurEdge->m_pPre));
				st.push(DoubleEdgeRep(pCurEdge));	
			}
			else
			{
				st.push(DoubleEdgeRep(pCurEdge->m_pNext));
				st.push(DoubleEdgeRep(pCurEdge));	
			}

		}
	}
	return true;
}



bool StatusLine::InsertNode(DoubleEdge* pEdge)
{
	StatusLineNode* pParentNode = &m_nullNode ;
	StatusLineNode* pCurNode = m_pRoot;

#if (_DEBUG_LEVEL==2)
	cout<<"Insert: "<<pEdge->m_edgeId.MajorIndex<<" "<<pEdge->m_edgeId.MinorIndex<<endl;
#endif

	int CompRes = 0;
	while (pCurNode!=&m_nullNode)
	{
		pParentNode = pCurNode;

		CompRes = CompareDoubleEdge(pEdge,pCurNode->m_pData);
		if (CompRes == 0)
		{
			return false; //the node have exist in the tree;
		}
		else if ( -1 == CompRes )
		{
			pCurNode = pCurNode->m_pLeft;
		}
		else if ( 1 == CompRes )				
		{
			pCurNode = pCurNode->m_pRight;
		}
	}

	if (pParentNode == &m_nullNode) //we have no nodes
	{
		_ASSERT(numSegs==0);
		m_pRoot = new StatusLineNode(pEdge,&m_nullNode,&m_nullNode,&m_nullNode,BLACK);
		m_nullNode.m_pLeft = m_nullNode.m_pRight = m_pRoot;
		++numSegs;
		return true;
	}
	else 
	{	//now the pCurNode parent Node ,under which we insert our new node
		StatusLineNode* pNewNode = new StatusLineNode(pEdge,pParentNode,&m_nullNode,&m_nullNode,RED);
		if (1 == CompRes)
		{
			pParentNode->m_pRight = pNewNode;
		}
		else
		{
			pParentNode->m_pLeft = pNewNode;
		}
		if (pParentNode->color==RED)
		{
			FixInsertNode(pNewNode);
		}
		++numSegs;
		return true;
	}

}


void StatusLine::FixInsertNode(StatusLineNode* pCurNode)
{
	StatusLineNode* pUncle = NULL;
	while(pCurNode->m_pParent->color==RED)//pCurNode->color==RED && pParent->color == RED
	{
		if ( pCurNode->m_pParent == pCurNode->m_pParent->m_pParent->m_pLeft)
		{
			pUncle =  pCurNode->m_pParent->m_pParent->m_pRight;
			if (pUncle->color == RED)
			{
				pCurNode->m_pParent->color = BLACK;
				pUncle->color = BLACK;
				pCurNode->m_pParent->m_pParent->color = RED;
				pCurNode = pCurNode->m_pParent->m_pParent;
			}
			else
			{
				if (pCurNode==pCurNode->m_pParent->m_pRight)
				{
					pCurNode = pCurNode->m_pParent;
					LeftRotation(pCurNode);
				}
				pCurNode->m_pParent->color = BLACK;
				pCurNode->m_pParent->m_pParent->color = RED;
				RightRotation(pCurNode->m_pParent->m_pParent);
			}
		}
		else
		{
			pUncle = pCurNode->m_pParent->m_pParent->m_pLeft;
			if (pUncle->color==RED)
			{
				pCurNode->m_pParent->color = BLACK;
				pUncle->color = BLACK;
				pCurNode->m_pParent->m_pParent->color = RED;
				pCurNode = pCurNode->m_pParent->m_pParent;
			}
			else
			{
				if (pCurNode==pCurNode->m_pParent->m_pLeft)
				{
					pCurNode = pCurNode->m_pParent;
					RightRotation(pCurNode);
				}
				pCurNode->m_pParent->color = BLACK;
				pCurNode->m_pParent->m_pParent->color = RED;
				LeftRotation(pCurNode->m_pParent->m_pParent);
			}
		}
	}
	m_pRoot->color = BLACK;
}

void StatusLine::LeftRotation(StatusLineNode* pCurNode)
{
	StatusLineNode* pParent = pCurNode->m_pParent; 
	StatusLineNode* pRight  = pCurNode->m_pRight;  // up this node
	StatusLineNode* pRightLeft = pRight->m_pLeft;
	_ASSERT(pRight!=&m_nullNode);

	if (pParent == &m_nullNode)
	{
		m_pRoot = pRight;
		m_nullNode.m_pLeft = m_pRoot;
	}
	else
	{
		pCurNode == pParent->m_pLeft? pParent->m_pLeft = pRight: pParent->m_pRight = pRight;
	}

	if (pRightLeft!=&m_nullNode)
	{
		pRightLeft->m_pParent = pCurNode;
	}
	else
	{
#if (_DEBUG_LEVEL==2)
		cout<<"Meet "<<endl;
#endif
	}

	pRight->m_pLeft = pCurNode;
	pRight->m_pParent = pParent;

	pCurNode->m_pRight = pRightLeft;
	pCurNode->m_pParent = pRight;
}

void StatusLine::RightRotation(StatusLineNode* pCurNode)
{
	StatusLineNode* pParent = pCurNode->m_pParent;
	StatusLineNode* pLeft   = pCurNode->m_pLeft;
	StatusLineNode* pLeftRight = pLeft->m_pRight;

	_ASSERT(pLeft!=&m_nullNode);

	if (pParent == &m_nullNode)
	{
		m_pRoot = pLeft;
		m_nullNode.m_pLeft = m_pRoot;
	}
	else
	{
		pCurNode == pParent->m_pLeft? pParent->m_pLeft = pLeft: pParent->m_pRight = pLeft;
	}

	if (pLeftRight!=&m_nullNode)
	{
		pLeftRight->m_pParent = pCurNode;
	}
	else
	{
#if (_DEBUG_LEVEL==2)
		cout<<"Meet "<<endl;
#endif
	}
	pLeft->m_pRight = pCurNode;
	pLeft->m_pParent = pParent;

	pCurNode->m_pLeft = pLeftRight;
	pCurNode->m_pParent = pLeft;
}

void StatusLine::Transplant(StatusLineNode* pU,StatusLineNode* pV)
{
	if (pU->m_pParent == &m_nullNode)
	{
		m_pRoot = pV;
	}
	else if ( pU == pU->m_pParent->m_pLeft)
	{
		pU->m_pParent->m_pLeft = pV;
	}
	else
	{
		pU->m_pParent->m_pRight = pV;
	}
	pV->m_pParent = pU->m_pParent;
}


bool StatusLine::EraseNode(DoubleEdge* pEdge)
{
	if (m_pRoot==&m_nullNode)//m_pRoot==NULL
	{
		_ASSERT(numSegs == 0);
		return false;
	}
	//1 find this node
	StatusLineNode* pCurNode = m_pRoot;

#if (_DEBUG_LEVEL==2)
	cout<<"EraseNode: "<<pEdge->m_edgeId.MajorIndex<<" "<<pEdge->m_edgeId.MinorIndex<<endl;
#endif

	do 
	{
		int CompRes = CompareDoubleEdge(pEdge,pCurNode->m_pData) ;
		if ( 0 == CompRes )
		{
			break;
		}
		else if ( CompRes == -1 )
		{
			if (pCurNode->m_pLeft!=&m_nullNode)
			{
				pCurNode = pCurNode->m_pLeft;
			}
			else
				return false;
		}
		else if ( CompRes == 1 )
		{
			if (pCurNode->m_pRight != &m_nullNode)
			{
				pCurNode = pCurNode->m_pRight;
			}
			else
				return false;
		}
	} while (true);

	StatusLineNode* pY = pCurNode;
	StatusLineNode* pX ;
	COLOR ycolor = pCurNode->color;

	if (pCurNode->m_pLeft == &m_nullNode)
	{
		pX = pCurNode->m_pRight;
		Transplant(pCurNode,pCurNode->m_pRight);
	}
	else if (pCurNode->m_pRight == &m_nullNode)
	{
		pX = pCurNode->m_pLeft;
		Transplant(pCurNode,pCurNode->m_pLeft);
	}
	else
	{
		pY = Minimum(pCurNode->m_pRight); 
		ycolor = pY->color;
		pX = pY->m_pRight;
		if (pY->m_pParent == pCurNode) //pY is the right child of pCurNode  (pY always have no left child)
		{
			pX->m_pParent = pY;
		}
		else
		{
			Transplant(pY,pY->m_pRight);
			pY->m_pRight = pCurNode->m_pRight;
			pY->m_pRight->m_pParent = pY;
		}
		Transplant(pCurNode,pY);
		pY->m_pLeft = pCurNode->m_pLeft;
		pY->m_pLeft->m_pParent = pY;
		pY->color = pCurNode->color;

	}
	delete pCurNode;
	if (ycolor==BLACK)
	{
		FixDeleteNode(pX);
	}

	--numSegs;

	return true;
}

void StatusLine::FixDeleteNode(StatusLineNode* pCurNode)
{	

	StatusLineNode* pUncle = NULL ;
	//pCurNode is not root and color is black
	while( pCurNode!=m_pRoot && pCurNode->color==BLACK)
	{
		if (pCurNode==pCurNode->m_pParent->m_pLeft)
		{
			pUncle = pCurNode->m_pParent->m_pRight;
			if (pUncle->color==RED)
			{
				pUncle->color = BLACK;
				pCurNode->m_pParent->color = RED;				
				LeftRotation(pCurNode->m_pParent);
				pUncle = pCurNode->m_pParent->m_pRight;
			}
			if (pUncle->m_pLeft->color == BLACK && pUncle->m_pRight->color == BLACK)
			{
				pUncle->color = RED;
				pCurNode = pCurNode->m_pParent;	
			}
			else 
			{
				if (pUncle->m_pRight->color == BLACK)
				{
					//ʱҪ֪ pUncle->m_pLeft->color == REDԻɫ				
					pUncle->m_pLeft->color = BLACK;
					pUncle->color = RED;
					RightRotation(pUncle);
					pUncle = pCurNode->m_pParent->m_pRight; //pUncle = pCurNode->m_pParent->m_pRight;			
				}
				pUncle->color = pCurNode->m_pParent->color;				
				pCurNode->m_pParent->color = BLACK;
				pUncle->m_pRight->color = BLACK;
				LeftRotation(pCurNode->m_pParent);
				pCurNode = m_pRoot;
			}
		}
		else
		{
			pUncle = pCurNode->m_pParent->m_pLeft;
			if (pUncle->color==RED)
			{
				pUncle->color = BLACK;
				pCurNode->m_pParent->color = RED;				
				RightRotation(pCurNode->m_pParent);
				pUncle = pCurNode->m_pParent->m_pLeft;
			}

			if (pUncle->m_pLeft->color==BLACK && pUncle->m_pRight->color==BLACK)
			{
				pUncle->color = RED; 
				pCurNode = pCurNode->m_pParent;
			}
			else
			{
				if (pUncle->m_pLeft->color == BLACK)
				{
					pUncle->color = RED; 
					pUncle->m_pRight->color = BLACK; 
					LeftRotation(pUncle);
					pUncle = pCurNode->m_pParent->m_pLeft; //pUncle = pCurNode->m_pParent->m_pLeft;
				}
				pUncle->color = pCurNode->m_pParent->color;
				pCurNode->m_pParent->color = BLACK; 
				pUncle->m_pLeft->color = BLACK;				
				RightRotation(pCurNode->m_pParent);
				pCurNode = m_pRoot;
			}				
		}
	}
	pCurNode->color = BLACK;
	//m_pRoot = m_nullNode.m_pLeft;
}

StatusLine::StatusLineNode* StatusLine::Minimum(StatusLineNode* pCurNode)
{
	while (pCurNode->m_pLeft!=&m_nullNode)
	{
		pCurNode = pCurNode->m_pLeft;
	}
	return pCurNode;
}

StatusLine::iterator StatusLine::Find(DoubleEdge* pEdge)
{
	if (m_pRoot==&m_nullNode)
	{
		return iterator(this,NULL); 
	}
	StatusLineNode* pCurNode = m_pRoot;

	do 
	{
		int CompRes = CompareDoubleEdge(pEdge,pCurNode->m_pData);
		if (CompRes == 0)
		{
			return iterator(this,pCurNode);
		}
		else if (CompRes==-1)
		{
			if (pCurNode->m_pLeft!=&m_nullNode)
			{
				pCurNode = pCurNode->m_pLeft;
			}
			else
				break; 
		}
		else if (CompRes==1)
		{
			if (pCurNode->m_pRight!=&m_nullNode)
			{
				pCurNode = pCurNode->m_pRight;
			}
			else
				break; 
		}
	} while (true);

	return iterator(this,NULL);
}

StatusLine::StatusLineNode* StatusLine::GetMinNode()
{
	if (m_pRoot==NULL)
	{
		return NULL;
	}
	StatusLineNode* pCurNode = m_pRoot;

	while (pCurNode->m_pLeft!=&m_nullNode)
	{
		pCurNode = pCurNode->m_pLeft;
	}
	return pCurNode;
}

StatusLine::StatusLineNode* StatusLine::GetMaxNode()
{
	if (m_pRoot== &m_nullNode)
	{
		return NULL;
	}
	StatusLineNode* pCurNode = m_pRoot;

	while (pCurNode->m_pRight!=&m_nullNode)
	{
		pCurNode = pCurNode->m_pRight;
	}
	return pCurNode;
}

StatusLine::StatusLineNode* StatusLine::GetPrev(StatusLineNode* pCurNode)
{
	if (pCurNode==NULL)
	{
		return NULL;
	}
	if (pCurNode->m_pLeft!=&m_nullNode)
	{
		StatusLineNode* pTmp = pCurNode->m_pLeft;
		while (pTmp->m_pRight!=&m_nullNode)
		{
			pTmp = pTmp->m_pRight;
		}
		return pTmp;
	}
	else
	{
		StatusLineNode* pParent = pCurNode->m_pParent;

NEXT:
		if (pParent==&m_nullNode)
		{
			return NULL;
		}
		else
		{
			if (pCurNode == pParent->m_pRight)
			{
				return pParent;
			}
			else
			{
				pCurNode = pParent;
				pParent = pParent->m_pParent;
				goto NEXT;
			}
		}
	}
}

StatusLine::StatusLineNode* StatusLine::GetNext(StatusLineNode* pCurNode)
{
	if (pCurNode->m_pRight!=&m_nullNode)
	{
		StatusLineNode* pTmp = pCurNode->m_pRight;
		while(pTmp->m_pLeft!=&m_nullNode)
		{
			pTmp = pTmp->m_pLeft;
		}
		return pTmp;
	}
	else
	{
		StatusLineNode* pParent = pCurNode->m_pParent;

NEXT:
		if (pParent==&m_nullNode)
		{
			return NULL;
		}
		else
		{
			if (pCurNode == pParent->m_pLeft)
			{
				return pParent;
			}
			else
			{
				pCurNode = pParent;
				pParent = pParent->m_pParent;
				goto NEXT;
			}
		}

	}
}

StatusLine::iterator StatusLine::begin()
{
	iterator ItBegin(this,GetMinNode());
	return ItBegin;
}

StatusLine::iterator StatusLine::end()
{
	iterator endIt(this,NULL);
	return endIt;
}

bool StatusLine::erase(iterator Iter)
{	
	return EraseNode(Iter.m_pCurNode->m_pData);
}

bool StatusLine::EraseAllNode()
{
	if (m_pRoot==&m_nullNode)
	{
		return true;
	}
	if (m_pRoot->m_pLeft!=&m_nullNode)
	{
		EraseSubTree(m_pRoot->m_pLeft);
	}
	if (m_pRoot->m_pRight!=&m_nullNode)
	{
		EraseSubTree(m_pRoot->m_pRight);
	}
	delete m_pRoot;
	return true;
}

bool StatusLine::EraseSubTree(StatusLineNode* pNode)
{
	if (pNode->m_pLeft!=&m_nullNode)
	{
		EraseSubTree(pNode->m_pLeft);
	}
	if (pNode->m_pRight!=&m_nullNode)
	{
		EraseSubTree(pNode->m_pRight);
	}
	delete pNode;
	return true;
}

DoubleEdge*  StatusLine::FindDirectLeft(Point_2 ptPoint)
{
	StatusLineNode* pParent = &m_nullNode;
	StatusLineNode* pCurNode = GetMaxNode();

	if (pCurNode==NULL)
	{
		return NULL;//
	}

	double curY = ptPoint.y;
	double IntersectionX ;

	while (pCurNode!=NULL)
	{	
		Point_2 ptStart = pCurNode->m_pData->m_ptStart;
		Point_2 ptEnd   = pCurNode->m_pData->m_pNext->m_ptStart;
        //bug may come from here
		if (DOUBLE_EQU(ptStart.y,ptEnd.y))//abs(ptStart.y-ptEnd.y)<1.e-9
		{
			IntersectionX = max(ptStart.x,ptEnd.x);
		}
		else
		{
			IntersectionX = (curY-ptStart.y)*(ptEnd.x-ptStart.x)/(ptEnd.y-ptStart.y) + ptStart.x;//
		}
		if (IntersectionX+1.e-9<ptPoint.x)
		{
			return pCurNode->m_pData;
		}
		else
		{
			pCurNode = GetPrev(pCurNode);
		}
	} 

	return NULL;
}

Triangulation::Triangulation(void)
{
}

Triangulation::~Triangulation(void)
{
	int n = (int)m_OutTriangle.size();
	for (int i=0;i<n;i++)
	{
		delete m_OutTriangle[i];
	}
	m_OutTriangle.clear();
	vecSurface.clear();
	m_setEdge.clear();
	m_InputSurface.m_nEdgeCount = 0;
	//DoubleEdge::ClearIdCount();
	//GeoGetDoubleEdgeFactory()->ClearMajorIdCount();
	//GeoGetDoubleEdgeFactory()->ClearMinorIdCount();
}

Surface* Triangulation::MakeYMonotone(int nSegmentNo_)
{
	m_InputSurface.MakeSurface(m_ptStart,vecInput,vecBulge);

	m_polyEdgeNum = (int)vecInput.size();

	//Insert all edges into set
	DoubleEdge* pStartEdge = m_InputSurface.m_ptEdge;
	DoubleEdge* pCurEdge = pStartEdge;
	
	do 
	{
		m_setEdge.insert(DoubleEdgeRep(pCurEdge));
		pCurEdge = pCurEdge->m_pNext;
	} while (pCurEdge!=pStartEdge);

   // m_polyEdgeNum = DoubleEdge::s_IdCount;

	int ret = 0;
	int cnt = 0;
	while (m_setEdge.size()>0)
	{
		++cnt;
		set<DoubleEdgeRep>::iterator Iter = m_setEdge.begin();
		DoubleEdgeRep& CurEdge = (*Iter);

		DoubleEdge* pCurEdge = CurEdge.m_pEdge;
		m_setEdge.erase(CurEdge);

		PointType pType = pCurEdge->m_eType;

		m_StatusLine.SetCurPosition(pCurEdge->m_ptStart);
		switch(pType)
		{
		case PT_BEGIN:
			{
				ret = ProcessBeginPoint(pCurEdge);
			}
			break;
		case PT_END:
			{
				ret = ProcessEndPoint(pCurEdge);
			}
			break;
		case PT_REGULARLEFT:
		case PT_REGULARRIGHT:
			{
				ret = ProcessRegularPoint(pCurEdge);
			}
			break;
		case PT_MERGE:
			{
				ret = ProcessMergePoint(pCurEdge);
			}
			break;
		case PT_SPLIT:
			{
				ret = ProcessSplitPoint(pCurEdge);
			}
			break;			
		}

//#ifdef _DEBUG
		if (ret==0)
		{
			cout<<cnt<<endl;
			return NULL;
		}
//#endif
	}

#if (_DEBUG_LEVEL>=1)
	cout<<"Polygone :"<<endl;
#endif
#if (_DEBUG_LEVEL==2)
	ofstream OutDebugFile;
	OutDebugFile.open("F:\\LineDebug.txt");
#endif
	for (int i=0;i<vecSurface.size();i++)
	{
		
#if (_DEBUG_LEVEL==2)
		vecSurface[i]->OutPutEdge(cout);
		vecSurface[i]->OutPutEdgeForDebug(OutDebugFile);
#endif
		Surface* curPoly = vecSurface[i];

		DoubleEdge* pStartEdge = curPoly->m_ptEdge;
		DoubleEdge* pCurEdge = pStartEdge;

		do 
		{
			if (pCurEdge->m_edgeId.MajorIndex == nSegmentNo_) // to do
			{
				curPoly->m_ptEdge = pCurEdge;
				return curPoly;
			}
			pCurEdge = pCurEdge->m_pNext;
		} while (pCurEdge!=pStartEdge);
	}
#if (_DEBUG_LEVEL==2)
	OutDebugFile.close();
#endif
	return NULL;
}

Surface* Triangulation::DoTriangle(int& nSegmentNo_)
{
	ReleaseOutTriangle();

	GeoGetDoubleEdgeFactory()->ClearMajorIdCount();
	GeoGetDoubleEdgeFactory()->ClearMinorIdCount();
	GeoGetDoubleEdgeFactory()->SetTriangulation(this);
	MakeYMonotone(nSegmentNo_);

	ReleaseOutTriangle();

	//ʷ֣뵽OutTriangle
	for (int i=0;i<vecSurface.size();i++)
	{
		Surface* pSurface = vecSurface[i];
		pSurface->DoTriangulate(m_OutTriangle);
	}

	if (nSegmentNo_!=-1)
	{	
		cout<<"Triangles :"<<endl;
		Point_2 ptIntroStart;
		for (int i=0;i<m_OutTriangle.size();i++)
		{
#if (_DEBUG_LEVEL>=1)
			m_OutTriangle[i]->OutPutEdge(cout);
#endif
			Surface* curTriangle = m_OutTriangle[i];
			DoubleEdge* Edge1 = curTriangle->m_ptEdge;
			DoubleEdge* Edge2 = Edge1->m_pNext;
			DoubleEdge* Edge3 = Edge2->m_pNext;
			if (Edge1->m_edgeId.MajorIndex == nSegmentNo_ || Edge2->m_edgeId.MajorIndex==nSegmentNo_ ||Edge3->m_edgeId.MajorIndex==nSegmentNo_ )
			{
				return curTriangle;
			}
		}
	}
	else //ѰҺʵ߶Σߡ
	{
		for(int i=0;i<m_OutTriangle.size();i++)
		{
			Surface* curTriangle = m_OutTriangle[i];
			double Area = curTriangle->Area();
			DoubleEdge* Edge1 = curTriangle->m_ptEdge;
			DoubleEdge* Edge2 = Edge1->m_pNext;
			DoubleEdge* Edge3 = Edge2->m_pNext;
			if (Edge1->m_edgeId.MajorIndex < m_polyEdgeNum || Edge2->m_edgeId.MajorIndex < m_polyEdgeNum ||Edge3->m_edgeId.MajorIndex < m_polyEdgeNum)
			{
				mp_AreaToID[Area] = i;
			}
		}
		map<double,int>::iterator Iter = mp_AreaToID.begin(),
			IterEnd = mp_AreaToID.end();
		--IterEnd;
		Surface* curTriangle  = NULL;
		map<double,int> mpLenToId;
		do
		{
			mpLenToId.clear();
			curTriangle = m_OutTriangle[(*IterEnd).second];
			DoubleEdge* Edge1 = curTriangle->m_ptEdge;
			DoubleEdge* Edge2 = Edge1->m_pNext;
			DoubleEdge* Edge3 = Edge2->m_pNext;
			//double  nLen = m_IODistance + m_InIntraLen + m_OutIntraLen;
			//double  nPowerLen = nLen*nLen;
			double  nEdgePowerLen1;
			double  nEdgePowerLen2;
			double  nEdgePowerLen3;
			
			if(Edge1->m_edgeId.MajorIndex < m_polyEdgeNum)
			{
				nEdgePowerLen1 = PowerDistance(Edge1->m_ptStart ,Edge2->m_ptStart);
				nSegmentNo_ = Edge1->m_edgeId.MajorIndex;
				
				//if(nEdgePowerLen>=nPowerLen)
				//{
				mpLenToId[nEdgePowerLen1] = nSegmentNo_;
					//nSegmentNo_ = Edge1->m_nId;
					//nEdgeLen = sqrt(nEdgePowerLen1);
					//Ratio = (nEdgeLen - m_IODistance)/2 / nEdgeLen;

					//return curTriangle;
				//}
			}
			if(Edge2->m_edgeId.MajorIndex < m_polyEdgeNum)
			{
				nEdgePowerLen2 = PowerDistance(Edge2->m_ptStart ,Edge3->m_ptStart);
				nSegmentNo_ = Edge2->m_edgeId.MajorIndex;

				mpLenToId[nEdgePowerLen2] = nSegmentNo_;
				//if(nEdgePowerLen>=nPowerLen)
				//{
					//nSegmentNo_ = Edge2->m_nId;
					//nEdgeLen = sqrt(nEdgePowerLen2);
					//Ratio = (nEdgeLen - m_IODistance)/2 / nEdgeLen;
					//return curTriangle;
				//}
			}

			if(Edge3->m_edgeId.MajorIndex < m_polyEdgeNum)
			{
				nEdgePowerLen3 = PowerDistance(Edge3->m_ptStart ,Edge1->m_ptStart);
				nSegmentNo_ = Edge3->m_edgeId.MajorIndex;
				mpLenToId[nEdgePowerLen3] = nSegmentNo_;
				//if(nEdgePowerLen>=nPowerLen)
				//{
					//nSegmentNo_ = Edge3->m_nId;
					//nEdgeLen = sqrt(nEdgePowerLen3);
					//Ratio = (nEdgeLen - m_IODistance)/2 / nEdgeLen;
					//return curTriangle;
				//}
			}
			if (Iter==IterEnd || mpLenToId.size()>=1)
			{
				break;
			}
		}while(1);

		//IterEnd = mp_AreaToID.end();
		//--IterEnd;
		//return curTriangle=m_OutTriangle[nSegmentNo_];
		map<double,int>::iterator EdgeIter = mpLenToId.end();
		--EdgeIter;
		nSegmentNo_ = (*EdgeIter).second;
		return curTriangle;
		
	}

	return NULL;
}

Surface* Triangulation::DoYMonotone(int& nSegmentNo_ , LLInformation* pLLInformation_, DestEdgeInfo& destEdgeInfo_)
{
	ReleaseOutTriangle();

	GeoGetDoubleEdgeFactory()->ClearMajorIdCount();
	GeoGetDoubleEdgeFactory()->ClearMinorIdCount();
	GeoGetDoubleEdgeFactory()->SetTriangulation(this);

	MakeYMonotone(nSegmentNo_);

	vector<Surface*> vecOutYMonotone ;//= vecSurface;	

	if (vecSurface.size()==0)
	{
		GeoGetDoubleEdgeFactory()->ClearMajorIdCount();
		GeoGetDoubleEdgeFactory()->ClearMinorIdCount();
		GeoGetDoubleEdgeFactory()->ReleaseEdges();
		m_InputSurface.m_nEdgeCount = 0;
		m_InputSurface.ReleaseEdges();
		m_InputSurface.MakeSurface(m_ptStart,vecInput,vecBulge);
		vecOutYMonotone.push_back(&m_InputSurface);
		//return NULL;
	}
	else{
		vecOutYMonotone = vecSurface;
	}



	if (nSegmentNo_!=-1)
	{	
		cout<<"YMonotone :"<<endl;
		Point_2 ptIntroStart;
		for (int i=0;i<vecOutYMonotone.size();i++)
		{
			Surface* curYMonotone = vecOutYMonotone[i];

			DoubleEdge* pObjEdge = curYMonotone->FindEdgeById(nSegmentNo_);

			if (pObjEdge!=NULL)
			{
				destEdgeInfo_.m_pDestEdge = pObjEdge;
				return curYMonotone;
			}
		}
		destEdgeInfo_.m_pDestEdge = NULL;
	}
	else //ѰҺʵ߶Σߡ
	{
		map<pair<double,Surface*>, set<pair<double,DoubleEdge*> > > mpAreaSurfaceToLengthEdges;
		map<DoubleEdge*, DestEdgeInfo> mpPEdgeToEdgeInfo;
		for(int i=0;i<vecOutYMonotone.size();i++)
		{
			Surface* curYMonotone = vecOutYMonotone[i];
#if (_DEBUG_LEVEL==2)
			ofstream Ofile;
			string strName = "F:\\SurfaceData";
			strName+=('0'+i);
			strName+=".txt";
			Ofile.open(strName.c_str());
			curYMonotone->OutPutEdgeForDebug(Ofile);
			Ofile.close();
#endif
			double Area = curYMonotone->Area();
			DoubleEdge* pCurEdge = curYMonotone->m_ptEdge;
			
			do 
			{
				if (pCurEdge->m_edgeId.MajorIndex < m_polyEdgeNum ) // to do
				{
					if (pCurEdge->m_edgeId.MinorIndex==-1)  //ǰΪֱ
					{
						double powlen = PowerDistance(pCurEdge->m_ptStart,pCurEdge->m_pNext->m_ptStart);
					    mpAreaSurfaceToLengthEdges[make_pair(Area,curYMonotone)].insert(make_pair(powlen,pCurEdge) ); //Area curYMonotone pair Ӧ 
					    pCurEdge = pCurEdge->m_pNext;
					}
					else
					{
						int ArcMajorIndex = pCurEdge->m_edgeId.MajorIndex;
						int ArcMinorIndex = pCurEdge->m_edgeId.MinorIndex;
						ArcInfo& arcInfo = m_mpIdxToArcInfo[ArcMajorIndex];
						double arclen = arcInfo.nLength;
						int ArcSegNum = arcInfo.nArcSegNum;
						int nTotalSeg = 1;

						//int tmpCnt = 1;
						DoubleEdge* pCenterEdge = NULL;
						DoubleEdge* pPreStart = pCurEdge;
						DoubleEdge* pAfterEnd = pCurEdge;
						//1 ͳƶٸ
						do 
						{
							pPreStart=pPreStart->m_pPre;
							if (pPreStart->m_edgeId.MajorIndex == ArcMajorIndex)
							{
								++nTotalSeg;
							}
							else
							{
								break;
							}
						} while (1);
						do 
						{
							pAfterEnd=pAfterEnd->m_pNext;
							if (pAfterEnd==curYMonotone->m_ptEdge)
							{
								goto NEXTYMONOTONE;
							}
							if (pAfterEnd->m_edgeId.MajorIndex == ArcMajorIndex)
							{
								++nTotalSeg;
							}
							else
							{
								break;
							}
						} while (1);
						double curArcLen = arclen*((double)nTotalSeg/(double)ArcSegNum);
						double powCurArcLen = curArcLen*curArcLen;
						int nHalf = nTotalSeg/2;

						pCenterEdge = pPreStart->m_pNext;
						
						for (int i=0;i<=nHalf;i++)
						{
							pCenterEdge = pCenterEdge->m_pNext;
						}
						
						//ڴԲʱǽԲɢ֮м߶α棬ΪĿ߶
						mpAreaSurfaceToLengthEdges[make_pair(Area,curYMonotone)].insert(make_pair(powCurArcLen,pCenterEdge) ); //Area curYMonotone pair Ӧ 

						DestEdgeInfo& _destEdgeInfo = mpPEdgeToEdgeInfo[pCenterEdge];
						_destEdgeInfo.m_nLength = curArcLen;
						_destEdgeInfo.m_pDestEdge = pCenterEdge;
						_destEdgeInfo.m_pArcStart = pPreStart->m_pNext;
						_destEdgeInfo.m_pArcEnd = pAfterEnd->m_pPre;

						ASSERT(_destEdgeInfo.m_pArcStart->m_edgeId.MajorIndex == _destEdgeInfo.m_pArcEnd->m_edgeId.MajorIndex);


						pCurEdge = pAfterEnd;
					}
				}
				else
				{
					pCurEdge = pCurEdge->m_pNext;
				}
				
			} while (pCurEdge!=curYMonotone->m_ptEdge);
NEXTYMONOTONE: ;
		}

		if (mpAreaSurfaceToLengthEdges.size()==0)
		{
			return NULL;
		}

		map<pair<double,Surface*>, set<pair<double,DoubleEdge*> > >::iterator Iter = mpAreaSurfaceToLengthEdges.begin(),
			IterEnd = mpAreaSurfaceToLengthEdges.end();		

		do //YMonotone С
		{
			--IterEnd;
			const pair<double,Surface*>&   SurfaceEdgePair = (*IterEnd).first;
			set<pair<double,DoubleEdge*> >& setLengthEdge = (*IterEnd).second;
			if (setLengthEdge.size()>1)
			{
				set<pair<double,DoubleEdge*> >::iterator setLengthIterEnd = setLengthEdge.end();
				--setLengthIterEnd;
				pair<double,DoubleEdge*> pr = (*setLengthIterEnd);
				nSegmentNo_ = pr.second->m_edgeId.MajorIndex; // to do
				destEdgeInfo_.m_pDestEdge = pr.second;
				if ((pr.first)*0.9 < pLLInformation_->nIODistance*pLLInformation_->nIODistance)
				{
					if (IterEnd==Iter)
					{
						break;
					}
					continue;
				}
				if (pr.second->m_edgeId.MinorIndex!=-1)
				{
					destEdgeInfo_ = mpPEdgeToEdgeInfo[destEdgeInfo_.m_pDestEdge];
				}
				return SurfaceEdgePair.second;//Ŀ߶ΡԲ£ԲɢĵǸС߶
			}
			if (IterEnd==Iter)
			{
				break;
			}
		}while(true);
	}
	return NULL;
}

Surface* Triangulation::DoMakeSurface()
{
	m_InputSurface.MakeSurface(m_ptStart,vecInput,vecBulge);
	return &m_InputSurface;
}

Surface* Triangulation::DoTriangleMax(int& nSegmentNo_)
{
	ReleaseOutTriangle();

	GeoGetDoubleEdgeFactory()->ClearMajorIdCount();
	GeoGetDoubleEdgeFactory()->ClearMinorIdCount();
	GeoGetDoubleEdgeFactory()->SetTriangulation(this);
	MakeYMonotone(nSegmentNo_);

	ReleaseOutTriangle();

	//ʷ֣뵽OutTriangle
	for (int i=0;i<vecSurface.size();i++)
	{
		Surface* pSurface = vecSurface[i];
		pSurface->DoTriangulate(m_OutTriangle);
	}


	for(int i=0;i<m_OutTriangle.size();i++)
	{
		Surface* curTriangle = m_OutTriangle[i];
		double Area = curTriangle->Area();
		DoubleEdge* Edge1 = curTriangle->m_ptEdge;
		DoubleEdge* Edge2 = Edge1->m_pNext;
		DoubleEdge* Edge3 = Edge2->m_pNext;
		if (Edge1->m_edgeId.MajorIndex < m_polyEdgeNum || Edge2->m_edgeId.MajorIndex < m_polyEdgeNum ||Edge3->m_edgeId.MajorIndex < m_polyEdgeNum)
		{
			mp_AreaToID[Area] = i;
		}
	}
	map<double,int>::iterator Iter = mp_AreaToID.begin(),
		IterEnd = mp_AreaToID.end();
	--IterEnd;
	do
	{
		Surface* curTriangle = m_OutTriangle[(*IterEnd).second];
		DoubleEdge* Edge1 = curTriangle->m_ptEdge;
		DoubleEdge* Edge2 = Edge1->m_pNext;
		DoubleEdge* Edge3 = Edge2->m_pNext;

		if(Edge1->m_edgeId.MajorIndex < m_polyEdgeNum)
		{
			nSegmentNo_ = Edge1->m_edgeId.MajorIndex;
			return curTriangle;
		}
		if(Edge2->m_edgeId.MajorIndex < m_polyEdgeNum)
		{
			nSegmentNo_ = Edge2->m_edgeId.MajorIndex;
			return curTriangle;
		}

		if(Edge3->m_edgeId.MajorIndex < m_polyEdgeNum)
		{
			nSegmentNo_ = Edge3->m_edgeId.MajorIndex;
			return curTriangle;
		}
		--IterEnd;
	}while(IterEnd != Iter);

	//˵ûкõλá
	IterEnd = mp_AreaToID.end();
	--IterEnd;
	nSegmentNo_ = 0;
	

	return NULL;
}

bool Triangulation::ProcessBeginPoint(DoubleEdge* pCurEdge)
{
	CheckPointer(pCurEdge);
	pCurEdge->UpdateHelper(pCurEdge);
	m_StatusLine.InsertNode(pCurEdge);
	return true;
}

bool Triangulation::ProcessEndPoint(DoubleEdge* pCurEdge)
{
	CheckPointer(pCurEdge);
	CheckPointer(pCurEdge->m_pPre);

	DoubleEdge* pPre = pCurEdge->m_pPre;
	CheckPointer(pPre->m_ptHelper);
	if(pPre->m_ptHelper->m_eType == PT_MERGE)
	{
		ConnectDiagonalEndPoint(pPre->m_ptHelper, pCurEdge);//
	}
	else
	{
		Surface* pNewSurface = new Surface(pCurEdge,true);
		CheckPointer(pNewSurface);
		pNewSurface->CheckSurface();
		vecSurface.push_back(pNewSurface);
	}
	m_StatusLine.EraseNode(pPre);
	return true;
}

bool Triangulation::ProcessRegularPoint(DoubleEdge* pCurEdge)
{
	CheckPointer(pCurEdge);
	if (pCurEdge->m_eType==PT_REGULARLEFT)
	{
		CheckPointer(pCurEdge->m_pPre);
		CheckPointer(pCurEdge->m_pNext);
		DoubleEdge* pPreRegular = pCurEdge->m_pPre;
		DoubleEdge* pNext = pCurEdge->m_pNext;
		CheckPointer(pPreRegular->m_ptHelper);
		if (pPreRegular->m_ptHelper->m_eType==PT_MERGE)
		{
			ConnectDiagonalLeftRegulerToMerge(pCurEdge,pPreRegular->m_ptHelper);
		}
		m_StatusLine.EraseNode(pPreRegular);

		CheckPointer(pNext->m_pPre);
		DoubleEdge* pNewCurEdge = pNext->m_pPre;
#if (_DEBUG_LEVEL==2)
		if (pNewCurEdge!=pCurEdge)
		{
			cout<<"ProcessRegularPoint LEFT pNewCurEdge!"<<endl;
			ASSERT(0);
		}
#endif
		//pCurEdge->m_ptHelper = pCurEdge;
		//m_StatusLine.InsertNode(pCurEdge);
		pNewCurEdge->UpdateHelper(pNewCurEdge);
		m_StatusLine.InsertNode(pNewCurEdge);
	}
	else 
	{
		CheckPointer(pCurEdge->m_pPre);
		DoubleEdge* pPreRegular = pCurEdge->m_pPre;
		DoubleEdge* pDirectLeft = m_StatusLine.FindDirectLeft(pCurEdge->m_ptStart);
		CheckPointer(pDirectLeft);
		//same bug
		DoubleEdge* pPre = pCurEdge->m_pPre;
		CheckPointer(pDirectLeft->m_ptHelper);
		if (pDirectLeft->m_ptHelper->m_eType==PT_MERGE)
		{
			ConnectDiagonalRightRegulerToMerge(pCurEdge,pDirectLeft->m_ptHelper);
		}
		DoubleEdge* pNewCurEdge = pPre->m_pNext;
		CheckPointer(pNewCurEdge);
		//pDirectLeft->UpdateHelper(pCurEdge);
		pDirectLeft->UpdateHelper(pNewCurEdge);
	}

	return true;
}

bool Triangulation::ProcessMergePoint(DoubleEdge* pCurEdge)
{
	CheckPointer(pCurEdge);
	DoubleEdge* pDirectLeft = m_StatusLine.FindDirectLeft(pCurEdge->m_ptStart);
	CheckPointer(pDirectLeft);
	CheckPointer(pCurEdge->m_pPre);
	DoubleEdge* pPre = pCurEdge->m_pPre;
	CheckPointer(pDirectLeft->m_ptHelper);
	if (pDirectLeft->m_ptHelper->m_eType == PT_MERGE)
	{
		ConnectDiagonalMergeToLeftMerge(pCurEdge,pDirectLeft->m_ptHelper);
	}

	CheckPointer(pPre->m_pNext);
	//update the helper of the direct left edge
	DoubleEdge* pNewCurEdge = pPre->m_pNext;
	//pDirectLeft->UpdateHelper(pCurEdge); //
	pDirectLeft->UpdateHelper(pNewCurEdge); //

	//
	if (pNewCurEdge!=pCurEdge)
	{
		cout<<"ProcessMergePoint new CurEdge"<<endl;
	}
	
	CheckPointer(pPre->m_ptHelper);
	//DoubleEdge* pPre = pCurEdge->m_pPre;
	if (pPre->m_ptHelper->m_eType == PT_MERGE)
	{
		ConnectDiagonalMergeToRightMerge(pCurEdge,pPre->m_ptHelper);
	}	
	//ɾpreEdge
	m_StatusLine.EraseNode(pPre);
	return true;
}

bool Triangulation::ProcessSplitPoint(DoubleEdge* pCurEdge)
{
	CheckPointer(pCurEdge);

	DoubleEdge* pDirectLeft = m_StatusLine.FindDirectLeft(pCurEdge->m_ptStart);

	CheckPointer(pDirectLeft);

	CheckPointer(pCurEdge->m_pPre);
	DoubleEdge* pPre = pCurEdge->m_pPre;

	CheckPointer(pDirectLeft->m_ptHelper);
	ConnectDiagonalSplit(pCurEdge,pDirectLeft->m_ptHelper);

	CheckPointer(pPre->m_pNext);
	DoubleEdge* pNewCurEdge = pPre->m_pNext;

	//pDirectLeft->UpdateHelper(pCurEdge);   
	//This bug is so tricky!!! ,ĸ²شеֻһʵӦúС
	pDirectLeft->UpdateHelper(pNewCurEdge); 
	pCurEdge->UpdateHelper(pCurEdge);
	m_StatusLine.InsertNode(pCurEdge);

	return true;
}

//endPoint  mergePoint2Surface
bool Triangulation::ConnectDiagonalEndPoint(DoubleEdge* pEdgeMerge,DoubleEdge* pEdgeEnd)
{
	DoubleEdge* pPreMerge = pEdgeMerge->m_pPre;
	DoubleEdge* pPreEnd = pEdgeEnd->m_pPre;

	DoubleEdge* pFromMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeMerge->m_ptStart);
	DoubleEdge* pFromEnd = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeEnd->m_ptStart);

	ConnectThreeEdge(pPreMerge,pFromMerge,pEdgeEnd);
	ConnectThreeEdge(pPreEnd,pFromEnd,pEdgeMerge);

	Surface* pNewSurface1 = new Surface(pFromMerge,true);
	Surface* pNewSurface2 = new Surface(pFromEnd,true);

	CheckPointer(pNewSurface1);
	CheckPointer(pNewSurface2);

	pNewSurface1->CheckSurface();
	pNewSurface2->CheckSurface();

	vecSurface.push_back(pNewSurface1);
	vecSurface.push_back(pNewSurface2);
	return true;
}

//Merge  Mergeһ汻ָȥ
bool Triangulation::ConnectDiagonalMergeToLeftMerge(DoubleEdge* pEdgeMerge,DoubleEdge* pEdgeLeftMerge)
{
	DoubleEdge* pMergeToLeftMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeMerge->m_ptStart);
	DoubleEdge* pLeftMergeToMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeLeftMerge->m_ptStart);

	DoubleEdge* pPreMerge = pEdgeMerge->m_pPre;
	DoubleEdge* pPreLeftMerge = pEdgeLeftMerge->m_pPre;

	ConnectThreeEdge(pPreLeftMerge,pLeftMergeToMerge,pEdgeMerge);
	ConnectThreeEdge(pPreMerge,pMergeToLeftMerge,pEdgeLeftMerge);

	Surface* pNewSurface = new Surface(pLeftMergeToMerge,true);  //the Surface which can be split out
	CheckPointer(pNewSurface);
	pNewSurface->CheckSurface();
	vecSurface.push_back(pNewSurface);

	return true;
}

//Merge  ұMergeһ汻ָȥ
bool Triangulation::ConnectDiagonalMergeToRightMerge(DoubleEdge* pEdgeMerge,DoubleEdge* pEdgeRightMerge)
{
	DoubleEdge* pMergeToRightMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeMerge->m_ptStart);
	DoubleEdge* pRightMergeToMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeRightMerge->m_ptStart);

	DoubleEdge* pPreMerge = pEdgeMerge->m_pPre;
	DoubleEdge* pPreRightMerge = pEdgeRightMerge->m_pPre;

	ConnectThreeEdge(pPreRightMerge,pRightMergeToMerge,pEdgeMerge);
	ConnectThreeEdge(pPreMerge,pMergeToRightMerge,pEdgeRightMerge);

	Surface* pNewSurface = new Surface(pMergeToRightMerge,true);  //the Surface which can be split out
	CheckPointer(pNewSurface);
	pNewSurface->CheckSurface();
	vecSurface.push_back(pNewSurface);
	return true;
}

//split  ,ÿһsplitȻµSurfaceɣʱ뵽vecSurfaceУΪȷǷ y 
bool Triangulation::ConnectDiagonalSplit(DoubleEdge* pEdgeSplit,DoubleEdge* pEdgeOther)
{
	DoubleEdge* pOtherToSplit = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeOther->m_ptStart);
	DoubleEdge* pSplitToOther = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeSplit->m_ptStart);

	DoubleEdge* pSplitPre = pEdgeSplit->m_pPre;
	DoubleEdge* pOtherPre = pEdgeOther->m_pPre;

	ConnectThreeEdge(pOtherPre,pOtherToSplit,pEdgeSplit);
	ConnectThreeEdge(pSplitPre,pSplitToOther,pEdgeOther);
	//pEdgeSplit = pSplitToOther;
	return true;
}

bool Triangulation::ConnectDiagonalLeftRegulerToMerge(DoubleEdge* pEdgeRegular,DoubleEdge* pEdgeMerge)
{
	DoubleEdge* pRegularToMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeRegular->m_ptStart);
	DoubleEdge* pMergeToRegular = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeMerge->m_ptStart);


	DoubleEdge* pPreRegular = pEdgeRegular->m_pPre;
	DoubleEdge* pPreMerge = pEdgeMerge->m_pPre;

	ConnectThreeEdge(pPreRegular,pRegularToMerge,pEdgeMerge);
	ConnectThreeEdge(pPreMerge,pMergeToRegular,pEdgeRegular);

	Surface* pNewSurface = new Surface(pRegularToMerge,true);
	CheckPointer(pNewSurface);
	pNewSurface->CheckSurface();
	vecSurface.push_back(pNewSurface);
	return true;
}

bool Triangulation::ConnectDiagonalRightRegulerToMerge(DoubleEdge* pEdgeRegular,DoubleEdge* pEdgeMerge)
{
	DoubleEdge* pRegularToMerge = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeRegular->m_ptStart);
	DoubleEdge* pMergeToRegular = GeoGetDoubleEdgeFactory()->ConstructLineEdge(pEdgeMerge->m_ptStart);

	DoubleEdge* pPreRegular = pEdgeRegular->m_pPre;
	DoubleEdge* pPreMerge = pEdgeMerge->m_pPre;

	ConnectThreeEdge(pPreRegular,pRegularToMerge,pEdgeMerge);
	ConnectThreeEdge(pPreMerge,pMergeToRegular,pEdgeRegular);

	Surface* pNewSurface = new Surface(pMergeToRegular,true);
	CheckPointer(pNewSurface);
	pNewSurface->CheckSurface();
	vecSurface.push_back(pNewSurface);
	return true;
}

void leadingline::ConnectThreeEdge(DoubleEdge* pFirst,DoubleEdge* pSecond,DoubleEdge* pThird)
{
	pFirst->m_pNext = pSecond;
	pSecond->m_pPre = pFirst;

	pSecond->m_pNext = pThird;
	pThird->m_pPre = pSecond;
}


//utility
void leadingline::DoRotationCCW(const Point_2& ptOrigin,const Point_2& ptStart,IN OUT Point_2& ptDest,double nAngle)
{
	double nCos = cos(nAngle);
	double nSin = sin(nAngle);

	double Vx = ptStart.x - ptOrigin.x;
	double Vy = ptStart.y - ptOrigin.y;

	double Dx = Vx*nCos - Vy*nSin;
	double Dy = Vy*nCos + Vx*nSin;

	ptDest.x = ptOrigin.x + Dx;
	ptDest.y = ptOrigin.y + Dy;
}

void leadingline::DoRotationCW(Point_2 ptOrigin,Point_2 ptStart,Point_2& ptDest,double nAngle)
{
	//double nNewAngle = 2*c_nPI - nAngle;
	//DoRotationCCW(ptOrigin, ptStart, ptDest,nNewAngle);
	double nCos = cos(nAngle);
	double nSin = sin(nAngle);

	double Vx = ptStart.x - ptOrigin.x;
	double Vy = ptStart.y - ptOrigin.y;

	double Dx = Vx*nCos + Vy*nSin;
	double Dy = Vy*nCos - Vx*nSin;

	ptDest.x = ptOrigin.x + Dx;
	ptDest.y = ptOrigin.y + Dy;
}

//ָǰnLength
void leadingline::GoFront(Point_2 ptCurPos,double nCos,double nSin,double nLength,Point_2& ptOut)
{
	ptOut.x = ptCurPos.x + nCos*nLength;
	ptOut.y = ptCurPos.y + nSin*nLength;
}

//жptPointǷpTriangle
bool leadingline::IsInTriangle(Surface* pTriangle,Point_2 ptPoint)
{
	DoubleEdge* pEdge = pTriangle->m_ptEdge;
	for (int i=0;i<3;i++)
	{
		Point_2 ptStart = pEdge->m_ptStart;
		Point_2 ptEnd = pEdge->m_pNext->m_ptStart;
		if (!IsLeftOf(ptPoint,ptStart,ptEnd))
		{
			return false;
		}
		pEdge = pEdge->m_pNext;
	}
	return true;
}

bool leadingline::IsLeftOf(Point_2 ptPoint,Point_2 ptStart,Point_2 ptEnd)
{
	double dx1 = ptEnd.x - ptStart.x;
	double dy1 = ptEnd.y - ptStart.y;
	double dx2 = ptPoint.x - ptStart.x;
	double dy2 = ptPoint.y - ptStart.y;
	double b = dx1*dy2 - dx2*dy1;
	if (b>0)
	{
		return true;
	}
	return false;
}

double leadingline::PowerDistance(Point_2 ptStart,Point_2 ptEnd)
{ 
	double dx = ptStart.x - ptEnd.x;
	double dy = ptStart.y - ptEnd.y;
	return dx*dx+dy*dy;
}

double leadingline::PointToSegmentDistance(const Point_2& ptSegStart,const Point_2& ptSegEnd,const Point_2& ptPoint)
{
	double area = 0;
	double BotLen = sqrt(PowerDistance(ptSegStart,ptSegEnd));
	vector<Point_2> vec;
	vec.push_back(ptSegStart),vec.push_back(ptSegEnd),vec.push_back(ptPoint),vec.push_back(ptSegStart);
	for (int i=0;i<vec.size()-1;i++)
	{
		area += ( vec[i].x * vec[i+1].y - vec[i+1].x * vec[i].y );
	}
	return (area/BotLen);
}

bool leadingline::IsPointInLine(IN const DPOINT2& ptPoint_,
						   IN const DPOINT2& ptLineStart_,
						   IN const DPOINT2& ptLineEnd_)
{
	ASSERT(IsValidDouble(ptPoint_.x));
	ASSERT(IsValidDouble(ptPoint_.y));

	// 㵽߶Լ߶յ㵽 ---  2010 09 14
	VectorD2D _vecPtToStart = ptPoint_ - ptLineStart_;
	VectorD2D _vecEndToPt = ptLineEnd_ - ptPoint_;

	double _nLength = (ptLineEnd_ - ptLineStart_).GetLength();
	double _nLengthStart = _vecPtToStart.GetLength();
	double _nLengthEnd = _vecEndToPt.GetLength();

	// ֮Ͳ߶γȣ˵߶ ---  2010 09 14
	if (!DOUBLE_EQU(_nLengthEnd + _nLengthStart, _nLength))
		return false;

	return true;
}

double leadingline::RegularizeRadian(IN const double& nRadian_)
{
	double _nRagularizeRadian = nRadian_;

	//С0ʱ  2010-1-20
	while (_nRagularizeRadian < 0.0)
	{
		_nRagularizeRadian += c_nPI * 2.0;
	}

	//ȴ2*piʱ  2010-1-20
	while (_nRagularizeRadian >= c_nPI * 2.0)
	{
		_nRagularizeRadian -= c_nPI * 2.0;
	}

	ASSERT(_nRagularizeRadian >= 0.0 && _nRagularizeRadian < c_nPI * 2.0);
	return _nRagularizeRadian;
}

bool leadingline::IsPointInArc(IN const DPOINT2& ptPoint_,
						  IN const DPOINT2& ptCenter_,
						  IN const double& nRadius_,
						  IN const double& nStartAngle_,
						  IN const double& nCentralAngle_,
						  IN const bool& bIsNormalDir_)
{
	ASSERT(IsValidDouble(ptPoint_.x));
	ASSERT(IsValidDouble(ptPoint_.y));
	ASSERT(IsValidDouble(ptCenter_.x));
	ASSERT(IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));
	ASSERT(DOUBLE_GREAT_ZERO(nCentralAngle_) &&
		DOUBLE_LE(nCentralAngle_, 2 * c_nPI));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) &&
		DOUBLE_LE(nStartAngle_, 2 * c_nPI));

	double _nFactor = bIsNormalDir_ ? 1 : -1;

	// 㵽Բĵڰ뾶 ---  2010 08 21
	DPOINT2 _vecPtToCenter = ptPoint_ - ptCenter_;
	double _nDis = _vecPtToCenter.GetLength();
	if (!DOUBLE_EQU(_nDis, nRadius_))
		return false;

	// ԲʼǵptPoint֮ԲĽ ---  2010 08 21
	double _nEndAngleTemp = D2GetNormalAngle(_vecPtToCenter);
	double _nCentralAngleTemp = (_nEndAngleTemp - nStartAngle_) * _nFactor;
	_nCentralAngleTemp = RegularizeRadian(_nCentralAngleTemp);

	// ȽԲĽǴСжǷԲ ---  2010 08 21
	return DOUBLE_LE(_nCentralAngleTemp, nCentralAngle_);
}

double leadingline::D2GetNormalAngle(IN const DPOINT2& ptX_)
{
	//ȡ㵽ԭľ ---  2010-1-20
	double _nDistance = sqrt(pow(ptX_.x, 2) + pow(ptX_.y, 2));

	//С㣬Ϊԭ㣬XϣɽǶΪ ---  2010-1-20
	if (DOUBLE_LE_ZERO(_nDistance))
		return 0.0;

	//Xɵֵ ---  2010-1-20
	ASSERT(_nDistance > 0.0);
	double _nTemp = ptX_.x / _nDistance;
	//_nTemp-11֮ ---  2010-1-20
	__CUT(_nTemp, -1., 1.);
	//㷴ңǶ ---  2010-1-20
	double _nAngle = acos(_nTemp);
	//ptX_.yΪֵ(ptX_.x,ptX_.y)ڵһڶ---  2010-1-20
	if (DOUBLE_GE_ZERO(ptX_.y)) 
		return _nAngle;
	//ptX_.yΪֵ(ptX_.x,ptX_.y)ڵ---  2010-1-20
	return 2.0 * c_nPI - _nAngle;
}

double leadingline::D2GetPointToLineDistance(IN DPOINT2 ptPoint_,
										IN DPOINT2 ptLineStart_,
										IN DPOINT2 ptLineEnd_)
{
	DPOINT2 _vecLine(ptLineEnd_ - ptLineStart_);
	ASSERT(DOUBLE_GREAT_ZERO(_vecLine.GetLength()));

	// ֱ߱׼̵ϵ
	double _a, _b, _c;
	_a = ptLineEnd_.y - ptLineStart_.y;
	_b = ptLineStart_.x - ptLineEnd_.x;
	_c = (ptLineEnd_.x - ptLineStart_.x) * ptLineStart_.y -
		(ptLineEnd_.y - ptLineStart_.y) * ptLineStart_.x;

	return (fabs(_a * ptPoint_.x + _b * ptPoint_.y + _c)) / _vecLine.GetLength();
}

double leadingline::D2GetProjectionDistance(IN const VectorD2D& vectorA_,IN const VectorD2D& vectorB_)
{
	double _nLength = vectorB_.GetLength();
	if (DOUBLE_NOT_ZERO(_nLength))
		return (vectorA_.x * vectorB_.x + vectorA_.y * vectorB_.y) / _nLength;
	return 0;
}

double D2GetProjectionDistance(IN const double& nPtAx_, IN const double& nPtAy_, IN const double& nPtBx_, IN const double& nPtBy_)
{
	VectorD2D _vectorA(nPtAx_, nPtAy_);
	VectorD2D _vectorB(nPtBx_, nPtBy_);
	return D2GetProjectionDistance(_vectorA, _vectorB);
}

namespace
{
	VectorD2D NormalizeVector(const VectorD2D& vecValue_)
	{
		double _nLength = vecValue_.GetLength();
		ASSERT(DOUBLE_GREAT_ZERO(_nLength));
		return VectorD2D(vecValue_.x / _nLength, vecValue_.y / _nLength);
	}
}

int leadingline::Intersect_BeelineWithBeeline(IN const DPOINT2& pt1_,
										 IN double nAngle1_,
										 IN const DPOINT2& pt2_,
										 IN double nAngle2_,
										 OUT DPOINT2* pptIntersect_ /* = NULL */)
{
	ASSERT(IsValidDouble(pt1_.x));
	ASSERT(IsValidDouble(pt1_.y));
	ASSERT(IsValidDouble(pt2_.x));
	ASSERT(IsValidDouble(pt2_.y));

	// һȡx = pt1.x + 1 ---  2010 09 07
	__point2<double> _pt11(pt1_.x, pt1_.y);
	__point2<double> _pt12(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);

	__point2<double> _pt21(pt2_.x, pt2_.y);
	__point2<double> _pt22(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);

	if (DOUBLE_EQU(nAngle1_, c_nPI / 2) || DOUBLE_EQU(nAngle1_, 3 * c_nPI / 2))
	{
		_pt12 .Set(pt1_.x, pt1_.y + 1);
	}
	else
	{
		_pt12.Set(pt1_.x + 1, tan(nAngle1_) + pt1_.y);
	}

	if (DOUBLE_EQU(nAngle2_, c_nPI / 2) || DOUBLE_EQU(nAngle2_, 3 * c_nPI / 2))
	{
		_pt22.Set(pt2_.x, pt2_.y + 1);
	}
	else
	{
		_pt22.Set(pt2_.x + 1, tan(nAngle2_) + pt2_.y);
	}

	// ԭ㵽ֱߵĶ˵ ---wanghong 2010 09 07
	VectorD2D _vec1x(_pt11), _vec1y(_pt12);
	VectorD2D _vec2x(_pt21), _vec2y(_pt22);
	// ֱߵ ---wanghong 2010 09 07
	VectorD2D _vecA(_vec1y - _vec1x), _vecB(_vec2y - _vec2x);
	// ڶֱߵĴֱ ---wanghong 2010 09 07
	VectorD2D _vecBPlumb(-_vecB.y, _vecB.x);

	if (DOUBLE_EQU_ZERO(NormalizeVector(_vecA) * NormalizeVector(_vecBPlumb)))
	{
		// ƽлغ ---wanghong 2010 09 07
		return 0;
	}

	double _n = (_vec2x * _vecBPlumb - _vec1x * _vecBPlumb) / (_vecA * _vecBPlumb);
	if (pptIntersect_ != NULL)
	{
		pptIntersect_->x = _vec1x.x + _n * _vecA.x;
		pptIntersect_->y = _vec1x.y + _n * _vecA.y;
	}

	return 1;
}

int leadingline::Intersect_LineWithLine(IN const DPOINT2& ptStart1_,
								   IN const DPOINT2& ptEnd1_,
								   IN const DPOINT2& ptStart2_,
								   IN const DPOINT2& ptEnd2_,
								   OUT DPOINT2* pptIntersect_ /* = NULL */)
{
	DPOINT2 _vecLine1(ptEnd1_ - ptStart1_);
	DPOINT2 _vecLine2(ptEnd2_ - ptStart2_);

	ASSERT(DOUBLE_GE_ZERO(_vecLine1.GetLength()));
	ASSERT(DOUBLE_GE_ZERO(_vecLine2.GetLength()));

	double _nAngle1 = D2GetNormalAngle(_vecLine1);
	double _nAngle2 = D2GetNormalAngle(_vecLine2);

	DPOINT2 _ptIntersect(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	int _nNum = Intersect_BeelineWithBeeline(ptStart1_, _nAngle1,
		ptStart2_, _nAngle2, &_ptIntersect);

	if (_nNum == 0 || !IsPointInLine(_ptIntersect, ptStart1_, ptEnd1_) ||
		!IsPointInLine(_ptIntersect, ptStart2_, ptEnd2_))
	{
		return 0;
	}

	if (pptIntersect_)
		*pptIntersect_ = _ptIntersect;

	return 1;
}

int leadingline::Intersect_BeelineWithCircle(IN const DPOINT2& ptBeeline_,
										IN double nBeelineAngle_,
										IN const DPOINT2& ptCenter_,
										IN double nRadius_,
										OUT DPOINT2* pptIntersect1_ /* = NULL */,
										OUT DPOINT2* pptIntersect2_ /* = NULL */)
{
	ASSERT(IsValidDouble(ptBeeline_.x));
	ASSERT(IsValidDouble(ptBeeline_.y));
	ASSERT(IsValidDouble(ptCenter_.x));
	ASSERT(IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));

	double x1, y1, x2, y2;		// ֱ12
	double x3, y3;				// Բĵֱ12ͶӰ;
	double _nCenterToLineDis;	// Բĵֱ12ľ
	double _nPrjToIntersectDis;	// ͶӰ㵽ľ
	double _nLineLength;		// ָ߶εĳ
	double dvx1c, dvy1c;		// ӵ1Բĵֱߵķʸ
	double prjc;				// ӵ1Բĵֱ12ϵͶӰ

	// Ϊֱ˵㸳ֵ
	x1 = ptBeeline_.x;
	y1 = ptBeeline_.y;
	x2 = c_nINVALID_DOUBLE;
	y2 = c_nINVALID_DOUBLE;

	if (DOUBLE_EQU(nBeelineAngle_, c_nPI / 2) ||
		DOUBLE_EQU(nBeelineAngle_, 3 * c_nPI / 2))
	{
		x2 = ptBeeline_.x;
		y2 = ptBeeline_.y + 1;
	}
	else
	{
		x2 = ptBeeline_.x + 1;
		y2 = ptBeeline_.y + tan(nBeelineAngle_);
	}

	DPOINT2 _ptBeelineEnd(DPOINT2(x2, y2));
	_nLineLength = (_ptBeelineEnd - ptBeeline_).GetLength();
	DPOINT2 _vecLine(_ptBeelineEnd - ptBeeline_);

	if (DOUBLE_EQU_ZERO(nRadius_))
		return -1;

	// Բĵֱߵľڰ뾶ʱ,ֱԲûн
	_nCenterToLineDis = D2GetPointToLineDistance(ptCenter_, ptBeeline_, _ptBeelineEnd);
	if (DOUBLE_GREAT(_nCenterToLineDis, nRadius_))
		return 0;

	dvx1c = ptCenter_.x - x1;
	dvy1c = ptCenter_.y - y1;
	prjc = D2GetProjectionDistance(dvx1c, dvy1c, _vecLine.x, _vecLine.y);

	_vecLine.Unify();
	x3 = x1 + prjc * _vecLine.x;
	y3 = y1 + prjc * _vecLine.y;

	double nTemp = nRadius_ * nRadius_ - _nCenterToLineDis * _nCenterToLineDis;
	if (nTemp < 0)
		nTemp = 0;
	_nPrjToIntersectDis = sqrt(nTemp);

	DPOINT2 _ptIntersect1(x3 - _nPrjToIntersectDis * _vecLine.x, y3 - _nPrjToIntersectDis * _vecLine.y);
	DPOINT2 _ptIntersect2(x3 + _nPrjToIntersectDis * _vecLine.x, y3 + _nPrjToIntersectDis * _vecLine.y);

	//  ---  2010 09 07
	if (pptIntersect1_ != NULL && DOUBLE_EQU_ZERO((_ptIntersect1 - _ptIntersect2).GetLength()))
	{
		*pptIntersect1_ = _ptIntersect1;
		return 1;
	}

	if (pptIntersect1_ != NULL)
		*pptIntersect1_ = _ptIntersect1;

	if (pptIntersect2_ != NULL)
		*pptIntersect2_ = _ptIntersect2;

	return 2;
}

int leadingline::Intersect_LineWithArc(IN const DPOINT2& ptStart_,
								  IN const DPOINT2& ptEnd_,
								  IN const DPOINT2& ptCenter_,
								  IN double nRadius_,
								  IN double nStartAngle_,
								  IN double nSweptAngle_,
								  OUT DPOINT2* pptIntersect1_/* = NULL*/,
								  OUT DPOINT2* pptIntersect2_/* = NULL*/)
{
	ASSERT(IsValidDouble(ptStart_.x) && IsValidDouble(ptStart_.y));
	ASSERT(IsValidDouble(ptEnd_.x) && IsValidDouble(ptEnd_.y));
	ASSERT(IsValidDouble(ptCenter_.x) && IsValidDouble(ptCenter_.y));
	ASSERT(DOUBLE_GREAT_ZERO(nRadius_));
	ASSERT(DOUBLE_GE_ZERO(nStartAngle_) && DOUBLE_LESS(nStartAngle_, 2 * c_nPI));
	ASSERT(DOUBLE_GREAT_ZERO(fabs(nSweptAngle_)) && DOUBLE_LE(fabs(nSweptAngle_), 2 * c_nPI));

	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	// ͨ߶ζӦֱԲӦԲ󽻣áǽ㡱ĵ㼯 --- wanghong 2010 09 08
	int _nImageIntersect = Intersect_BeelineWithCircle(ptStart_, D2GetNormalAngle(ptEnd_ - ptStart_),
		ptCenter_, nRadius_, &_ptIntersec1, &_ptIntersec2);
	if (_nImageIntersect < 1)
		return _nImageIntersect;

	bool _bPt1Exist = IsPointInLine(_ptIntersec1, ptStart_, ptEnd_) &&
		IsPointInArc(_ptIntersec1, ptCenter_, nRadius_, nStartAngle_, fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	bool _bPt2Exist = false;
	if (_nImageIntersect == 2)
	{	// ʱظ --- wanghong 2010 09 08
		_bPt2Exist = IsPointInLine(_ptIntersec2, ptStart_, ptEnd_) &&
			IsPointInArc(_ptIntersec2, ptCenter_, nRadius_, nStartAngle_, fabs(nSweptAngle_), DOUBLE_GREAT_ZERO(nSweptAngle_));
	}
	//  --- wanghong 2010 09 08
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}
	// һ --- wanghong 2010 09 08
	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}

	return 0;
}
}

int leadingline::Intersect_ArcWithArc(IN const DPOINT2& ptCenter1_, IN const DPOINT2& ptCenter2_,
						 IN double nRadius1_, IN double nRadius2_,
						 IN double nStartAngle1_, IN double nStartAngle2_,
						 IN double nSweptAngle1_, IN double nSweptAngle2_,
						 OUT DPOINT2* pptIntersect1_, OUT DPOINT2* pptIntersect2_)
{
	DPOINT2 _ptIntersec1(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	DPOINT2 _ptIntersec2(c_nINVALID_DOUBLE, c_nINVALID_DOUBLE);
	int _nImageIntersect = 0;
	DPOINT2 _vecCenter1ToCenter2(ptCenter2_.x - ptCenter1_.x, ptCenter2_.y - ptCenter1_.y);
	double _nAngleCenter1ToCenter2 = D2GetNormalAngle(_vecCenter1ToCenter2);
	double _nCenterDistance = sqrt(_vecCenter1ToCenter2.x * _vecCenter1ToCenter2.x + _vecCenter1ToCenter2.y * _vecCenter1ToCenter2.y);
	double _nRadiusSum = nRadius1_ + nRadius2_;
	double _nRadiusDif = abs(nRadius1_ - nRadius2_);

	if(_nCenterDistance - _nRadiusSum > 1e-3 || _nCenterDistance - _nRadiusDif < -(1e-3))
	{
		return 0;
	}

	// Բľڰ뾶ͻԲ
	if((_nCenterDistance - _nRadiusSum >= -(1e-3) && _nCenterDistance - _nRadiusSum <= 1e-3)
		|| (_nCenterDistance - _nRadiusDif >= -(1e-3) && _nCenterDistance - _nRadiusDif <= 1e-3))
	{
		_nImageIntersect = 1;
		if(_nCenterDistance - _nRadiusSum >= -(1e-3) && _nCenterDistance - _nRadiusSum <= 1e-3)
		{
			_ptIntersec1.x = ptCenter1_.x + nRadius1_ * _vecCenter1ToCenter2.x / _nCenterDistance;
			_ptIntersec1.y = ptCenter1_.y + nRadius1_ * _vecCenter1ToCenter2.y / _nCenterDistance;
		}
		if(_nCenterDistance - _nRadiusDif >= -(1e-3) && _nCenterDistance - _nRadiusDif <= 1e-3)
		{
			if(nRadius1_> nRadius2_)
			{
				_ptIntersec1.x = ptCenter1_.x + nRadius1_ * _vecCenter1ToCenter2.x / _nCenterDistance;
				_ptIntersec1.y = ptCenter1_.y + nRadius1_ * _vecCenter1ToCenter2.y / _nCenterDistance;
			}
			else 
			{
				_ptIntersec1.x = ptCenter2_.x - nRadius2_ * _vecCenter1ToCenter2.x / _nCenterDistance;
				_ptIntersec1.y = ptCenter1_.y - nRadius2_ * _vecCenter1ToCenter2.y / _nCenterDistance;
			}
		}
	}
	if(_nCenterDistance - _nRadiusSum < -(1e-3) && _nCenterDistance - _nRadiusDif > 1e-3)
	{
		_nImageIntersect = 2;
		double _nIntersectAngle = acos((nRadius1_ * nRadius1_ + _nCenterDistance * _nCenterDistance - nRadius2_ * nRadius2_) / (2 * nRadius1_ * _nCenterDistance));
		_ptIntersec1.x = ptCenter1_.x + nRadius1_ * cos(_nAngleCenter1ToCenter2 + _nIntersectAngle);
		_ptIntersec1.y = ptCenter1_.y + nRadius1_ * sin(_nAngleCenter1ToCenter2 + _nIntersectAngle);
		_ptIntersec2.x = ptCenter1_.x + nRadius1_ * cos(_nAngleCenter1ToCenter2 - _nIntersectAngle);
		_ptIntersec2.y = ptCenter1_.y + nRadius1_ * sin(_nAngleCenter1ToCenter2 - _nIntersectAngle);
	}



	// жԲĽǷڻ
	bool _bPt1Exist = IsPointInArc(_ptIntersec1, ptCenter1_, nRadius1_, nStartAngle1_, fabs(nSweptAngle1_), DOUBLE_GREAT_ZERO(nSweptAngle1_))
		&& IsPointInArc(_ptIntersec1, ptCenter2_, nRadius2_, nStartAngle2_, fabs(nSweptAngle2_), DOUBLE_GREAT_ZERO(nSweptAngle2_));
	bool _bPt2Exist = false;
	if(_nImageIntersect == 2)
	{
		_bPt2Exist = IsPointInArc(_ptIntersec2, ptCenter1_, nRadius1_, nStartAngle1_, fabs(nSweptAngle1_), DOUBLE_GREAT_ZERO(nSweptAngle1_))
			&& IsPointInArc(_ptIntersec2, ptCenter2_, nRadius2_, nStartAngle2_, fabs(nSweptAngle2_), DOUBLE_GREAT_ZERO(nSweptAngle2_));
	}
	
	if (_bPt1Exist && _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _ptIntersec1;
		if (pptIntersect2_ != NULL)
			*pptIntersect2_ = _ptIntersec2;
		return 2;
	}

	if (_bPt1Exist || _bPt2Exist)
	{
		if (pptIntersect1_ != NULL)
			*pptIntersect1_ = _bPt1Exist ? _ptIntersec1 : _ptIntersec2;
		return 1;
	}
	
	else
	{
		return 0;
	}
}