// ************************************************************************** //
// TubeToCadTranslator.cpp : implementation of CTubeToCadTranslator class.
// Copyright(c) 2017 WeiHong Company. All rights reserved.
//
// Abstract:
//   ļʵֽܲתΪCAD
// *********************************************** ---- ȫ 2017-02-28 ***** //
#include "stdafx.h"

using namespace translators;

//////////////////////////////////////////////////////////////////////////
// class CWires3DTranslator
CWires3DTranslator::CWires3DTranslator()
	: m_pTubeToCadTranslator(NULL)
{
}

void CWires3DTranslator::Initialize(CTubeToCadTranslator* pTubeToCadTranslator_)
{
	ASSERT(pTubeToCadTranslator_ != NULL);
	m_pTubeToCadTranslator = pTubeToCadTranslator_;
}

Standard_Boolean CWires3DTranslator::Translate(LISTWIRES3DONPLANE& listWires3DOnPlane_)
{
	ASSERT(listWires3DOnPlane_.empty());

	// ܲĺ߽
	if (!AdjustOuterBoundary())
		return false;

	LISTEDGES3DCITER _iterBegin = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.begin();
	LISTEDGES3DCITER _iterEnd = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		ASSERT((EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType) || (EdgeInfo::EDGE_OVAL == _iterBegin->nEdgeType));
		if (EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType)
		{
			TransWiresOnPlane(_iterBegin, listWires3DOnPlane_);
		}
		else if (DIST_EQU(_iterBegin->nLongRadius, _iterBegin->nShortRadius))
		{
			// ȻȡԲƽı/ߣٻȡԲϵı/
			Wire3DOnPlaneInfo _WireOnSurface;
			LISTEDGES3DCITER _iterNext = _iterBegin;
			if (++_iterNext != _iterEnd)
			{
				TransWiresOnPlane(_iterNext, listWires3DOnPlane_);
				GetWiresOnOvalSurface(*_iterBegin, _WireOnSurface);
				listWires3DOnPlane_.insert(--listWires3DOnPlane_.end(), _WireOnSurface);
				++_iterBegin;
			}
			else
			{
				GetWiresOnOvalSurface(*_iterBegin, _WireOnSurface);
				listWires3DOnPlane_.push_back(_WireOnSurface);
			}
		}
		else
		{
			ASSERT(!DIST_EQU(_iterBegin->nLongRadius, _iterBegin->nShortRadius));
			Wire3DOnPlaneInfo _WireOnSurface;
			GetWiresOnOvalSurface(*_iterBegin, _WireOnSurface);
			listWires3DOnPlane_.push_back(_WireOnSurface);
		}
	}

	if (listWires3DOnPlane_.empty())
		return false;

	ASSERT((listWires3DOnPlane_.size() == 1) || (listWires3DOnPlane_.size() == 4) || (listWires3DOnPlane_.size() == 8));
	DeleteRepeatEdges(listWires3DOnPlane_);

	return true;
}

Standard_Boolean CWires3DTranslator::AdjustOuterBoundary()
{
	ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.bClosed);

	switch (m_pTubeToCadTranslator->m_TubeParam.nTubeType)
	{
	case TubeInfo::rectangular:
		{
			// ܣ
			LISTEDGES3D& _listEdges = m_pTubeToCadTranslator->m_OuterBoundary.listEdges;
			LISTEDGES3DCITER _iterBegin = _listEdges.begin();
			if (_iterBegin->nEdgeType != EdgeInfo::EDGE_LINE)
				++_iterBegin;

			LISTEDGES3DCITER _iterNext = _iterBegin;
			if ((++_iterNext)->nEdgeType != EdgeInfo::EDGE_LINE)
				++_iterNext;

			ASSERT((EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType) && (EdgeInfo::EDGE_LINE == _iterNext->nEdgeType));
			ASSERT(IsEqualVectorNoDir(_iterBegin->ptDir ^ _iterNext->ptDir, m_pTubeToCadTranslator->m_ptAxleWireDir));
			if (!(_iterBegin->ptDir ^ _iterNext->ptDir).IsEqual(m_pTubeToCadTranslator->m_ptAxleWireDir, g_nDistPrecision))
			{
				_listEdges.reverse();
				LISTEDGES3D::iterator _iterBegin = _listEdges.begin();
				LISTEDGES3D::iterator _iterEnd = _listEdges.end();
				for (; _iterBegin != _iterEnd; ++_iterBegin)
					_iterBegin->Reverse();
			}

			// ܣıΪһ
			Standard_Integer _nSize = static_cast<Standard_Integer>(_listEdges.size());
			Standard_Integer _nIndex = 0;
			while (_nIndex++ < _nSize)
			{
				const Edge3DInfo& _edge = _listEdges.front();
				if ((EdgeInfo::EDGE_LINE == _edge.nEdgeType)
					&& DIST_EQU((_edge.ptEnd - _edge.ptStart).Modulus(), m_pTubeToCadTranslator->m_TubeParam.nRectWidth
					- 2.0 * m_pTubeToCadTranslator->m_TubeParam.nRectRadius))
					break;

				_listEdges.push_back(_edge);
				_listEdges.pop_front();
			}

			ASSERT(_nIndex < _nSize);
			return (_nIndex < _nSize);
		}
		break;

	case TubeInfo::circular:
		{
			// Բܣ߿
			AdjustCircularOuterBoundary();
		}
	case TubeInfo::oval:
		{
			// Բ/Բܣ
			ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 1);
			Edge3DInfo& _edge = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.front();
			ASSERT(IsEqualVectorNoDir(_edge.ptXDir ^ _edge.ptYDir, m_pTubeToCadTranslator->m_ptAxleWireDir));
			if ((_edge.ptXDir ^ _edge.ptYDir).IsEqual(m_pTubeToCadTranslator->m_ptAxleWireDir, g_nDistPrecision)
				!= (_edge.nLastParam > _edge.nFirstParam))
			{
				_edge.Reverse();
			}
		}
		break;

	case TubeInfo::waist:
		{
			// ܣ
			LISTEDGES3D::iterator _iterBegin = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.begin();
			LISTEDGES3D::iterator _iterTemp = _iterBegin;
			if (EdgeInfo::EDGE_LINE == _iterTemp->nEdgeType)
				++_iterTemp;

			ASSERT(EdgeInfo::EDGE_OVAL == _iterTemp->nEdgeType);
			ASSERT(IsEqualVectorNoDir(_iterTemp->ptXDir ^ _iterTemp->ptYDir, m_pTubeToCadTranslator->m_ptAxleWireDir));
			if ((_iterTemp->ptXDir ^ _iterTemp->ptYDir).IsEqual(m_pTubeToCadTranslator->m_ptAxleWireDir, g_nDistPrecision)
				!= (_iterTemp->nLastParam > _iterTemp->nFirstParam))
			{
				m_pTubeToCadTranslator->m_OuterBoundary.listEdges.reverse();
				LISTEDGES3D::iterator _iterEnd = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.end();
				for (; _iterBegin != _iterEnd; ++_iterBegin)
				{
					_iterBegin->Reverse();
				}
			}

			// ܣֱ߱Ϊһ
			Standard_Integer _nSize = static_cast<Standard_Integer>(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());
			ASSERT(_nSize == 4);
			Standard_Integer _nIndex = 0;
			while (_nIndex++ < _nSize)
			{
				const Edge3DInfo& _edge = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.front();
				if ((EdgeInfo::EDGE_LINE == _edge.nEdgeType)
					&& DIST_EQU((_edge.ptEnd - _edge.ptStart).Modulus(), m_pTubeToCadTranslator->m_TubeParam.nWaistWidth))
					break;

				m_pTubeToCadTranslator->m_OuterBoundary.listEdges.push_back(_edge);
				m_pTubeToCadTranslator->m_OuterBoundary.listEdges.pop_front();
			}

			ASSERT(_nIndex < _nSize);
			return (_nIndex < _nSize);
		}
		break;

	default:
		ASSERT(false);
		return false;
	}

	return true;
}

void CWires3DTranslator::AdjustCircularOuterBoundary()
{
	ASSERT(TubeInfo::circular == m_pTubeToCadTranslator->m_TubeParam.nTubeType);
	ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 1);

	Edge3DInfo& _edge = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.front();
	if ((_edge.nEdgeType != EdgeInfo::EDGE_OVAL) || !ANGLE_EQU((_edge.nLastParam - _edge.nFirstParam), 2.0 * g_nPIE))
		return;

	// Χ
	Standard_Real _nParamDir = (_edge.nLastParam > _edge.nFirstParam) ? 1.0 : -1.0;
	_edge.nFirstParam = 0.;
	_edge.nLastParam = _nParamDir * 2.0 * g_nPIE;

	// 
	gp_XYZ _ptNormalVector = _edge.ptXDir ^ _edge.ptYDir;
	ASSERT(IsEqualVectorNoDir(_ptNormalVector, m_pTubeToCadTranslator->m_ptAxleWireDir));
	ASSERT(DIGIT_EQU(max(max(fabs(_ptNormalVector.X()), fabs(_ptNormalVector.Y())), fabs(_ptNormalVector.Z())), 1.0));

	gp_XYZ _ptNewXDir;
	Standard_Real _nMaxAxle = max(max(fabs(_edge.ptXDir.X()), fabs(_edge.ptXDir.Y())), fabs(_edge.ptXDir.Z()));
	if (DIGIT_EQU(_nMaxAxle, fabs(_edge.ptXDir.X())))
	{
		_ptNewXDir.SetX(_copysign(1.0, _edge.ptXDir.X()));
	}
	else if (DIGIT_EQU(_nMaxAxle, fabs(_edge.ptXDir.Y())))
	{
		_ptNewXDir.SetY(_copysign(1.0, _edge.ptXDir.Y()));
	}
	else
	{
		ASSERT(DIGIT_EQU(_nMaxAxle, fabs(_edge.ptXDir.Z())));
		_ptNewXDir.SetZ(_copysign(1.0, _edge.ptXDir.Z()));
	}

	_edge.ptXDir = _ptNewXDir;
	_edge.ptYDir = _ptNormalVector ^ _edge.ptXDir;
}

void CWires3DTranslator::TransWiresOnPlane(const LISTEDGES3DCITER& curIter_, LISTWIRES3DONPLANE& listWiresOnPlane_)
{
	ASSERT(EdgeInfo::EDGE_LINE == curIter_->nEdgeType);
	gp_XYZ _ptAimPoint = (curIter_->StartPoint() + curIter_->EndPoint()) * 0.5;
	gp_XYZ _ptNormalVector = curIter_->ptDir ^ m_pTubeToCadTranslator->m_ptAxleWireDir;

	Wire3DOnPlaneInfo _WireOnPlane;
	_WireOnPlane.ptNormalVector = _ptNormalVector;

	LISTWIRES3D::iterator _iterWireBegin = m_pTubeToCadTranslator->m_listDiceyWires.begin();
	LISTWIRES3D::iterator _iterWireEnd = m_pTubeToCadTranslator->m_listDiceyWires.end();
	while (_iterWireBegin != _iterWireEnd)
	{
		Wire3DInfo _wire;
		LISTEDGES3D::iterator _iterEdgeBegin = _iterWireBegin->listEdges.begin();
		LISTEDGES3D::iterator _iterEdgeEnd = _iterWireBegin->listEdges.end();
		while (_iterEdgeBegin != _iterEdgeEnd)
		{
			if (!JudgeEdgeOnPlane(_ptAimPoint, _ptNormalVector, *_iterEdgeBegin))
			{
				++_iterEdgeBegin;
				continue;
			}

			// ñǰı߷β
			if (!_wire.listEdges.empty() && !_wire.listEdges.back().EndPoint().IsEqual(_iterEdgeBegin->StartPoint(), g_nDistPrecision))
			{
				CheckRepeatWire(_wire, _iterWireBegin->bExclude);
				_WireOnPlane.listWires.push_back(_wire);
				_wire.listEdges.clear();
			}

			_wire.listEdges.push_back(*_iterEdgeBegin);
			LISTEDGES3D::iterator _iterTemp = _iterEdgeBegin++;
			_iterWireBegin->listEdges.erase(_iterTemp);
		}

		if (!_wire.listEdges.empty())
		{
			CheckRepeatWire(_wire, _iterWireBegin->bExclude);
			_WireOnPlane.listWires.push_back(_wire);
		}

		if (!_iterWireBegin->listEdges.empty())
		{
			++_iterWireBegin;
			continue;
		}

		LISTWIRES3D::iterator _iterTemp = _iterWireBegin++;
		m_pTubeToCadTranslator->m_listDiceyWires.erase(_iterTemp);
	}

	listWiresOnPlane_.push_back(_WireOnPlane);
}

void CWires3DTranslator::GetWiresOnOvalSurface(const Edge3DInfo& OvalEdge_, Wire3DOnPlaneInfo& WireOnSurface_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == OvalEdge_.nEdgeType);

	LISTWIRES3D::iterator _iterWireBegin = m_pTubeToCadTranslator->m_listDiceyWires.begin();
	LISTWIRES3D::iterator _iterWireEnd = m_pTubeToCadTranslator->m_listDiceyWires.end();
	while (_iterWireBegin != _iterWireEnd)
	{
		Wire3DInfo _wire;
		LISTEDGES3D::iterator _iterEdgeBegin = _iterWireBegin->listEdges.begin();
		LISTEDGES3D::iterator _iterEdgeEnd = _iterWireBegin->listEdges.end();
		while (_iterEdgeBegin != _iterEdgeEnd)
		{
			if (!JudgeEdgeOnOvalSurface(OvalEdge_, *_iterEdgeBegin))
			{
				++_iterEdgeBegin;
				continue;
			}

			// ñǰı߷β
			if (!_wire.listEdges.empty() && !_wire.listEdges.back().EndPoint().IsEqual(_iterEdgeBegin->StartPoint(), g_nDistPrecision))
			{
				CheckRepeatWire(_wire, _iterWireBegin->bExclude);
				WireOnSurface_.listWires.push_back(_wire);
				_wire.listEdges.clear();
			}

			_wire.listEdges.push_back(*_iterEdgeBegin);
			LISTEDGES3D::iterator _iterTemp = _iterEdgeBegin++;
			_iterWireBegin->listEdges.erase(_iterTemp);
		}

		if (!_wire.listEdges.empty())
		{
			CheckRepeatWire(_wire, _iterWireBegin->bExclude);
			WireOnSurface_.listWires.push_back(_wire);
		}

		if (!_iterWireBegin->listEdges.empty())
		{
			++_iterWireBegin;
			continue;
		}

		LISTWIRES3D::iterator _iterTemp = _iterWireBegin++;
		m_pTubeToCadTranslator->m_listDiceyWires.erase(_iterTemp);
	}
}

Standard_Boolean CWires3DTranslator::JudgeEdgeOnPlane(const gp_XYZ& ptAimPoint_, const gp_XYZ& ptNormalVector_, const Edge3DInfo& edge_) const
{
	if (EdgeInfo::EDGE_LINE == edge_.nEdgeType)
	{
		return (JudgePointOnPlane(ptAimPoint_, ptNormalVector_, edge_.StartPoint())
			&& JudgePointOnPlane(ptAimPoint_, ptNormalVector_, edge_.EndPoint()));
	}

	ASSERT(EdgeInfo::EDGE_OVAL == edge_.nEdgeType);
	return (JudgePointOnPlane(ptAimPoint_, ptNormalVector_, edge_.ptCenter)
		&& IsEqualVectorNoDir(edge_.ptXDir ^ edge_.ptYDir, ptNormalVector_));
}

Standard_Boolean CWires3DTranslator::JudgeEdgeOnOvalSurface(const Edge3DInfo& OvalEdge_, const Edge3DInfo& edge_) const
{
	ASSERT(EdgeInfo::EDGE_OVAL == OvalEdge_.nEdgeType);

	if (EdgeInfo::EDGE_LINE == edge_.nEdgeType)
	{
		gp_XYZ _ptMiddlePoint = 0.5 * (edge_.ptStart + edge_.ptEnd);
		return (JudgePointOnOvalSurface(OvalEdge_, edge_.ptStart)
			&& JudgePointOnOvalSurface(OvalEdge_, _ptMiddlePoint)
			&& JudgePointOnOvalSurface(OvalEdge_, edge_.ptEnd));
	}

	// ȡԲ/Բϵ5ж
	ASSERT(EdgeInfo::EDGE_OVAL == edge_.nEdgeType);
	for (Standard_Integer i = 0; i < 5; ++i)
	{
		Standard_Real _nAlpha = static_cast<Standard_Real>(i) / 4.0;
		Standard_Real _nParam = edge_.nFirstParam * (1.0 - _nAlpha) + edge_.nLastParam * _nAlpha;
		gp_XYZ _ptPoint = edge_.ptCenter + edge_.nLongRadius * cos(_nParam) * edge_.ptXDir + edge_.nShortRadius * sin(_nParam) * edge_.ptYDir;
		if (!JudgePointOnOvalSurface(OvalEdge_, _ptPoint))
			return false;
	}

	return true;
}

Standard_Boolean CWires3DTranslator::JudgePointOnOvalSurface(const Edge3DInfo& OvalEdge_, const gp_XYZ& ptPoint_) const
{
	ASSERT(EdgeInfo::EDGE_OVAL == OvalEdge_.nEdgeType);

	Standard_Real _nHeight = (ptPoint_ - OvalEdge_.ptCenter) * m_pTubeToCadTranslator->m_ptAxleWireDir;
	gp_XYZ _ptProjection = ptPoint_ - _nHeight * m_pTubeToCadTranslator->m_ptAxleWireDir;

	// Բ
	if (DIST_EQU(OvalEdge_.nLongRadius, OvalEdge_.nShortRadius))
		return fabs((_ptProjection - OvalEdge_.ptCenter).Modulus() - OvalEdge_.nLongRadius) < 2.0 * g_nEdgeDeflection;

	// ԲܣԲ㣬Բ壨ΪֵжͶӰǷԲ
	gp_XYZ _ptLeftFocus, _ptRightFocus;
	GetOvalFocus(OvalEdge_, _ptLeftFocus, _ptRightFocus);
	Standard_Real _nError = (_ptProjection - _ptLeftFocus).Modulus() + (_ptProjection - _ptRightFocus).Modulus() - 2.0 * OvalEdge_.nLongRadius;

	return (fabs(_nError) < 2.0 * g_nEdgeDeflection);
}

void CWires3DTranslator::CheckRepeatWire(Wire3DInfo& wire_, Standard_Boolean bExclude_) const
{
	wire_.bExclude = bExclude_;

	// ǵֱ߱ߣΪظ
	if ((wire_.listEdges.size() == 1) && (EdgeInfo::EDGE_LINE == wire_.listEdges.front().nEdgeType))
		wire_.bExclude = true;
}

void CWires3DTranslator::DeleteRepeatEdges(LISTWIRES3DONPLANE& listWires3DOnPlane_) const
{
	LISTWIRES3DONPLANE::iterator _iterPlaneBegin = listWires3DOnPlane_.begin();
	LISTWIRES3DONPLANE::iterator _iterPlaneEnd = listWires3DOnPlane_.end();
	for (; _iterPlaneBegin != _iterPlaneEnd; ++_iterPlaneBegin)
	{
		LISTWIRES3D::iterator _iterWireBegin = _iterPlaneBegin->listWires.begin();
		LISTWIRES3D::iterator _iterWireEnd = _iterPlaneBegin->listWires.end();
		while (_iterWireBegin != _iterWireEnd)
		{
			if (!_iterWireBegin->bExclude || !JudgeRepeatEdges(*_iterWireBegin, _iterPlaneBegin->listWires))
			{
				++_iterWireBegin;
				continue;
			}

			LISTWIRES3D::iterator _iterTemp = _iterWireBegin++;
			_iterPlaneBegin->listWires.erase(_iterTemp);
		}
	}
}

Standard_Boolean CWires3DTranslator::JudgeRepeatEdges(Wire3DInfo& wire_, LISTWIRES3D& listWires_) const
{
	ASSERT(wire_.bExclude);

	LISTWIRES3D::iterator _iterWireBegin = listWires_.begin();
	LISTWIRES3D::iterator _iterWireEnd = listWires_.end();
	for (; _iterWireBegin != _iterWireEnd; ++_iterWireBegin)
	{
		if (_iterWireBegin->bExclude)
			continue;

		LISTEDGES3DCITER _iterEdgeBegin = wire_.listEdges.begin();
		LISTEDGES3DCITER _iterEdgeEnd = wire_.listEdges.end();
		for (; _iterEdgeBegin != _iterEdgeEnd; ++_iterEdgeBegin)
		{
			if (!JudgeRepeatEdges(*_iterEdgeBegin, _iterWireBegin->listEdges))
				break;
		}

		if (_iterEdgeBegin == _iterEdgeEnd)
		{
			wire_.bExclude = false;
			return true;
		}
	}

	wire_.bExclude = false;
	return false;
}

Standard_Boolean CWires3DTranslator::JudgeRepeatEdges(const Edge3DInfo& edge_, const LISTEDGES3D& listEdges_) const
{
	// ֱߣ㡢е㼰յǷҪ
	if (EdgeInfo::EDGE_LINE == edge_.nEdgeType)
	{
		gp_XYZ _ptMiddle = 0.5 * (edge_.ptStart + edge_.ptEnd);
		if (!JudgePointOnEdges(edge_.ptStart, listEdges_)
			|| !JudgePointOnEdges(_ptMiddle, listEdges_)
			|| !JudgePointOnEdges(edge_.ptEnd, listEdges_))
			return false;
	}
	else
	{
		ASSERT(EdgeInfo::EDGE_OVAL == edge_.nEdgeType);
		Standard_Real _nAimCentralAngle = fabs(edge_.nLastParam - edge_.nFirstParam);
		Standard_Real _nAimMiddleAngle = 0.5 * (edge_.nFirstParam + edge_.nLastParam);
		gp_XYZ _ptAimMiddlePoint = edge_.ptCenter + edge_.nLongRadius * cos(_nAimMiddleAngle) * edge_.ptXDir
			+ edge_.nShortRadius * sin(_nAimMiddleAngle) * edge_.ptYDir;
		gp_XYZ _ptAimStartPoint = edge_.StartPoint();
		gp_XYZ _ptAimEndPoint = edge_.EndPoint();

		LISTEDGES3DCITER _iterBegin = listEdges_.begin();
		LISTEDGES3DCITER _iterEnd = listEdges_.end();
		for (; _iterBegin != _iterEnd; ++_iterBegin)
		{
			if ((_iterBegin->nEdgeType != EdgeInfo::EDGE_OVAL)
				|| !edge_.ptCenter.IsEqual(_iterBegin->ptCenter, g_nDistPrecision)
				|| !DIST_EQU(edge_.nLongRadius, _iterBegin->nLongRadius)
				|| !DIST_EQU(edge_.nShortRadius, _iterBegin->nShortRadius)
				|| !IsEqualVectorNoDir(edge_.ptXDir ^ edge_.ptYDir, _iterBegin->ptXDir ^ _iterBegin->ptYDir))
				continue;

			if (!DIST_EQU(edge_.nLongRadius, edge_.nShortRadius) && !IsEqualVectorNoDir(edge_.ptXDir, _iterBegin->ptXDir))
				continue;

			Standard_Real _nCentralAngle = fabs(_iterBegin->nLastParam - _iterBegin->nFirstParam);
			if (!ANGLE_EQU(_nAimCentralAngle, _nCentralAngle) && (_nAimCentralAngle > _nCentralAngle))
				continue;

			if (!JudgePointOnArc(_ptAimStartPoint, *_iterBegin)
				|| !JudgePointOnArc(_ptAimMiddlePoint, *_iterBegin)
				|| !JudgePointOnArc(_ptAimEndPoint, *_iterBegin))
				continue;

			break;
		}

		if (_iterBegin == _iterEnd)
			return false;
	}

	return true;
}

Standard_Boolean CWires3DTranslator::JudgePointOnEdges(const gp_XYZ& ptPoint_, const LISTEDGES3D& listEdges_) const
{
	LISTEDGES3DCITER _iterBegin = listEdges_.begin();
	LISTEDGES3DCITER _iterEnd = listEdges_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		if (EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType)
		{
			if (ptPoint_.IsEqual(_iterBegin->ptStart, g_nDistPrecision) || ptPoint_.IsEqual(_iterBegin->ptEnd, g_nDistPrecision))
				return true;

			// 㵽߶εС
			Standard_Real _nMinDistance = 0.;
			gp_XYZ _ptStartDiff = ptPoint_ - _iterBegin->ptStart;
			gp_XYZ _ptEndDiff = ptPoint_ - _iterBegin->ptEnd;
			if (_ptStartDiff * _iterBegin->ptDir < -1.0 * g_nCosinPrecision)
				_nMinDistance = _ptStartDiff.Modulus();
			else if (_ptEndDiff * _iterBegin->ptDir.Reversed() < -1.0 * g_nCosinPrecision)
				_nMinDistance = _ptEndDiff.Modulus();
			else
				_nMinDistance = (_ptStartDiff ^ _iterBegin->ptDir).Modulus();

			if (_nMinDistance < 2.0 * g_nEdgeDeflection)
				return true;
		}
		else
		{
			ASSERT(EdgeInfo::EDGE_OVAL == _iterBegin->nEdgeType);
			if (JudgePointOnArc(ptPoint_, *_iterBegin))
				return true;
		}
	}

	return false;
}

Standard_Boolean CWires3DTranslator::JudgePointOnArc(const gp_XYZ& ptPoint_, const Edge3DInfo& edge_) const
{
	ASSERT(EdgeInfo::EDGE_OVAL == edge_.nEdgeType);

	// жϵǷԲ/Բ
	if (!COSIN_EQU((ptPoint_ - edge_.ptCenter) * (edge_.ptXDir ^ edge_.ptYDir), 0.))
		return false;

	if (DIST_EQU(edge_.nLongRadius, edge_.nShortRadius))
	{
		Standard_Real _nDistError = (ptPoint_ - edge_.ptCenter).Modulus() - edge_.nLongRadius;
		if (fabs(_nDistError) > 2.0 * g_nDistPrecision)
			return false;
	}
	else
	{
		gp_XYZ _ptLeftFocus, _ptRightFocus;
		GetOvalFocus(edge_, _ptLeftFocus, _ptRightFocus);
		Standard_Real _nDistError = (ptPoint_ - _ptLeftFocus).Modulus() + (ptPoint_ - _ptRightFocus).Modulus() - 2.0 * edge_.nLongRadius;
		if (fabs(_nDistError) > 2.0 * g_nDistPrecision)
			return false;
	}

	// жϵǷԲ/Բ
	if (ANGLE_EQU(fabs(edge_.nLastParam - edge_.nFirstParam), 2.0 * g_nPIE))
		return true;

	Standard_Real _nAngle = GetElipsAngle(edge_.ptXDir, edge_.ptYDir, edge_.ptCenter, ptPoint_);
	Standard_Real _nMinAngle = min(edge_.nFirstParam, edge_.nLastParam);
	while (!ANGLE_EQU(_nAngle, _nMinAngle) && (_nAngle < _nMinAngle))
		_nAngle += 2.0 * g_nPIE;
	while (!ANGLE_EQU(_nAngle, _nMinAngle) && (ANGLE_EQU(_nAngle - 2.0 * g_nPIE, _nMinAngle) || (_nAngle - 2.0 * g_nPIE > _nMinAngle)))
		_nAngle -= 2.0 * g_nPIE;

	if (ANGLE_EQU(_nAngle, edge_.nFirstParam)
		|| ANGLE_EQU(_nAngle, edge_.nLastParam)
		|| ((_nAngle - edge_.nFirstParam) * (_nAngle - edge_.nLastParam) < 0.))
		return true;

	return false;
}

//////////////////////////////////////////////////////////////////////////
// class CWires2DTranslator
CWires2DTranslator::CWires2DTranslator()
	: m_nCurXBorder(0.)
	, m_nCurXOffset(0.)
	, m_pTubeToCadTranslator(NULL)
{
}

void CWires2DTranslator::Initialize(CTubeToCadTranslator* pTubeToCadTranslator_)
{
	ASSERT(pTubeToCadTranslator_ != NULL);
	m_pTubeToCadTranslator = pTubeToCadTranslator_;
	m_nCurXBorder = 0.;
	m_nCurXOffset = 0.;
}

Standard_Boolean CWires2DTranslator::Translate(const LISTWIRES3DONPLANE& listWires3DOnPlane_, LISTWIRES2DONSURFACE& listWires2DOnSurface_)
{
	ASSERT(listWires3DOnPlane_.size() == m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());
	if (listWires3DOnPlane_.size() != m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size())
		return false;

	LISTDOUBLE _listXBorder;
	LISTDOUBLE _listXOffset;
	if (!GetExpandXBorder(_listXBorder) || !GetExpandXOffset(_listXOffset))
		return false;

	LISTEDGES3DCITER _iterBaseEdge = m_pTubeToCadTranslator->m_OuterBoundary.listEdges.begin();
	LISTDOUBLE::const_iterator _iterXBorder = _listXBorder.begin();
	LISTDOUBLE::const_iterator _iterXOffset = _listXOffset.begin();

	LISTWIRES3DONPLANECITER _iterPlaneBegin = listWires3DOnPlane_.begin();
	LISTWIRES3DONPLANECITER _iterPlaneEnd = listWires3DOnPlane_.end();
	for (; _iterPlaneBegin != _iterPlaneEnd; ++_iterPlaneBegin, ++_iterBaseEdge, ++_iterXBorder, ++_iterXOffset)
	{
		// ǰXչı߽缰ƫ
		m_nCurXBorder = *_iterXBorder;
		m_nCurXOffset = *_iterXOffset;

		Wire2DOnSurfaceInfo _wires2DOnSurface;
		if (EdgeInfo::EDGE_LINE == _iterBaseEdge->nEdgeType)
		{
			Trans3DWiresTo2DWiresOnPlane(*_iterBaseEdge, _iterPlaneBegin->listWires, _wires2DOnSurface.listWires);
		}
		else
		{
			ASSERT(EdgeInfo::EDGE_OVAL == _iterBaseEdge->nEdgeType);
			Trans3DWiresTo2DWiresOnOvalSurface(*_iterBaseEdge, _iterPlaneBegin->listWires, _wires2DOnSurface.listWires);
		}

		listWires2DOnSurface_.push_back(_wires2DOnSurface);
	}

	ASSERT(listWires2DOnSurface_.size() == m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());

	return true;
}

Standard_Boolean CWires2DTranslator::GetExpandXBorder(LISTDOUBLE& listXBorder_) const
{
	ASSERT(listXBorder_.empty());

	TubeInfo& _TubeParam = m_pTubeToCadTranslator->m_TubeParam;
	if ((TubeInfo::circular == _TubeParam.nTubeType) || (TubeInfo::oval == _TubeParam.nTubeType))
	{
		listXBorder_.push_back(0.);
		return true;
	}
	else if (TubeInfo::rectangular == _TubeParam.nTubeType)
	{
		if (m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 4)
		{
			// ֱǾ
			for (Standard_Integer i = 0; i < 2; ++i)
			{
				listXBorder_.push_back(0.5 * _TubeParam.nRectWidth);
				listXBorder_.push_back(0.5 * _TubeParam.nRectHeight);
			}
		}
		else
		{
			// ԲǾ
			ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 8);
			for (Standard_Integer i = 0; i < 2; ++i)
			{
				listXBorder_.push_back(0.5 * _TubeParam.nRectWidth - _TubeParam.nRectRadius);
				listXBorder_.push_back(0.);
				listXBorder_.push_back(0.5 * _TubeParam.nRectHeight - _TubeParam.nRectRadius);
				listXBorder_.push_back(0.);
			}
		}

		return (listXBorder_.size() == m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());
	}
	else if (TubeInfo::waist == _TubeParam.nTubeType)
	{
		// 
		ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 4);
		for (Standard_Integer i = 0; i < 2; ++i)
		{
			listXBorder_.push_back(0.5 * _TubeParam.nWaistWidth);
			listXBorder_.push_back(0.);
		}

		return (listXBorder_.size() == m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());
	}

	ASSERT(false);
	return false;
}

Standard_Boolean CWires2DTranslator::GetExpandXOffset(LISTDOUBLE& listXOffset_) const
{
	ASSERT(listXOffset_.empty());

	TubeInfo& _TubeParam = m_pTubeToCadTranslator->m_TubeParam;
	if ((TubeInfo::circular == _TubeParam.nTubeType) || (TubeInfo::oval == _TubeParam.nTubeType))
	{
		listXOffset_.push_back(0.);
		return true;
	}
	else if (TubeInfo::rectangular == _TubeParam.nTubeType)
	{
		if (m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 4)
		{
			// ֱǾ
			Standard_Real _nStep = 0.5 * (_TubeParam.nRectWidth + _TubeParam.nRectHeight);
			for (Standard_Integer i = 0; i < 4; ++i)
				listXOffset_.push_back(_nStep * i);
		}
		else
		{
			// ԲǾ
			ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 8);
			Standard_Real _nStep = _TubeParam.nRectWidth + _TubeParam.nRectHeight;
			for (Standard_Integer i = 0; i < 2; ++i)
			{
				listXOffset_.push_back(_nStep * i);
				listXOffset_.push_back(_nStep * i + 0.5 * _TubeParam.nRectWidth - _TubeParam.nRectRadius);
				listXOffset_.push_back(_nStep * i + 0.5 * _TubeParam.nRectWidth + 0.5 * _TubeParam.nRectHeight);
				listXOffset_.push_back(_nStep * i + 0.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight - _TubeParam.nRectRadius);
			}
		}

		return (listXOffset_.size() == m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());
	}
	else if (TubeInfo::waist == _TubeParam.nTubeType)
	{
		// 
		ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 4);
		Standard_Real _nStep = _TubeParam.nWaistWidth + 0.5 * g_nPIE * _TubeParam.nWaistHeight;
		for (Standard_Integer i = 0; i < 2; ++i)
		{
			listXOffset_.push_back(_nStep * i);
			listXOffset_.push_back(_nStep * i + 0.5 * _TubeParam.nWaistWidth);
		}

		return (listXOffset_.size() == m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size());
	}

	ASSERT(false);
	return false;
}

void CWires2DTranslator::Trans3DWiresTo2DWiresOnPlane(const Edge3DInfo& baseEdge_, const LISTWIRES3D& listWires3D_, LISTWIRES2D& listWires2D_)
{
	ASSERT(EdgeInfo::EDGE_LINE == baseEdge_.nEdgeType);

	LISTWIRES3DCITER _iterWireBegin = listWires3D_.begin();
	LISTWIRES3DCITER _iterWireEnd = listWires3D_.end();
	for (; _iterWireBegin != _iterWireEnd; ++_iterWireBegin)
	{
		Wire2DInfo _wire2D;
		LISTEDGES3DCITER _iterEdge3DBegin = _iterWireBegin->listEdges.begin();
		LISTEDGES3DCITER _iterEdge3DEnd = _iterWireBegin->listEdges.end();
		for (; _iterEdge3DBegin != _iterEdge3DEnd; ++_iterEdge3DBegin)
		{
			Trans3DEdgeTo2DEdgeOnPlane(baseEdge_, *_iterEdge3DBegin, _wire2D);
		}

		// ӵǷڱ߽
		LISTEDGES2D::iterator _iterEdge2DBegin = _wire2D.listEdges.begin();
		LISTEDGES2D::iterator _iterEdge2DEnd = _wire2D.listEdges.end();
		for (; _iterEdge2DBegin != _iterEdge2DEnd; ++_iterEdge2DBegin)
		{
			if (DIST_EQU(fabs(_iterEdge2DBegin->StartPoint().X() - m_nCurXOffset), m_nCurXBorder))
			{
				_wire2D.bOnBorder = true;
				break;
			}
		}

		if (!_wire2D.bOnBorder && DIST_EQU(fabs(_wire2D.listEdges.back().EndPoint().X() - m_nCurXOffset), m_nCurXBorder))
			_wire2D.bOnBorder = true;

		listWires2D_.push_back(_wire2D);
	}
}

void CWires2DTranslator::Trans3DEdgeTo2DEdgeOnPlane(const Edge3DInfo& baseEdge_, const Edge3DInfo& edge3D_, Wire2DInfo& wire2D_)
{
	ASSERT(EdgeInfo::EDGE_LINE == baseEdge_.nEdgeType);
	gp_XYZ _ptStandardPoint = (baseEdge_.StartPoint() + baseEdge_.EndPoint()) * 0.5;

	Edge2DInfo _edge2D;
	_edge2D.nEdgeType = edge3D_.nEdgeType;
	if (EdgeInfo::EDGE_LINE == edge3D_.nEdgeType)
	{
		_edge2D.ptStart = GetProjectionOnPlane(baseEdge_, _ptStandardPoint, edge3D_.ptStart);
		_edge2D.ptEnd = GetProjectionOnPlane(baseEdge_, _ptStandardPoint, edge3D_.ptEnd);
		wire2D_.listEdges.push_back(_edge2D);
		return;
	}

	ASSERT(EdgeInfo::EDGE_OVAL == edge3D_.nEdgeType);
	_edge2D.ptCenter = GetProjectionOnPlane(baseEdge_, _ptStandardPoint, edge3D_.ptCenter);
	_edge2D.nLongRadius = edge3D_.nLongRadius;
	_edge2D.nShortRadius = edge3D_.nShortRadius;

	gp_XYZ _ptNormalVector = baseEdge_.ptDir ^ m_pTubeToCadTranslator->m_ptAxleWireDir;
	Standard_Real _nRotateAngle = gp_Dir(baseEdge_.ptDir).Angle(gp_Dir(edge3D_.ptXDir));
	gp_XYZ _ptTestNormalVector = baseEdge_.ptDir ^ edge3D_.ptXDir;
	if (!DIST_EQU(_ptTestNormalVector.Modulus(), 0.) && !_ptNormalVector.IsEqual(_ptTestNormalVector.Normalized(), g_nDistPrecision))
		_nRotateAngle = 2.0 * g_nPIE - _nRotateAngle;
	_edge2D.nDipAngle = _nRotateAngle;

	Standard_Real _nRatio = _ptNormalVector.IsEqual(edge3D_.ptXDir ^ edge3D_.ptYDir, g_nDistPrecision) ? 1.0 : -1.0;
	_edge2D.nFirstParam = edge3D_.nFirstParam * _nRatio;
	_edge2D.nLastParam = edge3D_.nLastParam * _nRatio;

	wire2D_.listEdges.push_back(_edge2D);
}

gp_XY CWires2DTranslator::GetProjectionOnPlane(const Edge3DInfo& baseEdge_, const gp_XYZ& ptStandardPoint_, const gp_XYZ& ptAimPoint_)
{
	ASSERT(EdgeInfo::EDGE_LINE == baseEdge_.nEdgeType);

	Standard_Real _nRelativeY = (ptAimPoint_ - ptStandardPoint_) * m_pTubeToCadTranslator->m_ptAxleWireDir;
	gp_XYZ _ptProjectionDiff = (ptAimPoint_ - _nRelativeY * m_pTubeToCadTranslator->m_ptAxleWireDir) - ptStandardPoint_;
	ASSERT(DIST_EQU(_ptProjectionDiff.Modulus(), 0.) || IsEqualVectorNoDir(_ptProjectionDiff.Normalized(), baseEdge_.ptDir));
	Standard_Real _nRelativeX = _ptProjectionDiff * baseEdge_.ptDir;

	return gp_XY(_nRelativeX + m_nCurXOffset, _nRelativeY);
}

void CWires2DTranslator::Trans3DWiresTo2DWiresOnOvalSurface(const Edge3DInfo& baseEdge_, const LISTWIRES3D& listWires3D_, LISTWIRES2D& listWires2D_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == baseEdge_.nEdgeType);

	LISTWIRES3DCITER _iterWireBegin = listWires3D_.begin();
	LISTWIRES3DCITER _iterWireEnd = listWires3D_.end();
	for (; _iterWireBegin != _iterWireEnd; ++_iterWireBegin)
	{
		Wire2DInfo _wire2D;
		LISTEDGES3DCITER _iterEdgeBegin = _iterWireBegin->listEdges.begin();
		LISTEDGES3DCITER _iterEdgeEnd = _iterWireBegin->listEdges.end();
		for (; _iterEdgeBegin != _iterEdgeEnd; ++_iterEdgeBegin)
		{
			if ((EdgeInfo::EDGE_LINE == _iterEdgeBegin->nEdgeType)
				|| IsEqualVectorNoDir(_iterEdgeBegin->ptXDir ^ _iterEdgeBegin->ptYDir, m_pTubeToCadTranslator->m_ptAxleWireDir))
			{
				Trans3DEdgeTo2DEdgeOnOvalSurface(baseEdge_, *_iterEdgeBegin, _wire2D);
				continue;
			}

			// Բ/Բվɢ
			LISTEDGES3D _listEdges3D;
			SplitArcEdgeToLineEdges(*_iterEdgeBegin, _listEdges3D);

			LISTEDGES3DCITER _iterTempBegin = _listEdges3D.begin();
			LISTEDGES3DCITER _iterTempEnd = _listEdges3D.end();
			for (; _iterTempBegin != _iterTempEnd; ++_iterTempBegin)
			{
				ASSERT(EdgeInfo::EDGE_LINE == _iterTempBegin->nEdgeType);
				Trans3DEdgeTo2DEdgeOnOvalSurface(baseEdge_, *_iterTempBegin, _wire2D);
			}
		}

		_wire2D.bOnBorder = true;
		ScatterWireByJoin(_wire2D, listWires2D_);
	}
}

void CWires2DTranslator::Trans3DEdgeTo2DEdgeOnOvalSurface(const Edge3DInfo& baseEdge_, const Edge3DInfo& edge3D_, Wire2DInfo& wire2D_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == baseEdge_.nEdgeType);
	ASSERT((EdgeInfo::EDGE_LINE == edge3D_.nEdgeType)
		|| IsEqualVectorNoDir(edge3D_.ptXDir ^ edge3D_.ptYDir, m_pTubeToCadTranslator->m_ptAxleWireDir));

	gp_XYZ _ptStandardPoint = baseEdge_.ptCenter;

	Edge2DInfo _edge2D;
	_edge2D.nEdgeType = EdgeInfo::EDGE_LINE;
	_edge2D.ptStart = GetProjectionOnOvalSurface(baseEdge_, _ptStandardPoint, edge3D_.StartPoint());
	_edge2D.ptEnd = GetProjectionOnOvalSurface(baseEdge_, _ptStandardPoint, edge3D_.EndPoint());

	const TubeInfo& _TubeParam = m_pTubeToCadTranslator->m_TubeParam;
	if ((_TubeParam.nTubeType != TubeInfo::circular) && (_TubeParam.nTubeType != TubeInfo::oval))
	{
		wire2D_.listEdges.push_back(_edge2D);
		return;
	}

	// Բ/Բ
	Standard_Real _nPerimeter = 0.;
	if (TubeInfo::circular == _TubeParam.nTubeType)
	{
		ASSERT(DIST_EQU(_TubeParam.nCircleRadius, baseEdge_.nLongRadius) && DIST_EQU(baseEdge_.nLongRadius, baseEdge_.nShortRadius));
		_nPerimeter = baseEdge_.nLongRadius * 2.0 * g_nPIE;
	}
	else
	{
		ASSERT(TubeInfo::oval == _TubeParam.nTubeType);
		ASSERT(DIST_EQU(_TubeParam.nOvalLongRadius, baseEdge_.nLongRadius) && DIST_EQU(_TubeParam.nOvalShortRadius, baseEdge_.nShortRadius));
		_nPerimeter = _TubeParam.nOvalPerimeter;
	}

	if (IsEqualVectorNoDir(edge3D_.ptXDir ^ edge3D_.ptYDir, m_pTubeToCadTranslator->m_ptAxleWireDir)
		&& ANGLE_EQU(fabs(edge3D_.nLastParam - edge3D_.nFirstParam), 2.0 * g_nPIE))
	{
		ASSERT(edge3D_.StartPoint().IsEqual(edge3D_.EndPoint(), g_nDistPrecision));
		_edge2D.ptStart.SetX(0.);
		_edge2D.ptEnd.SetX(_nPerimeter);
		wire2D_.listEdges.push_back(_edge2D);
		return;
	}

	// ǷԽ߽
	if (EdgeInfo::EDGE_LINE == edge3D_.nEdgeType)
		JudgeLineAcrossBorder(baseEdge_, edge3D_, _edge2D, wire2D_);
	else
	{
		ASSERT(EdgeInfo::EDGE_OVAL == edge3D_.nEdgeType);
		JudgeArcAcrossBorder(baseEdge_, edge3D_, _edge2D, wire2D_);
	}
}

gp_XY CWires2DTranslator::GetProjectionOnOvalSurface(const Edge3DInfo& baseEdge_, const gp_XYZ& ptStandardPoint_, const gp_XYZ& ptAimPoint_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == baseEdge_.nEdgeType);

	Standard_Real _nRelativeY = (ptAimPoint_ - ptStandardPoint_) * m_pTubeToCadTranslator->m_ptAxleWireDir;
	gp_XYZ _ptProjectionDiff = (ptAimPoint_ - _nRelativeY * m_pTubeToCadTranslator->m_ptAxleWireDir) - ptStandardPoint_;
	ASSERT(COSIN_EQU(_ptProjectionDiff * m_pTubeToCadTranslator->m_ptAxleWireDir, 0.));

	gp_XYZ _ptStartDiff = baseEdge_.StartPoint() - ptStandardPoint_;
	Standard_Real _nAngle = gp_Dir(_ptProjectionDiff).Angle(gp_Dir(_ptStartDiff));
	Standard_Real _nRelativeX = 0.;
	switch (m_pTubeToCadTranslator->m_TubeParam.nTubeType)
	{
	case TubeInfo::rectangular:
		{
			ASSERT((fabs(_nAngle) > -g_nEdgeDeflection) && (_nAngle < g_nPIE * 0.5 + g_nEdgeDeflection));
			__CUT(_nAngle, 0., g_nPIE * 0.5);
			_nRelativeX = 2.0 * baseEdge_.nLongRadius * _nAngle / (g_nPIE * 0.5);
		}
		break;

	case TubeInfo::circular:
	case TubeInfo::oval:
		{
			if (DIGIT_EQU(_nAngle, 0.))
				_nAngle = 0.;
			else if (DIGIT_EQU(_nAngle, g_nPIE))
				_nAngle = g_nPIE;
			else
			{
				// ػת45ֵ
				Standard_Real _nRotateAngle = 0.25 * g_nPIE * _copysign(1.0, baseEdge_.nLastParam - baseEdge_.nFirstParam);
				gp_XYZ _ptMiniDeltaStart = baseEdge_.ptCenter + baseEdge_.nLongRadius * cos(baseEdge_.nFirstParam + _nRotateAngle) * baseEdge_.ptXDir
					+ baseEdge_.nShortRadius * sin(baseEdge_.nFirstParam + _nRotateAngle) * baseEdge_.ptYDir;
				gp_XYZ _ptNormalVector = (_ptStartDiff ^ (_ptMiniDeltaStart - ptStandardPoint_)).Normalized();
				gp_XYZ _ptAimNormalVector = _ptStartDiff ^ _ptProjectionDiff;
				if (!DIST_EQU(_ptAimNormalVector.Modulus(), 0.) && !_ptNormalVector.IsEqual(_ptAimNormalVector.Normalized(), g_nDistPrecision))
				{
					ASSERT(_ptNormalVector.IsEqual(_ptAimNormalVector.Normalized().Reversed(), g_nDistPrecision));
					_nAngle = 2.0 * g_nPIE - _nAngle;
				}
			}

			_nRelativeX = GetOvalArcLength(baseEdge_, _nAngle);
		}
		break;

	case TubeInfo::waist:
		{
			ASSERT((fabs(_nAngle) > -g_nEdgeDeflection) && (_nAngle < g_nPIE + g_nEdgeDeflection));
			__CUT(_nAngle, 0., g_nPIE);
			_nRelativeX = GetOvalArcLength(baseEdge_, _nAngle);
		}
		break;

	default:
		ASSERT(false);
		return gp_XY(0., 0.);
	}

	return gp_XY(_nRelativeX + m_nCurXOffset, _nRelativeY);
}

Standard_Real CWires2DTranslator::GetOvalArcLength(const Edge3DInfo& baseEdge_, Standard_Real nCentralAngle_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == baseEdge_.nEdgeType);

	const TubeInfo& _TubeParam = m_pTubeToCadTranslator->m_TubeParam;
	if (TubeInfo::circular == _TubeParam.nTubeType)
	{
		ASSERT(DIST_EQU(_TubeParam.nCircleRadius, baseEdge_.nLongRadius) && DIST_EQU(baseEdge_.nLongRadius, baseEdge_.nShortRadius));
		return baseEdge_.nLongRadius * nCentralAngle_;
	}

	if (TubeInfo::waist == _TubeParam.nTubeType)
	{
		ASSERT(DIST_EQU(_TubeParam.nWaistHeight * 0.5, baseEdge_.nLongRadius) && DIST_EQU(baseEdge_.nLongRadius, baseEdge_.nShortRadius));
		return baseEdge_.nLongRadius * nCentralAngle_;
	}

	// ֵּԲ
	ASSERT(TubeInfo::oval == _TubeParam.nTubeType);
	ASSERT(DIST_EQU(_TubeParam.nOvalLongRadius, baseEdge_.nLongRadius) && DIST_EQU(_TubeParam.nOvalShortRadius, baseEdge_.nShortRadius));
	Standard_Real _nStartRadian = 0.;
	Standard_Real _nEndRadian = 0.;
	if (baseEdge_.nLastParam > baseEdge_.nFirstParam)
	{
		_nStartRadian = baseEdge_.nFirstParam;
		_nEndRadian = baseEdge_.nFirstParam + nCentralAngle_;
	}
	else
	{
		_nStartRadian = baseEdge_.nFirstParam - nCentralAngle_;
		_nEndRadian = baseEdge_.nFirstParam;
	}

	ASSERT(_nStartRadian <= _nEndRadian);
	if (ANGLE_EQU(_nStartRadian, _nEndRadian))
		return 0.;
	if (ANGLE_EQU(_nEndRadian - _nStartRadian, 2.0 * g_nPIE))
		return _TubeParam.nOvalPerimeter;

	// Բ
	m_nEccentricity = sqrt(pow(_TubeParam.nOvalLongRadius, 2) - pow(_TubeParam.nOvalShortRadius, 2)) / _TubeParam.nOvalLongRadius;

	// ֵּԲ
	int _nLevel = 4;
	double _nLength1 = CotesFormula(_nStartRadian, _nEndRadian, _nLevel);
	double _nLength2 = CotesFormula(_nStartRadian, _nEndRadian, _nLevel * 2);
	while (true)
	{
		// ߻Ϊ߳ȵǧ֮һ
		if (fabs(_nLength1 - _nLength2) < min(1.e-6, _nLength2 * 0.001))
			return min(_nLength2, _TubeParam.nOvalPerimeter);

		_nLevel *= 2;
		_nLength1 = _nLength2;
		_nLength2 = CotesFormula(_nStartRadian, _nEndRadian, _nLevel * 2);
	}
}

Standard_Real CWires2DTranslator::CotesFormula(Standard_Real nLower_, Standard_Real nUpper_, Standard_Integer nLevel_) const
{
	// nLevel_׸Ͽ˹ʽS = L*_{}^{}sqrt(1-k^2*(sin)^2)d
	// ЦΪޣΪޣLΪ뾶kΪ
	double _nStep = (nUpper_ - nLower_) / nLevel_;
	ASSERT(!DIGIT_EQU(_nStep, 0.));

	double _nLength = 0.;
	for (int i = 0; i < nLevel_; ++i)
	{
		double _nBaseRadian = nLower_ + i * _nStep;
		_nLength += 32.0 * GetIntegralTerm(_nBaseRadian + 0.25 * _nStep);
		_nLength += 12.0 * GetIntegralTerm(_nBaseRadian + 0.5 * _nStep);
		_nLength += 32.0 * GetIntegralTerm(_nBaseRadian + 0.75 * _nStep);
		if (i > 0)
			_nLength += 14.0 * GetIntegralTerm(_nBaseRadian);
	}

	_nLength += 7.0 * GetIntegralTerm(nLower_);
	_nLength += 7.0 * GetIntegralTerm(nUpper_);

	_nLength = _nLength * _nStep / 90.0;
	return _nLength * m_pTubeToCadTranslator->m_TubeParam.nOvalLongRadius;
}

Standard_Real CWires2DTranslator::GetIntegralTerm(Standard_Real nRadian_) const
{
	// sqrt(1-k^2*(sin)^2)
	return sqrt(1.0 - pow(m_nEccentricity * sin(nRadian_), 2));
}

void CWires2DTranslator::JudgeLineAcrossBorder(const Edge3DInfo& baseEdge_, const Edge3DInfo& edge3D_, const Edge2DInfo& edge2D_, Wire2DInfo& wire2D_)
{
	ASSERT((EdgeInfo::EDGE_OVAL == baseEdge_.nEdgeType) && (EdgeInfo::EDGE_LINE == edge3D_.nEdgeType));

	Standard_Real _nMinXPos = min(edge2D_.ptStart.X(), edge2D_.ptEnd.X());
	Standard_Real _nMaxXPos = max(edge2D_.ptStart.X(), edge2D_.ptEnd.X());
	if (DIST_EQU(_nMinXPos, _nMaxXPos))
	{
		wire2D_.listEdges.push_back(edge2D_);
		return;
	}

	//βձȡ㲢תȻת
	const static Standard_Real _s_nRatio = 0.01;
	gp_XYZ _ptFrontPoint3D = (1.0 - _s_nRatio) * edge3D_.ptStart + _s_nRatio * edge3D_.ptEnd;
	gp_XYZ _ptBackPoint3D = _s_nRatio * edge3D_.ptStart + (1.0 - _s_nRatio) * edge3D_.ptEnd;
	gp_XY _ptFrontPoint2D = GetProjectionOnOvalSurface(baseEdge_, baseEdge_.ptCenter, _ptFrontPoint3D);
	gp_XY _ptBackPoint2D = GetProjectionOnOvalSurface(baseEdge_, baseEdge_.ptCenter, _ptBackPoint3D);

	if ((min(_ptFrontPoint2D.X(), _ptBackPoint2D.X()) >= _nMinXPos) && (max(_ptFrontPoint2D.X(), _ptBackPoint2D.X()) <= _nMaxXPos))
	{
		wire2D_.listEdges.push_back(edge2D_);
		return;
	}

	// Բ/Բܳ
	Standard_Real _nPerimeter = (TubeInfo::oval == m_pTubeToCadTranslator->m_TubeParam.nTubeType)
		? m_pTubeToCadTranslator->m_TubeParam.nOvalPerimeter
		: 2.0 * g_nPIE * m_pTubeToCadTranslator->m_TubeParam.nCircleRadius;
	ASSERT(fabs(edge2D_.ptStart.X() - edge2D_.ptEnd.X()) > 0.5 * _nPerimeter);
	if (edge2D_.ptStart.X() > edge2D_.ptEnd.X())
	{
		Standard_Real _nRatio = (_nPerimeter - edge2D_.ptStart.X()) / (_nPerimeter + edge2D_.ptEnd.X() - edge2D_.ptStart.X());
		Standard_Real _nNewYPos = (1.0 - _nRatio) * edge2D_.ptStart.Y() + _nRatio * edge2D_.ptEnd.Y();
		Edge2DInfo _edge2D = edge2D_;
		_edge2D.ptEnd.SetX(_nPerimeter);
		_edge2D.ptEnd.SetY(_nNewYPos);
		if (!DIST_EQU((_edge2D.ptEnd - _edge2D.ptStart).Modulus(), 0.))
			wire2D_.listEdges.push_back(_edge2D);

		_edge2D.ptStart = _edge2D.ptEnd;
		if (DIST_EQU((_edge2D.ptStart - edge2D_.ptEnd).Modulus(), 0.))
			return;

		_edge2D.ptStart.SetX(0.);
		_edge2D.ptEnd = edge2D_.ptEnd;
		if (!DIST_EQU((_edge2D.ptEnd - _edge2D.ptStart).Modulus(), 0.))
			wire2D_.listEdges.push_back(_edge2D);
	}
	else
	{
		Standard_Real _nRatio = edge2D_.ptStart.X() / (edge2D_.ptStart.X() - edge2D_.ptEnd.X() + _nPerimeter + edge2D_.ptEnd.X());
		Standard_Real _nNewYPos = (1.0 - _nRatio) * edge2D_.ptStart.Y() + _nRatio * edge2D_.ptEnd.Y();
		Edge2DInfo _edge2D = edge2D_;
		_edge2D.ptEnd.SetX(0.);
		_edge2D.ptEnd.SetY(_nNewYPos);
		if (!DIST_EQU((_edge2D.ptEnd - _edge2D.ptStart).Modulus(), 0.))
			wire2D_.listEdges.push_back(_edge2D);

		_edge2D.ptStart = _edge2D.ptEnd;
		if (DIST_EQU((_edge2D.ptStart - edge2D_.ptEnd).Modulus(), 0.))
			return;

		_edge2D.ptStart.SetX(_nPerimeter);
		_edge2D.ptEnd = edge2D_.ptEnd;
		if (!DIST_EQU((_edge2D.ptEnd - _edge2D.ptStart).Modulus(), 0.))
			wire2D_.listEdges.push_back(_edge2D);
	}
}

void CWires2DTranslator::JudgeArcAcrossBorder(const Edge3DInfo& baseEdge_, const Edge3DInfo& edge3D_, const Edge2DInfo& edge2D_, Wire2DInfo& wire2D_)
{
	ASSERT((EdgeInfo::EDGE_OVAL == baseEdge_.nEdgeType) && (EdgeInfo::EDGE_OVAL == edge3D_.nEdgeType));

	// Բ/Բܳͻ
	Standard_Real _nPerimeter = 0.;
	Standard_Real _nOvalArcLength = 0.;
	if (TubeInfo::circular == m_pTubeToCadTranslator->m_TubeParam.nTubeType)
	{
		_nPerimeter = 2.0 * g_nPIE * baseEdge_.nLongRadius;
		_nOvalArcLength = fabs(edge3D_.nLastParam - edge3D_.nFirstParam) * baseEdge_.nLongRadius;
	}
	else
	{
		ASSERT(TubeInfo::oval == m_pTubeToCadTranslator->m_TubeParam.nTubeType);
		_nPerimeter = m_pTubeToCadTranslator->m_TubeParam.nOvalPerimeter;
		_nOvalArcLength = GetOvalArcLength(baseEdge_, fabs(edge3D_.nLastParam - edge3D_.nFirstParam));
	}

	Standard_Boolean _bSameDir = (baseEdge_.ptXDir ^ baseEdge_.ptYDir).IsEqual(edge3D_.ptXDir ^ edge3D_.ptYDir, g_nDistPrecision)
		== ((baseEdge_.nLastParam - baseEdge_.nFirstParam) * (edge3D_.nLastParam - edge3D_.nFirstParam) > 0.);
	if (_bSameDir)
	{
		Standard_Real _nRealEndX = edge2D_.ptStart.X() + _nOvalArcLength;
		if (DIST_EQU(_nRealEndX, _nPerimeter) || (_nRealEndX > _nPerimeter))
		{
			Edge2DInfo _edge2D = edge2D_;
			_edge2D.ptEnd.SetX(_nPerimeter);
			if (!DIST_EQU(_edge2D.ptStart.X(), _edge2D.ptEnd.X()))
				wire2D_.listEdges.push_back(_edge2D);

			_edge2D.ptStart = _edge2D.ptEnd;
			if (DIST_EQU(_edge2D.ptStart.X(), edge2D_.ptEnd.X()))
				return;

			_edge2D.ptStart.SetX(0.);
			_edge2D.ptEnd = edge2D_.ptEnd;
			if (!DIST_EQU(_edge2D.ptStart.X(), _edge2D.ptEnd.X()))
				wire2D_.listEdges.push_back(_edge2D);

			return;
		}
	}
	else
	{
		Standard_Real _nRealEndX = edge2D_.ptStart.X() - _nOvalArcLength;
		if (DIST_EQU(_nRealEndX, 0.) || (_nRealEndX < 0.))
		{
			Edge2DInfo _edge2D = edge2D_;
			_edge2D.ptEnd.SetX(0.);
			if (!DIST_EQU(_edge2D.ptStart.X(), _edge2D.ptEnd.X()))
				wire2D_.listEdges.push_back(_edge2D);

			_edge2D.ptStart = _edge2D.ptEnd;
			if (DIST_EQU(_edge2D.ptStart.X(), edge2D_.ptEnd.X()))
				return;

			_edge2D.ptStart.SetX(_nPerimeter);
			_edge2D.ptEnd = edge2D_.ptEnd;
			if (!DIST_EQU(_edge2D.ptStart.X(), _edge2D.ptEnd.X()))
				wire2D_.listEdges.push_back(_edge2D);

			return;
		}
	}

	wire2D_.listEdges.push_back(edge2D_);
}

void CWires2DTranslator::ScatterWireByJoin(const Wire2DInfo& wire2D_, LISTWIRES2D& listWires2D_)
{
	const TubeInfo::TubeType& _nTubeType = m_pTubeToCadTranslator->m_TubeParam.nTubeType;
	if ((_nTubeType != TubeInfo::circular) && (_nTubeType != TubeInfo::oval))
	{
		listWires2D_.push_back(wire2D_);
		return;
	}

	Wire2DInfo _wire2D;
	_wire2D.bOnBorder = wire2D_.bOnBorder;

	LISTEDGES2D::const_iterator _iterEnd = wire2D_.listEdges.end();
	LISTEDGES2D::const_iterator _iterCur = wire2D_.listEdges.begin();
	LISTEDGES2D::const_iterator _iterNext = ++wire2D_.listEdges.begin();
	LISTEDGES2D::const_iterator _iterFlag = _iterCur;
	while (_iterCur != _iterEnd && _iterNext != _iterEnd)
	{
		if (!_iterCur->EndPoint().IsEqual(_iterNext->StartPoint(), g_nDistPrecision))
		{
			_wire2D.listEdges.assign(_iterFlag, _iterNext);
			listWires2D_.push_back(_wire2D);
			_wire2D.listEdges.clear();
			_iterFlag = _iterNext;
		}

		++_iterCur;
		++_iterNext;
	}

	if (_iterFlag == wire2D_.listEdges.begin())
		listWires2D_.push_back(wire2D_);
	else
	{
		_wire2D.listEdges.assign(_iterFlag, _iterNext);
		listWires2D_.push_back(_wire2D);
	}
}

//////////////////////////////////////////////////////////////////////////
// class CCad2DTranslator
CCad2DTranslator::CCad2DTranslator()
	: m_nPerimeter(0.)
	, m_pTubeToCadTranslator(NULL)
{
}

void CCad2DTranslator::Initialize(CTubeToCadTranslator* pTubeToCadTranslator_)
{
	ASSERT(pTubeToCadTranslator_ != NULL);
	m_pTubeToCadTranslator = pTubeToCadTranslator_;

	m_listEdgesOnArris.clear();
	m_listArrisBoundary.clear();
	m_nPerimeter = 0.;
}

Standard_Boolean CCad2DTranslator::Translate(LISTWIRES2DONSURFACE& listWiresOnSurface_)
{
	// ȥܲ⴦ı߽
	if (!RemoveArrisBoundary(listWiresOnSurface_))
		return false;

	// ȡβ
	LISTWIRES2DONSURFACE _listWiresOnSurface;
	GetConjoint2DWires(listWiresOnSurface_, _listWiresOnSurface);

	// ܲı߽ϵߺϲչ
	LISTWIRES2DONSURFACE::iterator _iterSurfaceBegin = _listWiresOnSurface.begin();
	LISTWIRES2DONSURFACE::iterator _iterSurfaceEnd = _listWiresOnSurface.end();
	for (; _iterSurfaceBegin != _iterSurfaceEnd; ++_iterSurfaceBegin)
	{
		LISTWIRES2D::iterator _iterWireBegin = _iterSurfaceBegin->listWires.begin();
		LISTWIRES2D::iterator _iterWireEnd = _iterSurfaceBegin->listWires.end();
		while (_iterWireBegin != _iterWireEnd)
		{
			if (_iterWireBegin->bOnBorder
				&& CheckIsOnlyLeftBoundary(*_iterWireBegin)
				&& CombineWireOnArris(*_iterWireBegin, m_nPerimeter, _listWiresOnSurface.back().listWires))
			{
				LISTWIRES2D::iterator _iterTemp = _iterWireBegin++;
				_iterSurfaceBegin->listWires.erase(_iterTemp);
			}
			else
				++_iterWireBegin;
		}

		_iterWireBegin = _iterSurfaceBegin->listWires.begin();
		_iterWireEnd = _iterSurfaceBegin->listWires.end();
		while (_iterWireBegin != _iterWireEnd)
		{
			if (_iterWireBegin->bOnBorder
				&& CheckIsOnlyRightBoundary(*_iterWireBegin)
				&& CombineWireOnArris(*_iterWireBegin, -m_nPerimeter, _listWiresOnSurface.front().listWires))
			{
				LISTWIRES2D::iterator _iterTemp = _iterWireBegin++;
				_iterSurfaceBegin->listWires.erase(_iterTemp);
			}
			else
				++_iterWireBegin;
		}
	}

	// άչתΪάռCAD
	_iterSurfaceBegin = _listWiresOnSurface.begin();
	_iterSurfaceEnd = _listWiresOnSurface.end();
	for (; _iterSurfaceBegin != _iterSurfaceEnd; ++_iterSurfaceBegin)
	{
		LISTWIRES2D::iterator _iterWireBegin = _iterSurfaceBegin->listWires.begin();
		LISTWIRES2D::iterator _iterWireEnd = _iterSurfaceBegin->listWires.end();
		while (_iterWireBegin != _iterWireEnd)
		{
			if (!_iterWireBegin->bOnBorder)
			{
				// ƽ
				Trans2DWireTo2DCad(_iterWireBegin->listEdges);
				LISTWIRES2D::iterator _iterTemp = _iterWireBegin++;
				_iterSurfaceBegin->listWires.erase(_iterTemp);
				continue;
			}

			LISTWIRES2DONSURFACE::iterator _iterSurfaceTemp = _iterSurfaceBegin;
			++_iterSurfaceTemp;
			if (_iterSurfaceTemp == _iterSurfaceEnd)
			{
				Trans2DWireTo2DCad(_iterWireBegin->listEdges);
				LISTWIRES2D::iterator _iterTemp = _iterWireBegin++;
				_iterSurfaceBegin->listWires.erase(_iterTemp);
				continue;
			}

			if (!CombineWireOnArris(*_iterWireBegin, 0., _iterSurfaceTemp->listWires))
				Trans2DWireTo2DCad(_iterWireBegin->listEdges);

			LISTWIRES2D::iterator _iterTemp = _iterWireBegin++;
			_iterSurfaceBegin->listWires.erase(_iterTemp);
		}
	}

	return true;
}

Standard_Boolean CCad2DTranslator::RemoveArrisBoundary(LISTWIRES2DONSURFACE& listWiresOnSurface_)
{
	// ȡܲ⴦Xֵ߽
	ASSERT(m_listArrisBoundary.empty());

	const TubeInfo& _TubeParam = m_pTubeToCadTranslator->m_TubeParam;
	switch (_TubeParam.nTubeType)
	{
	case TubeInfo::circular:
	case TubeInfo::oval:
		{
			m_nPerimeter = (TubeInfo::circular == _TubeParam.nTubeType) ? 2.0 * g_nPIE * _TubeParam.nCircleRadius : _TubeParam.nOvalPerimeter;
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0., true));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.25 * m_nPerimeter, false));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * m_nPerimeter, false));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.75 * m_nPerimeter, false));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(m_nPerimeter, true));
		}
		break;

	case TubeInfo::rectangular:
		{
			m_nPerimeter = 2.0 * (_TubeParam.nRectWidth + _TubeParam.nRectHeight);
			if (m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 4)
			{
				// ֱǾ
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(-0.5 * _TubeParam.nRectWidth, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nRectWidth, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nRectWidth + 2.0 * _TubeParam.nRectHeight, true));
			}
			else
			{
				// ԲǾ
				ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 8);
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(-0.5 * _TubeParam.nRectWidth + _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nRectWidth - _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nRectWidth + _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight - _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight + _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight - _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nRectWidth + _TubeParam.nRectHeight + _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nRectWidth + 2.0 * _TubeParam.nRectHeight - _TubeParam.nRectRadius, true));
				m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nRectWidth + 2.0 * _TubeParam.nRectHeight + _TubeParam.nRectRadius, true));
			}
		}
		break;

	case TubeInfo::waist:
		{
			//ι
			m_nPerimeter = 2.0 * _TubeParam.nWaistWidth + g_nPIE * _TubeParam.nWaistHeight;
			ASSERT(m_pTubeToCadTranslator->m_OuterBoundary.listEdges.size() == 4);
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(-0.5 * _TubeParam.nWaistWidth, true));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nWaistWidth, true));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nWaistWidth + 0.25 * g_nPIE * _TubeParam.nWaistHeight, false));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(0.5 * _TubeParam.nWaistWidth + 0.5 * g_nPIE * _TubeParam.nWaistHeight, true));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nWaistWidth + 0.5 * g_nPIE * _TubeParam.nWaistHeight, true));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nWaistWidth + 0.75 * g_nPIE * _TubeParam.nWaistHeight, false));
			m_listArrisBoundary.push_back(ArrisBoundaryInfo(1.5 * _TubeParam.nWaistWidth + g_nPIE * _TubeParam.nWaistHeight, true));
		}
		break;

	default:
		ASSERT(false);
		return false;
	}

	// ȥܲ⴦ı߽
	LISTWIRES2DONSURFACE::iterator _iterSurfaceBegin = listWiresOnSurface_.begin();
	LISTWIRES2DONSURFACE::iterator _iterSurfaceEnd = listWiresOnSurface_.end();
	for (; _iterSurfaceBegin != _iterSurfaceEnd; ++_iterSurfaceBegin)
	{
		RemoveArrisBoundary(_iterSurfaceBegin->listWires);
	}

	return true;
}

void CCad2DTranslator::RemoveArrisBoundary(LISTWIRES2D& listWires2D_)
{
	LISTWIRES2D::iterator _iterWireBegin = listWires2D_.begin();
	LISTWIRES2D::iterator _iterWireEnd = listWires2D_.end();
	while (_iterWireBegin != _iterWireEnd)
	{
		if (!_iterWireBegin->bOnBorder)
		{
			++_iterWireBegin;
			continue;
		}

		LISTEDGES2D::iterator _iterEdgeBegin = _iterWireBegin->listEdges.begin();
		LISTEDGES2D::iterator _iterEdgeEnd = _iterWireBegin->listEdges.end();
		LISTEDGES2D::iterator _iterEdgeFlag = _iterEdgeBegin;
		while (_iterEdgeBegin != _iterEdgeEnd)
		{
			if (_iterEdgeBegin->nEdgeType != EdgeInfo::EDGE_LINE)
			{
				++_iterEdgeBegin;
				continue;
			}

			gp_XY _ptCrossProduct = (_iterEdgeBegin->ptEnd - _iterEdgeBegin->ptStart);
			if (!DIST_EQU(_ptCrossProduct.X(), 0.))
			{
				++_iterEdgeBegin;
				continue;
			}

			if (!JudgeOnAllArrisBoundary(_iterEdgeBegin->ptEnd.X()))
			{
				++_iterEdgeBegin;
				continue;
			}

			// ռϵıߣЩǲɾģҪں油
			m_listEdgesOnArris.push_back(*_iterEdgeBegin);

			// 
			Wire2DInfo _wire2D;
			_wire2D.listEdges = LISTEDGES2D(_iterEdgeFlag, _iterEdgeBegin++);
			_wire2D.bOnBorder = _iterWireBegin->bOnBorder;
			if (!_wire2D.listEdges.empty())
				listWires2D_.push_front(_wire2D);
			_iterEdgeFlag = _iterEdgeBegin;
		}

		if (_iterEdgeFlag == _iterWireBegin->listEdges.begin())
		{
			++_iterWireBegin;
			continue;
		}

		if (_iterEdgeFlag != _iterWireBegin->listEdges.end())
		{
			Wire2DInfo _wire2D;
			_wire2D.listEdges = LISTEDGES2D(_iterEdgeFlag, _iterWireBegin->listEdges.end());
			_wire2D.bOnBorder = _iterWireBegin->bOnBorder;
			listWires2D_.push_front(_wire2D);
		}

		LISTWIRES2D::iterator _iterWireTemp = _iterWireBegin++;
		listWires2D_.erase(_iterWireTemp);
	}
}

Standard_Boolean CCad2DTranslator::JudgeOnRealArrisBoundary(Standard_Real nAimXPos_) const
{
	LISTARRISBOUNDARY::const_iterator _iterBegin = m_listArrisBoundary.begin();
	LISTARRISBOUNDARY::const_iterator _iterEnd = m_listArrisBoundary.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		if (_iterBegin->bRealArris && DIST_EQU(nAimXPos_, _iterBegin->nArrisXPos))
			return true;
	}

	return false;
}

Standard_Boolean CCad2DTranslator::JudgeOnAllArrisBoundary(Standard_Real nAimXPos_) const
{
	// Բ/Բ/ιܻϹ⣬ɾı
	LISTARRISBOUNDARY::const_iterator _iterBegin = m_listArrisBoundary.begin();
	LISTARRISBOUNDARY::const_iterator _iterEnd = m_listArrisBoundary.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		if (DIST_EQU(nAimXPos_, _iterBegin->nArrisXPos))
			return true;
	}

	return false;
}

void CCad2DTranslator::GetConjoint2DWires(LISTWIRES2DONSURFACE& listWiresOnSurfaceIn_, LISTWIRES2DONSURFACE& listWiresOnSurfaceOut_) const
{
	LISTWIRES2DONSURFACE::iterator _iterBegin = listWiresOnSurfaceIn_.begin();
	LISTWIRES2DONSURFACE::iterator _iterEnd = listWiresOnSurfaceIn_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		Wire2DOnSurfaceInfo _wires2DOnSurface;
		GetConjoint2DWires(_iterBegin->listWires, _wires2DOnSurface.listWires);
		listWiresOnSurfaceOut_.push_back(_wires2DOnSurface);
	}
}

void CCad2DTranslator::GetConjoint2DWires(LISTWIRES2D& listWires2DIn_, LISTWIRES2D& listWires2DOut_) const
{
	while (!listWires2DIn_.empty())
	{
		Wire2DInfo _wire2D;
		_wire2D.listEdges = listWires2DIn_.front().listEdges;
		listWires2DIn_.pop_front();

		LISTWIRES2D::iterator _iterWireBegin = listWires2DIn_.begin();
		LISTWIRES2D::iterator _iterWireEnd = listWires2DIn_.end();
		while (_iterWireBegin != _iterWireEnd)
		{
			if (!Combine2DWires(_iterWireBegin->listEdges, _wire2D.listEdges))
			{
				++_iterWireBegin;
				continue;
			}

			LISTWIRES2D::iterator _iterTemp = _iterWireBegin++;
			listWires2DIn_.erase(_iterTemp);
		}

		gp_XY& _ptStartPoint = _wire2D.listEdges.front().StartPoint();
		gp_XY& _ptEndPoint = _wire2D.listEdges.back().EndPoint();
		if (JudgeOnRealArrisBoundary(_ptStartPoint.X()) || JudgeOnRealArrisBoundary(_ptEndPoint.X()))
			_wire2D.bOnBorder = true;

		// ϲ
		Merge2DEdges(_wire2D.listEdges);

		// Բ/Բ޸Ĳֵ֤ͬһͼα﷽ʽһ
		if ((_wire2D.listEdges.size() == 1) && (_wire2D.listEdges.back().nEdgeType == EdgeInfo::EDGE_OVAL))
		{
			Edge2DInfo& _edge2D = _wire2D.listEdges.back();
			if (ANGLE_EQU(fabs(_edge2D.nLastParam - _edge2D.nFirstParam), 2.0 * g_nPIE))
			{
				_edge2D.nFirstParam = 0.;
				_edge2D.nLastParam = 2.0 * g_nPIE;

				if (DIST_EQU(_edge2D.nLongRadius, _edge2D.nShortRadius))
					_edge2D.nDipAngle = 0.;
				else if (_edge2D.nDipAngle > g_nPIE)
					_edge2D.nDipAngle -= g_nPIE;
			}
		}

		// תΪԲ
		CArcSplineHelper _arcSplineHelper;
		Edge2DInfo _transEdge;
		if (_arcSplineHelper.TransPolylineToCircle(_wire2D.listEdges, _transEdge))
		{
			_wire2D.listEdges.clear();
			_wire2D.listEdges.push_back(_transEdge);
		}

		if (!CheckWireExist(_wire2D, listWires2DOut_))
			listWires2DOut_.push_back(_wire2D);
	}
}

Standard_Boolean CCad2DTranslator::Combine2DWires(LISTEDGES2D& listEdgesSrc_, LISTEDGES2D& listEdgesDest_) const
{
	ASSERT(!listEdgesSrc_.empty() && !listEdgesDest_.empty());

	gp_XY _ptSrcStart = listEdgesSrc_.front().StartPoint();
	gp_XY _ptSrcEnd = listEdgesSrc_.back().EndPoint();
	gp_XY _ptDestStart = listEdgesDest_.front().StartPoint();
	gp_XY _ptDestEnd = listEdgesDest_.back().EndPoint();

	// ߵķSrcյ  Dest Խ
	if (_ptSrcEnd.IsEqual(_ptDestEnd, g_nDistPrecision))
		ReverseEdges(listEdgesDest_);
	else if (_ptSrcStart.IsEqual(_ptDestStart, g_nDistPrecision))
		ReverseEdges(listEdgesSrc_);
	else if (_ptSrcStart.IsEqual(_ptDestEnd, g_nDistPrecision))
	{
		ReverseEdges(listEdgesSrc_);
		ReverseEdges(listEdgesDest_);
	}
	else if (!_ptSrcEnd.IsEqual(_ptDestStart, g_nDistPrecision))
		return JudgeInclude2DWires(listEdgesSrc_, listEdgesDest_);

	// ɾظ
	DeleteRepeatEdges(listEdgesSrc_, listEdgesDest_);
	Combine2DEdges(listEdgesSrc_, listEdgesDest_);

	return listEdgesSrc_.empty();
}

void CCad2DTranslator::ReverseEdges(LISTEDGES2D& listEdges_) const
{
	LISTEDGES2D _listEdges;
	LISTEDGES2D::const_reverse_iterator _iterRBegin = listEdges_.rbegin();
	LISTEDGES2D::const_reverse_iterator _iterREnd = listEdges_.rend();
	for (; _iterRBegin != _iterREnd; ++_iterRBegin)
	{
		_listEdges.push_back(*_iterRBegin);
		_listEdges.back().Reverse();
	}

	listEdges_ = _listEdges;
}

void CCad2DTranslator::DeleteRepeatEdges(LISTEDGES2D& listCurEdges_, LISTEDGES2D& listNextEdges_) const
{
	ASSERT(listCurEdges_.back().EndPoint().IsEqual(listNextEdges_.front().StartPoint(), g_nDistPrecision));

	// listCurEdges_ĩβʼǷظ
	LISTEDGES2D::iterator _iterRBegin = --listCurEdges_.end();
	LISTEDGES2D::iterator _iterREnd = listCurEdges_.begin();
	while (true)
	{
		Standard_Boolean _bBreak = false;
		LISTEDGES2D::iterator _iterBegin = listNextEdges_.begin();
		LISTEDGES2D::iterator _iterEnd = listNextEdges_.end();
		for (; _iterBegin != _iterEnd; ++_iterBegin)
		{
			if (!JudgeInclude2DEdge(*_iterBegin, *_iterRBegin))
				continue;

			if (_iterRBegin == _iterREnd)
			{
				listCurEdges_.erase(_iterRBegin);
				_bBreak = true;
			}
			else
			{
				LISTEDGES2D::iterator _iterBak = _iterRBegin--;
				listCurEdges_.erase(_iterBak);
			}

			break;
		}

		if (_bBreak || (_iterBegin == _iterEnd))
			break;
	}

	// listNextEdges_ĿͷʼǷظ
	LISTEDGES2D::iterator _iterBegin = listNextEdges_.begin();
	LISTEDGES2D::iterator _iterEnd = listNextEdges_.end();
	while (_iterBegin != _iterEnd)
	{
		LISTEDGES2D::reverse_iterator _iterRBegin = listCurEdges_.rbegin();
		LISTEDGES2D::reverse_iterator _iterREnd = listCurEdges_.rend();
		for (; _iterRBegin != _iterREnd; ++_iterRBegin)
		{
			if (JudgeInclude2DEdge(*_iterRBegin, *_iterBegin))
			{
				LISTEDGES2D::iterator _iterBak = _iterBegin++;
				listNextEdges_.erase(_iterBak);
				break;
			}
		}

		if (_iterRBegin == _iterREnd)
			break;
	}
}

void CCad2DTranslator::Combine2DEdges(LISTEDGES2D& listEdgesSrc_, LISTEDGES2D& listEdgesDest_) const
{
	ASSERT(!listEdgesSrc_.empty() || !listEdgesDest_.empty());
	if (listEdgesSrc_.empty())
		return;

	if (listEdgesDest_.empty())
	{
		listEdgesDest_ = listEdgesSrc_;
		listEdgesSrc_.clear();
		return;
	}

	gp_XY _ptSrcStart = listEdgesSrc_.front().StartPoint();
	gp_XY _ptSrcEnd = listEdgesSrc_.back().EndPoint();
	gp_XY _ptDestStart = listEdgesDest_.front().StartPoint();
	gp_XY _ptDestEnd = listEdgesDest_.back().EndPoint();
	if (_ptDestEnd.IsEqual(_ptSrcStart, g_nDistPrecision))
	{
		listEdgesDest_.splice(listEdgesDest_.end(), listEdgesSrc_);
		ASSERT(listEdgesSrc_.empty());
	}
	else if (_ptDestEnd.IsEqual(_ptSrcEnd, g_nDistPrecision))
	{
		LISTEDGES2D::const_reverse_iterator _iterRBegin = listEdgesSrc_.rbegin();
		LISTEDGES2D::const_reverse_iterator _iterREnd = listEdgesSrc_.rend();
		for (; _iterRBegin != _iterREnd; ++_iterRBegin)
		{
			listEdgesDest_.push_back(*_iterRBegin);
			listEdgesDest_.back().Reverse();
		}

		listEdgesSrc_.clear();
	}
	else if (_ptDestStart.IsEqual(_ptSrcStart, g_nDistPrecision))
	{
		LISTEDGES2D::const_iterator _iterBegin = listEdgesSrc_.begin();
		LISTEDGES2D::const_iterator _iterEnd = listEdgesSrc_.end();
		for (; _iterBegin != _iterEnd; ++_iterBegin)
		{
			listEdgesDest_.push_front(*_iterBegin);
			listEdgesDest_.front().Reverse();
		}

		listEdgesSrc_.clear();
	}
	else if (_ptDestStart.IsEqual(_ptSrcEnd, g_nDistPrecision))
	{
		listEdgesDest_.splice(listEdgesDest_.begin(), listEdgesSrc_);
		ASSERT(listEdgesSrc_.empty());
	}
}

Standard_Boolean CCad2DTranslator::JudgeInclude2DEdge(const Edge2DInfo& standardEdge_, const Edge2DInfo& checkEdge_) const
{
	if ((standardEdge_.nEdgeType != checkEdge_.nEdgeType) || (EdgeInfo::EDGE_NONE == standardEdge_.nEdgeType))
		return false;

	// ֱ
	if (EdgeInfo::EDGE_LINE == standardEdge_.nEdgeType)
	{
		Standard_Real _nCrossProduct = (standardEdge_.ptEnd - standardEdge_.ptStart) ^ (checkEdge_.ptEnd - checkEdge_.ptStart);
		if (!COSIN_EQU(_nCrossProduct, 0.))
			return false;

		// ߶γ
		Standard_Real _nStdEdgeLength = (standardEdge_.ptEnd - standardEdge_.ptStart).Modulus();
		Standard_Real _nCheckEdgeLength = (checkEdge_.ptEnd - checkEdge_.ptStart).Modulus();
		if (!DIST_EQU(_nStdEdgeLength, _nCheckEdgeLength) && (_nCheckEdgeLength > _nStdEdgeLength))
			return false;

		Standard_Real _nScalarProduct = (standardEdge_.ptEnd - standardEdge_.ptStart) * (checkEdge_.ptEnd - checkEdge_.ptStart);
		if ((_nScalarProduct > 0.)
			&& (standardEdge_.ptStart.IsEqual(checkEdge_.ptStart, g_nDistPrecision)
			|| standardEdge_.ptEnd.IsEqual(checkEdge_.ptEnd, g_nDistPrecision)))
			return true;

		if ((_nScalarProduct < 0.)
			&& (standardEdge_.ptStart.IsEqual(checkEdge_.ptEnd, g_nDistPrecision)
			|| standardEdge_.ptEnd.IsEqual(checkEdge_.ptStart, g_nDistPrecision)))
			return true;

		return false;
	}

	// Բ/Բ
	ASSERT(EdgeInfo::EDGE_OVAL == standardEdge_.nEdgeType);
	if (!standardEdge_.ptCenter.IsEqual(checkEdge_.ptCenter, g_nDistPrecision))
		return false;
	if (!DIST_EQU(standardEdge_.nLongRadius, checkEdge_.nLongRadius))
		return false;
	if (!DIST_EQU(standardEdge_.nShortRadius, checkEdge_.nShortRadius))
		return false;
	if (!DIST_EQU(standardEdge_.nLongRadius, checkEdge_.nShortRadius)
		&& !ANGLE_EQU(standardEdge_.nDipAngle, checkEdge_.nDipAngle)
		&& !ANGLE_EQU(fabs(standardEdge_.nDipAngle - checkEdge_.nDipAngle), g_nPIE))
		return false;

	// ԲĽ
	Standard_Real _nStdCentralAngle = fabs(standardEdge_.nLastParam - standardEdge_.nFirstParam);
	Standard_Real _nCheckCentralAngle = fabs(checkEdge_.nLastParam - checkEdge_.nFirstParam);
	if (!ANGLE_EQU(_nStdCentralAngle, _nCheckCentralAngle) && (_nStdCentralAngle < _nCheckCentralAngle))
		return false;

	// 
	Standard_Boolean _bSameDir = (standardEdge_.nLastParam - standardEdge_.nFirstParam) * (checkEdge_.nLastParam - checkEdge_.nFirstParam) > 0.;

	if (_bSameDir
		&& (standardEdge_.StartPoint().IsEqual(checkEdge_.StartPoint(), g_nDistPrecision)
		|| standardEdge_.EndPoint().IsEqual(checkEdge_.EndPoint(), g_nDistPrecision)))
		return true;

	if (!_bSameDir
		&& (standardEdge_.StartPoint().IsEqual(checkEdge_.EndPoint(), g_nDistPrecision)
		|| standardEdge_.EndPoint().IsEqual(checkEdge_.StartPoint(), g_nDistPrecision)))
		return true;

	return false;
}

Standard_Boolean CCad2DTranslator::JudgeInclude2DWires(LISTEDGES2D& listEdgesSrc_, LISTEDGES2D& listEdgesDest_) const
{
	if (listEdgesSrc_.size() > listEdgesDest_.size())
	{
		if (JudgeInclude2DWires(listEdgesDest_, listEdgesSrc_))
		{
			listEdgesDest_ = listEdgesSrc_;
			return true;
		}

		return false;
	}

	LISTEDGES2D::iterator _iterSrcBegin = listEdgesSrc_.begin();
	LISTEDGES2D::iterator _iterSrcEnd = listEdgesSrc_.end();
	for (; _iterSrcBegin != _iterSrcEnd; ++_iterSrcBegin)
	{
		LISTEDGES2D::iterator _iterDestBegin = listEdgesDest_.begin();
		LISTEDGES2D::iterator _iterDestEnd = listEdgesDest_.end();
		for (; _iterDestBegin != _iterDestEnd; ++_iterDestBegin)
		{
			if (JudgeInclude2DEdge(*_iterDestBegin, *_iterSrcBegin))
				break;
		}

		if (_iterDestBegin == _iterDestEnd)
			return false;
	}

	return true;
}

Standard_Boolean CCad2DTranslator::CheckWireExist(const Wire2DInfo& wire2D_, const LISTWIRES2D& listWires_) const
{
	ASSERT(!wire2D_.listEdges.empty());

	LISTEDGES2D::const_iterator _iterBegin = wire2D_.listEdges.begin();
	LISTEDGES2D::const_iterator _iterEnd = wire2D_.listEdges.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		if (!CheckEdgeExist(*_iterBegin, listWires_))
			return false;
	}

	return true;
}

Standard_Boolean CCad2DTranslator::CheckEdgeExist(const Edge2DInfo& edge2D_, const LISTWIRES2D& listWires_) const
{
	LISTWIRES2D::const_iterator _iterWireBegin = listWires_.begin();
	LISTWIRES2D::const_iterator _iterWireEnd = listWires_.end();
	for (; _iterWireBegin != _iterWireEnd; ++_iterWireBegin)
	{
		LISTEDGES2D::const_iterator _iterEdgeBegin = _iterWireBegin->listEdges.begin();
		LISTEDGES2D::const_iterator _iterEdgeEnd = _iterWireBegin->listEdges.end();
		for (; _iterEdgeBegin != _iterEdgeEnd; ++_iterEdgeBegin)
		{
			if (JudgeSame2DEdge(edge2D_, *_iterEdgeBegin))
				return true;
		}
	}

	return false;
}

Standard_Boolean CCad2DTranslator::JudgeSame2DEdge(const Edge2DInfo& leftEdge_, const Edge2DInfo& rightEdge_) const
{
	if ((leftEdge_.nEdgeType != rightEdge_.nEdgeType) || (EdgeInfo::EDGE_NONE == leftEdge_.nEdgeType))
		return false;

	// ֱ
	if (EdgeInfo::EDGE_LINE == leftEdge_.nEdgeType)
	{
		if (leftEdge_.ptStart.IsEqual(rightEdge_.ptStart, g_nDistPrecision) && leftEdge_.ptEnd.IsEqual(rightEdge_.ptEnd, g_nDistPrecision))
			return true;
		else if (leftEdge_.ptStart.IsEqual(rightEdge_.ptEnd, g_nDistPrecision) && leftEdge_.ptEnd.IsEqual(rightEdge_.ptStart, g_nDistPrecision))
			return true;

		return false;
	}

	// Բ
	ASSERT(EdgeInfo::EDGE_OVAL == leftEdge_.nEdgeType);
	if (!leftEdge_.ptCenter.IsEqual(rightEdge_.ptCenter, g_nDistPrecision))
		return false;
	if (!DIST_EQU(leftEdge_.nLongRadius, rightEdge_.nLongRadius))
		return false;
	if (!DIST_EQU(leftEdge_.nShortRadius, rightEdge_.nShortRadius))
		return false;
	if (!ANGLE_EQU(fabs(leftEdge_.nLastParam - leftEdge_.nFirstParam), fabs(rightEdge_.nLastParam - rightEdge_.nFirstParam)))
		return false;
	if (!DIST_EQU(leftEdge_.nLongRadius, leftEdge_.nShortRadius) && !ANGLE_EQU(leftEdge_.nDipAngle, rightEdge_.nDipAngle)
		&& !ANGLE_EQU(fabs(leftEdge_.nDipAngle - rightEdge_.nDipAngle), g_nPIE))
		return false;

	gp_XY _ptLeftStart = leftEdge_.StartPoint();
	gp_XY _ptLeftEnd = leftEdge_.EndPoint();
	gp_XY _ptRightStart = rightEdge_.StartPoint();
	gp_XY _ptRightEnd = rightEdge_.EndPoint();
	if (_ptLeftStart.IsEqual(_ptRightStart, g_nDistPrecision) && _ptLeftEnd.IsEqual(_ptRightEnd, g_nDistPrecision)
		&& ANGLE_EQU(leftEdge_.nLastParam - leftEdge_.nFirstParam, rightEdge_.nLastParam - rightEdge_.nFirstParam))
		return true;
	else if (_ptLeftStart.IsEqual(_ptRightEnd, g_nDistPrecision) && _ptLeftEnd.IsEqual(_ptRightStart, g_nDistPrecision)
		&& ANGLE_EQU(leftEdge_.nLastParam - leftEdge_.nFirstParam, rightEdge_.nFirstParam - rightEdge_.nLastParam))
		return true;

	return false;
}

Standard_Boolean CCad2DTranslator::CheckIsOnlyLeftBoundary(const Wire2DInfo& wire2D_) const
{
	ASSERT(!m_listArrisBoundary.empty());
	ASSERT(m_listArrisBoundary.front().bRealArris && m_listArrisBoundary.back().bRealArris);

	Standard_Real _nStartX = wire2D_.listEdges.front().StartPoint().X();
	Standard_Real _nEndX = wire2D_.listEdges.back().EndPoint().X();
	if (!DIST_EQU(_nStartX, m_listArrisBoundary.front().nArrisXPos) && !DIST_EQU(_nEndX, m_listArrisBoundary.front().nArrisXPos))
		return false;

	Standard_Real _nNextBoundary = 0.;
	const TubeInfo::TubeType& _nTubeType = m_pTubeToCadTranslator->m_TubeParam.nTubeType;
	if ((TubeInfo::circular == _nTubeType) || (TubeInfo::oval == _nTubeType))
		_nNextBoundary = m_listArrisBoundary.back().nArrisXPos;
	else
	{
		LISTARRISBOUNDARY::const_iterator _iterBegin = ++m_listArrisBoundary.begin();
		LISTARRISBOUNDARY::const_iterator _iterEnd = m_listArrisBoundary.end();
		for (; _iterBegin != _iterEnd; ++_iterBegin)
		{
			if (_iterBegin->bRealArris)
			{
				_nNextBoundary = _iterBegin->nArrisXPos;
				break;
			}
		}

		ASSERT(_iterBegin != _iterEnd);
	}

	if (DIST_EQU(_nStartX, _nNextBoundary) || DIST_EQU(_nEndX, _nNextBoundary))
		return false;
	if ((_nStartX > _nNextBoundary) || (_nEndX > _nNextBoundary))
		return false;

	return true;
}

Standard_Boolean CCad2DTranslator::CheckIsOnlyRightBoundary(const Wire2DInfo& wire2D_) const
{
	ASSERT(!m_listArrisBoundary.empty());
	ASSERT(m_listArrisBoundary.front().bRealArris && m_listArrisBoundary.back().bRealArris);

	Standard_Real _nStartX = wire2D_.listEdges.front().StartPoint().X();
	Standard_Real _nEndX = wire2D_.listEdges.back().EndPoint().X();
	if (!DIST_EQU(_nStartX, m_listArrisBoundary.back().nArrisXPos) && !DIST_EQU(_nEndX, m_listArrisBoundary.back().nArrisXPos))
		return false;

	Standard_Real _nLastBoundary = 0.;
	const TubeInfo::TubeType& _nTubeType = m_pTubeToCadTranslator->m_TubeParam.nTubeType;
	if ((TubeInfo::circular == _nTubeType) || (TubeInfo::oval == _nTubeType))
		_nLastBoundary = m_listArrisBoundary.front().nArrisXPos;
	else
	{
		LISTARRISBOUNDARY::const_reverse_iterator _iterRBegin = ++m_listArrisBoundary.rbegin();
		LISTARRISBOUNDARY::const_reverse_iterator _iterREnd = m_listArrisBoundary.rend();
		for (; _iterRBegin != _iterREnd; ++_iterRBegin)
		{
			if (_iterRBegin->bRealArris)
			{
				_nLastBoundary = _iterRBegin->nArrisXPos;
				break;
			}
		}

		ASSERT(_iterRBegin != _iterREnd);
	}

	if (DIST_EQU(_nStartX, _nLastBoundary) || DIST_EQU(_nEndX, _nLastBoundary))
		return false;
	if ((_nStartX < _nLastBoundary) || (_nEndX < _nLastBoundary))
		return false;

	return true;
}

Standard_Boolean CCad2DTranslator::CombineWireOnArris(const Wire2DInfo& wire2D_, Standard_Real nXOffset_, LISTWIRES2D& listWires_)
{
	// ϵ߼ƫúԺĿߺϲϲĿ
	gp_XY _ptStartPoint = wire2D_.listEdges.front().StartPoint() + gp_XY(nXOffset_, 0.);
	gp_XY _ptEndPoint = wire2D_.listEdges.back().EndPoint() + gp_XY(nXOffset_, 0.);

	LISTWIRES2D::iterator _iterWireEnd = listWires_.end();
	LISTWIRES2D::iterator _iterWireCur = listWires_.begin();
	while (_iterWireCur != _iterWireEnd)
	{
		if (_iterWireCur->listEdges.front().StartPoint().IsEqual(_ptStartPoint, g_nDistPrecision))
		{
			LISTEDGES2D::const_iterator _iterEdgeBegin = wire2D_.listEdges.begin();
			LISTEDGES2D::const_iterator _iterEdgeEnd = wire2D_.listEdges.end();
			for (; _iterEdgeBegin != _iterEdgeEnd; ++_iterEdgeBegin)
			{
				_iterWireCur->listEdges.push_front(*_iterEdgeBegin);
				_iterWireCur->listEdges.front().Reverse();
				AddXOffsetForEdge(_iterWireCur->listEdges.front(), nXOffset_);
			}
		}
		else if (_iterWireCur->listEdges.front().StartPoint().IsEqual(_ptEndPoint, g_nDistPrecision))
		{
			LISTEDGES2D::const_reverse_iterator _iterEdgeRBegin = wire2D_.listEdges.rbegin();
			LISTEDGES2D::const_reverse_iterator _iterEdgeREnd = wire2D_.listEdges.rend();
			for (; _iterEdgeRBegin != _iterEdgeREnd; ++_iterEdgeRBegin)
			{
				_iterWireCur->listEdges.push_front(*_iterEdgeRBegin);
				AddXOffsetForEdge(_iterWireCur->listEdges.front(), nXOffset_);
			}
		}
		else if (_iterWireCur->listEdges.back().EndPoint().IsEqual(_ptStartPoint, g_nDistPrecision))
		{
			LISTEDGES2D::const_iterator _iterEdgeBegin = wire2D_.listEdges.begin();
			LISTEDGES2D::const_iterator _iterEdgeEnd = wire2D_.listEdges.end();
			for (; _iterEdgeBegin != _iterEdgeEnd; ++_iterEdgeBegin)
			{
				_iterWireCur->listEdges.push_back(*_iterEdgeBegin);
				AddXOffsetForEdge(_iterWireCur->listEdges.back(), nXOffset_);
			}
		}
		else if (_iterWireCur->listEdges.back().EndPoint().IsEqual(_ptEndPoint, g_nDistPrecision))
		{
			LISTEDGES2D::const_reverse_iterator _iterEdgeRBegin = wire2D_.listEdges.rbegin();
			LISTEDGES2D::const_reverse_iterator _iterEdgeREnd = wire2D_.listEdges.rend();
			for (; _iterEdgeRBegin != _iterEdgeREnd; ++_iterEdgeRBegin)
			{
				_iterWireCur->listEdges.push_back(*_iterEdgeRBegin);
				_iterWireCur->listEdges.back().Reverse();
				AddXOffsetForEdge(_iterWireCur->listEdges.back(), nXOffset_);
			}
		}
		else
		{
			++_iterWireCur;
			continue;
		}

		break;
	}

	if (_iterWireCur == _iterWireEnd)
		return false;

	// Ѿڸ򱻺ϲߣѰҸǷ񻹴ڿԺϲ
	_ptStartPoint = _iterWireCur->listEdges.front().StartPoint();
	_ptEndPoint = _iterWireCur->listEdges.back().EndPoint();
	LISTWIRES2D::iterator _iterWireTemp = listWires_.begin();
	while (_iterWireTemp != _iterWireEnd)
	{
		if (_iterWireTemp == _iterWireCur)
		{
			++_iterWireTemp;
			continue;
		}

		if (_iterWireTemp->listEdges.front().StartPoint().IsEqual(_ptStartPoint, g_nDistPrecision))
		{
			LISTEDGES2D::iterator _iterEdgeBegin = _iterWireTemp->listEdges.begin();
			LISTEDGES2D::iterator _iterEdgeEnd = _iterWireTemp->listEdges.end();
			for (; _iterEdgeBegin != _iterEdgeEnd; ++_iterEdgeBegin)
			{
				_iterWireCur->listEdges.push_front(*_iterEdgeBegin);
				_iterWireCur->listEdges.front().Reverse();
			}
		}
		else if (_iterWireTemp->listEdges.front().StartPoint().IsEqual(_ptEndPoint, g_nDistPrecision))
		{
			_iterWireCur->listEdges.splice(_iterWireCur->listEdges.end(), _iterWireTemp->listEdges);
		}
		else if (_iterWireTemp->listEdges.back().EndPoint().IsEqual(_ptStartPoint, g_nDistPrecision))
		{
			_iterWireCur->listEdges.splice(_iterWireCur->listEdges.begin(), _iterWireTemp->listEdges);
		}
		else if (_iterWireTemp->listEdges.back().EndPoint().IsEqual(_ptEndPoint, g_nDistPrecision))
		{
			LISTEDGES2D::reverse_iterator _iterEdgeRBegin = _iterWireTemp->listEdges.rbegin();
			LISTEDGES2D::reverse_iterator _iterEdgeREnd = _iterWireTemp->listEdges.rend();
			for (; _iterEdgeRBegin != _iterEdgeREnd; ++_iterEdgeRBegin)
			{
				_iterWireCur->listEdges.push_back(*_iterEdgeRBegin);
				_iterWireCur->listEdges.back().Reverse();
			}
		}
		else
		{
			++_iterWireTemp;
			continue;
		}

		listWires_.erase(_iterWireTemp);
		break;
	}

	return true;
}

void CCad2DTranslator::AddXOffsetForEdge(Edge2DInfo& edge2D_, Standard_Real nXOffset_)
{
	if (EdgeInfo::EDGE_LINE == edge2D_.nEdgeType)
	{
		edge2D_.ptStart.SetX(edge2D_.ptStart.X() + nXOffset_);
		edge2D_.ptEnd.SetX(edge2D_.ptEnd.X() + nXOffset_);
	}
	else
	{
		ASSERT(EdgeInfo::EDGE_OVAL == edge2D_.nEdgeType);
		edge2D_.ptCenter.SetX(edge2D_.ptCenter.X() + nXOffset_);
	}
}

void CCad2DTranslator::Trans2DWireTo2DCad(LISTEDGES2D& listEdges_)
{
	ASSERT(m_pTubeToCadTranslator != NULL);

	ReplenishMissEdge(listEdges_);
	MoveWireToExtendedPlane(listEdges_);
	m_pTubeToCadTranslator->Trans2DWireTo2DCad(listEdges_);
}

void CCad2DTranslator::ReplenishMissEdge(LISTEDGES2D& listEdges_)
{
	// ֻԿͼЧ
	gp_XY _ptStartPoint = listEdges_.front().StartPoint();
	gp_XY _ptEndPoint = listEdges_.back().EndPoint();
	if (_ptStartPoint.IsEqual(_ptEndPoint, g_nDistPrecision))
		return;

	if ((listEdges_.size() == 1) && (EdgeInfo::EDGE_LINE == listEdges_.back().nEdgeType))
		return;

	// ʹͼαպϵı
	Edge2DInfo _edge2D;
	_edge2D.nEdgeType = EdgeInfo::EDGE_LINE;
	_edge2D.ptStart = _ptEndPoint;
	_edge2D.ptEnd = _ptStartPoint;

	// жյǷ߶
	LISTEDGES2D::const_iterator _iterBegin = m_listEdgesOnArris.begin();
	LISTEDGES2D::const_iterator _iterEnd = m_listEdgesOnArris.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		ASSERT(EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType);

		gp_XY _ptStartTemp = _iterBegin->StartPoint();
		gp_XY _ptEndTemp = _iterBegin->EndPoint();
		if (JudgePointOnSegment(_ptStartTemp, _ptEndTemp, _ptStartPoint) && JudgePointOnSegment(_ptStartTemp, _ptEndTemp, _ptEndPoint))
		{
			listEdges_.push_back(_edge2D);
			return;
		}

		_ptStartTemp.SetX(_ptStartTemp.X() + m_nPerimeter);
		_ptEndTemp.SetX(_ptEndTemp.X() + m_nPerimeter);
		if (JudgePointOnSegment(_ptStartTemp, _ptEndTemp, _ptStartPoint) && JudgePointOnSegment(_ptStartTemp, _ptEndTemp, _ptEndPoint))
		{
			listEdges_.push_back(_edge2D);
			return;
		}
	}
}

Standard_Boolean CCad2DTranslator::JudgePointOnSegment(const gp_XY& ptStart_, const gp_XY& ptEnd_, const gp_XY& ptPoint_)
{
	if (ptPoint_.IsEqual(ptStart_, g_nDistPrecision) || ptPoint_.IsEqual(ptEnd_, g_nDistPrecision))
		return true;

	gp_XY _ptStartVector = (ptPoint_ - ptStart_).Normalized();
	gp_XY _ptEndVector = (ptPoint_ - ptEnd_).Normalized();

	return _ptStartVector.IsEqual(_ptEndVector.Reversed(), g_nDistPrecision);
}

void CCad2DTranslator::MoveWireToExtendedPlane(LISTEDGES2D& listEdges_)
{
	// ֻԲ/ԲЧ
	const TubeInfo::TubeType& _nTubeType = m_pTubeToCadTranslator->m_TubeParam.nTubeType;
	if ((_nTubeType != TubeInfo::circular) && (_nTubeType != TubeInfo::oval))
		return;

	Standard_Boolean _bOnBoundary = false;
	LISTEDGES2D::iterator _iterBegin = listEdges_.begin();
	LISTEDGES2D::iterator _iterEnd = listEdges_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		Standard_Real _nStartX = _iterBegin->StartPoint().X();
		if (!DIST_EQU(_nStartX, m_nPerimeter) && (_nStartX > m_nPerimeter))
		{
			_bOnBoundary = true;
			break;
		}

		Standard_Real _nEndX = _iterBegin->EndPoint().X();
		if (!DIST_EQU(_nEndX, m_nPerimeter) && (_nEndX > m_nPerimeter))
		{
			_bOnBoundary = true;
			break;
		}
	}

	if (!_bOnBoundary)
		return;

	for (_iterBegin = listEdges_.begin(); _iterBegin != _iterEnd; ++_iterBegin)
	{
		AddXOffsetForEdge(*_iterBegin, -1.0 * m_nPerimeter);
	}
}

//////////////////////////////////////////////////////////////////////////
// class CTubeToCadTranslator
CTubeToCadTranslator::CTubeToCadTranslator()
	: m_plistCadObjects(NULL)
	, m_pFirstCadObject(NULL)
	, m_nMinYPos(DBL_MAX)
	, m_nMaxYPos(-1.0 * DBL_MAX)
{
}

void CTubeToCadTranslator::SetAxleWireDir(const gp_XYZ& ptAxleWireDir_)
{
	m_ptAxleWireDir = ptAxleWireDir_;
	ASSERT(DIST_EQU(m_ptAxleWireDir.Modulus(), 1.0));

	m_listDiceyWires.clear();
	m_OuterBoundary = Wire3DInfo();
	m_TubeParam = TubeInfo();

	m_Wires3DTranslator.Initialize(this);
	m_Wires2DTranslator.Initialize(this);
	m_Cad2DTranslator.Initialize(this);

	m_plistCadObjects = NULL;
	m_pFirstCadObject = NULL;

	m_nMinYPos = DBL_MAX;
	m_nMaxYPos = -1.0 * DBL_MAX;
}

Standard_Boolean CTubeToCadTranslator::Translate(nce::CadObjectList* plistCadObjects_)
{
	ASSERT(plistCadObjects_ != NULL);
	if (!plistCadObjects_)
		return false;

	LISTWIRES3DONPLANE _listWires3DOnPlane;
	if (!m_Wires3DTranslator.Translate(_listWires3DOnPlane))
		return false;

	LISTWIRES2DONSURFACE _listWires2DOnSurface;
	if (!m_Wires2DTranslator.Translate(_listWires3DOnPlane, _listWires2DOnSurface))
		return false;

	m_plistCadObjects = plistCadObjects_;
	if (!m_Cad2DTranslator.Translate(_listWires2DOnSurface) || (NULL == m_pFirstCadObject))
		return false;

	return true;
}

void CTubeToCadTranslator::Trans2DWireTo2DCad(LISTEDGES2D& listEdges_)
{
	ASSERT(!listEdges_.empty());
	if (listEdges_.size() == 0)
		return;

	// תǰȽкϲ
	Merge2DEdges(listEdges_);

	// ȡYı߽
	GetYBoundary(listEdges_);

	if (listEdges_.size() == 1)
	{
		TransPreprocess2DWireTo2DCad(listEdges_);
		return;
	}

	// аԲ򽫶߽зָֹԲԲת
	LISTEDGES2D _listEdges;
	LISTEDGES2D::const_iterator _iterBegin = listEdges_.begin();
	LISTEDGES2D::const_iterator _iterEnd = listEdges_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		if ((_iterBegin->nEdgeType != EdgeInfo::EDGE_OVAL) || DIST_EQU(_iterBegin->nLongRadius, _iterBegin->nShortRadius))
		{
			_listEdges.push_back(*_iterBegin);
			continue;
		}

		// ԲǰĶ߽ת
		TransPreprocess2DWireTo2DCad(_listEdges);
		_listEdges.clear();

		// תǰԲ
		_listEdges.push_back(*_iterBegin);
		TransPreprocess2DWireTo2DCad(_listEdges);
		_listEdges.clear();
	}

	TransPreprocess2DWireTo2DCad(_listEdges);
}

void CTubeToCadTranslator::TransPreprocess2DWireTo2DCad(LISTEDGES2D& listEdges_)
{
	if (listEdges_.empty())
		return;

	// һƷתһȦƽ
	while (true)
	{
		if (m_TubeParam.nTubeType != TubeInfo::rectangular)
			break;

		// ֻXƽеֱߵ
		if ((listEdges_.front().nEdgeType != EdgeInfo::EDGE_LINE) || (listEdges_.back().nEdgeType != EdgeInfo::EDGE_LINE))
			break;

		Edge2DInfo& _frontEdge = listEdges_.front();
		Edge2DInfo& _backEdge = listEdges_.back();
		if (!DIST_EQU(_frontEdge.ptStart.Y(), _frontEdge.ptEnd.Y())
			|| !DIST_EQU(_backEdge.ptStart.Y(), _backEdge.ptEnd.Y())
			|| !DIST_EQU(_frontEdge.ptStart.Y(), _backEdge.ptEnd.Y()))
			break;

		const static Standard_Real _s_nMinThickness = 0.01;
		Standard_Real _nXOffset = max(m_TubeParam.nRectRadius, max(m_TubeParam.nRectThickness, _s_nMinThickness));
		Standard_Real _nXMinBorder = -0.5 * m_TubeParam.nRectWidth;
		Standard_Real _nXMaxBorder = _nXMinBorder + 2.0 * (m_TubeParam.nRectWidth + m_TubeParam.nRectHeight);
		if ((!DIST_EQU(_frontEdge.ptStart.X(), _nXMinBorder) || !DIST_EQU(_backEdge.ptEnd.X(), _nXMaxBorder))
			&& (!DIST_EQU(_frontEdge.ptStart.X(), _nXMaxBorder) || !DIST_EQU(_backEdge.ptEnd.X(), _nXMinBorder)))
			break;
		if (((_frontEdge.ptEnd - _frontEdge.ptStart).Modulus() < _nXOffset) || ((_backEdge.ptEnd - _backEdge.ptStart).Modulus() < _nXOffset))
			break;

		// ʼֹյ
		_frontEdge.ptStart.SetX(_frontEdge.ptStart.X() + _nXOffset);
		_backEdge.ptEnd.SetX(_backEdge.ptEnd.X() + _nXOffset);

		if (DIST_EQU((_frontEdge.ptEnd - _frontEdge.ptStart).Modulus(), 0.))
			listEdges_.pop_front();
		if (!listEdges_.empty() && DIST_EQU((_backEdge.ptEnd - _backEdge.ptStart).Modulus(), 0.))
			listEdges_.pop_back();

		break;
	}

	if (listEdges_.empty())
		return;

	// 
	if (listEdges_.size() > 1)
	{
		if (CheckIsRectangle(listEdges_))
			Trans2DWireToRectangle(listEdges_);
		else
			Trans2DWireToPolyline(listEdges_);

		return;
	}

	ASSERT(listEdges_.size() == 1);
	const Edge2DInfo& _edge2D = listEdges_.back();
	if (EdgeInfo::EDGE_LINE == _edge2D.nEdgeType)
	{
		// ֱ
		Trans2DWireToLine(_edge2D);
		return;
	}

	ASSERT(EdgeInfo::EDGE_OVAL == _edge2D.nEdgeType);
	Standard_Boolean _bEqualRadius = DIST_EQU(_edge2D.nLongRadius, _edge2D.nShortRadius);
	Standard_Boolean _bRound = ANGLE_EQU(fabs(_edge2D.nLastParam - _edge2D.nFirstParam), 2.0 * g_nPIE);
	if (_bEqualRadius && _bRound)
	{
		// Բ
		Trans2DWireToCircle(_edge2D);
	}
	else if (_bEqualRadius && !_bRound)
	{
		// Բ
		Trans2DWireToArc(_edge2D);
	}
	else if (!_bEqualRadius && _bRound)
	{
		// Բ
		Trans2DWireToEllipse(_edge2D);
	}
	else
	{
		// Բ
		Trans2DWireToEllipseArc(_edge2D);
	}
}

void CTubeToCadTranslator::Trans2DWireToLine(const Edge2DInfo& edge2D_)
{
	ASSERT(EdgeInfo::EDGE_LINE == edge2D_.nEdgeType);

	nce::CCadLine* _pCadLine = nce::access::NewCadLine();
	_pCadLine->SetStartPoint(DPOINT2(edge2D_.ptStart.X(), edge2D_.ptStart.Y()));
	_pCadLine->SetEndPoint(DPOINT2(edge2D_.ptEnd.X(), edge2D_.ptEnd.Y()));

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadLine);
	StoreFirstCadObject(_pCadLine);
}

void CTubeToCadTranslator::Trans2DWireToArc(const Edge2DInfo& edge2D_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == edge2D_.nEdgeType);
	ASSERT(DIST_EQU(edge2D_.nLongRadius, edge2D_.nShortRadius));
	ASSERT(!ANGLE_EQU(fabs(edge2D_.nLastParam - edge2D_.nFirstParam), 2.0 * g_nPIE));

	nce::CCadArc* _pCadArc = nce::access::NewCadArc();
	_pCadArc->SetCenterPoint(DPOINT2(edge2D_.ptCenter.X(), edge2D_.ptCenter.Y()));
	_pCadArc->SetRadius(edge2D_.nLongRadius);

	Standard_Real _nStartAngle = NormalizeAngle(min(edge2D_.nFirstParam, edge2D_.nLastParam) + edge2D_.nDipAngle);
	Standard_Real _nCentralAngle = fabs(edge2D_.nLastParam - edge2D_.nFirstParam);
	_pCadArc->SetStartAngle(_nStartAngle);
	_pCadArc->SetCentralAngle(_nCentralAngle);

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadArc);
	StoreFirstCadObject(_pCadArc);
}

void CTubeToCadTranslator::Trans2DWireToCircle(const Edge2DInfo& edge2D_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == edge2D_.nEdgeType);
	ASSERT(DIST_EQU(edge2D_.nLongRadius, edge2D_.nShortRadius));
	ASSERT(ANGLE_EQU(fabs(edge2D_.nLastParam - edge2D_.nFirstParam), 2.0 * g_nPIE));

	nce::CCadCircle* _pCadCircle = nce::access::NewCadCircle();
	_pCadCircle->SetCenterPoint(DPOINT2(edge2D_.ptCenter.X(), edge2D_.ptCenter.Y()));
	_pCadCircle->SetRadius(edge2D_.nLongRadius);

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadCircle);
	StoreFirstCadObject(_pCadCircle);
}

void CTubeToCadTranslator::Trans2DWireToEllipse(const Edge2DInfo& edge2D_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == edge2D_.nEdgeType);
	ASSERT(!DIST_EQU(edge2D_.nLongRadius, edge2D_.nShortRadius));
	ASSERT(ANGLE_EQU(fabs(edge2D_.nLastParam - edge2D_.nFirstParam), 2.0 * g_nPIE));

	nce::CCadEllipse* _pCadEllipse = nce::access::NewCadEllipse();
	_pCadEllipse->SetCenterPoint(DPOINT2(edge2D_.ptCenter.X(), edge2D_.ptCenter.Y()));
	_pCadEllipse->SetLongRadius(edge2D_.nLongRadius);
	_pCadEllipse->SetShortRadius(edge2D_.nShortRadius);
	_pCadEllipse->SetAngle(edge2D_.nDipAngle);

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadEllipse);
	StoreFirstCadObject(_pCadEllipse);
}

void CTubeToCadTranslator::Trans2DWireToEllipseArc(const Edge2DInfo& edge2D_)
{
	ASSERT(EdgeInfo::EDGE_OVAL == edge2D_.nEdgeType);
	ASSERT(!DIST_EQU(edge2D_.nLongRadius, edge2D_.nShortRadius));
	ASSERT(!ANGLE_EQU(fabs(edge2D_.nLastParam - edge2D_.nFirstParam), 2.0 * g_nPIE));

	nce::CCadEllipseArc* _pCadEllipseArc = nce::access::NewCadEllipseArc();
	_pCadEllipseArc->SetCenterPoint(DPOINT2(edge2D_.ptCenter.X(), edge2D_.ptCenter.Y()));
	_pCadEllipseArc->SetLongRadius(edge2D_.nLongRadius);
	_pCadEllipseArc->SetShortRadius(edge2D_.nShortRadius);
	_pCadEllipseArc->SetAngle(edge2D_.nDipAngle);

	Standard_Real _nStartAngle = NormalizeAngle(min(edge2D_.nFirstParam, edge2D_.nLastParam));
	Standard_Real _nCentralAngle = fabs(edge2D_.nLastParam - edge2D_.nFirstParam);
	_pCadEllipseArc->SetStartAngle(_nStartAngle);
	_pCadEllipseArc->SetCentralAngle(_nCentralAngle);

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadEllipseArc);
	StoreFirstCadObject(_pCadEllipseArc);
}

void CTubeToCadTranslator::Trans2DWireToRectangle(const LISTEDGES2D& listEdges_)
{
	ASSERT(listEdges_.size() == 4);
	ASSERT(listEdges_.front().StartPoint().IsEqual(listEdges_.back().EndPoint(), g_nDistPrecision));

	LISTEDGES2D::const_iterator _iterBegin = listEdges_.begin();
	ASSERT(EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType);
	DPOINT2 _ptPoint1(_iterBegin->ptStart.X(), _iterBegin->ptStart.Y());

	++_iterBegin;
	ASSERT(EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType);
	DPOINT2 _ptPoint2(_iterBegin->ptStart.X(), _iterBegin->ptStart.Y());

	++_iterBegin;
	ASSERT(EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType);
	DPOINT2 _ptPoint3(_iterBegin->ptStart.X(), _iterBegin->ptStart.Y());

	++_iterBegin;
	ASSERT(EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType);
	DPOINT2 _ptPoint4(_iterBegin->ptStart.X(), _iterBegin->ptStart.Y());
	ASSERT(++_iterBegin == listEdges_.end());

	nce::CCadRectangle* _pCadRectangle = nce::access::NewCadRectangle();
	_pCadRectangle->SetVertexes(_ptPoint1, _ptPoint2, _ptPoint3, _ptPoint4);

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadRectangle);
	StoreFirstCadObject(_pCadRectangle);
}

void CTubeToCadTranslator::Trans2DWireToPolyline(const LISTEDGES2D& listEdges_)
{
	ASSERT(listEdges_.size() > 1);

	nce::CCadPolyline* _pCadPolyline = nce::access::NewCadPolyline();
	gp_XY _ptStartPoint = listEdges_.front().StartPoint();
	_pCadPolyline->SetStartPoint(DPOINT2(_ptStartPoint.X(), _ptStartPoint.Y()));

	LISTEDGES2D::const_iterator _iterBegin = listEdges_.begin();
	LISTEDGES2D::const_iterator _iterEnd = listEdges_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		ASSERT(DIST_EQU(_pCadPolyline->GetEndPoint().x, _iterBegin->StartPoint().X()));
		ASSERT(DIST_EQU(_pCadPolyline->GetEndPoint().y, _iterBegin->StartPoint().Y()));
		gp_XY _ptTempEnd = _iterBegin->EndPoint();
		if (EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType)
		{
			_pCadPolyline->AddLine(DPOINT2(_ptTempEnd.X(), _ptTempEnd.Y()));
		}
		else if (DIGIT_EQU(fabs(_iterBegin->nLastParam - _iterBegin->nFirstParam), 2.0 * g_nPIE))
		{
			// ԲֳԲ
			ASSERT(EdgeInfo::EDGE_OVAL == _iterBegin->nEdgeType);
			ASSERT(DIST_EQU(_iterBegin->nLongRadius, _iterBegin->nShortRadius));
			gp_XY _ptTempMiddle = 2.0 * _iterBegin->ptCenter - _ptTempEnd;
			Standard_Real _nBulge = (_iterBegin->nLastParam > _iterBegin->nFirstParam) ? 1.0 : -1.0;
			_pCadPolyline->AddArc(DPOINT2(_ptTempMiddle.X(), _ptTempMiddle.Y()), _nBulge);
			_pCadPolyline->AddArc(DPOINT2(_ptTempEnd.X(), _ptTempEnd.Y()), _nBulge);
		}
		else
		{
			ASSERT(EdgeInfo::EDGE_OVAL == _iterBegin->nEdgeType);
			ASSERT(DIST_EQU(_iterBegin->nLongRadius, _iterBegin->nShortRadius));
			Standard_Real _nBulge = tan(0.25 * (_iterBegin->nLastParam - _iterBegin->nFirstParam));
			_pCadPolyline->AddArc(DPOINT2(_ptTempEnd.X(), _ptTempEnd.Y()), _nBulge);
		}
	}

	gp_XY _ptEndPoint = listEdges_.back().EndPoint();
	if (_ptStartPoint.IsEqual(_ptEndPoint, g_nDistPrecision))
		_pCadPolyline->SetClosed(true);

	nce::access::CadObjectList_push_back(m_plistCadObjects, _pCadPolyline);
	StoreFirstCadObject(_pCadPolyline);
}

void CTubeToCadTranslator::StoreFirstCadObject(nce::CCadObject* pCadObject_)
{
	if (NULL == m_pFirstCadObject)
	{
		m_pFirstCadObject = pCadObject_;
		ASSERT(m_pFirstCadObject != NULL);
	}
}

Standard_Boolean CTubeToCadTranslator::CheckIsRectangle(const LISTEDGES2D& listEdges_) const
{
	// պͼ
	if (!listEdges_.front().StartPoint().IsEqual(listEdges_.back().EndPoint(), g_nDistPrecision))
		return false;

	// ߣڶ㴦
	if (listEdges_.size() != 4)
		return false;

	LISTEDGES2D::const_iterator _iterBegin = listEdges_.begin();
	LISTEDGES2D::const_iterator _iterEnd = listEdges_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		if (_iterBegin->nEdgeType != EdgeInfo::EDGE_LINE)
			return false;
	}

	_iterBegin = listEdges_.begin();
	_iterEnd = listEdges_.end();
	gp_XY _ptPoint1 = _iterBegin->ptStart;
	gp_XY _ptPoint2 = (++_iterBegin)->ptStart;
	gp_XY _ptPoint3 = (++_iterBegin)->ptStart;
	gp_XY _ptPoint4 = (++_iterBegin)->ptStart;
	ASSERT(++_iterBegin == _iterEnd);

	// жĸǷɾ
	if (!(_ptPoint1 - _ptPoint2).IsEqual(_ptPoint4 - _ptPoint3, g_nDistPrecision))
		return false;
	if (!(_ptPoint1 - _ptPoint4).IsEqual(_ptPoint2 - _ptPoint3, g_nDistPrecision))
		return false;
	if (DIST_EQU((_ptPoint1 - _ptPoint2).Modulus(), 0.) || DIST_EQU((_ptPoint1 - _ptPoint4).Modulus(), 0.))
		return false;
	if (!DIST_EQU((_ptPoint1 - _ptPoint3).Modulus(), (_ptPoint2 - _ptPoint4).Modulus()))
		return false;

	return true;
}

Standard_Real CTubeToCadTranslator::NormalizeAngle(Standard_Real nAngle_) const
{
	if (nAngle_ < 0.)
		return (nAngle_ + 2.0 * g_nPIE);

	if (nAngle_ >= 2.0 * g_nPIE)
		return (nAngle_ - 2.0 * g_nPIE);

	return nAngle_;
}

void CTubeToCadTranslator::GetYBoundary(const LISTEDGES2D& listEdges_)
{
	LISTEDGES2D::const_iterator _iterBegin = listEdges_.begin();
	LISTEDGES2D::const_iterator _iterEnd = listEdges_.end();
	for (; _iterBegin != _iterEnd; ++_iterBegin)
	{
		// ֱ
		if (EdgeInfo::EDGE_LINE == _iterBegin->nEdgeType)
		{
			GetYBoundary(_iterBegin->ptStart.Y());
			GetYBoundary(_iterBegin->ptEnd.Y());
			continue;
		}

		// Բܲһ㶼ǱպͼΣԲ
		gp_XY _ptPole;
		ASSERT(EdgeInfo::EDGE_OVAL == _iterBegin->nEdgeType);
		ASSERT(ANGLE_EQU(fabs(_iterBegin->nDipAngle), 2.0 * g_nPIE) || (fabs(_iterBegin->nDipAngle) < 2.0 * g_nPIE));
		if (ANGLE_EQU(fabs(_iterBegin->nDipAngle), 0.5 * g_nPIE) || ANGLE_EQU(fabs(_iterBegin->nDipAngle), 1.5 * g_nPIE))
			_ptPole.SetX(_iterBegin->nLongRadius);
		else
		{
			// תǰоԲһе
			Standard_Real _nSlope = tan(-1.0 * _iterBegin->nDipAngle);
			Standard_Real _nDenominator = sqrt(pow(_iterBegin->nLongRadius * _nSlope, 2) + pow(_iterBegin->nShortRadius, 2));
			_ptPole.SetX(pow(_iterBegin->nLongRadius, 2) * _nSlope / _nDenominator);
			_ptPole.SetY(-1.0 * pow(_iterBegin->nShortRadius, 2) / _nDenominator);
		}

		Standard_Real _nRotateY = _ptPole.X() * sin(_iterBegin->nDipAngle) + _ptPole.Y() * cos(_iterBegin->nDipAngle);
		GetYBoundary(_iterBegin->ptCenter.Y() + _nRotateY);
		GetYBoundary(_iterBegin->ptCenter.Y() - _nRotateY);
	}
}

void CTubeToCadTranslator::GetYBoundary(Standard_Real nAimYPos_)
{
	if (nAimYPos_ < m_nMinYPos)
		m_nMinYPos = nAimYPos_;
	if (nAimYPos_ > m_nMaxYPos)
		m_nMaxYPos = nAimYPos_;
}